#include <stdio.h>
#include "compilerApi.h"

enum FLAG {off, on};

FLAG in_main_program;
FLAG in_procedure;

int tabnumber=0;

void puttab(int number)
{
  for(int i=0; i<number; i++)
    printf("  ");
}

void action(SgStatement * s)
{
  puttab(tabnumber);
  printf("id:%4d   variant:%4d\n", s->id(), s->variant());
}

void transform(SgStatement * s) 
{
  switch(s->variant()) {
  case GLOBAL:
    {
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      for (int i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren; i++)
	transform(children[i]);
      
      delete [] children;
      break;
    };

  case PROG_HEDR:
    {
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      for (int i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      in_main_program = on;

      action(s);

      tabnumber++;
      
      for (i=0; i<org_numberOfChildren; i++)
	transform(children[i]);

      tabnumber--;

      in_main_program=off;
      
      delete [] children;

      break;
    }

  case PROC_HEDR:
    {
      SgProcHedrStmt * prochedr = (SgProcHedrStmt *)s;

      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];
      for (int i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      in_procedure=on;

      action(s);

      tabnumber++;

      for (i=0; i<org_numberOfChildren; i++)
	transform(children[i]);
      
      tabnumber--;

      in_procedure = off;

      delete [] children;

      break;
    };

  case VAR_DECL:
  case ASSIGN_STAT:
    action(s);
    break;

  case FOR_NODE:
    {

      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      for (int i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      action(s);

      tabnumber++;

      for (i=0; i<org_numberOfChildren; i++)
	transform(children[i]);

      tabnumber--;

      delete [] children;
      
      break;
    }

  case CONTROL_END:
  case ENDDO_STMT:
    break;

  default:
    printf("Not touched statement in transform phase on line %d\n", s->lineNumber());
    break;
  };
};

int main(int argc, char **argv){

  SgProject * project = new SgProject("test", argc, argv);

  for (int i=0; i< project->numberOfFiles(); i++) {

    SgFile * file = &(project->file(i));
    
    SgStatement * s = file->firstStatement();

    transform(s);

  };

  // save a copy of the input file, if you like.
  // f->saveDepFile("debug.dep");

  return 0;
}







