/*******************************************************************************
    Semantic check about HPF align directive
    written by Wang Zhenlin
*******************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include <stdio.h>

#include "tags.h"
#   include "typedef.h"
#   include "bif.h"
#   include "ll.h"
#   include "symb.h"

#include "defnodes.h"
#include "macro.h"
#include "decl.h"
 
#include "exglobal.h"

#define REFERD_BIT              0x1

PTR_SYMB align_dummy_table[7];
int      align_table_size ;

int is_align_dummy(sym)
PTR_SYMB sym;
{
	int i;
	for (i=0; i<align_table_size ; i++)
		if (!strcasecmp(sym->ident , align_dummy_table[i]->ident))
			return 1;
	return 0;
}

/*
	align-directive        is ALIGN alignee align-directive-stuff
	align-directive-stuff  is (align-source-list) align-with-clause

			BIF
			 |
		 +-------+--------+
		 |       |        |
	     >var_list        EXPR_LIST
				  |
			>align_attr(ALIGN_OP)
				  |
			      +--------+
			      |        |
			  align_source align_with
				           |
				      +----------+
						 |
					     array_ref (align_target)
*/
int align_attr_smtc(var_list,align_attr,isAlign)
PTR_LLND var_list;
PTR_LLND align_attr;
int      isAlign;
{
     PTR_LLND varlist=var_list;
     PTR_SYMB alignee,align_target,align_target_leaf,sym,symbol;
     PTR_LLND target;
     int      alignee_dims,collapse_dims,align_source_num,align_target_dims;
     PTR_LLND align_source,align_with,cur_llnd;
     PTR_LLND source_ele,with_ele;
     int      have_star;
     PTR_LLND llndA, llndB; /* added by LY, get align-subscript */
     PTR_LLND exprList,expr,node;
     int      err_code,no;
     int      dim1,dim2;
     int      i,count;
	
	align_table_size=0;

	err_code = 0;

	llndA = (PTR_LLND) malloc(sizeof(PTR_LLND));
	llndB = (PTR_LLND) malloc(sizeof(PTR_LLND));
     
      /*get dimension of align_target*/
     if (align_attr == LLNULL) /*no align source and align target*/
         return 1;

     align_source = NODE_OPERAND0(align_attr);
     align_with =   NODE_OPERAND1(align_attr);
		
     target = NODE_OPERAND1(align_with);
     /*align_target= NODE_SYMB(NODE_OPERAND1(align_with));*/
     align_target = get_var_name(target);
	/* check for HPF 2.0 the align target is structure componenet */
     align_target_leaf=align_target;
     if (target->variant == RECORD_REF )
     {
	PTR_LLND part;
	part=NODE_OPERAND1(target);
        for ( ; part && NODE_OPERAND1(part) ; part=NODE_OPERAND1(part) )
                if (!is_scalar(NODE_OPERAND0(part)->type))
                        error3(657,NODE_SYMB(NODE_OPERAND0(part))->ident);
	align_target_leaf=NODE_SYMB(NODE_OPERAND0(part));
     }
     
     if (align_source == LLNULL && align_with == LLNULL)
        return 1;

     /* add by LY, check for global symbol */
/*					modified by zheng.gb
     check_leaf_sym(align_source);
     check_leaf_sym(align_with);
     check_leaf_sym(var_list);
*/

     /* added by LY, to check PURE function */
        if ( is_in_pure() && !isAlign)
	{
		err_code = 325;
                error3(325);
	}

     /*There exist starts before align_source or align_with*/     
     have_star = 0;
     if (align_with != LLNULL && align_with->entry.Template.ll_ptr1 != LLNULL)  
         have_star = 1;

     /* star can not be in front of align_source or align_with in realign stmt*/
     if (have_star && !isAlign)
     {
	err_code = 310;
        error1(310);
     }


     if (align_source == LLNULL ){
        align_source_num = -1;
     }
     else {
        align_source_num = 0;
	collapse_dims = 0;
	while (align_source != LLNULL) {
                align_source_num++;
		if ( NODE_CODE(NODE_OPERAND0(align_source)) == STAR_OP )
			collapse_dims++;
		if (  NODE_CODE(NODE_OPERAND0(align_source)) != STAR_OP
		   || NODE_CODE(NODE_OPERAND0(align_source)) != COLON_OP)
  	           cur_llnd = NODE_OPERAND0(NODE_OPERAND0(align_source));

		/* add by zheng.gb it is a dummy */      
		if (NODE_CODE(NODE_OPERAND0(align_source))==VAR_REF){
/*
     			check_leaf_sym(NODE_OPERAND0(align_source));
			symbol=NODE_SYMB(NODE_OPERAND0(align_source));
				  align dummy is of integer type ? 
			if (!isIntegerType(symbol->type))
                        {
				err_code = 224;
				error3(224,symbol->ident);
			}
*/
			/* make another symbol , and make local */
			sym=NODE_SYMB(NODE_OPERAND0(align_source));
			symbol=make_symb(fi,VARIABLE_NAME,sym->ident);
			symbol->scope=cur_bfnd;
			symbol->type = global_int;
			symbol->decl = HARD;
			NODE_SYMB(NODE_OPERAND0(align_source))=symbol;
			/* if (!var_declared(sym)) */
			if (sym->decl != HARD)
				sym->scope=BFNULL;
	
			if (is_align_dummy(symbol))
				error3(330,symbol->ident);
			else
				align_dummy_table[align_table_size++]=symbol;
		}

                align_source = NODE_OPERAND1(align_source);
	 } /*while ; go through every align_source*/
     }  

     /*check every alignees*/
     while (var_list != LLNULL){ 

        /* get dimensions of alignee*/
	PTR_LLND  node = var_list->entry.Template.ll_ptr1;
        /* add by LY, check for global symbol */
        check_leaf_sym(node);

        alignee=get_var_name(node);

	/* check for HPF 2.0 */
        /* check if the type of alignee is explicit mapped ! */
        {
        PTR_TYPE base_type=get_base_type(alignee->type);
        if (base_type->variant == T_RECORD)
                if (base_type->entry.re_decl.mapped_flag == 1)
                        error3(655,alignee->ident);

        if (node->variant == RECORD_REF) 
                if (isAlign == 1)   /* in redistribute stmt */
                        error3(656,alignee->ident);
		else {
			PTR_LLND part=NODE_OPERAND1(node);
			for ( ; part && NODE_OPERAND1(part) ; part=NODE_OPERAND1(part) )
				if (!is_scalar(NODE_OPERAND0(part)->type))
					error3(657,NODE_SYMB(NODE_OPERAND0(part))->ident);
		}
        }

	/*subobject can not alignd , this is for HPF 1.0*/
/*
        if (isSubobject(var_list->entry.Template.ll_ptr1)) 
	{
	  err_code = 301;
          error3(301,alignee->ident);
	}
       else
*/
          if (node->type->variant == T_ARRAY)
            alignee_dims=node->type->entry.ar_decl.num_dimensions;
         else
            alignee_dims=0;   /*scalar*/
      
         /*if existing star infront of align_source or align_with, alignee 
          should have INHERIT attribute*/
/*				comment by zheng.gb 
       if (isAlign && have_star && !have_attribute(alignee,INHERIT_BIT)) 
          error3(302,alignee->ident);
*/
       
       /*(re)alignee should not have DISTRIBUTE attribute*/
       if (have_attribute(alignee,DISTRIBUTE_BIT)) 
	{
	  err_code = 307;
          error3(307,alignee->ident);
	}
       
       /*realignee should have DYNAMIC attribute*/
       if (!isAlign && !have_dynamic_attribute(node))
       {
	  err_code = 311;
          error3(311,alignee->ident);
       }

        /* added by LY, to check PURE function */
        if (is_in_pure() && is_pure_local_var(alignee) && alignee->scope!=align_target->scope)
	{
		err_code = 323;
                error3(323, alignee->ident);
	}

       if (align_source_num != -1 && align_source_num != alignee_dims)
	{
	  err_code = 312;
          error3(312,alignee->ident);
	}

       /*add attr pointer in symbol table*/
       alignee->entry.var_decl.hpfattr = align_attr;

       var_list = NODE_OPERAND1(var_list);
     }


     /* added by LY, call Zheng G.B.'s function to check align-subscript */
     no = 0;
     for (exprList=NODE_OPERAND0(NODE_OPERAND1(align_with));
		exprList!=LLNULL; exprList=NODE_OPERAND1(exprList))
     {
		/* add by zheng.gb for check if is specification expression */	
	expr = NODE_OPERAND0(exprList);
	no++;
	
	/* if not expression do not count a*x + b */
	if (expr->variant == STAR_OP ) continue;

	/* check if align subscript triplet contain dummys */
	if (expr->variant == DDOT)
	{
		for (i=0; i<align_table_size ; i++)
			if (IsSymbolInExpression(expr,align_dummy_table[i]))
			{
				error3(331,align_dummy_table[i]->ident);
				err_code=331;
			}
	}

	/* if not expression do not count a*x + b */
	if (expr->variant == DDOT ) continue;

	/*  only 2 cases: int_expr , align_suscript_use */
	count = 0;
	for (i=0; i<align_table_size ; i++)
		if (node=IsSymbolInExpression(expr,align_dummy_table[i]))
		{
			if (align_dummy_table[i]->attr & REFERD_BIT)
			{
				err_code = 333;
				error3(333,align_dummy_table[i]->ident);
			}
			else
				align_dummy_table[i]->attr |= REFERD_BIT;
			NODE_SYMB(node) = align_dummy_table[i];
			count ++;
		}
	/* if (count > 1)  { error1(332); err_code = 332 ; }  */

	check_expr_type(expr);
	if (!isInteger(exprList->entry.Template.ll_ptr1))
		error2(225,no);

     	err_code = get_a_b(align_attr->entry.Template.ll_ptr1,
		exprList->entry.Template.ll_ptr1, &llndA, &llndB);
	if (err_code) error1(err_code);
	if (llndA) {
		if (check_initialization_expr(llndA)) {
			comput_const_expr(llndA,0);
		}
	}
	if (llndB) {
		if (check_initialization_expr(llndB)) {
			comput_const_expr(llndB,0);
		}
	}
/* for test */
/*
	exprList->entry.Template.ll_ptr1=llndA;
	exprList->entry.Template.ll_ptr2=llndB;
	break;
*/
     }
     if (no && no != NODE_TYPE(NODE_SYMB(NODE_OPERAND1(align_with)))->entry.ar_decl.num_dimensions)
     {
	err_code = 318;
	error3(318,SYMB_IDENT(NODE_SYMB(NODE_OPERAND1(align_with))));
     }

	/* check every dimension of alignee with the align target */
     if (err_code) return;

     align_source = NODE_OPERAND0(align_attr);
     align_with =   NODE_OPERAND1(align_attr);
     align_with =   NODE_OPERAND0(NODE_OPERAND1(align_with));

     dim1=1; dim2=1;
     while (align_source && align_with )
     {
	while (align_source && NODE_CODE(NODE_OPERAND0(align_source)) != COLON_OP)  {
     		align_source = NODE_OPERAND1(align_source);
		dim1++;
	}
	if (align_source)
     		source_ele = NODE_OPERAND0(align_source);
	while (align_with && NODE_CODE(NODE_OPERAND0(align_with)) != DDOT)  {
     		align_with = NODE_OPERAND1(align_with);
		dim2++;
	}
	if (align_with)
		with_ele   = NODE_OPERAND0(align_with);

        if (!align_with || !align_source)
     	{
		if (!align_with && !align_source)  ;
		else error3(316,align_target->ident);
		break;
     	}

	for (expr=varlist;expr; expr=NODE_OPERAND1(expr) )
	{
/*
		sym=NODE_SYMB(NODE_OPERAND0(expr));
		if (!check_alignee_var_dim(sym,dim1,align_target,dim2,with_ele))
*/
		/* check for HPF 2.0 */
		if (expr->variant == RECORD_REF)
		{
			PTR_LLND part=NODE_OPERAND1(expr);
        		for ( ; part && NODE_OPERAND1(part) ; part=NODE_OPERAND1(part) );
			sym=NODE_SYMB(NODE_OPERAND0(part));
		}
		else
			sym=NODE_SYMB(NODE_OPERAND0(expr));
		if (!check_alignee_var_dim(sym,dim1,align_target_leaf,dim2,with_ele))
			error4(317,sym->ident,align_target->ident);
	}
     	align_source = NODE_OPERAND1(align_source);
     	align_with = NODE_OPERAND1(align_with);
	dim1++; dim2++;
     }

}

int is_pure_local_var(pVar)
PTR_SYMB pVar;
{
	if (pVar->scope->entry.procedure.proc_symb->attr & PURE_BIT)
		return 1;
	else return 0;
}

/* 1: ok , 0: error */
int check_alignee_var_dim(var,dim1,target,dim2,ddot)
PTR_SYMB var,target;
int  dim1,dim2;
PTR_LLND ddot;
{
	int vardim,targetdim;
	int UA,LA;
	int UT,LT,ST;

	if (!get_array_lbound(var->type,dim1,&LA) ) return 1;
	if (!get_array_ubound(var->type,dim1,&UA) ) return 1;
	vardim=UA - LA + 1;
	
	if (NODE_OPERAND1(ddot))
	{
		if (!check_initialization_expr(NODE_OPERAND1(ddot))) return 1;
		if (!isInteger(NODE_OPERAND1(ddot))) return 1;
		comput_const_expr(NODE_OPERAND1(ddot),0);
		ST=NODE_OPERAND1(ddot)->const_val;
	}
	else
		ST=1;
	ddot=NODE_OPERAND0(ddot);
	if (ddot && NODE_OPERAND0(ddot))
	{
		if (!check_initialization_expr(NODE_OPERAND0(ddot))) return 1;
		if (!isInteger(NODE_OPERAND0(ddot))) return 1;
		comput_const_expr(NODE_OPERAND0(ddot),0);
		LT=NODE_OPERAND0(ddot)->const_val;
	}
	else
	{
		if (!get_array_lbound(target->type,dim2,&LT) ) return 1;
	}
	if (ddot && NODE_OPERAND1(ddot))
	{
		if (!check_initialization_expr(NODE_OPERAND1(ddot))) return 1;
		if (!isInteger(NODE_OPERAND1(ddot))) return 1;
		comput_const_expr(NODE_OPERAND1(ddot),0);
		UT=NODE_OPERAND1(ddot)->const_val;
	}
	else
	{
		if (!get_array_ubound(target,dim2,&UT) ) return 1;
	}
	targetdim= (2*(UT-LT+1)+ST)/(2*ST);

	if (vardim != targetdim) return 0;
	return 1;
}

