/*****************************************************************************
        Semantic check about the parser of expression
			&
		computation of constant expr

written by Zheng Gengbin
 		Peking University , P.R.China
*****************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#if defined(STDC_HEADERS) && (STDC_HEADERS==1)
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include <stdio.h>

#   include "typedef.h"
#   include "bif.h"
#   include "ll.h"
#   include "symb.h"

#include "defnodes.h"
#include "tags.h"
#include "decl.h"
#include "macro.h"

#include "exglobal.h"

extern PTR_SYMB search_defined_intrinsic_op();
PTR_SYMB is_defined_op();
extern int debug1;


/*           
		ARRAY_REF    ----------------> type
		   | (llnd)
	     ------
	     |
	    EXPR_LIST
	      |   |
	           ----------
	    		     |
			  EXPR_LIST
*/
int check_array_dim(ll_ptr)
PTR_LLND ll_ptr;
{
	int num,n;
	PTR_LLND expr_list;
	
	num=ll_ptr->type->entry.ar_decl.num_dimensions;
	expr_list=ll_ptr->entry.Template.ll_ptr1;
	n=0;
	for ( ; expr_list ;expr_list=expr_list->entry.Template.ll_ptr2)
		n++;
	if (n>num) error3(101,ll_ptr->entry.Template.symbol->ident);
}

/* only numeric computation */	
int  decide_number_kind(l_type,r_type,l_kind,r_kind)
PTR_TYPE l_type,r_type;
int      l_kind,r_kind;
{
	if (l_kind == r_kind) return l_kind;
	if (l_type->variant == T_INT && r_type->variant == T_INT )
		return  l_kind > r_kind ? l_kind:r_kind;
	if (l_type->variant == T_INT) return r_kind;
	if (r_type->variant == T_INT) return l_kind;
	return l_kind > r_kind ? l_kind:r_kind;
}

/* node must be a operator */
int check_node_constant(node)
PTR_LLND node;
{
	PTR_LLND lp,rp;

	lp=node->entry.Template.ll_ptr1;
	rp=node->entry.Template.ll_ptr2;
	if (!lp && !rp ) return 0;
	if (lp && !rp)
		if (lp->attr & PARAMETER_BIT ) 
			node->attr |= PARAMETER_BIT;
	if (lp && rp)
		if ((lp->attr & PARAMETER_BIT)  && (rp->attr & PARAMETER_BIT) ) 
			node->attr |= PARAMETER_BIT;
}

static int check_expr_type_node(ll_head)
PTR_LLND ll_head;
{
	PTR_LLND  lp ,rp;
	PTR_LLND  l_kindexp,r_kindexp;
	int       n1,n2,n,status;
	PTR_TYPE  l_type, r_type,result,newtype;
	int       l_rank=0 ,r_rank=0,result_rank;
	int       l_kind=-1 ,r_kind=-1,result_kind;
	PTR_SYMB  intrin_op,sym;
	

	switch (ll_head->variant) {
	case BINARY_VAL:
	case OCTN_VAL:
	case HEX_VAL:
		if (cur_bfnd->variant != DATA_DECL) error1(239);
	case INT_VAL:
	case FLOAT_VAL:
	case DOUBLE_VAL:
	case BOOL_VAL:
	case STRING_VAL:
	case COMPLEX_VAL:
		ll_head->attr |= PARAMETER_BIT;
	}

	lp=ll_head->entry.Template.ll_ptr1;
	rp=ll_head->entry.Template.ll_ptr2;
	
	switch (ll_head->variant) {
	case VAR_REF:
		/* intrinsic_bit var can as a argument can have no type */
		if (NODE_SYMB(ll_head)->attr & INTRINSIC_BIT)
			return;
	case KIND_VAL:
	case COMPLEX_VAL:
	case CONST_REF:
	case RECORD_REF:
	case ARRAY_REF:
	case SUB_STRING:
	case ARRAY_CONSTRUCTOR:
	case STRUCTURE_CONSTRUCTOR:
	case CONSTRUCTOR_REF:
	case FUNC_CALL:
	case INTRIN_FUNC_CALL:
		if (!ll_head->type) 
		{
			sym=get_var_name(ll_head);
			error3(107,sym->ident);
			if (sym)
				sym->type = global_int;
			ll_head->type= global_int;
		}
		return;
	}

	if (!lp && !rp) {
		if (!ll_head->type) 
		{
			sym=get_var_name(ll_head);
			error3(107,sym->ident);
			if (sym)
				sym->type = global_int;
			ll_head->type= global_int;
		}
		return;
	}

	if (lp) check_expr_type_node(lp);
	if (rp) check_expr_type_node(rp);

	l_type=lp? lp->type: TYNULL;
	r_type=rp? rp->type: TYNULL;
	

	/* for separate compilation by pass the checking */
	if (l_type && l_type==global_default || 
		r_type && r_type==global_default )
	{
		ll_head->type=global_default;
		return;
	}

/* get base type */
	if (l_type && l_type->variant == T_ARRAY)
        {
                l_rank = l_type->entry.ar_decl.num_dimensions;
                l_type = l_type->entry.ar_decl.base_type;
        }
	if (l_type && l_type->variant == T_KIND){
		l_kindexp=l_type->entry.kind_decl.kindexp; 
		l_kind=get_kind(l_type);
		l_type = l_type->entry.kind_decl.base_type;
	}

        if (r_type && r_type->variant == T_ARRAY)
        {
                r_rank = r_type->entry.ar_decl.num_dimensions;
                r_type = r_type->entry.ar_decl.base_type;
        }
	if (r_type && r_type->variant == T_KIND) {
		r_kindexp=r_type->entry.kind_decl.kindexp;
		r_kind=get_kind(r_type);
		r_type = r_type->entry.kind_decl.base_type;
	}

	if (l_type && r_type)
           if (r_rank && l_rank && r_rank !=l_rank) error1(113);

	result_rank= r_rank > l_rank ?r_rank : l_rank;

	if (is_operator(ll_head->variant) )
		check_node_constant(ll_head);

	switch (ll_head->variant) {
	case ADD_OP:
	case MINUS_OP:
		switch (ll_head->variant) {
		case ADD_OP:
			if (intrin_op=is_defined_op("+"))
			break;
		case MINUS_OP:
			if (intrin_op=is_defined_op("-"))
			break;
		}
		if (intrin_op) 
			if (check_defined_intrinsic_op(ll_head,intrin_op,
							FUNCTION_NAME)==0)
                		return;
		
		if (lp && !rp) {
			switch (l_type->variant) {
                	case T_INT:
                	case T_FLOAT:
			case T_DOUBLE:
			case T_DCOMPLEX:
                	case T_COMPLEX:
                       		result=l_type;
                        	break;
                	default:
                        	error3(103,ll_head->variant==ADD_OP?"+":"-");
                       		result=l_type;
                        	break;
                	}
			ll_head->type=lp->type;
			return;
		}	/* if */
	case MULT_OP:
	case DIV_OP:
	case EXP_OP:
		switch (ll_head->variant) {
                case MULT_OP:
                        intrin_op=is_defined_op("*");
                        break;
                case DIV_OP:
                        intrin_op=is_defined_op("/");
                        break;
                case EXP_OP:
                        intrin_op=is_defined_op("**");
                        break;
                }
                if (intrin_op)
                        if (check_defined_intrinsic_op(ll_head,intrin_op,
							FUNCTION_NAME)==0)
                        {
                        return;
                        }

		n1=isNumber(lp);
		n2=isNumber(rp);
		if (n1==0 || n2==0) {
			switch (ll_head->variant) {
			case ADD_OP :  error3(103,"+"); break;
			case MINUS_OP: error3(103,"-"); break;
			case MULT_OP:  error3(103,"*"); break;
        		case DIV_OP:   error3(103,"/"); break;
        		case EXP_OP:   error3(103,"**"); break;
			}
			result =global_int;
			result_kind=-1;
		}
		else {
			n= n1>n2? n1: n2;
			result = n==n1 ? l_type : r_type;
				/*  get out the kind of the result */
			result_kind=decide_number_kind(l_type,r_type,l_kind,r_kind);
		}
		break;
	case EQ_OP:
	case NOTEQL_OP:
		switch (ll_head->variant) {
		case EQ_OP:
			if (intrin_op=is_defined_op(".eq.")) break;
			if (intrin_op=is_defined_op("=="))  break;
			break;
		case NOTEQL_OP:
			if (intrin_op=is_defined_op(".ne.")) break;
			if (intrin_op=is_defined_op("/=")) break;
			break;
		}
		if (intrin_op) 
			if (check_defined_intrinsic_op(ll_head,intrin_op,
							FUNCTION_NAME)==0)
                		return;
		result= global_bool;
		if (isNumber(lp) && !isNumber(rp))
			error3(103,ll_head->variant==EQ_OP? ".EQ.":".NE.");
		else if(isChar(lp) && !isChar(rp))
			error3(103,ll_head->variant==EQ_OP? ".EQ.":".NE.");
		else if(isString(lp) && !isString(rp))
			error3(103,ll_head->variant==EQ_OP? ".EQ.":".NE.");
		else if (!isNumber(lp) && !isChar(lp) && !isString(lp) )
			error3(103,ll_head->variant==EQ_OP? ".EQ.":".NE.");
		else {
			result=global_bool;
		}
		result_kind = -1; 	/* relative op */
		break;
	case CONCAT_OP:
		switch (ll_head->variant) {
		case CONCAT_OP:
			if (intrin_op=is_defined_op("//")) break;
		}
		if (intrin_op) 
			if (check_defined_intrinsic_op(ll_head,intrin_op,
							FUNCTION_NAME)==0)
                		return;
		/*  string ?  no T_CHAR ?*/
		if (l_kind != r_kind) error1(193);
		result = global_string;
		result_kind = l_kind;
		break;
	case GT_OP:
	case GTEQL_OP:
	case LT_OP:
	case LTEQL_OP:
		switch (ll_head->variant) {
		case GT_OP:
			if (intrin_op=is_defined_op(".gt.")) break;
			if (intrin_op=is_defined_op(">")) break;
			break;
		case GTEQL_OP:
			if (intrin_op=is_defined_op(".ge.")) break;
			if (intrin_op=is_defined_op(">=")) break;
			break;
		case LT_OP:
			if (intrin_op=is_defined_op(".lt.")) break;
			if (intrin_op=is_defined_op("<")) break;
			break;
		case LTEQL_OP:
			if (intrin_op=is_defined_op(".le.")) break;
			if (intrin_op=is_defined_op("<=")) break;
			break;
		}
		if (intrin_op) 
			if (check_defined_intrinsic_op(ll_head,intrin_op,
							FUNCTION_NAME)==0)
                		return;
		result=global_bool;
		if ( (isInteger(lp) || isFloat(lp)) && 
			(isInteger(rp) || isFloat(rp)) )
			result=global_bool;
		else switch (ll_head->variant) {
                        case GT_OP :  error3(103,">"); break;
                        case GTEQL_OP: error3(103,">="); break;
                        case LT_OP:  error3(103,"<"); break;
                        case LTEQL_OP:   error3(103,"<="); break;
                        }
		result_kind = -1; 	/* relative op */
		break;
	case NOT_OP:
		intrin_op=is_defined_op(".not.") ;
		if (intrin_op) 
			if (check_defined_intrinsic_op(ll_head,intrin_op,
							FUNCTION_NAME)==0)
                		return;
		result=global_bool;
		if (isBool(lp))  result=global_bool;
		else error3(103,".NOT.");
		result_kind = l_kind;
		break;
	case AND_OP:
	case OR_OP:
	case EQV_OP:
	case NEQV_OP:
		switch (ll_head->variant) {
		case AND_OP:
			if (intrin_op=is_defined_op(".and.")) break;
			break;
		case OR_OP:
			if (intrin_op=is_defined_op(".or.")) break;
			break;
		case EQV_OP:
			if (intrin_op=is_defined_op(".eqv.")) break;
			break;
		case NEQV_OP:
			if (intrin_op=is_defined_op(".neqv.")) break;
			break;
		}
		if (intrin_op) 
			if (check_defined_intrinsic_op(ll_head,intrin_op,
							FUNCTION_NAME)==0)
                		return;
		result=global_bool;
		if (isBool(lp) && isBool(rp)) result=global_bool;
		else switch (ll_head->variant) {
                        case AND_OP:  error3(103,".AND."); break;
                        case OR_OP: error3(103,".OR."); break;
                        case EQV_OP:  error3(103,".EQV."); break;
                        case NEQV_OP:   error3(103,".NEQV."); break;
                        }
		result_kind = l_kind > r_kind ? l_kind:r_kind;
		break;
	case DEFINED_OP:
		check_defined_op(ll_head);
		return;
	default:
printf("here wrong llnd type:%d\n",ll_head->variant);
		error1(104);	
		return;
	}

	if (result_kind != -1) {
if (debug1) printf("Line:%d L_kind:%d,R_kind:%d,result:%d\n",cur_bfnd->g_line,l_kind,r_kind,result_kind);
		newtype=make_type(fi,T_KIND);
		newtype->entry.kind_decl.base_type=result;
		if (result_kind == l_kind)
			newtype->entry.kind_decl.kindexp=l_kindexp;
		else
			newtype->entry.kind_decl.kindexp=r_kindexp;
		result=newtype;
	}
	if (result_rank == 0) 
		ll_head->type= result;
	else{
		newtype=make_type(fi,T_ARRAY);
                newtype->entry.ar_decl.base_type= result;
                newtype->entry.ar_decl.num_dimensions = result_rank;
                ll_head->type = newtype;
	}
		
}

/*  if an expr checked , return 1 */
int is_expr_checked(expr)
PTR_LLND expr;
{
	return  (expr->attr & CHECKED_EXPR_TYPE) ? 1:0 ;
}

int set_expr_checked(expr)
PTR_LLND expr;
{
	expr->attr |= CHECKED_EXPR_TYPE;
}

/*****************************************************************************

			CHECK_EXPR_TYPE

		IN:   a expression llnd tree;
		
		RESULT: result type placed in root node of the llnd tree.

******************************************************************************/

int check_expr_type(llnd)
PTR_LLND llnd;
{
	if (!llnd) return;

		/* if have checked */
	if (!is_linking && is_expr_checked(llnd) )
		return 0;

	check_leaf_sym(llnd);
	check_expr_type_node(llnd);

	set_expr_checked(llnd);
}


/*   wrap check_expr_type saving some information */
int Check_expr_type(llnd)
PTR_LLND llnd;
{
	push_record();
	check_expr_type(llnd);
	pop_record();
}

/****************************************************************************
 *	function name :    check_defined_op                                 *
 *	input   :          a llnd tree of defined op with actual arg        *
 *       output  :          none.                                           *
 *	function:          reconstruct tree to func call , and check the    *
 *			    acutual argument                                *
*****************************************************************************/

int check_defined_op(func)
PTR_LLND func;
{
	PTR_LLND ptr;

/* reconstruct tree */
	if (func->entry.Template.ll_ptr2)	/* binary op */
		ptr=makeLlnd(fi,EXPR_LIST,func->entry.Template.ll_ptr2, 
				LLNULL,func->entry.Template.symbol);
	else 
		ptr=LLNULL;

	ptr=makeLlnd(fi,EXPR_LIST,func->entry.Template.ll_ptr1,
			ptr,func->entry.Template.symbol);
        
        call_or_func(fi,cur_bfnd,func->entry.Template.symbol,
                        ptr,FUNCTION_NAME,1,&func->type);
}


int test_initialization_bit(llnd)
PTR_LLND llnd;
{
	if ((llnd->attr & PARAMETER_BIT)==0)
	    if (g_state & NEED_INITIALIZATION_EXPR_BIT)
		g_state &= ~NEED_INITIALIZATION_EXPR_BIT;
}

int check_initialization_expr(llnd)
PTR_LLND llnd;
{
        g_state = NEED_INITIALIZATION_EXPR_BIT; 
	if (is_expr_checked(llnd))
		test_initialization_bit(llnd);
	else
        	check_expr_type(llnd); 
        if (g_state & NEED_INITIALIZATION_EXPR_BIT) return 1;  
        else return 0; 
}

int test_spec_expr_bit(llnd)
PTR_LLND llnd;
{
	PTR_SYMB sym;
	int set =0;

	if ((g_state & NEED_SPECIFICATION_EXPR_BIT) == 0) return;
	/* not consider the  spec of intrinsic call */
	if (in_intrinsic_call) return;

	if (llnd->attr & PARAMETER_BIT) return;

	switch (llnd->variant ) {
	case RECORD_REF:
	case SUB_STRING:
		llnd = llnd->entry.Template.ll_ptr1;
	}

	switch (llnd->variant ) {
	case VAR_REF:
	case ARRAY_REF:
		sym = llnd->entry.Template.symbol;
		if ((sym->attr & OPTIONAL_BIT) ==0 && (sym->attr & OUT_BIT)==0
			&& sym->entry.var_decl.local == IO )
			return;
		if (sym->entry.Template.common_link.common_item) 
			return;
		if (sym->scope != cur_scope() )
			return;
		break;
	case ARRAY_CONSTRUCTOR:
	case STRUCTURE_CONSTRUCTOR:
	case CONSTRUCTOR_REF:
		return;
	case INTRIN_FUNC_CALL:
	case FUNC_CALL:
		return;
	}
		
	g_state &= ~NEED_SPECIFICATION_EXPR_BIT;
}

/*****************************************************************************

	CHECK_SPEC_EXPR:
		check if llnd is a specifiaction expression;
		  must be scalar and integer type ;

	return  0:  yes
		>0: err_code;
******************************************************************************/

int check_spec_expr(llnd)
PTR_LLND llnd;
{
	PTR_TYPE atype;

	g_state = NEED_SPECIFICATION_EXPR_BIT;
						/* spec_err set */
        check_expr_type(llnd);
        if (g_state & NEED_SPECIFICATION_EXPR_BIT) {
		if (llnd->type->variant == T_ARRAY) return 204;
		atype=get_base_type(llnd->type);
		if (atype->variant != T_INT) return 205;
		return 0;
	}
        else return 203;
}

int test_leaf_sym(llnd)
PTR_LLND llnd;
{
	test_initialization_bit(llnd);
        test_spec_expr_bit(llnd);
}

PTR_SYMB is_defined_op(op)
char *op;
{
	return search_defined_intrinsic_op(op);
}


/*			op
		        |
		 +--------------+
		 |	        |
		var1	      var2
	return 
		0: is defined op
		>0 : not
*/
int check_defined_intrinsic_op(func,symb,kind)
PTR_LLND func;
PTR_SYMB  symb;
int	  kind;			/*  FUNCTION_NAME or  PROCEDURE_NAME */
{
        PTR_LLND ptr;
        PTR_TYPE type;
	int	 status;

/* reconstruct tree */
        if (func->entry.Template.ll_ptr2)       /* binary op */
                ptr=makeLlnd(fi,EXPR_LIST,func->entry.Template.ll_ptr2,
                                LLNULL,symb);
        else
                ptr=LLNULL;
        ptr=makeLlnd(fi,EXPR_LIST,func->entry.Template.ll_ptr1,
                        ptr,symb);
        status=call_or_func(fi,cur_bfnd,symb,ptr,kind,0,&func->type);
/*
        if(status==0) {
		func->type=type;
	}
*/
	return status;
}



/****************************************************************************

		constant exprssion checking

******************************************************************************/
/********
	ln,rn : left expr & right expr operant;
	op    : operator ;
	result: root node
******/
static int count(result,op,ln,rn)
PTR_LLND result;
int op;
PTR_LLND ln,rn;
{
	int n1,n2;
	int v1,v2;

		/*  get the priority of the type of ln & rn */
	n1=isNumber(ln);      n2=isNumber(rn);
/*
	v1=ln->type->variant; v2=rn->type->variant;
*/
	if (ln->type->variant == T_KIND) 
		 v1=ln->type->entry.kind_decl.base_type->variant;
	else 	 v1=ln->type->variant;
	if (rn->type->variant == T_KIND) 
		 v2=rn->type->entry.kind_decl.base_type->variant;
	else 	 v2=rn->type->variant;

	switch (op) {
	case ADD_OP:
		switch (result->type->variant) {
		case T_INT:
			/*result->entry.ival=ln->entry.ival+rn->entry.ival; */
			result->attr |=PARAMETER_BIT;
			result->const_val=ln->const_val+rn->const_val;
			break;
		case T_FLOAT:
			result->entry.dval=((v1==T_INT ? ln->entry.ival: 
				ln->entry.dval) +(v2==T_INT ? ln->entry.ival:
				ln->entry.dval));
			break;
		}
		break;
	case MINUS_OP:
                switch (result->type->variant) {
                case T_INT:
			result->attr |=PARAMETER_BIT;
			result->const_val=ln->const_val-rn->const_val;
                        break;
/*
                case T_FLOAT:
                        result->entry.dval=((v1==T_INT ? ln->entry.ival: 
                                ln->entry.dval) -(v2==T_INT ? ln->entry.ival: 
                                ln->entry.dval));
                        break;
*/
                }
                break; 
	case MULT_OP:
                switch (result->type->variant) {
                case T_INT:
                        /*result->entry.ival=ln->entry.ival*rn->entry.ival; */
			result->attr |=PARAMETER_BIT;
			result->const_val=ln->const_val*rn->const_val;
                        break;
/*
                case T_FLOAT:
                        result->entry.dval=((v1==T_INT ? ln->entry.ival: 
                                ln->entry.dval) *(v2==T_INT ? ln->entry.ival: 
                                ln->entry.dval));
                        break;
*/
                }
                break;
	case DIV_OP:
                switch (result->type->variant) {
                case T_INT:
			result->attr |=PARAMETER_BIT;
			result->const_val=ln->const_val/rn->const_val;
                        break;
/*
                case T_FLOAT:
                        v1=ln->type; v2=rn->type;
                        result->entry.dval=((v1==T_INT ? ln->entry.ival: 
                                ln->entry.dval) /(v2==T_INT ? ln->entry.ival: 
                                ln->entry.dval));
                        break;
*/
                }
                break;
	}
}

static int const_expr(expr)
PTR_LLND expr;
{
	PTR_LLND  lp ,rp;
	PTR_TYPE  l_type,r_type,newtype;
        int       n1,n2,n;
	int       l_kind=-1 ,r_kind=-1,result_kind;
	PTR_SYMB  sym;


        lp=expr->entry.Template.ll_ptr1;
        rp=expr->entry.Template.ll_ptr2;

					/* zheng.gb 4.29 1996 */
        if (!lp && !rp || expr->variant == KIND_VAL ||
			expr->variant == INTRIN_FUNC_CALL ) 
	{				/*leaf node must have value*/
		sym = expr->entry.Template.symbol;
		if (!isInteger(expr) ) {
			error1(105);
		}
		else {
			switch (expr->variant) {
			case VAR_REF:
				if (expr->entry.Template.symbol->attr & 
					PARAMETER_BIT) {
				  expr->attr|=PARAMETER_BIT;
				  comput_const_expr(expr->entry.Template.symbol->entry.Template.const_value,1);
				  expr->const_val=expr->entry.Template.symbol->entry.Template.const_value->const_val;
				}
				else error3(231,expr->entry.Template.symbol->ident);
				break;
			case INT_VAL:
				expr->attr |=PARAMETER_BIT;
				expr->const_val=expr->entry.ival;
				break;
			case KIND_VAL:
				expr->attr |=PARAMETER_BIT;
				expr->const_val=expr->entry.Template.ll_ptr1
								->entry.ival;
                                break;
			case INTRIN_FUNC_CALL:
				expr->attr |=PARAMETER_BIT;
                                expr->const_val= 1000;
                                break;
			case BINARY_VAL:
				expr->const_val=get_binary_val(expr);
				expr->attr |=PARAMETER_BIT;
				break;
			case OCTN_VAL:
				expr->const_val=get_octn_val(expr);
				expr->attr |=PARAMETER_BIT;
				break;
			case HEX_VAL:
				expr->const_val=get_hex_val(expr);
				expr->attr |=PARAMETER_BIT;
				break;
			default:
				error1(106);
			}
		}
		return;
	}

        if (lp) const_expr(lp);
        if (rp) const_expr(rp);

	l_type=lp?lp->type:TYNULL;
        r_type=rp?rp->type:TYNULL;

	if (l_type) l_kind=get_kind(l_type);
        if (r_type) r_kind=get_kind(r_type);

	if (l_type && l_type->variant == T_KIND)
                l_type = l_type->entry.kind_decl.base_type;

	if (r_type && r_type->variant == T_KIND)
                r_type = r_type->entry.kind_decl.base_type;


        switch (expr->variant) {
	case ADD_OP:
        case MINUS_OP:
		if (lp && !rp) {
			if ((lp->attr & PARAMETER_BIT) ==0) return;
                        if (isNumber(lp)) {
                                expr->type=lp->type;
				switch (l_type->variant) {
				case T_INT:  
					expr->attr=1;
					expr->const_val=(expr->variant==ADD_OP?
						 lp->const_val: -lp->const_val);
						break;
/*
				case T_DOUBLE:
				case T_FLOAT: expr->entry.dval=
						expr->variant==ADD_OP ? 
						lp->entry.dval:-lp->entry.dval;
						break;
*/
				}
				expr->attr |=PARAMETER_BIT;
			}
                        else error3(103,expr->variant==ADD_OP?"+":"-");
                        return;
                }
	case MULT_OP:
        case DIV_OP:
        case EXP_OP:
		if ((lp->attr&PARAMETER_BIT)==0 || (rp->attr&PARAMETER_BIT==0))
			 return;
                n1=isNumber(lp);
                n2=isNumber(rp);
                if (n1==0 || n2==0) 
			switch (expr->variant) {
                        case ADD_OP :  error3(103,"+"); break;
                        case MINUS_OP: error3(103,"-"); break;
                        case MULT_OP:  error3(103,"*"); break;
                        case DIV_OP:   error3(103,"/"); break;
                        case EXP_OP:   error3(103,"**"); break;
                        }
                else {
                        n= n1>n2? n1: n2;
                        expr->type = (n==n1 ? l_type : r_type);
			count(expr,expr->variant,lp,rp);
                }
		result_kind=decide_number_kind(l_type,r_type,l_kind,r_kind);
                break;
	}

	if (result_kind > -1) {
if (debug1) printf("%d\n",result_kind);
                newtype=make_type(fi,T_KIND);
                newtype->entry.kind_decl.base_type=expr->type;
                if (result_kind == l_kind)
                        newtype->entry.kind_decl.kindexp=
                                        lp->type->entry.kind_decl.kindexp;
                else
                        newtype->entry.kind_decl.kindexp=
                                        rp->type->entry.kind_decl.kindexp;
		expr->type = newtype;
        }

}

/*****************************************************************************

		compute the result of a initialization expression 

	IN:
		expr : a llnd expression;
		kind:  1: report error;
		       0: not;

	out:
		0:   success ;
		>0 : error code;

	the result is stored in the root node of the llnd expression tree.
*****************************************************************************/

int comput_const_expr(expr,kind)
PTR_LLND expr;
int      kind;    /*   1: report error 0: no */
{

		/* if have caculated the value ? */
	/* if (expr->attr & COUNTED_EXPR) return 0; */

	if (expr == LLNULL) {
		if (kind) error1(197);
		return 197;
	}

	if (!check_initialization_expr(expr)) {
		if (kind) error1(177);
		return 177;
	}
	else if (!is_scalar(expr->type)) {
		if (kind) error1(188);
		return 188;
	     }
	     else if (!isInteger(expr)) {
		    if (kind) error1(105);
		    return 105;
	          }
	          else const_expr(expr);
	/* expr->attr |= COUNTED_EXPR; */
	return 0;
}


