/*********************************************************************/
/*             Main Control Program Used to Semantic Checking        */ 
/*                   PACT Group, Peking University                   */ 
/*                                                                   */
/*   rewritten by Zheng Gengbin  1996.5.15                           */
/*********************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include <stdio.h>

#   include "typedef.h"
#   include "bif.h"
#   include "ll.h"
#   include "symb.h"

#include "defnodes.h"
#include "dep.h"
#include "dep_struct.h"
#include "tags.h"
#include "decl.h"

#include "check.h"
#include "global.h"

#define   DEFAULT_PROJ    "a.dep.out"

int	debug;
int     debug1=0;
int     intrinc_flag=0;

int     g_state;
int     scope_status = NON_PURE;
int 	checking_part;
int     check_seq;

SYMB_LIST   rename_list,undef_symlist;

PTR_TYPE retrieve_global_type();

PTR_PROJ  proj;
char     proj_name[100];

char    *cur_fname;

static void
usage()
{
	printf("Usage: hpfsc [-vhcdoiDV] HPF_file.dep \n");
        printf("       -v    : version          \n");
        printf("       -h    : usage          \n");
        printf("       -m    : write .mod files for module . \n");
        printf("       -c    : compile the source files, but do not link. \n");
        printf("       -d    : debugging parser \n");
        printf("       -o    : output filename. \n");
        printf("       -i    : standard library compilation \n");
        printf("       -l    : generate log files for every file \n");
        printf("       -D    : display the module dependence .\n");
        printf("       -V    : verbosely display the module compliation .\n");
}


int main(argc, argv, envp)
    int	  argc;
    char *argv[];
    char *envp[];
{
	FILE	 *fd;
	int i;
	int   ac,no;
	char **av,*cur_logname;
	PTR_BLOB new,chain;
	int    proj_Error=0;

	strcpy(proj_name,DEFAULT_PROJ);

	init_errMsg();

	/* Check if there is required .dep FILE name */
	if (argc==1)
        {
		usage();
        	exit(0);
        }
	else
	{
        argv++;
	argc--;
        while ((argc > 0) && ((*argv)[0] == '-') )
        {
	   	i=1;
	   	while (argv[0][i] != '\0') {
			switch (argv[0][i]) {
			case 'v' :
				printf("hpfsc version : %s \n",VERSION);exit(0);
			case 'h' :
				usage();  exit(0);
			case 'd' :
				debug1=1; break;
			case 'i' :
				intrinc_flag =1; break;
			case 'c' :
				cc_flag=1;   break;
			case 'm' :
				mod_flag=1; break;
			case 'l' :
				log_flag=1; break;
			case 'D' :
				show_mod_flag=1; break;
			case 'V' :
				is_verbose=1;	break;
			case 'o' :
				if (argv[0][i+1]!='\0') {
					strcpy(proj_name,argv[0]+i+1);
				   	if (cc_flag==1) cc_flag=0;
					i=strlen(argv[0]);
					continue;
				}
				else {
				   if (argc==1) {
					printf("warning 422: Unknown option \"o\" ignored.\n");
				   }
				   else {
           			   	argc--;
           			   	argv++;
				   	strcpy(proj_name,argv[0]);
				   	i=strlen(argv[0])-1;   /* quit loop */
				   	if (cc_flag==1) cc_flag=0;
				   }
				}
				break;
			default:
				printf("warning 422: Unknown option \"%c\" ignored.\n",argv[0][i]);
			}
	    		i++;
           	}   /* while */
           argc--;
           argv++;
         }		/* while */
        }    /* else */

	if (intrinc_flag==1) cc_flag=1;

		/* av now point to the first of file input list */
	ac=argc;
	av=argv;

	if ((proj = (PTR_PROJ) calloc(1, sizeof(struct proj_obj))) == NULL) {
                fprintf(stderr, "main -- No more space\n");
                exit(1);
        }

	proj->proj_name = proj_name;

	no=ac;
	while (ac>0) {
	if (no>1)
		printf("\n%s:\n",*av);           /* print out the file name */

	cur_fname=*av;
		/* if log flag is specified , then open .log file */
	if (log_flag)  {
		cur_logname = gen_log_name(cur_fname);
		log_fd = fopen(cur_logname,"w");
		if (log_fd == NULL) {
			fprintf(stderr,"Cannot open file \"%s\"\n", cur_logname);
			exit(1);
		}
		fprintf(log_fd,"====== First parse  ========\n");
	}

	if ((fd=fopen(*av, "r")) == NULL) {
		fprintf(stderr,"Cannot open file \"%s\"\n", *av);
		exit(1);
	}

	if ((fi = (PTR_FILE) calloc(1, sizeof(struct file_obj))) == NULL) {
		fprintf(stderr, "main -- No more space\n");
		exit(1);
	}

	/* The following program is try to restore  the AST & ST structure
	   from the specified .dep */

	fi->fid = fd;

	if (readNodes(fi) >= 0) {
			/* some initialization */
		fi->global_bfnd->control_parent = NULL;
			/* clear the error flag */
		globeErr_g=0;
			/* show mod dependence */
		if (show_mod_flag) {
			build_hash(fi->head_symb,hash_table);
			printf("** module dependence  **\n");
			root = (PTR_MOD)NULL;
			load_modules();
			goto jump0;
		}
			   /* if have been semantic checked ? */
		if (fi->global_bfnd->checked==SOFT) {         /* no check */
                	semantic_check(); 
			if (globeErr_g==0) 
				fi->global_bfnd->checked=HARD;   /* checked */
			else
				proj_Error++;
			writeNodes(fi, *av);
		}      /* end of SOFT */
		else {                       /* have been checked before */
        		build_hash(fi->head_symb,hash_table);
			printf("# checked #\n");
		}
			/* create mod_files */
		if (mod_flag && globeErr_g==0 )
			write_modules();
	}

jump0:	new = (struct blob *) chkalloc(sizeof(struct blob));
	new->ref=(PTR_BFND)fi;
	new->next=BLNULL;
	if (ac==no)	
		proj->file_chain = new;
	else {
		for (chain=proj->file_chain; chain->next;chain= chain->next);
		chain->next=new;
	}
		/* next input file in argument list */
	ac--;
	av++;
	}

	if (show_mod_flag) return 0;

	if (proj_Error) {
		if (cc_flag==0)
			printf("\n *** hpfsc: Errors detect , no link *** \n");
		return 0;
	}

	/* end of sematic routine */

	/* begin link session */
	link_session(proj);
	exit (0);
}

int link_session(proj)
PTR_PROJ  proj;
{
	if (cc_flag==0)  {
		/* initialize */
		is_linking=1;
		undef_symlist = (SYMB_LIST)NULL;

		printf("\n *** Begin link session . *** \n");
		if (log_flag)
			fprintf(log_fd,"====== Link parse  ========\n");
		link_dep(proj);
		link_processing(proj);
		if (globeErr_g) {
			printf("\n *** hpfsc/link failed . *** \n");
			printf(" *** End of compilation of hpfsc . *** \n");
			exit (1);
		}
		change_file_name((PTR_FILE)proj->file_chain->ref,proj->proj_name);
		printf("\nproject: %s\n",proj->proj_name);
		writeNodes((PTR_FILE)proj->file_chain->ref, proj->proj_name);
		printf("___The project file '%s' have been written ___ \n",proj->proj_name);
		printf("\n *** hpfsc/link succeed. *** \n");
		printf(" *** End of compilation of hpfsc . *** \n");
	}
	exit (0);
}

int semantic_check()
{
     PTR_BLOB program_unit_blob;
     PTR_BFND next_bfnd;

           	/* reget global type , no need when merge hpf2dep */
	reget_type(fi);
		/* load intrinsic function lib */
	init_intrinsic_function();	/* add by zheng.gb */
		/* build hash table */
	build_hash(fi->head_symb,hash_table);

		/* preprocess */
	if (!intrinc_flag) {
		load_modules();
		preprocess(); 
	}
		/* do initialization */
        initialization();    

        cur_bfnd=fi->global_bfnd;  
        program_unit_blob=cur_blob=fi->global_bfnd->entry.Template.bl_ptr1;  
	
		/* travel through the program unit of AST */
        while (program_unit_blob) {
             check_one_program_unit(program_unit_blob);   /*check one program unit*/             
             program_unit_blob=program_unit_blob->next; /*next program unit*/
	}
        
        /*this procedure is used to deal with the check which should be done after scan all the AST*/
        end_processing();       
}

/* travel through the procedure units in the contain block */
int check_contains(contains_blob)
PTR_BLOB  contains_blob;
{	
        contains_blob = contains_blob->next; /*point to first procedure in conatins*/
        while (contains_blob) {
		 /*check one procedure*/
		if (is_main_header(contains_blob->ref))
             		check_one_contain_procedure(contains_blob);
             	contains_blob=contains_blob->next; /*next program unit*/
	}
}

int reget_type(fi)
PTR_FILE fi;
{
/*    recover global type , remove after we join with hpfc */
   global_int      = retrieve_global_type(fi,T_INT);
   global_float    = retrieve_global_type(fi,T_FLOAT);
   global_double   = retrieve_global_type(fi,T_DOUBLE);
   global_char     = retrieve_global_type(fi,T_CHAR);
   global_string   = retrieve_global_type(fi,T_STRING);
   global_bool     = retrieve_global_type(fi,T_BOOL);
   global_complex  = retrieve_global_type(fi,T_COMPLEX);
   global_dcomplex = retrieve_global_type(fi,T_DCOMPLEX);
   global_template = retrieve_global_type(fi,T_TEMPLATE);
   global_processors = retrieve_global_type(fi,T_PROCESSORS);
   global_default  = retrieve_global_type(fi,DEFAULT);
}

/* initialization before we do semantic checking */
int initialization()
{
	PTR_SYMB sym;
	PTR_LABEL label;

if (debug1) printf("***Initialization ! ***\n");
	init_scope_table();
	top_implicit_level=-1;
					/* scope_table start from 1 */
	add_scope_level(fi->global_bfnd,0);
					/* init record table */
	init_record_table();

	init_sub();/*initiate function or saubroutine name*/
 	isrecursive_modu(); /*added by Yanmi, module is recursive?*/

	globeErr_g=0;
	default_access=0;
	in_module = 0; /* added by Feng Xiaobing */
	in_block_data = 0; /* added by Feng Xiaobing */
	empty_access_stmt = 0; /* added by Feng Xiaobing */
	in_type_define = 0; /* added by Feng Xiaobing */
	in_interface_body = 0;
	common_table = CNTNULL; /* added by Feng Xiaobing */
	rename_list = (SYMB_LIST)NULL;   
	root = (PTR_MOD)NULL;

			/* reset decl */
	for (sym = fi->head_symb->thread ; sym ; sym=sym->thread) 
	{
		if (sym->scope->checked == VIRTUAL_BFND) break;
		sym->decl = SOFT;
	}
			/* reset label isused */
	for (label = fi->head_lab; label; label=label->next )
	{
		label->labused=0;
	}
}

int check_one_program_unit(unit_blob)           /* program header */
PTR_BLOB unit_blob;
{
	PTR_BLOB cur_blob,contains_blob;

	checking_part=1; /*if checking specification part,checking_part = 1;
			  if checking execution part,checking_part = 2;
			  if checking contain part,checking_part =3;*/

	cur_bfnd=unit_blob->ref;
	check_header(cur_bfnd);
	check_seq = 0xFF ;
        check_specification_stmt(cur_bfnd);  

	cur_blob = cur_bfnd->entry.Template.bl_ptr1;
	contains_blob = get_contain_blob(cur_bfnd);
	
	cur_bfnd = cur_blob->ref;

        while (cur_blob->next && is_specification_stmt(cur_bfnd)) 
	{
            	check_specification_stmt(cur_bfnd);  
		cur_blob = cur_blob->next;
		cur_bfnd = cur_blob->ref;
	}

	if (debug1) printf("*********************%d**********************\n",checking_part);
	if (checking_part == 1) {    /* at intersection of specificatin part*/
		check_at_end_of_spec();			/*and executin  part*/
                checking_part = 2;
        }

	while (cur_blob->next) {
		if (!execution_part()) break;
		cur_blob = cur_blob->next;
		cur_bfnd = cur_blob->ref;
	}

        if (contains_blob) {
		checking_part=3;
    		check_contains(contains_blob);
         }

	for ( ; cur_blob->next ; cur_blob=cur_blob->next);
	cur_bfnd = cur_blob->ref;   /* cur_bfnd point to the last CONTROL_END */

	check_end();
}

int check_one_contain_procedure(unit_blob)
PTR_BLOB unit_blob;                        /* procedure header */
{
	PTR_BLOB  cur_blob,contains_blob;
	int       part;

	part=1; /*if checking specification part,checking_part = 1;
			  if checking execution part,checking_part = 2;
			  if checking contain part,checking_part =3;*/

    
	cur_bfnd=unit_blob->ref;
	check_header(cur_bfnd);
	check_seq = 0xFF ;
	cur_bfnd=unit_blob->ref;
        check_specification_stmt(cur_bfnd);  

	cur_blob = cur_bfnd->entry.Template.bl_ptr1;
	contains_blob = get_contain_blob(cur_bfnd);
	
	cur_bfnd = cur_blob->ref;

        while (cur_blob->next && is_specification_stmt(cur_bfnd)) 
	{
            	check_specification_stmt(cur_bfnd);  
		cur_blob = cur_blob->next;
		cur_bfnd = cur_blob->ref;
	}

	if (part == 1){     /* at intersection of specificatin part*/
                check_at_end_of_spec();			/*and executin  part*/
                part = 2;
        }

	while (cur_blob->next) {
		if (!execution_part()) break;
		cur_blob = cur_blob->next;
		cur_bfnd = cur_blob->ref;
	}

        if (contains_blob) {
		part=3;
    		check_contains(contains_blob);
         }

	for ( ; cur_blob->next ; cur_blob=cur_blob->next);
	cur_bfnd = cur_blob->ref;
	check_end();
}

int check_header(cur_bfnd)
PTR_BFND cur_bfnd;
{
	PTR_LLND interface_list;
	SYMB_LIST tmp;
	int i;

	push_scope(fi,cur_bfnd);

	if (top_implicit_level== -1)
	{
	top_implicit_level++;	
	for (i=0;i<27;i++) 
		if (i>= 'i'-'a' && i<= 'n'-'a')
			implicit_table[i][top_implicit_level]=global_int;
		else 
			implicit_table[i][top_implicit_level]=global_float;
	}
	else {
	   top_implicit_level++;
	   for (i=0;i<27;i++)
		implicit_table[i][top_implicit_level]=
				implicit_table[i][top_implicit_level-1];
	}

	implicit_flag = 0; /* added by Feng Xiaobing */
	save_flag = 0; /* added by Feng Xiaobing */
	empty_save_flag = 0; /* added by Feng Xiaobing */	

	if (debug1) show_cur_scope();

		/* free the rename_list */
	while (rename_list) {
		tmp=rename_list;
		rename_list=rename_list->next;
		free(tmp);
	}
	rename_list= (SYMB_LIST) NULL;
}

int is_specification_stmt(cur_bfnd)
PTR_BFND cur_bfnd;
{
	check_stmt_function();			/* check if stmt_function */

	switch (cur_bfnd->variant ){
	case  CONTROL_END:
	case  MODULE_STMT:
	case  PROG_HEDR:
	case  FUNC_HEDR:
	case  PROC_HEDR:
	case  BLOCK_DATA:

	case  PARAM_DECL:
	case  STRUCT_DECL:
	case  VAR_DECL :
	case  SAVE_DECL :  
	case  COMM_STAT :
	case  USE_STMT:
	case  DIM_STAT:
	case  DIRSEQUENCE_STMT :
	case  NOSEQUENCE_STMT :
	case  IMPL_DECL :
	case  EQUI_STAT :
	case  INTENT_STMT :
	case  OPTIONAL_STMT :
	case  ACCESS_STMT :
	case  ALLOCATABLE_STMT :
	case  POINTER_STMT :
	case  TARGET_STMT :
	case  DATA_DECL :
	case  PRIVATE_STMT :
	case  SEQUENCE_STMT :
	case  NAMELIST_STAT :	
	/* by Yanmi*/
	case    INTERFACE_STMT:
	case	EXTERN_STAT:
	case	INTRIN_STAT:
	case 	ENTRY_STAT:
        case    FORMAT_STAT:
        case    MODULE_PROC_STMT:
	case    STMTFN_STAT:
	/* by WZL-LY, 03-12 */
        case    DISTRIBUTE_DECL:
        case    ALIGN_DECL:
        case    COMBINED_DIRECTIVE_DECL:
	case    RANGE_STMT:
	case    SHADOW_STMT:
        case    PROCESSORS_STMT:
        case    TEMPLATE_STMT:
        case    INHERIT_DECL:
        case    DYNAMIC_DECL:
			return 1;
	}
	return 0;
}

int check_specification_stmt(cur_bfnd)
PTR_BFND cur_bfnd;
{
	illegal_stmt_label();		/* illegal llabel if there is */

	switch (cur_bfnd->variant ){
	case  CONTROL_END:
			control_end();
			break;
/* added by Yanmi*/
	case   USE_STMT:	
			if ((check_seq & HAS_USE)==0) error1(12);
			use_stmt();
			break;
	case	INTERFACE_STMT:
					/* add by zheng.gb  4.29 1996 */
			check_seq = 0xFF ;
			interface_stmt();
			interface_block();         /* by zheng.gb 1996.5.13*/
			break;
	case	MODULE_PROC_STMT:
			module_proc_stmt();
                       break;
	case	EXTERN_STAT:
			external_stmt();
			break;
	case	INTRIN_STAT:
			intrinsic_stmt();
			break;
	case    PROG_HEDR:
			program_stmt();
			break;
	case  MODULE_STMT:
			module_stmt();
			break;
	case	FUNC_HEDR:
			check_seq = 0xFF ;
			function_stmt();
			break;
	case	PROC_HEDR:
			check_seq = 0xFF ;
			subroutine_stmt();
			break;
	case	ENTRY_STAT:
			entry_stmt();
				/* need not to check the sequence of stmt */
			return;
        case    FORMAT_STAT:
				/* need not to check the sequence of stmt */
			return;
	case    STMTFN_STAT:
			stmt_function_stmt();
			break;
/* the following statements added by Feng Xiaobing */
	case   	VAR_DECL :
			check_var_decl_stmt();
			break;
	case 	DIM_STAT :
			check_dim_stmt();
			break;
	case 	COMM_STAT :
			check_comm_stmt();
			break;
	case 	SAVE_DECL : 
			check_save_decl_stmt();
			break;
	case 	SEQUENCE_STMT :
			check_sequence_stmt();
			break;
	case 	DIRSEQUENCE_STMT :
			check_dirsequence_stmt();
			break;
	case	NOSEQUENCE_STMT :
			check_nosequence_stmt();
			break;
	case	PARAM_DECL :
			check_param_decl();
			break;
	case	IMPL_DECL : 
			if ((check_seq & HAS_IMPLICIT)==0) error1(13);
			check_impl_decl();
			if (implicit_table[26][top_implicit_level]==NULL &&
				(check_seq & HAS_IMPLICIT_NONE)==0 )
				error1(13);
			break;
	case	EQUI_STAT :
			check_equi_stmt();
			break;
	case	INTENT_STMT :	
			check_intent_stmt();
			break;
	case	OPTIONAL_STMT :	
			check_optional_stmt();
			break;
	case	ACCESS_STMT :	
			check_access_stmt();
			break;
	case	ALLOCATABLE_STMT :
			check_allocatable_stmt();
			break;
	case	POINTER_STMT :	
			check_pointer_stmt();
			break;
	case	TARGET_STMT :
			check_target_stmt();
			break;
	case 	DATA_DECL :
			check_data_decl();
			break;
	case 	STRUCT_DECL :
			check_struct_decl();
			strcut_decl_block();           /* zheng.gb */
			break;
	case	PRIVATE_STMT :
			check_private_stmt();
			break;
	case	NAMELIST_STAT :	
			check_namelist_stat();
			break;
	case	BLOCK_DATA : 
			check_block_data();
			break;
	/* by WZL-LY, 03-12 */
        case    DISTRIBUTE_DECL:
                        distribute_decl_smtc(cur_bfnd);
                        break;
        case    ALIGN_DECL:
                        align_decl_smtc(cur_bfnd);
                        break;
        case    RANGE_STMT:
                        range_decl_smtc(cur_bfnd);
                        break;
        case    SHADOW_STMT:
                        shadow_decl_smtc(cur_bfnd);
                        break;
        case    COMBINED_DIRECTIVE_DECL:
                        combined_directive_decl_smtc(cur_bfnd);
                        break;
        case    INHERIT_DECL:
                        inherit_decl_smtc(cur_bfnd);
                        break;
        case    DYNAMIC_DECL:
                        dynamic_decl_smtc(cur_bfnd);
                        break;
	case    PROCESSORS_STMT:
			processors_decl_stmtc(cur_bfnd);
			break;
        case    TEMPLATE_STMT:
			template_stmt(cur_bfnd);
			break;
	}

	if (is_main_header(cur_bfnd)) return;

	/* USE 
	   IMPLICIT NONE
	   PARAMETER         IMPLICIT
        */

	if ((check_seq & HAS_USE)  && cur_bfnd->variant != USE_STMT)
		check_seq &= ~HAS_USE;
	if ((check_seq & HAS_IMPLICIT_NONE) && (check_seq & HAS_USE) ==0 ) 
	{
		if (cur_bfnd->variant!=IMPL_DECL)
			check_seq &= ~HAS_IMPLICIT_NONE;
		else 
			if (implicit_table[26][top_implicit_level]!=NULL)
				check_seq &= ~HAS_IMPLICIT_NONE;
	}
	if (check_seq & HAS_IMPLICIT)
		if ((check_seq & 3 )== 0)
			if (cur_bfnd->variant!=IMPL_DECL &&
				cur_bfnd->variant!=PARAM_DECL )
			  check_seq &= ~HAS_IMPLICIT;

}


int execution_part()
{
        if (executable_construct()) return 1;   /* 1 means true a executable */
        if (execution_part_construct()) return 1;
	if (cur_bfnd->variant == CONTAINS_STMT ||
		cur_bfnd->variant == CONTROL_END ) return 0;
        error2(10,cur_bfnd->variant);
        cur_bfnd=cur_bfnd->thread;
	return 0;
}

int executable_construct()
{
        if ( action_stmt() )  return 1;

        switch (cur_bfnd->variant)
        {
        case LOGIF_NODE:
        	stmt_label();           /* check statement label */
                if_construct();
                break;
	case SWITCH_NODE:
        	stmt_label();           /* check statement label */
		case_stmt();
        	cur_bfnd=cur_bfnd->thread;
		break;
	case FOR_NODE:
        	stmt_label();           /* check statement label */
		do_stmt();
        	cur_bfnd=cur_bfnd->thread;
		break;
	case DO_WHILE_NODE:
        	stmt_label();           /* check statement label */
		do_while_stmt();
        	cur_bfnd=cur_bfnd->thread;
		break;
	case FORALL_CONSTRUCT_STMT:
        	stmt_label();           /* check statement label */
		forall_construct_stmt();
        	cur_bfnd=cur_bfnd->thread;
		break;
	case WHERE_BLOCK_STMT:
        	stmt_label();           /* check statement label */
		where_block_stmt();
        	cur_bfnd=cur_bfnd->thread;
		break;
        default:
                return 0;
        }

        return 1;
}

int action_stmt()
{
	PTR_SYMB sub_name;

        switch (cur_bfnd->variant)
        {
        case ASSIGN_STAT:
        		stmt_label();           /* check statement label */
                        check_assign_stmt();
                        break;
        case  POINTER_ASSIGN_STAT:
        		stmt_label();           /* check statement label */
                        check_pointer_assign_stmt();
                        break;
	case ALLOCATE_STMT:
        		stmt_label();           /* check statement label */
			allocate_stmt();
			break;
	case DEALLOCATE_STMT:
        		stmt_label();           /* check statement label */
			deallocate_stmt();
			break;
	case NULLIFY_STMT:
        		stmt_label();           /* check statement label */
			nullify_stmt();
			break;
	case WHERE_NODE:
        		stmt_label();           /* check statement label */
			where_stmt();
			break;
        case LOGIF_STMT:
        		stmt_label();           /* check statement label */
                	if_stmt();
                	break;
        case FORALL_STMT:
        		stmt_label();           /* check statement label */
                	forall_stmt();
                	break;
        case GOTO_NODE:
        		stmt_label();           /* check statement label */
                        goto_stmt();
                        break;
	case COMGOTO_NODE:
        		stmt_label();           /* check statement label */
			comput_goto_stmt();
			break;
	case ASSLAB_STAT:
        		stmt_label();           /* check statement label */
			assign_to_stmt();
			break;
	case ASSGOTO_NODE:
        		stmt_label();           /* check statement label */
			assign_goto_stmt();
			break;
	case ARITHIF_NODE:
        		stmt_label();           /* check statement label */
			arithmetic_if_stmt();
	case CYCLE_STMT:
	case EXIT_STMT:
        		stmt_label();           /* check statement label */
			exit_stmt();
			break;
	case STOP_STAT:
        		stmt_label();           /* check statement label */
			stop_stmt();
			break;
	case PAUSE_NODE:
        		stmt_label();           /* check statement label */
			pause_stmt();
			break;
	case CONTINUE_NODE:
        		stmt_label();           /* check statement label */
			break;
	case RETURN_STAT:
        		stmt_label();           /* check statement label */
                        return_stmt();
                        break;
        case PROC_STAT:
        		stmt_label();           /* check statement label */
			sub_name = cur_bfnd->entry.call.call_exp->
						entry.Template.symbol;
			if (is_intrinsic_call(sub_name)!=0)
				check_intrinsic_func(cur_bfnd->entry.call.call_exp,PROCEDURE_NAME);
			else {
				/* do not want to do some extra check */
				g_state=IS_VARIABLE;
                        	call_or_func(fi,cur_bfnd,sub_name, cur_bfnd->entry.call.call_exp->entry.Template.ll_ptr1,PROCEDURE_NAME,1,TYNULL);
				g_state=0;
				/* pointer to the right symbol entry */
				if (sub_name->variant != INTERFACE_NAME && cur_subname != sub_name)
					cur_bfnd->entry.call.call_exp->entry.Template.symbol=cur_subname;

				/* record the matched function in interface */
/*printf("sub_name:%d,cursub:%d.\n ",sub_name->id,cur_subname->id);*/
				if (sub_name->variant == INTERFACE_NAME && sub_name!=cur_subname) {
					PTR_LLND matched;
					matched=makeLlnd(fi,VAR_REF,LLNULL,LLNULL,cur_subname);
					cur_bfnd->entry.call.call_exp->entry.Template.ll_ptr2=matched;
				}
			}
                        break;
	case  OPEN_STAT:
			open_close();
			break;
	case  READ_STAT:	
			read_stmt();
			break;
	case WRITE_STAT:
			write_stmt();
			break;
	case PRINT_STAT:	
			print_stmt();
			break;
	case CLOSE_STAT:	
			open_close();
			break;
	case INQUIRE_STAT:
			inquire_stmt();
			break;
	case BACKSPACE_STAT:
	case REWIND_STAT:
	case ENDFILE_STAT:
			backspace_endfile_rewind_stmt();
			break;		

	/* by WZL-LY, 03-12 */
        case DISTRIBUTE_STMT:
        		stmt_label();           /* check statement label */
                        distribute_decl_smtc(cur_bfnd);
                        break;
	case REDISTRIBUTE_STMT:
	case CBREDISTRIBUTE_STMT:
        		stmt_label();           /* check statement label */
                        redistribute_stmt_smtc(cur_bfnd);
                        break;
        case ALIGN_STMT:
        		stmt_label();           /* check statement label */
                        align_decl_smtc(cur_bfnd);
                        break;
	case REALIGN_STMT:
	case CBREALIGN_STMT:
        		stmt_label();           /* check statement label */
                        realign_stmt_smtc(cur_bfnd);
                        break;
	case INDEPENDENT_STMT:
        		stmt_label();           /* check statement label */
			independent_stmt_smtc(cur_bfnd);
			break;
        default:
                        return 0;
        }

        cur_bfnd=cur_bfnd->thread;

        return 1;
}

int execution_part_construct()          /* execution_part_construct */
{
        if ( executable_construct() ) return 1;

        switch (cur_bfnd->variant)
        {
          case ENTRY_STAT:
		illegal_stmt_label();		/* illegal llabel if there is */
                break;
          case FORMAT_STAT:
        	stmt_label();           /* check statement label */
                break;
	  case DATA_DECL:
		illegal_stmt_label();
		check_data_decl();
		break;
          default:
                return 0;
        }
        cur_bfnd=cur_bfnd->thread;
        return 1;
}

int block()
{
        return execution_part_construct();
}

int check_end()
{
	check_control_end();

	check_scope_labels();

	top_implicit_level--;
	delete_beyond_scope_level(0);

	if (debug1) show_cur_scope();
}

int end_processing()
{
	isrecursive_sub(fi);
	check_interface();
	
}

/*
 *   link_processing() ------- link time global checking
 */
int link_processing(proj)
PTR_PROJ	proj;
{	

	fi=(PTR_FILE)proj->file_chain->ref;

	is_dup_globle(proj);
	is_checked_bfnd(proj);
	isrecursive_sub(fi);
	check_interface();
	report_undef_symbol();           /* report undef symbol */
}

PTR_TYPE retrieve_global_type(fi,var)
PTR_FILE fi;
int var;
{
	PTR_TYPE  tp,kind;

	for (tp=fi->head_type; tp; tp=tp->thread)
		if (tp->variant == var)
			return tp;
}


show_cur_scope()
{
	int i;
/*	printf("Scope level:%d,line:%d\n",top_scope_level,cur_bfnd->g_line);*/
        for (i=1; i<= top_scope_level ; i++)
	{
		printf("%d:",i);
		if (scope_table[i]->entry.Template.symbol)
			printf("%s ",scope_table[i]->entry.Template.symbol->ident);
		else
			printf("-- ");
	}
	printf(" **** %d **** \n",cur_bfnd->g_line);
}

check_at_end_of_spec()
{
	common_equi_process();
	check_dirsequence_attribute();
	check_common_unity();   
}

