/*******************************************************************************
    Semantic check for HPF2.0 independent directive
    written by Lan Sai, 1997.3.12 

Subroutines defined:
.independent_stmt_smtc 
	.simple_independent_stmt
	.independent_with_NEW
		.basic_constraint_check
			.basic_symb_check
	.independent_with_REDUCTION		
		.reduction_var_check
			.intrinsic_type_check
		.is_same_variable
		.appear_in_scope_check
			.appear_again_check		
		.reduction_stmt_in_do_check	
			.reduction_stmt_of_var_check
				.reduction_stmt_check
			.red_var_appear 	
			.reduction_op_check	
				.is_lexically_same
			.reduction_func_check	

*******************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include <stdio.h>

#include "tags.h"
#include "typedef.h"
#include "bif.h"
#include "ll.h"
#include "symb.h" 

#include "defnodes.h"
#include "macro.h"
#include "decl.h"
#include "exglobal.h"

#define HAS_NEW 1
#define HAS_REDUCTION 2
int category;		/*indicate this INDEPENDENT directive's category of usage
			  it falls in the three: INDEPENDENT only, HAS_NEW, HAS_REDUCTION
			*/

int op=0,op1=0;		/* top level operator in reduction statement
	   		   1 for '+' & '-', 2 for '*' and '/', 
			   3,4,5,6,7... for others separately 
			*/
 
int independent_stmt_smtc(cur_bfnd) 
PTR_BFND cur_bfnd;
{
 PTR_BFND this_bfnd;
 PTR_LLND llnd1,llnd2;


	category = 0;
        this_bfnd = cur_bfnd;

	llnd1 = this_bfnd->entry.Template.ll_ptr1;
	llnd2 = this_bfnd->entry.Template.ll_ptr2;
      	if(llnd1 != LLNULL)  category |= HAS_NEW; 		/* NEW is present */	
	if(llnd2 != LLNULL)  category |= HAS_REDUCTION;		/* REDUCTION is present */


	/* The following part deals with all the three situations */ 
	if(category == 0) 
		simple_independent_stmt(this_bfnd);
	else {
		if(category&HAS_NEW) independent_with_NEW(this_bfnd); 
		if(category&HAS_REDUCTION) independent_with_REDUCTION(this_bfnd); 
	     }	


	return 0;		

}


/*
//In the simple case, there are two things to check:
//1. the statement following the INDEPENDENT directive should be one of 
//	do_stmt, forall_stmt and forall_construct
//2. if the statement follows is a do_stmt, then there should be a do_var in it
//result: 0-----OK, 1-----error detected 
*/
int simple_independent_stmt(this_bfnd)
PTR_BFND this_bfnd;
{
 PTR_SYMB do_var;
 int is_do;

	is_do=0;		/* false */
	this_bfnd = this_bfnd->thread;
		
	switch(this_bfnd->variant) {
	case FOR_NODE:
		is_do=1;	/* true */
		break;
	case FORALL_STMT:
	case FORALL_CONSTRUCT_STMT:
		is_do=0;	/* false */
		break;
	default:
		error1(800);
	}/* end of switch */

	
	if(is_do) {
		do_var=this_bfnd->entry.do_stmt.control_var;
		if(do_var == SMNULL) { 
			error1(801);
	 	}
	}/* end of if(is_do) */


	return 0;

}


/*
//When NEW directive is present, we have to ensure two things:
//1. the statement immediately following INDEPENDENT must be a do_stmt, with do-variable
//2. the variable named in NEW clause should meet several basic constraints  
//	e.g. it can't be 
//		dummy argument,
//		have the SAVE or TARGET attribute
//		in COMMON or storage associated with objects in COMMON
//		use & host associated, or accessed via host association	
*/
int independent_with_NEW(this_bfnd)
PTR_BFND this_bfnd;
{
 PTR_BFND the_bfnd;
 PTR_SYMB do_var;
 PTR_LLND llnd;

	the_bfnd = this_bfnd->thread;
	
	/* check 1. */
	if(the_bfnd->variant != FOR_NODE) { 
		error1(802);
	}
	else {
                do_var=the_bfnd->entry.do_stmt.control_var;
                if(do_var == SMNULL) {
       			error1(801);
         }                  		
	}/* end of check 1. */

	
	/* check 2. */

	/* variable list in NEW clause */
	llnd = this_bfnd->entry.Template.ll_ptr1;  		

	/* check 2 on variables in NEW clause one by one */
	for( ; llnd; llnd=llnd->entry.list_s.next) {
		basic_constraint_check(llnd->entry.list_s.item);	
	}

}


/*
//There are three kinds of valid llnd variant here: VAR_REF,ARRAY_REF,RECORD_REF.
//We will deal with all the three situations separately.
*/
int basic_constraint_check(var)
PTR_LLND var;
{
 PTR_BFND scope;
 PTR_SYMB symb,part_symb;
 PTR_LLND part;

	check_leaf_sym(var);	/* to ensure the legal ref of variable */


	/* check to see whether the var meet the basic constraints in NEW or REDUCTION clause */
	
	scope = cur_scope();
	symb = get_var_name(var);
	
	/* check to see whether symb is a dummy argument */
	if(is_dummy(scope,symb)) error3(804,symb->ident); 	

	/* other basic check */
	switch(var->variant) {
	case VAR_REF:	
	case ARRAY_REF:
		basic_symb_check(symb);
		break;

	case RECORD_REF:	/* 
				// we have to check the struct name for 
				// all the basic constraints;
				// and the field name for only the SAVE or TARGET attr
				*/
		basic_symb_check(symb);

		part = var->entry.record_ref_s.field;
		for( ; part; part=part->entry.record_ref_s.field) {
			part_symb = part->entry.Template.ll_ptr1->entry.Template.symbol;
			/* attribute check */
        		if( (part_symb->attr & SAVE_BIT) || (part_symb->attr & TARGET_BIT) )
                		error3(806,part_symb->ident);
		}
		break;
	
	case CONST_REF:
		error3(805,symb->ident);
		break;

	default:
		error1(803);
		return 1;
	}

	return 0;

}


/*
// Perform the check common to VAR_REF,ARRAY_REF and RECORD_REF 
*/
int basic_symb_check(symb)
PTR_SYMB symb;
{
 PTR_SYMB eq_symb,ultimate_symb; 	
	 
	/* attribute check */
        if( (symb->attr & SAVE_BIT) || (symb->attr & TARGET_BIT) )
                error3(806,symb->ident);


	/*
        // storage association check
	// offset != 0 means that this symb is in COMMON 
	*/
	eq_symb=symb->entry.Template.eq_link.next; 	/* eq_symb: symbols in the eq_link */

  	if(symb->entry.Template.common_link.common_item != CNTNULL) error3(807,symb->ident);   
        else {
                for(; eq_symb && (symb != eq_symb); eq_symb=eq_symb->entry.Template.eq_link.next) {   
			/* check to see if this symb is storage associated with objects in COMMON */
                        if(eq_symb->entry.Template.common_link.common_item != CNTNULL) error3(807,symb->ident);   
                }
        }


	/*
        //use & host association check
        //method: check the scope of symb, if it's not cur_scope(), then error
        //      using the field---declared_name--- to find the ultimate variable refered to
        */
	if(symb->entry.Template.declared_name != SMNULL) {     /* this symb is use associated with other symbol */ 
                ultimate_symb = symb->entry.Template.declared_name;
                while(ultimate_symb->entry.Template.declared_name != SMNULL) {
                        ultimate_symb = ultimate_symb->entry.Template.declared_name;
                }
                if(ultimate_symb->scope != cur_scope()) error3(808,ultimate_symb->ident); 
        } 
	else {
                if(symb->scope != cur_scope()) error3(808,symb->ident);
        }

	return 0;
}

 

/*
//When REDUCTION is present, we have to ensure three things: 
//1. the statement immediately following INDEPENDENT must be a do_stmt, with do-variable
//2. the reduction-variable in REDUCTION clause should meet several basic constraints  
//	2.1. a reduction-variable 	is 	array-variable-name
//					or	scalar-variable-name
//					or 	struct-component without subscript-section-list
//		and: the variale must be of intrinsic type other than CHARACTER
//	2.2. all the same requirements for variables in NEW clause.
//	2.3. the reduction-variable should not appear in a NEW clause in the same INDEPENDENT directive,
//		nor can it appear in any NEW or REDUCTION clause in the scope of the do_stmt follows 
//3. the reduction-variable can only appear in a reduction-stmt in the do loop
//      the syntax for a reduction-stmt is defined on page 57 
*/
int independent_with_REDUCTION(this_bfnd)
PTR_BFND this_bfnd;
{
 PTR_BFND the_bfnd,in_bfnd;
 PTR_SYMB do_var;
 PTR_LLND this_llnd,comp_llnd,this_var,comp_var;
 PTR_BLOB blob;

	the_bfnd = this_bfnd->thread;		/* FOR_NODE */
	
	/* check 1. */
	if(the_bfnd->variant != FOR_NODE) { 
		error1(802);
	}
	else {
                do_var = the_bfnd->entry.do_stmt.control_var;
                if(do_var == SMNULL) {
       			error1(801);
         	}                  		
	}/* end of check 1. */

	
	/* check 2. */

	/* variable list in REDUCTION clause */
	this_llnd = this_bfnd->entry.Template.ll_ptr2;  

	if(category&HAS_NEW) 
		comp_llnd = this_bfnd->entry.Template.ll_ptr1; 	/* variable list in NEW clause */ 
	else
		comp_llnd = LLNULL;

	/* check 2 on variables in REDUCTION clause one by one */
	for( ; this_llnd; this_llnd=this_llnd->entry.list_s.next) {
		
		/* check 2.1 */
		reduction_var_check(this_llnd->entry.list_s.item);

		/* check 2.2 */
		basic_constraint_check(this_llnd->entry.list_s.item);

		/* check 2.3 */
		this_var = this_llnd->entry.list_s.item;
		if(comp_llnd != LLNULL) {	/* compare with vars in NEW clause in the same INDEPENDENT */
			for( ; comp_llnd; comp_llnd=comp_llnd->entry.list_s.next){
				comp_var = comp_llnd->entry.list_s.item;
				if(is_same_variable(this_var,comp_var,1)) break; 
			}/* end of for */
		}/* end if */

		/* look into the do_stmt that follows to check duplicate occurance */
		/* the efficiency can be improved if taken out of for */
		appear_in_scope_check(this_var,the_bfnd);

	}/* end for*/

	/* check 3 */
	reduction_stmt_in_do_check(this_bfnd);
}

/*
//make sure the red_var is a valid reduction-variable
//i.e. it satisfies the above 2.1 constraint
*/
int reduction_var_check(red_var)
PTR_LLND red_var;
{
 PTR_SYMB symb;  
 PTR_TYPE type;
 PTR_LLND llnd,llnd1,var_list;


	switch(red_var->variant) {
	case VAR_REF:
		symb = red_var->entry.Template.symbol;		/* equal to get_var_name(red_var) */
		type = get_base_type(symb->type);
		intrinsic_type_check(type,symb);
		break;

	case ARRAY_REF:
		symb = red_var->entry.Template.symbol;		/* equal to get_var_name(red_var) */
		type = get_base_type(symb->type);
		intrinsic_type_check(type,symb);
		if(red_var->entry.array_ref_s.subscripts != LLNULL)
			error3(824,red_var->entry.array_ref_s.symbol->ident);
		break;

	case RECORD_REF:
		for(llnd=red_var; llnd; llnd=llnd->entry.record_ref_s.field) {
			llnd1 = llnd->entry.record_ref_s.rec;
		}
		symb = llnd1->entry.Template.symbol;
		type = get_base_type(symb->type);
		intrinsic_type_check(type,symb);

		/* check subscript-section-list */
		for(var_list=red_var->entry.Template.ll_ptr2; var_list; 
			var_list=var_list->entry.Template.ll_ptr2) {

			llnd = var_list->entry.Template.ll_ptr1;
			if(llnd->variant == ARRAY_REF) {
				if(llnd->entry.array_ref_s.subscripts != LLNULL)
					error3(811,llnd->entry.array_ref_s.symbol->ident);
			}
		}/* end for */
		break;

	default:	/* not necessary? */
		symb = get_var_name(red_var);
		error3(812,symb->ident);
	}

	return 0;
}

/* return: 1--intrinsic type other than CHAR; 0--CHAR or not intrinsic type */
int intrinsic_type_check(red_type,red_symb)
PTR_TYPE red_type;
PTR_SYMB red_symb;
{

	switch(red_type->variant) {
	case T_INT:
	case T_FLOAT:
	case T_DOUBLE:
	case T_COMPLEX:
	case T_DCOMPLEX: 
	case T_BOOL:
		return 1;
	case T_CHAR:
	case T_STRING:
		error3(809,red_symb->ident);
		break;
	default:
		error3(810,red_symb->ident);
	}

	return 0;
}



/*
//check to see whether two reduction-variables are the same,
//i.e.,if they are VAR_REF or ARRAY_REF,then their symbol is the same;
//     if they are RECORD_REF,then their part-ref should be the same in corresponding positions.
//
//input:  var1,var2--the two variables to compare, 
//	 place--in the NEW clause in the same INDEPENDENT_STMT(1) or 
//	        in the NEW or REDUCTION clause in the INDEPENDENT_STMT amid its lexical body(2) or
//		just to get the compare result(3). 
//return: 0--no, 1--yes
*/
int is_same_variable(var1,var2,place)
PTR_LLND var1,var2;
int place;
{
 PTR_SYMB symb1,symb2;
 PTR_LLND var1_list,var2_list;
 char name[100];
 int part_eq = 0;		/* indicate the equality of two records */


	switch(var1->variant) {
	case VAR_REF:
	case ARRAY_REF:
		if((var2->variant != VAR_REF) && (var2->variant != ARRAY_REF)) 
			return 0; 
		else {
			symb1 = var1->entry.Template.symbol;
			symb2 = var2->entry.Template.symbol;
			if(strcmp(symb1->ident,symb2->ident) == 0) {	/* same variable name */
				if(place == 1)		/* check the same INDEPENDENT_STMT */  
					error3(813,symb1->ident);
				else if(place == 2)	/* check the INDEPENDENT_STMT in its lexiacal body */
					error3(814,symb1->ident); 

				return 1;
			}
		}
		break;

	case RECORD_REF:
		if(var2->variant != RECORD_REF) return 0;
		else {
			var1_list = var1;
			var2_list = var2;
			symb1 = get_var_name(var1_list); 
			strcat(name,symb1->ident); 
			symb2 = get_var_name(var2_list);
			if(strcmp(symb1->ident,symb2->ident) == 0) part_eq = 1;	

			/* compare the record part_ref one by one */
			while(part_eq && var1_list->entry.Template.ll_ptr2 && var2_list->entry.Template.ll_ptr2) {
				var1_list = var1_list->entry.Template.ll_ptr2;
				var2_list = var2_list->entry.Template.ll_ptr2;
				symb1 = var1_list->entry.Template.ll_ptr1->entry.Template.symbol;
				strcat(name,"%");
				strcat(name,symb1->ident);
				symb2 = var2_list->entry.Template.ll_ptr1->entry.Template.symbol;
 				if(strcmp(symb1->ident,symb2->ident) != 0) 
					part_eq = 0;
			}

			/* check the compare result */
			if(!part_eq) return 0; 			/* var1 & var2 not same */ 
			else {		/* part_eq == 1 until now, so at least 
					   one of var1_list->... & var2_list->... must be LLNULL */
				if((var1_list->entry.Template.ll_ptr2 == LLNULL) 
					&& (var2_list->entry.Template.ll_ptr2 == LLNULL)) {
					if(place ==1)
						error3(813,name);	/* same RECORD_REF */
					else if(place == 2)
						error3(814,name);

					return 1;
				}/* end if */

			}/* end if-else */

		}/* end if-else */
		break;

	default:
		return 0;
	}/* end of switch */

	return 0;
}

/*
//Recursively check the duplicate appearance of red_var in INDEPEDENT_STMT inside bfnd's lexical body
*/
int appear_in_scope_check(red_var,bfnd)
PTR_LLND red_var;
PTR_BFND bfnd;
{
 PTR_BLOB blob1,blob2;
 PTR_BFND in_bfnd;

 	blob1 = bfnd->entry.Template.bl_ptr1;
	if(blob1) {
		for( ; blob1; blob1=blob1->next) {
			in_bfnd = blob1->ref;
			if(in_bfnd->variant == INDEPENDENT_STMT) 
				appear_again_check(red_var,in_bfnd);
			else
				appear_in_scope_check(red_var,in_bfnd);

		}/* end for */
	}/* end if */

 	blob2 = bfnd->entry.Template.bl_ptr2;
	if(blob2) {
		for( ; blob2; blob2=blob2->next) {
			in_bfnd = blob2->ref;
			if(in_bfnd->variant == INDEPENDENT_STMT) 
				appear_again_check(red_var,in_bfnd);
			else
				appear_in_scope_check(red_var,in_bfnd);

		}/* end for */
	}/* end if */

	return 0;

}


/*
//check if the reduction-variable(var) appear again in the INDEPEDENT_STMT in its lexical body(bfnd)
*/
int appear_again_check(var,bfnd)
PTR_LLND var;
PTR_BFND bfnd;
{
 PTR_LLND in_var;
 PTR_LLND llnd1,llnd2,var_list;

	/* first check var_list in the NEW clause */
	llnd1 = bfnd->entry.Template.ll_ptr1;

	if(llnd1) {
		var_list = llnd1;
		for( ; var_list; var_list=var_list->entry.list_s.next) {
			in_var = var_list->entry.list_s.item;
			if(is_same_variable(var,in_var,2)) break;
		}/* end for */
	}/* end if */


	/* second check var_list in the REDCUTION clause */
	llnd2 = bfnd->entry.Template.ll_ptr2;

	if(llnd2) {
		var_list = llnd2;
		for( ; var_list; var_list=var_list->entry.list_s.next) {
			in_var = var_list->entry.list_s.item;
			if(is_same_variable(var,in_var,2)) break;
		}/* end for */
	}/* end if */

	return 0;

}


/*
//check the reduction-stmt in the do loop following bfnd(it's a INDEPENDENT STMT)
//Method: for each red_var, check its reduction-stmt
//	also check all of its appearance in other stmts.
*/
int reduction_stmt_in_do_check(bfnd)
PTR_BFND bfnd;
{
 PTR_BFND do_bfnd,this_bfnd;
 PTR_LLND red_var,red_var_list;

	do_bfnd = bfnd->thread;

	/* establish a red_var list */
	red_var_list = bfnd->entry.Template.ll_ptr2;

	for( ; red_var_list; red_var_list=red_var_list->entry.list_s.next) {
		red_var = red_var_list->entry.list_s.item;
		op = op1 = 0; 
		reduction_stmt_of_var_check(red_var,do_bfnd);
	}/* end for */

	return 0;
}


/*
//For red_var, check all of its reduction-stmt in the scope of bfnd
*/
int reduction_stmt_of_var_check(red_var,bfnd)
PTR_LLND red_var;
PTR_BFND bfnd;
{
 PTR_BLOB blob1,blob2;
 PTR_BFND this_bfnd;
 PTR_LLND left_var,llnd;
 PTR_SYMB red_var_symb;

 	blob1 = bfnd->entry.Template.bl_ptr1;
	if(blob1) {
		for( ; blob1; blob1=blob1->next) {

			this_bfnd = blob1->ref;
			left_var = this_bfnd->entry.assign_s.l_value;

			if((this_bfnd->variant == ASSIGN_STAT) 
				&& is_lexically_same(left_var,red_var)) {	/* reduction-stmt of red_var */ 
			
				/* check for operator mixing */
 				if(!op)
					op = reduction_stmt_check(this_bfnd);
				else {
					op1 = reduction_stmt_check(this_bfnd);
					if(op != op1) error1(820);
				}

			}
			else { 		/* check appearance in other stmts, and report error */

				red_var_symb = get_var_name(red_var);

				llnd = this_bfnd->entry.Template.ll_ptr1;
				if(llnd) {
					if(red_var_appear(red_var,llnd)) 
						error3(821,red_var_symb->ident);
				}
				
				llnd = this_bfnd->entry.Template.ll_ptr2;
				if(llnd) {
					if(red_var_appear(red_var,llnd)) 
						error3(821,red_var_symb->ident);
				}
	
				llnd = this_bfnd->entry.Template.ll_ptr3;
				if(llnd) {
					if(red_var_appear(red_var,llnd)) 
						error3(821,red_var_symb->ident);
				}

				reduction_stmt_of_var_check(red_var,this_bfnd);

			}/* end if-else */

		}/* end for */
	}/* end if */

 	blob2 = bfnd->entry.Template.bl_ptr2;
	if(blob2) {
		for( ; blob2; blob2=blob2->next) {

			this_bfnd = blob2->ref;
			left_var = this_bfnd->entry.assign_s.l_value;

			if((this_bfnd->variant == ASSIGN_STAT) 
				&& is_lexically_same(left_var,red_var)) {	/* reduction-stmt of red_var */ 
			
				/* check for operator mixing */
 				if(!op)
					op = reduction_stmt_check(this_bfnd);
				else {
					op1 = reduction_stmt_check(this_bfnd);
					if(op != op1) error1(820);
				}

			}
			else { 		/* check appearance in other stmts, and report error */

				red_var_symb = get_var_name(red_var);

				llnd = this_bfnd->entry.Template.ll_ptr1;
				if(llnd) {
					if(red_var_appear(red_var,llnd)) 
						error3(821,red_var_symb->ident);
				}
				
				llnd = this_bfnd->entry.Template.ll_ptr2;
				if(llnd) {
					if(red_var_appear(red_var,llnd)) 
						error3(821,red_var_symb->ident);
				}
	
				llnd = this_bfnd->entry.Template.ll_ptr3;
				if(llnd) {
					if(red_var_appear(red_var,llnd)) 
						error3(821,red_var_symb->ident);
				}

				reduction_stmt_of_var_check(red_var,this_bfnd);

			}/* end if-else */

		}/* end for */
	}/* end if */

	return 0;
}



/*
//Check to see if "red_stmt" satisfies the demand for a reduction-stmt. 
*/
int reduction_stmt_check(red_stmt)
PTR_BFND red_stmt;
{
 PTR_LLND left_var,expr;
 int op_kind=0;

	left_var = red_stmt->entry.assign_s.l_value;
	expr = red_stmt->entry.assign_s.r_value;

	check_expr_type(expr);

	switch(expr->variant) {
	case ADD_OP:
	case MINUS_OP:
		op_kind = 1;
		reduction_op_check(left_var,expr);
		break;
	case MULT_OP:
	case DIV_OP:
		op_kind = 2;
		reduction_op_check(left_var,expr);
		break;
	case AND_OP:
		op_kind = 3;
		reduction_op_check(left_var,expr);
		break;
	case OR_OP:
		op_kind = 4;
		reduction_op_check(left_var,expr);
		break;
	case EQV_OP:
		op_kind = 5;
		reduction_op_check(left_var,expr);
		break;
	case NEQV_OP:			/* these are allowed operator's */
		op_kind = 6;
		reduction_op_check(left_var,expr);
		break;

	case INTRIN_FUNC_CALL:
		op_kind = reduction_func_check(left_var,expr);
		break;
	default:
		error1(815);
	}/* end switch */

	return op_kind;
}


/* 
//Check the 1 and 2 kind of reduction-stmt.
//input: the reduction-variable and the expression on the right of assign operator
*/
int  reduction_op_check(red_var,expr)
PTR_LLND red_var,expr;
{
 PTR_LLND l_value,r_value;
 PTR_SYMB symb;

	symb = get_var_name(red_var);
	l_value = expr->entry.Template.ll_ptr1;
	r_value = expr->entry.Template.ll_ptr2;

	if(is_lexically_same(red_var,r_value)) {		/* kind 1 */
		if((expr->variant == MINUS_OP) || (expr->variant == DIV_OP))
			error3(818,symb->ident);
		if(red_var_appear(red_var,l_value))
			error3(819,symb->ident);
	}
	else if(is_lexically_same(red_var,l_value)){		/* kind 2 */
		if(red_var_appear(red_var,r_value))
			error3(819,symb->ident);
	}
	else {
		if(is_same_variable(red_var,r_value,3) || is_same_variable(red_var,l_value,3))
			error3(817,symb->ident);
		else 		/* red_var not found at top level in reduction-stmt */ 
			error3(816,symb->ident);
	}

	return 0;
}

/*
//Check whether two variables are lexically the same
//It's a recursive check.
*/
int is_lexically_same(var1,var2)
PTR_LLND var1,var2;
{
 PTR_LLND left1,right1,left2,right2;
 PTR_SYMB symb1,symb2;

	if(var1->variant != var2->variant) return 0;
	else {
		symb1 = var1->entry.Template.symbol;
		symb2 = var2->entry.Template.symbol;
		if(symb1 && symb2) {
			if(strcmp(symb1->ident,symb2->ident) != 0) 	/* not same symbol */
				return 0;
		}
		else if((!symb1 && symb2) || (!symb2 && symb1)) 	/* only one of them has symbol */
				return 0;
	}


	/* check left and right branches */

	left1 = var1->entry.Template.ll_ptr1;
	right1 = var1->entry.Template.ll_ptr2;

	left2 = var2->entry.Template.ll_ptr1;
	right2 = var2->entry.Template.ll_ptr2;

	if(left1 && left2) {
		if(!is_lexically_same(left1,left2)) return 0;
	} 
	else if((!left1 && left2) || (!left2 && left1)) return 0;

	if(right1 && right2) {
		if(!is_lexically_same(right1,right2)) return 0;
	}
	else if((!right1 && right2) || (!right2 && right1)) return 0;


	return 1;	/* lexically the same */
}


/*
//Check whether the reduction-variable appear in expr.
//It's recursive.
*/
int red_var_appear(red_var,expr)
PTR_LLND red_var,expr;
{
 PTR_LLND left,right;

	if(red_var->variant == expr->variant) 
		if(is_lexically_same(red_var,expr)) return 1;

	/* check left & right branches */
	left = expr->entry.Template.ll_ptr1;
	right = expr->entry.Template.ll_ptr2;

	if(left != LLNULL) 
		if(red_var_appear(red_var,left)) return 1;

	if(right != LLNULL)
		if(red_var_appear(red_var,right)) return 1;

	return 0;
}


/*
//Check the 3 and 4 kind of reduction-stmt.
//input: the reduction variable and the function call expr.
*/
int reduction_func_check(red_var,expr)
PTR_LLND red_var,expr;
{
 PTR_LLND arg_list,arg1,arg2;
 PTR_SYMB symb,func_symb;
 int kind=0;

	symb = get_var_name(red_var);
	func_symb = expr->entry.func_call.symbol;

	arg_list = expr->entry.func_call.param_list;
	arg1 = arg_list->entry.Template.ll_ptr1;
	arg2 = arg_list->entry.Template.ll_ptr2->entry.Template.ll_ptr1;
	
	if(arg2->entry.Template.ll_ptr2 != LLNULL) error1(822);

	if(strcasecmp(func_symb->ident,"max") == 0) {
		kind = 7;
	}
	else if(strcasecmp(func_symb->ident,"min") == 0) {
		kind = 8;
	}
	else if(strcasecmp(func_symb->ident,"iand") == 0) {
		kind = 9;
	}
	else if(strcasecmp(func_symb->ident,"ior") == 0) {
		kind = 10;
	}
	else if(strcasecmp(func_symb->ident,"ieor") == 0) {
		kind = 11;
	}
	else error3(823,func_symb->ident);


	if(is_lexically_same(red_var,arg1)) {		/* kind 3 */
		if(red_var_appear(red_var,arg2))
			error3(819,symb->ident);
 
	}
	else if(is_lexically_same(red_var,arg2)){	/* kind 4 */
		if(red_var_appear(red_var,arg1))
			error3(819,symb->ident);

	}
	else {
		if(is_same_variable(red_var,arg1,3) || is_same_variable(red_var,arg2,3))
			error3(817,symb->ident); 
		else 		/* red_var not found in reduction-stmt */ 
			error3(816,symb->ident);
	}

	return kind;
	
}

