/*************************************************************************

	readmod.c -- Separate compilation
		for read mod file

written by Zheng Gengbin
*************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include <stdio.h>

#   include "typedef.h"
#   include "bif.h"
#   include "ll.h"
#   include "symb.h"
#include "defnodes.h"
#include "dep.h"
#include "mod_struct.h"
#include "tags.h"
#include "decl.h"

#include "exglobal.h"


#define NULL_CHECK(BASE,VALUE) ((VALUE) ? (BASE + (VALUE-1)): 0)

/*
 * Local variables
 */

static struct _locs      floc;   /* used to read in preamble "floc" */
static struct _preamble  head;   /* used to read in preamble "head" */
static struct _ll_nd     ll;     /* used to read in ll nodes */
static struct _sym_nd    sym;    /* used to read in symbol nodes */
static struct _typ_nd    typ;    /* used to read in type nodes */

static PTR_BFND head_bfnd, cur_module;
static PTR_LLND head_llnd, cur_llnd;
static PTR_SYMB head_symb, cur_symb;
static PTR_TYPE head_type, cur_type;
static PTR_DEP  head_dep,  cur_dep;
static PTR_LABEL head_lab, cur_lab;
static PTR_CMNT head_cmnt, cur_cmnt;


static char **strtbl;           /* starting address of string table */
static int tmp[2000];       /* temp working area */
static FILE *fd;                /* local copy of file id for the dep file */
static PTR_FILE lfi;
static int need_swap = 0;       /* set to 1 if we need to swap bytes */

static int     ntype=0;
static int     nsymb=0;
static int     nllnd=0;

/********************************************************
 *                      swap_w                          *
 *                                                      *
 *           Swap bytes of one word (2 bytes)           *
 ********************************************************/
static void
swap_w(p)
        char *p;
{
    char c;

    c = *(p+1);
    *(p+1) = *p;
    *p = c;
}


/********************************************************
 *                      swap_i                          *
 *                                                      *
 *           Swap bytes of an integer (4 bytes)         *
 ********************************************************/
static void
swap_i(p)
        char *p;
{
    char c;

    c = *(p+3);                 /* swap the 1st and 4th bytes */
    *(p+3) = *p;
    *p++ = c;
    c = *p;                     /* swap the 2nd and 3rd bytes */
    *p = *(p+1);
    *(p+1) = c;
}

/*------------------------------------------------------*
		read_string
--------------------------------------------------------*/
char *read_string(fd)
FILE *fd;
{
	u_short u;
	int n;
	char *s;

	if ((int)fread( (char *) &u, sizeof(u_short), 1, fd) < 0)
            return NULL;
        if (need_swap)
            swap_w((char *)&u);
        n = (int) u;
        if ((s = malloc(n+1)) == NULL) {
            fprintf(stderr, "read_str_tbl: No more space\n");
            exit(1);
        }
        if ((int)fread(s, sizeof(char), n, fd) < 0) /* now the content */
            return NULL;
        *(s+n) = '\0';
	return s;
}

/*------------------------------------------------------*
 *		   read_str_tbl				*
 *							*
 *	Read in the string table in dep file		*
 *------------------------------------------------------*/
static int
read_str_tbl()
{
    int	     i, n, sz;
    u_short  u;
    char    *s;
    char   **cp;

    /*
     * Fast forward to where the string table starts
     */
    if (fseek(fd, floc.strs, 0) < 0)
	return -1;

    /*
     * The first word is the total number of strings in the dep file
     */

    /* get size of string table */
    if ((int)fread( (char *) &u, sizeof(u_short), 1, fd) < 0) 
	return -1;

    if (need_swap)
	swap_w((char *)&u);
    sz = (int) u;
    if((cp = strtbl = (char **)malloc(sz*sizeof(char *))) == NULL) {
	fprintf(stderr, "read_str_tbl: No more space\n");
	exit(1);
    }

    /*
     * Then followed by strings in the form of
     *		-------------------------
     *		| str length | contents	|
     *		-------------------------
     */
    for (i = 0; i < sz; i++) {
        /* get string length */
	if ((int)fread( (char *) &u, sizeof(u_short), 1, fd) < 0) 

	    return -1;
	if (need_swap)
	    swap_w((char *)&u);
	n = (int) u;
	if ((s = malloc(n+1)) == NULL) {
	    fprintf(stderr, "read_str_tbl: No more space\n");
	    exit(1);
	}
	if ((int)fread(s, sizeof(char), n, fd) < 0) /* now the content */
	    return -1;
	*(s+n) = '\0';
	*cp++ = s;
    }
    return 0;
}


int read_uselist()
{
	int 	i;
	u_short sum;
	char   *str;

	if ((int)fread(&sum, sizeof(u_short), 1, fd) < 0) 
	    return -1;

	for (i=0; i<sum ; i++) {
		str=read_string(fd);
/*
printf("Read use list: %s\n",str);
*/
	}
	return 0;
}


/*
 *  check_if_mod --- check if a file is a mod file
 */
int check_if_mod(fd)
FILE *fd;
{
	u_short mno[2];

	/* The first 8 bytes is the file magic (see /etc/magic) PHB */
        if ((int)fread( (char *) mno, sizeof(u_short), 1, fd) < 0)
                return -1;
        if (*mno != MAGIC_NO) {      /* Is this a dep file? */
        need_swap = 1;          /* No... */
        swap_w((char *)mno);            /*  ... Maybe we need to swap bytes */
                if(*tmp != MAGIC_NO) {   /* Try again */
                fprintf(stderr, "Are you sure this is a legal mod file? %x\n",*mno);
                return -2;
                }
        }
	return 0;
}


/*--------------------------------------------------------------*
 *                      read_preamble                           *
 *          Read in the preamble part of the dep file           *
 *--------------------------------------------------------------*/
int
read_preamble(fd)
FILE *fd;
{
	int i;
	char *fname;
	long pos;

	i=check_if_mod(fd);
	if (i != 0 ) return i;

	fname=read_string(fd);

	read_uselist();

	/* Some more data */
	if ((int)fread( (char *) &floc, sizeof(struct _locs), 1, fd) < 0)
        return -1;

	pos=ftell(fd);

	/* Reconstruct the string table first */
	if (read_str_tbl() < 0)
        return -1;

    /* rewind back to the point after "locs" information (8 is filemagic) */
	if (fseek(fd, pos, 0) < 0)
        return -1;

    /*
     * Read in the second part of preamble.  Here we have numbers of
     * all nodes (bif, low level, etc.) for this dep file
     */
    if ((int)fread( (char *) &head, sizeof(struct _preamble), 1, fd) < 0)
        return -1;
    if (need_swap)
        swab((char *)&head, (char *)&head, sizeof(struct _preamble));

	nsymb=head.num_symbs;
        nllnd=head.num_llnds;
        ntype=head.num_types;
/*
printf("read -- nsymb:%d,nllnd:%d,ntype:%d\n",nsymb,nllnd,ntype);
*/

        head_symb = (PTR_SYMB)(nsymb>0? calloc(nsymb, sizeof(struct symb)): NULL);
	head_type = (PTR_TYPE)(ntype>0? calloc(ntype, sizeof( struct data_type)): NULL);
	head_llnd = (PTR_LLND)(nllnd>0? calloc(nllnd, sizeof( struct llnd)): NULL);
}


/*------------------------------------------------------------------------

		retrieve_symb()

--------------------------------------------------------------------------*/

PTR_BFND
retrieve_bfnd(no)
int no;
{
	if (no==0) return lfi->global_bfnd;
	else return cur_module;
}


/*
 *   get symbol by the number 
 */
PTR_SYMB
retrieve_sym(no)
int no;
{
	PTR_SYMB sym;
	PTR_BFND mod;
	char    *str,*modstr;

	if (no==0) return SMNULL;
	if (no>0)
		return head_symb+(no-1);
	
	str=*(strtbl+(-no));
	modstr=*(strtbl+(-no+1));
	if (str[0]=='-' || modstr[0]=='-') return SMNULL;
	mod=search_head_name(fi,modstr);
	if (mod==BFNULL) {
		error3(50,modstr);
		return SMNULL;
	}
	sym=look_up_symb_at_scope_in_hash(str,mod);
	if (sym==SMNULL) error3(51,str);
	return sym;
}


PTR_TYPE
retrieve_type(no)
int no;
{
	PTR_SYMB type_name;

	switch (no) {
        case T_INT:
		return global_int;
        case T_FLOAT:
		return global_float;
        case T_DOUBLE:
		return global_double;
        case T_COMPLEX:
		return global_complex;
        case T_DCOMPLEX:
		return global_dcomplex;
        case T_BOOL:
		return global_bool;
        case T_STRING:
		return global_string;
        }
	if (no>0)  {
		no--;
		return head_type+no;
	}
	return TYNULL;
}

PTR_LLND
retrieve_llnd(no)
int no;
{
        if (no==0) return LLNULL;
        if (no>0)
                return head_llnd+(no-1);
        return LLNULL;
}


/*--------------------------------------------------------------*
 *			read_ll_nodes				*
 *								*
 *		      routines to read ll_nodes			*
 *--------------------------------------------------------------*/
static int
read_mod_ll_nodes()
{
    PTR_LLND llnd_ptr;
    int i;

    for(i = 0; i < nllnd; i++) {
	if ((int)fread( (char *) &ll, sizeof(struct _ll_nd), 1, fd) < 0)
	    return -1;
	if (need_swap)
	    swab((char *)&ll, (char *)&ll, sizeof(struct _ll_nd));

	llnd_ptr	  = head_llnd + i;
	llnd_ptr->variant = (int) ll.variant; 
	llnd_ptr->type	  = retrieve_type(ll.type); /*NULL_CHECK(head_type, ll.type);*/
	llnd_ptr->attr=ll.attr;
	llnd_ptr->const_val=ll.const_val;


	switch(llnd_ptr->variant) {

	case INT_VAL :
	    if ((int)fread( (char *) &llnd_ptr->entry.ival, sizeof(int), 1, fd) < 0)
		return -1;
	    if (need_swap)
		swap_i((char *)&llnd_ptr->entry.ival);
	    break;
	case BOOL_VAL :
	    if ((int)fread( (char *) tmp, sizeof(int), 1, fd) < 0)
		return -1;
	    if (need_swap)
		swap_w((char *)tmp);
	    llnd_ptr->entry.bval = (int)(*tmp);
	    break;
	case CHAR_VAL :
	    if ((int)fread( (char *) tmp, sizeof(int), 1, fd) < 0)
		return -1;
	    if (need_swap)
		swap_w((char *)tmp);
	    llnd_ptr->entry.cval = (char)(*tmp);
	    break;
					/* modified by zheng.gb */
        case BINARY_VAL:
        case OCTN_VAL:
        case HEX_VAL:
	case DOUBLE_VAL:
	case FLOAT_VAL :
	case STMT_STR  :
	case STRING_VAL:
	case KEYWORD_VAL:
	    if ((int)fread( (char *) tmp, sizeof(int), 1, fd) < 0)
		return -1;
	    if (need_swap)
		swap_w((char *)tmp);
	    llnd_ptr->entry.string_val = *(strtbl+(*tmp));
	    break;
	case LABEL_REF :
				/* add by zheng.gb */
	case ERR_LABEL:
	case END_LABEL:
	    if ((int)fread( (char *) tmp, sizeof(int), 1, fd) < 0)
		return -1;
	    if (need_swap)
		swap_w((char *)tmp);

	/**    llnd_ptr->entry.label_list.lab_ptr= NULL_CHECK(head_lab,(*tmp));  **/
 /* Modify for HPF */
	    llnd_ptr->entry.label_ref.lab_ptr= NULL_CHECK(head_lab,(*tmp));

	    break;

/***  This add for HPF 1 ***/

        case LABEL_LIST:
            if ((int)fread( (char *) tmp, sizeof(int), 3, fd) < 0)
                return -1;
            if (need_swap)
                swab((char *)tmp, (char *)tmp, 3*sizeof(int));
            llnd_ptr->entry.label_list_s.lab_ptr =NULL_CHECK(head_lab,(*tmp));
            llnd_ptr->entry.Template.ll_ptr1=NULL_CHECK(head_llnd,tmp[1]);
            llnd_ptr->entry.Template.ll_ptr2=NULL_CHECK(head_llnd,tmp[2]);

            break;
	default:
	    if ((int)fread( (char *) tmp, sizeof(int), 3, fd) < 0)
		return -1;
	    if (need_swap)
		swab((char *)tmp, (char *)tmp, 3*sizeof(int));
	    llnd_ptr->entry.Template.symbol =retrieve_sym(*tmp);
	    llnd_ptr->entry.Template.ll_ptr1=retrieve_llnd(tmp[1]);
	    llnd_ptr->entry.Template.ll_ptr2=retrieve_llnd(tmp[2]);
	}


    }     /*  end of for */
    return 0;
}


/*----------------------------------------------------------------------*
 *									*
 *		    routines to read type table				*
 *									*
 *----------------------------------------------------------------------*/
static int
read_mod_type_nodes()
{
    PTR_TYPE type_ptr;
    int i, uss1, uss2;

    for(i = 0; i < ntype; i++) {
	if ((int)fread( (char *) &typ, sizeof(struct _typ_nd), 1, fd) < 0)
	    return -1;
	if (need_swap)
	    swab((char *)&typ, (char *)&typ, sizeof(struct _typ_nd));

	type_ptr		= head_type + i;
	type_ptr->variant	= (int)typ.variant;
	type_ptr->name		= retrieve_sym(typ.name);

	switch (type_ptr->variant) {
	case T_INT	:
	case T_FLOAT	:
	case T_DOUBLE	:
	case T_CHAR	:
	case T_BOOL	:
        case T_COMPLEX  :
        case T_DCOMPLEX  :
	case T_STRING	:
	    if ((int)fread( (char *) tmp, sizeof(int), 1, fd) < 0)
		return -1;
	    if (need_swap)
		swap_w((char *)tmp); 
		/* swab((char *)tmp, (char *)tmp, sizeof(u_short)); */
	    type_ptr->entry.Template.ranges    = retrieve_llnd(tmp[0]);
	    break;
	case DEFAULT	:
	case T_VOID	:	/* NEW ADDED FOR VPC */
        case T_UNKNOWN  :
	case T_ENUM_FIELD:
	    break;

	case T_ARRAY	:
	    if ((int)fread( (char *) tmp, sizeof(int), 3, fd) < 0)
		return -1;
	    if (need_swap)
		swab((char *)tmp, (char *)tmp, 3*sizeof(int));
	    type_ptr->entry.ar_decl.num_dimensions = (int)tmp[0];
	    type_ptr->entry.ar_decl.base_type = retrieve_type(tmp[1]);
	    type_ptr->entry.ar_decl.ranges    = retrieve_llnd(tmp[2]);
	    break;

	case T_KIND:			
	    if ((int)fread( (char *) tmp, sizeof(int), 2, fd) < 0)
		return -1;
	    if (need_swap)
		swab((char *)tmp, (char *)tmp, 2*sizeof(u_short));
	    type_ptr->entry.kind_decl.base_type = retrieve_type(tmp[0]);
	    type_ptr->entry.kind_decl.kindexp = retrieve_llnd(tmp[1]);
	    break;

	case T_RECORD	:
	    if ((int)fread( (char *) tmp, sizeof(int), 3, fd) < 0)
		return -1;
	    if (need_swap)
		swab((char *)tmp, (char *)tmp, 3*sizeof(u_short));
	    type_ptr->entry.re_decl.num_fields = (int)(*tmp);
	    type_ptr->entry.re_decl.first      = retrieve_sym(tmp[1]);
	    type_ptr->entry.re_decl.mapped_flag = (int)tmp[2];
	    break;


	default :
	    break;
	}
    }
    return 0;
}


/*--------------------------------------------------------------*
 *								*
 *		  routines to read symbol table			*
 *								*
 *--------------------------------------------------------------*/
static int
read_mod_symb_nodes()
{
    PTR_SYMB symb_ptr;
    int i;

    for(i = 0; i < nsymb; i++) {
	if ((int)fread( (char *) &sym, sizeof(struct _sym_nd), 1, fd) < 0)
	    return -1;
	if (need_swap)
	    swab((char *)&sym, (char *)&sym, sizeof(struct _sym_nd));

	symb_ptr	    = head_symb + i;
	symb_ptr->variant   = (int) sym.variant;
	symb_ptr->type	    = retrieve_type(sym.type);
	symb_ptr->attr	    = (int) sym.attr;
	symb_ptr->next_symb = SMNULL; /*retrieve_sym(sym.next);*/
	symb_ptr->scope	    = retrieve_bfnd(i);
	symb_ptr->ident	    = *(strtbl + sym.ident);
		/* add by zheng.gb */
	symb_ptr->decl      = HARD;

	switch (symb_ptr->variant) {
	case DEFAULT	:
	case TYPE_NAME	:
	    break;
	case CONST_NAME	:
	    if ((int)fread( (char *) tmp, sizeof(int), 1, fd) < 0)
		return -1;
	    if (need_swap)
		swap_w((char *)tmp);
	    symb_ptr->entry.Template.const_value = retrieve_llnd((*tmp));
	    break;
	case ENUM_NAME	:
	case FIELD_NAME	:
	    if ((int)fread( (char *) tmp, sizeof(int), 3, fd) < 0)
		return -1;
	    if (need_swap)
		swab((char *)tmp, (char *)tmp, 5*sizeof(int));
	    symb_ptr->entry.field.tag	= (int)(*tmp);
	    symb_ptr->entry.field.next	= retrieve_sym(tmp[1]);
	    symb_ptr->entry.field.base_name= retrieve_sym(tmp[2]);
/**  Modify for HPF 1	    symb_ptr->entry.field.declared_name = NULL_CHECK(head_symb,tmp[3]);  **/
/**  Modify for HPF 2	    symb_ptr->entry.field.restricted_bit= NULL_CHECK(head_llnd,tmp[4]);  **/
	    break;
	case VARIABLE_NAME:
	    if ((int)fread( (char *) tmp, sizeof(int), 6, fd) < 0)
		return -1;
	    if (need_swap)
		swab((char *)tmp, (char *)tmp, 6*sizeof(int));
	    symb_ptr->entry.var_decl.local  = (int)(*tmp);
	    symb_ptr->entry.var_decl.next_in= retrieve_sym(tmp[1]);
	    symb_ptr->entry.var_decl.next_out=retrieve_sym(tmp[2]);
/* 		add by zheng.gb */
	    symb_ptr->entry.Template.declared_name=retrieve_sym(tmp[3]);
	    symb_ptr->entry.Template.const_value = retrieve_llnd(tmp[4]);
	    symb_ptr->entry.var_decl.hpfattr = retrieve_llnd(tmp[5]);
	    break;
	case PROGRAM_NAME:
	    if ((int)fread( (char *) tmp, sizeof(int), 2, fd) < 0)
		return -1;
	    if (need_swap)
		swab((char *)tmp, (char *)tmp, 2*sizeof(int));

	    symb_ptr->entry.prog_decl.symb_list = NULL_CHECK(head_symb,(*tmp));
	    symb_ptr->entry.prog_decl.prog_hedr = NULL_CHECK(head_bfnd,tmp[1]);
	    break;
	case MODULE_NAME:
            break;
	case PROCEDURE_NAME :
        case PROCESS_NAME:
	case FUNCTION_NAME:	 
	    if ((int)fread( (char *) tmp, sizeof(int),10, fd) < 0)
		return -1;
	    if (need_swap)
		swab((char *)tmp, (char *)tmp, 10*sizeof(int));

	    symb_ptr->entry.proc_decl.num_input	 = (int)(*tmp);
	    symb_ptr->entry.proc_decl.num_output = (int)tmp[1];
	    symb_ptr->entry.proc_decl.num_io	 = (int)tmp[2];
	    symb_ptr->entry.proc_decl.in_list =retrieve_sym(tmp[3]);
	    symb_ptr->entry.proc_decl.out_list =retrieve_sym(tmp[4]);
	    symb_ptr->entry.proc_decl.symb_list=retrieve_sym(tmp[5]);
	    symb_ptr->entry.proc_decl.proc_hedr=NULL_CHECK(head_bfnd,tmp[6]);
            symb_ptr->entry.proc_decl.local_size = (int)tmp[7];
	    symb_ptr->entry.Template.declared_name=retrieve_sym(tmp[8]);
	    symb_ptr->entry.Template.next=retrieve_sym(tmp[9]);
	    break;
	case INTERFACE_NAME:
	    if ((int)fread( (char *) tmp, sizeof(int), 2, fd) < 0)
		return -1;
	    symb_ptr->entry.Template.declared_name=retrieve_sym(tmp[0]);
	    symb_ptr->entry.Template.next=retrieve_sym(tmp[1]);
	    break;
	default:
	    break;
	}
    }
    return 0;
}


/*----------------------------------------------------------------------------

			read_mod_nodes(fi,name)

		read mod file of name and append information at the
		end of the fi.

----------------------------------------------------------------------------*/
int
read_mod_nodes(fi,name)
PTR_FILE fi;
char *name;
{
	PTR_BFND module;
	PTR_BLOB blob,mod_blob;
	PTR_BFND bfnd,end;
	PTR_SYMB sym;
	PTR_TYPE type;
	int      i;
	extern PTR_HASH  hash_table[];

	printf("Loading module file '%s'...\n",name);

	lfi=fi;

	if ((fd = fopen (name, "r")) == NULL) {
                fprintf(stderr, "Could not open %s for read\n", name);
                return (-1);
        }
	
	if (read_preamble(fd) < 0)
        	return -1;
	
	/* make module bfnd */
	module=makeBfnd(fi,MODULE_STMT);
	module->control_parent=fi->global_bfnd;
	module->checked = VIRTUAL_BFND;
	/* link to the BFND */
	for (bfnd=fi->head_bfnd;bfnd->thread; bfnd=bfnd->thread);
	bfnd->thread=module;
	module->thread=BFNULL;

	mod_blob=make_blob(fi,module,BLNULL);
	for (blob=fi->global_bfnd->entry.Template.bl_ptr1;blob->next; blob=blob->next);
	blob->next=mod_blob;
		/* control_end */
	end=makeBfnd(fi,CONTROL_END);
	end->control_parent=module;
	module->thread=end;
	mod_blob=make_blob(fi,end,BLNULL);
	bfnd->entry.Template.bl_ptr1=mod_blob;

	cur_module=bfnd;
	read_mod_symb_nodes();
	module->entry.Template.symbol=head_symb;

	/*  merge */
	merge_symb_at_end(fi,head_symb,nsymb);

	/* begin to read type table */
	if (ftell(fd) != floc.type) {
        	fprintf(stderr,"readNodes: wrong location of type nodes\n");
        	if(fseek(fd, floc.type, 0) < 0)
            	return -1;
    	}

    	if (read_mod_type_nodes() < 0)
        	return -1;

	/* merge type */
	merge_type_at_end(fi,head_type,ntype);

	if (ftell(fd) != floc.llnd) {
        	fprintf (stderr,"readNodes: wrong location of low level nodes\n");
        	if (fseek(fd, floc.llnd, 0) < 0)
            	return -1;
    	}

    	if (read_mod_ll_nodes() < 0) {
        	perror("read_ll_nodes:");
        	return -1;
    	}

	/* merge type */
	merge_llnd_at_end(fi,head_llnd,nllnd);

	return 0;
}



int merge_type_at_end(fi,new_head,ntype)
PTR_FILE fi;
PTR_TYPE new_head;
int ntype;
{
	PTR_TYPE type;
	int 	i;

	/* merge type */
	fi->num_types += ntype;
	for (type=fi->head_type; type->thread; type=type->thread);
	for (i = 0; i < ntype; i++) {
        	(new_head + i)->id     = i + 1 +type->id;
        	(new_head + i)->thread=(i==ntype-1)?TYNULL:new_head + (i + 1);
    	}
	type->thread=new_head;
	for (type=fi->head_type ; type->thread; type=type->thread);
	fi->cur_type=type;
}


int merge_symb_at_end(fi,head_symb,nsymb)
PTR_FILE fi;
PTR_SYMB head_symb;
int      nsymb;
{
	PTR_SYMB sym;
	int	 i;

/*  merge */
        fi->num_symbs += nsymb;
        for (sym=fi->head_symb; sym->thread; sym=sym->thread);
        for (i = 0; i < nsymb; i++) {
                (head_symb + i)->id     = i + 1 + sym->id;
                (head_symb + i)->thread=(i==nsymb-1)?SMNULL:head_symb + (i + 1);
                /* insert into hash table */
                insert_hash(head_symb+i,hash_table);
        }
        sym->thread=head_symb;
        for (sym=fi->head_symb; sym->next_symb; sym=sym->next_symb);
        sym->next_symb=head_symb;
	for (sym=fi->head_symb ; sym->thread; sym=sym->thread);
	fi->cur_symb=sym;
}

int merge_llnd_at_end(fi,head_llnd,nllnd)
PTR_FILE fi;
PTR_LLND head_llnd;
int nllnd;
{
	PTR_LLND llnd;
	int	 i;

	fi->num_llnds += nllnd;
        /*for (llnd=fi->head_llnd; llnd->thread; llnd=llnd->thread);*/
	llnd=fi->cur_llnd;
        for (i = 0; i < nllnd; i++) {
                (head_llnd + i)->id     = i + 1 + (llnd?llnd->id:0);
                (head_llnd + i)->thread=(i==nllnd-1)?LLNULL:head_llnd + (i + 1);
        }
	if (llnd)
        	llnd->thread=head_llnd;
	else
		fi->head_llnd=head_llnd;

	/* cur_llnd */
	for (llnd=fi->head_llnd ; llnd->thread; llnd=llnd->thread);
	fi->cur_llnd=llnd;
}



