/*************************************************************************

	writemod.c -- Separate compilation

written by Zheng Gengbin
	Peking University , P.R.China
*************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include <stdio.h>

#   include "typedef.h"
#   include "bif.h"
#   include "ll.h"
#   include "symb.h"
#include "defnodes.h"
#include "dep.h"
#include "mod_struct.h"     /* zheng.gb */
#include "tags.h"
#include "decl.h"

#include "check.h"

extern char *cur_fname;

#define FOLLOW_BIF_POINTER_TO_ID(VAR) \
        (bf_ptr->entry.Template.VAR? bf_ptr-> entry.Template.VAR->id: 0)

#define FOLLOW_LL_POINTER_TO_ID(VAR) \
        (ll_ptr-> entry.Template.VAR? ll_ptr-> entry.Template.VAR->id: 0)

#define FOLLOW_SYMB_POINTER_1_TO_ID(VAR) \
        (sy_ptr->VAR? sy_ptr->VAR->id: 0)

#define FOLLOW_SYMB_POINTER_2_TO_ID(VAR) \
        (sy_ptr->entry.VAR? sy_ptr->entry.VAR->id: 0)

#define FOLLOW_TYPE_POINTER_TO_ID(VAR) \
        (ty_ptr->entry.VAR? ty_ptr->entry.VAR->id: 0)

#define FOLLOW_DEP_TO_ID(VAR) \
        (dep->VAR? dep->VAR->id: 0)

static PTR_BFND head_bfnd, cur_bfnd;
static PTR_LLND head_llnd, cur_llnd;
static PTR_SYMB head_symb, cur_symb;
static PTR_TYPE head_type, cur_type;
static PTR_DEP head_dep, cur_dep;
static PTR_LABEL head_label, cur_label;
static PTR_CMNT head_cmnt, cur_cmnt;
static PTR_FNAME head_file;
static PTR_BFND global_bfnd;

static int num_llnds;
static int num_symbs;
static int num_types;

/*
 * Local variables
 */
static struct _preamble  head;
static struct _ll_nd     ll;
static struct _sym_nd    sym;
static struct _typ_nd    typ;
static struct _locs      loc;

static FILE  *fd;               /* file pointer of the dep file */
static char **strtbl,           /* start of string table */
            **endtbl,           /* end of string table */
            **cp;               /* current pointer */
static int nstr = 0;            /* no of string stored so far */
static int tblsz = 2000;        /* initial string table size */

static int need_swap = 0;       /* set to 1 if we need to swap bytes */

static int tmp[2000];       /* some work space */

static SYMB_LIST   symb_list;   /* symbol needed to save */
static int	   nsymb=0;
static TYPE_LIST   type_list;   /* type needed to save */
static int	   ntype=0;
static LLND_LIST   llnd_list;   /* llnd needed to save */
static int	   nllnd=0;

static PTR_BFND    mod_scope;   /* current module scope */

/*------------------------------------------------------*
 *              write_string                            *
 *------------------------------------------------------*/
static int
write_string(str)
        char *str;
{
        int l1;
	u_short len;

        if(!str) l1 = 0;
        else l1 = strlen(str);
        len = (u_short) l1;
        if ((int)fwrite( (char *)&len, sizeof(u_short), 1, fd) >= 0)
                if ((int)fwrite( (char *) str, sizeof(char), l1, fd) >= 0)
                        return 0;
        return -1;
}

/*------------------------------------------------------*
 *                      store_str                       *
 *                                                      *
 *       put the given string into string table         *
 *------------------------------------------------------*/
static u_short
store_str(str)
        char *str;
{
        if (nstr >= tblsz) {
                tblsz += 1000;
                if(!(strtbl = (char **)realloc(strtbl, tblsz*sizeof(char **))))
{
                        fprintf(stderr,"store_str: No more space\n");
                        exit (1);
                }
                endtbl = strtbl + tblsz;
                cp = strtbl + nstr;
        }
        *cp++ = str;
        return (u_short)nstr++;
}


/*------------------------------------------------------*
 *              write_str_tbl                           *
 *------------------------------------------------------*/
static int
write_str_tbl(str, n)
        char **str;
        int    n;
{
        register char **p = str;
        register int i;
        u_short u;

        u = (u_short) n;
        if ((int)fwrite( (char *) &u, sizeof(u_short), 1, fd) < 0) /* output no
of strings */
                return -1;
        for (i = 0; i < n; i++)
                if (write_string(*p++) < 0) {
                        perror("write_str_tbl:");
                        return -1;
                }
        return 0;
}


/*--------------------------------------------------------------------------

			write_uselist()

--------------------------------------------------------------------------*/
write_uselist(bfnd)
PTR_BFND bfnd;
{
	PTR_LLND use_list;
	u_short	 sum;
	long     pos,curpos;
	
	use_list=bfnd->entry.Template.symbol->entry.Template.use_list;
	sum=0;
	pos=ftell(fd);
        if ((int)fwrite( (char *) &sum, sizeof(u_short), 1, fd) < 0) 
		return -1;
        for ( ; use_list; use_list=use_list->entry.Template.ll_ptr2)
        {
                write_string(use_list->entry.Template.ll_ptr1->entry.Template.symbol->ident);
/*
printf("write:%s\n",use_list->entry.Template.ll_ptr1->entry.Template.symbol->ident);
*/

		sum++;
        }
	if (sum) {
		curpos=ftell(fd);
		fseek(fd,pos,0);
        	if ((int)fwrite( (char *) &sum, sizeof(u_short), 1, fd) < 0) 
			return -1;
		fseek(fd,curpos,0);
	}
	return 0;
}

int store_sym(sym)
PTR_SYMB sym;
{
	int no;
	SYMB_LIST tmp;

	if (sym==SMNULL) return 0;
/*
printf("stored symbol:%s\n",sym->ident);
*/
	no=1;
	for (tmp=symb_list; tmp; tmp=tmp->next){
		if (tmp->sym == sym) return no;
		no++;
	}
	/* not in current scope */
	if (sym->ident==NULL)  no=-store_str("-");
	else no=-store_str(sym->ident);
	if (sym->scope==BFNULL) store_str("-");
	else if (sym->scope->entry.Template.symbol==SMNULL)
		store_str("-");
	     else
		store_str(sym->scope->entry.Template.symbol->ident);
	return no;
}

/*------------------------------------------------------*
			store type nodes in list
 *------------------------------------------------------*/
int
store_type(type)
PTR_TYPE type;
{
	TYPE_LIST new_list,tmp;
	int       no;

	if (type==TYNULL) return 0;

	switch (type->variant) {
        case T_INT:
        case T_FLOAT:
        case T_DOUBLE:
        case T_COMPLEX:
        case T_DCOMPLEX:
        case T_BOOL:
        case T_STRING:
		return type->variant;
        }

	/* search if alread in list */
	no = 1;
	for (new_list = type_list; new_list ; new_list=new_list->next) 
	{
		if (new_list->type == type) return no;
		no++;
	}
		/* type defined in other scope */
	if (type->variant==T_RECORD && type->name->scope!=mod_scope) {
/*
printf("store a type of other scope '%s'.\n",type->name->ident);
*/
		no=-store_sym(type->name);
		store_sym(type->name->scope->entry.Template.symbol);
		return no;
	}
	new_list=(TYPE_LIST) chkalloc(sizeof(struct _type_list));
        new_list->type=type;
        new_list->next=NULL;
	if (type_list == NULL) {
		type_list=new_list;
		return ++ntype;
	}
	for (tmp=type_list ; tmp->next ; tmp=tmp->next) ;
	tmp->next=new_list;
	return ++ntype;
}

/*------------------------------------------------------*
			store llnd nodes in list
 *------------------------------------------------------*/
int
store_llnd(llnd)
PTR_LLND llnd;
{
	LLND_LIST new_list,tmp;
	int       no;

	if (llnd==LLNULL) return 0;
/*
printf("store LLND:%d\n",llnd->variant);
*/
	/* search if alread in list */
	no = 1;
	for (new_list = llnd_list; new_list ; new_list=new_list->next) 
	{
		if (new_list->llnd == llnd) return no;
		no++;
	}
	new_list=(LLND_LIST) chkalloc(sizeof(struct _llnd_list));
        new_list->llnd=llnd;
        new_list->next=NULL;
	if (llnd_list == NULL) {
		llnd_list=new_list;
		return ++nllnd;
	}
	for (tmp=llnd_list ; tmp->next ; tmp=tmp->next) ;
	tmp->next=new_list;
	return ++nllnd;
}

/*-----------------------------------------------------------------------
		expand_llnd_list()

		expand all the leaf of the llnd list
------------------------------------------------------------------------*/

static int expand_llnds(llnd)
PTR_LLND llnd;
{
	PTR_LLND lp,rp;
	
	lp=llnd->entry.Template.ll_ptr1;
	rp=llnd->entry.Template.ll_ptr2;
	if (lp) {
		store_llnd(lp);
		expand_llnds(lp);
	}
	if (rp) {
		store_llnd(rp);
		expand_llnds(rp);
	}
}

static int expand_llnd_list()
{
	LLND_LIST node;

	for (node=llnd_list; node; node=node->next)
		expand_llnds(node->llnd);
}

/*------------------------------------------------------*
 *		     write_ll_node			*
 *							*
 *	       print out one low level node		*
 *------------------------------------------------------*/
static int 
write_mod_ll_node(ll_ptr)
PTR_LLND ll_ptr;
{
	int n = 0;

	ll.id	   = (u_short) ll_ptr->id;
	ll.variant = (u_short) ll_ptr->variant;
	ll.type	   = store_type(ll_ptr->type);

	ll.attr=ll_ptr->attr;
	ll.const_val=ll_ptr->const_val;


	if ((int)fwrite( (char *) &ll, sizeof(struct _ll_nd), 1, fd) < 0)
		return -1;

	switch (ll_ptr->variant) {
	    case INT_VAL:
		return (int)fwrite( (char *) &ll_ptr->entry.ival, sizeof(int), 1, fd);
	    case BOOL_VAL:
		tmp[0] = (u_short) ll_ptr->entry.bval;
		n = 1;
		break;
	    case CHAR_VAL:
		tmp[0] = (u_short) ll_ptr->entry.cval;
		n = 1;
		break;
					/* modified by zheng.gb */
	    case BINARY_VAL:
            case OCTN_VAL:
            case HEX_VAL:	
	    case DOUBLE_VAL:
	    case FLOAT_VAL:
	    case STMT_STR:
	    case STRING_VAL:
	    case KEYWORD_VAL:
		tmp[0] = store_str(ll_ptr->entry.string_val);
		n = 1;
		break;
	    case LABEL_REF:
                                    /* add by zheng.gb */
        case ERR_LABEL:
        case END_LABEL:
                tmp[0] = (u_short) ll_ptr->entry.label_ref.lab_ptr->id;

		n = 1;
		break;


            case LABEL_LIST:
                tmp[0] = (u_short) (ll_ptr->entry.label_list_s.lab_ptr ?
				ll_ptr->entry.label_list_s.lab_ptr->id:0 );
                tmp[1] = store_llnd(ll_ptr->entry.Template.ll_ptr1);
                tmp[2] = store_llnd(ll_ptr->entry.Template.ll_ptr2);
                n = 3;
                break;

	    case ARITH_ASSGN_OP:	/* New added for VPC++ */
/* The next line is a _REAL_ hack, I added the cast (PHB) */
		tmp[0] = (u_short) ((int) ll_ptr->entry.Template.symbol);
		tmp[1] = (u_short) FOLLOW_LL_POINTER_TO_ID(ll_ptr1);
		tmp[2] = (u_short) FOLLOW_LL_POINTER_TO_ID(ll_ptr2);
		n = 3;
		break;
	   default:
		n=0;
		tmp[n] = store_sym(ll_ptr->entry.Template.symbol); n++;
		tmp[n] = store_llnd(ll_ptr->entry.Template.ll_ptr1); n++;
		tmp[n] = store_llnd(ll_ptr->entry.Template.ll_ptr2); n++;
		break;
	}

	return (n? (int)fwrite( (char *) tmp, sizeof(int), n, fd): 0);
}


/*------------------------------------------------------*
 *		  write_ll_nodes			*
 *							*
 *		dump low level nodes			*
 *------------------------------------------------------*/
static int 
write_mod_ll_nodes()
{
	register PTR_LLND ll_ptr;
	LLND_LIST         tmp;

        for (tmp = llnd_list; tmp; tmp = tmp->next)
                if (write_mod_ll_node(tmp->llnd) < 0) {
                        perror("write_llnd_nodes:");
                        return -1;
                }
	return 0;
}



/*------------------------------------------------------*
 *		  write_type_node			*
 *							*
 *	      print out one type node			*
 *------------------------------------------------------*/
static int 
write_mod_type_node(ty_ptr)
	PTR_TYPE ty_ptr;
{       
	int	n = 0;
        int uss1;
	typ.id	    = (u_short) ty_ptr->id;
	typ.variant = (u_short) ty_ptr->variant;
	typ.name    = store_sym(ty_ptr->name);
/*
printf("store atype:%d,%s,%d\n",ty_ptr->variant,ty_ptr->name->ident,typ.name );
*/

	if ((int)fwrite( (char *) &typ, sizeof(struct _typ_nd), 1, fd) < 0)
		return -1;

	switch (ty_ptr->variant) {
	    case T_INT:
	    case T_FLOAT:
	    case T_DOUBLE:
	    case T_CHAR:
	    case T_BOOL:
	    case T_STRING:
	    case T_COMPLEX:
	    case T_DCOMPLEX:
		tmp[0] = store_llnd(ty_ptr->entry.Template.ranges);
                n = 1;
		break;

/**  Modify for HPF  2  **/

	    case T_ARRAY:
		tmp[0] = ty_ptr->entry.ar_decl.num_dimensions;
		tmp[1] = store_type(ty_ptr->entry.ar_decl.base_type);
		tmp[2] = store_llnd(ty_ptr->entry.ar_decl.ranges);
		n = 3;
		break;

/**  Modify for HPF  1  **/

					/* add by zheng.gb */
	    case T_KIND:
		tmp[0] =store_type(ty_ptr->entry.kind_decl.base_type);
		tmp[1] =store_llnd(ty_ptr->entry.kind_decl.kindexp);
		n=2;
		break;
/**  Modify for HPF  2  **/

	    case T_RECORD:
		tmp[0] = ty_ptr->entry.re_decl.num_fields;
		tmp[1] = store_sym(ty_ptr->entry.re_decl.first);
		tmp[2] = (u_short) ty_ptr->entry.re_decl.mapped_flag;
		n = 3;
		break;

/**  Modify for HPF  2  **/
	
	    default:
		break;
	}
	return (n? (int)fwrite( (char *) tmp, sizeof(int), n, fd): 0);
}


/*------------------------------------------------------*
 *		    write_type_nodes			*
 *------------------------------------------------------*/
static int 
write_mod_type_nodes()
{
	register PTR_TYPE ty_ptr;
	TYPE_LIST   tmp;

        for (tmp = type_list; tmp; tmp = tmp->next)
                if (write_mod_type_node(tmp->type) < 0) {
                        perror("write_type_nodes:");
                        return -1;
                }
	return 0;
}



/*------------------------------------------------------*
 *		    write_symb_node			*
 *							*
 *		print out one symbol node		*
 *------------------------------------------------------*/
static int 
write_mod_symb_node(sy_ptr)
	PTR_SYMB sy_ptr;
{
	int n = 0;

	sym.id	    = (u_short) sy_ptr->id;
	sym.variant = (u_short) sy_ptr->variant;
	sym.type    =  store_type(sy_ptr->type);
	sym.attr    = (int) sy_ptr->attr;
	sym.next    =  store_sym(sy_ptr->next_symb);
	sym.scope   =  store_sym(sy_ptr->scope->entry.Template.symbol);
	sym.ident   = store_str(sy_ptr->ident);
/*
printf("put symbol:%d,%s.\n",sym.ident,sy_ptr->ident);
*/

	if ((int)fwrite( (char *) &sym, sizeof(struct _sym_nd), 1, fd) < 0)
		return -1;

	switch (sy_ptr->variant) {
	    case CONST_NAME:
		tmp[0] =  store_llnd(sy_ptr->entry.Template.const_value);
		n = 1;
		break;
	    case ENUM_NAME:
	    case FIELD_NAME:
		tmp[0] = (u_short)sy_ptr->entry.field.tag;
		tmp[1] = store_sym(sy_ptr->entry.field.next);
        	tmp[2] = store_sym(sy_ptr->entry.field.base_name);
		n = 3;
		break;
	    case VARIABLE_NAME:
		tmp[0] = sy_ptr->entry.var_decl.local;
		tmp[1] = store_sym(sy_ptr->entry.var_decl.next_in);
		tmp[2] = store_sym(sy_ptr->entry.var_decl.next_out);
		tmp[3] = store_sym(sy_ptr->entry.Template.declared_name);
		tmp[4] = store_llnd(sy_ptr->entry.Template.const_value);
		tmp[5] = store_llnd(sy_ptr->entry.var_decl.hpfattr);
		n = 6;
		break;
	    case PROGRAM_NAME:
		tmp[0] = store_llnd(sy_ptr->entry.prog_decl.symb_list);
		tmp[1] = store_sym(sy_ptr->entry.prog_decl.prog_hedr);
		n = 2;
		break;
/*   add by zheng.gb  */
	    case MODULE_NAME:
		n=0;
		break;
	    case PROCEDURE_NAME:
            case PROCESS_NAME:
	    case FUNCTION_NAME:
		tmp[0] = (u_short) sy_ptr->entry.proc_decl.num_input;
		tmp[1] = (u_short) sy_ptr->entry.proc_decl.num_output;
		tmp[2] = (u_short) sy_ptr->entry.proc_decl.num_io;
		tmp[3] =  store_sym(sy_ptr->entry.proc_decl.in_list);
		tmp[4] =  store_sym(sy_ptr->entry.proc_decl.out_list);
		tmp[5] =  store_sym(sy_ptr->entry.proc_decl.symb_list);
		tmp[6] = (u_short) 0;   /* ??? proc_hedr */
                tmp[7] = (u_short) sy_ptr->entry.func_decl.local_size;
		tmp[8] = store_sym(sy_ptr->entry.Template.declared_name);
		tmp[9] = store_sym(sy_ptr->entry.Template.next);
		n = 10;
		break;
	    case INTERFACE_NAME:
		tmp[0] = store_sym(sy_ptr->entry.Template.declared_name);
		tmp[1] = store_sym(sy_ptr->entry.Template.next);
		n=2;
	    default:
		break;
	}
	return (n? (int)fwrite( (char *) tmp, sizeof(int), n, fd): 0);
}

/*------------------------------------------------------*
 *              write_symb_nodes                        *
 *                                                      *
 *              dump symbol table                       *
 *------------------------------------------------------*/
int
write_mod_symb_nodes()
{
	register SYMB_LIST tmp;

        for (tmp = symb_list; tmp; tmp = tmp->next)
                if (write_mod_symb_node(tmp->sym) < 0) {
                        perror("write_symb_nodes:");
                        return -1;
                }
        return 0;
}

int insert_sym(sym)
PTR_SYMB sym;
{
	SYMB_LIST new_list,tmp;

/*
printf("Insert symb:%s\n",sym->ident);
*/
	new_list=(SYMB_LIST) chkalloc(sizeof(struct _symbol_list));
        new_list->sym=sym;
        new_list->next=NULL;
	if (symb_list == NULL) {
		symb_list=new_list;
		return ++nsymb;
	}
	for (tmp=symb_list ; tmp->next ; tmp=tmp->next) ;
	tmp->next=new_list;
	return ++nsymb;
}


/*-----------------------------------------------------------------------
		create_sym_list()
 *                                                                       *
 *               create a list for save 
 *-----------------------------------------------------------------------*/
int 
create_sym_list(bfnd)
PTR_BFND bfnd;
{
	PTR_SYMB head,end,symbol,tmp;

	head=bfnd->entry.Template.symbol;
	end =head->next_symb;
	insert_sym(head);
	symbol=head->thread;
	while (symbol && symbol != end)
	{
		if (symbol->scope != bfnd || symbol->decl != HARD ) {
			/* if is an intrinsic call , same the function name */
			if ( is_intrinsic_call(symbol))  {
			}
			else {
				symbol=symbol->next_symb;
				continue;
			}
		}
		insert_sym(symbol);
		switch (symbol->variant) {
		case TYPE_NAME:
			for (tmp=symbol->thread;tmp;tmp=tmp->next_symb)
				insert_sym(tmp);
			break;
		case PROCEDURE_NAME:
		case FUNCTION_NAME:
/*
printf("func or proc :%s\n",symbol->ident);
*/
			for (tmp=symbol->entry.proc_decl.symb_list;tmp;
				tmp=tmp->entry.var_decl.next_in) 
			{
/*
printf("insert dummy:%s\n",tmp->ident);
*/
				insert_sym(tmp);
			}
			break;
		} /* switch */
		symbol=symbol->next_symb;
	}
}

/*  write module symbol table which is re-allocatable */
int write_mod_nodes(fi,modu,name)
PTR_FILE fi;
PTR_BFND modu;
char *name;
{
	PTR_SYMB mod_sym;
	u_short      magic_no=MAGIC_NO; /*2672;*/
	int	 loc_pos,pos,i,tsymb,ttype,tllnd;


	mod_sym=modu->entry.Template.symbol;
	if ((fd = fopen (name, "w")) == NULL) {
                fprintf(stderr, "Could not open %s for write\n", name);
                return (-1);
        }

	/* write magic number */
	fwrite( (char *)&magic_no, sizeof(u_short), 1, fd);

	/* write file name to produce the modfile */
	write_string(cur_fname);

	/* write use list */
	write_uselist(modu);

	loc_pos = ftell(fd);
	if ((int)fwrite( (char *) &loc, sizeof(struct _locs), 1, fd) < 0)
                return -1;

	pos = ftell(fd);
	/* write nsymb , ntype , nllnd */
	head.num_symbs    = (u_short) nsymb;
        head.num_llnds  = (u_short) nllnd;
        head.num_types  = (u_short) ntype;
        fwrite( (char *) &head, sizeof(struct _preamble), 1, fd);

	mod_scope = modu;
	head_symb = mod_sym;
        cur_symb  = mod_sym;

	symb_list=NULL;
	type_list=NULL;
	llnd_list=NULL;
	nstr = 0;
	nsymb= 0;
	ntype= 0;
	nllnd= 0;
        if (strtbl == NULL)
          if ( !(strtbl = (char **) calloc(tblsz, sizeof(char *)))) {
            perror("writeNodes(): calloc() error");
            return (-1);
          }

        cp = strtbl;
        endtbl = strtbl + tblsz;

	create_sym_list(modu);

	if (write_mod_symb_nodes() < 0) {
          perror("writeNodes(): write_symb_nodes() failed");
          return (-1);
        }

	if ((loc.type = ftell(fd)) < 0) {
          perror("writeNodes(): ftell() failed (2)");
          return (-1);
        }

        if (write_mod_type_nodes() < 0) {
          perror("writeNodes(): write_type_nodes() failed");
          return (-1);
        }


	if ((loc.llnd = ftell(fd)) < 0) {
          perror("writeNodes(): ftell() failed (0)");
          return (-1);
        }

	/* expand llnds */
	expand_llnd_list();

        if (write_mod_ll_nodes() < 0) {
          perror("writeNodes(): write_ll_nodes() failed");
          return (-1);
        }
/*
for (;llnd_list;llnd_list=llnd_list->next)
printf("ID:%d,var:%d\n",llnd_list->llnd->id,llnd_list->llnd->variant);
*/
	if ((loc.strs = ftell(fd)) < 0) {
          perror("writeNodes(): ftell() failed (0)");
          return (-1);
        }

	if (write_str_tbl(strtbl, nstr) < 0) {
          perror("writeNodes(): write_str_tbl() failed");
          return (-1);
        }

	/* Rewind to beginning of data segment  PHB */
        if (fseek(fd, loc_pos, 0) < 0) {
          perror("writeNodes(): fseek");
          return -1;
        }
        /* write out the offsets */
        if ((int)fwrite( (char *) &loc, sizeof(struct _locs), 1, fd) < 0) {
          perror("writeNodes(): Could not write out offsets");
          return -1;
        }

	fseek(fd,pos,0);
	head.num_symbs    = (u_short) nsymb;
        head.num_llnds  = (u_short) nllnd;
        head.num_types  = (u_short) ntype;
/*
printf("nsym:%d,nllnd:%d,ntype:%d.\n",nsymb,nllnd,ntype);
*/
        fwrite( (char *) &head, sizeof(struct _preamble), 1, fd);

	if (fclose(fd) < 0) {
          perror("writeNodes(): Could not close dep file");
          return -1;
        }

	printf("___The module file '%s' have been written ___ \n",name);
        return 0;
}


