/* #################################################################### */
/* #   Copyright (c) 1996 NPACT Working Group. You may copy,          # */
/* #   distribute, and modify it freely as long as you preserve this  # */
/* #   copyright and permission notice.                               # */
/* #                                                                  # */
/* #   NPACT Working Group is a joint force of                        # */
/* #      NPAC at Syracuse University,                                # */
/* #      PACT at Harbin Institute of Technology, and                 # */
/* #      PACT at Peking University                                   # */
/* #################################################################### */

/**@name The frontEnd system

\begin{center}
  
{\Large{\bf Preface}}

\end{center}

The frontEnd system is an extendible software package for constructing
programming language processing tools, such as compiler, interpreter
and language to language translator. The system analyzes the input
source language, and converts it to a uniform intermediate
representation (IR), which later can be manipulated by a set of
functions provided in a \Ref{Class library}.

In this document, section \Ref{Brief on the system} gives a outlook of
the system; Section \Ref{Functional modules} and \Ref{Intermediate representation} 
intoduce more about its composition; Finally, section
\Ref{Package interface} explains the package usage in detail.

{\bf Distribution } The latest version of the package is public domain
\URL[available]{http://www.npac.syr.edu/users/zgs/frontEndDistribution}.

{\bf About the document} This document is created automatically from
the source code of the package with 
\URL[doc++]{http://www.zib.de/Visual/software/doc++/index.html} system, 
having both LaTeX and HTML version. It is designed to
provide a thorough and online help for the serious users of the
package.

{\small Document maitained by 
\URL[Guansong Zhang]{http://www.npac.syr.edu/users/zgs}.}

@memo - an extendable package for programming language processing.

*/

//@{

  /** @name Brief on the system

At present time, the package includes:

\begin{itemize}

\item {\bf Language parsers}

HPF language parser: to convert HPF source code to the IR, including a
semantics checking on all constrains in language definition.

Java language parser: to convert java source code to the IR.

\item {\bf Language tools for displaying the IR} 

This includes unparsing the IR back to source code (mainly HPF 1.0,
since the package is originally designed to help building HPF
compiler.) and dumping IR in plain text and HTML format.

\item {\bf C++ Class} 

corresponding to language component.

\item {\bf Testsuite} 

including an extensive one on HPF language parser. (The HPF parser has
passed a "first order" exhaustive syntax test, all example programs
from "Migrating to Fortran 90" and "HPF handbook", and real
application programs from NAG.)
 
\end{itemize}

We are currently working to improve the support for Java language.

Physically, the package is organized in the following directory
structure.

\begin{verbatim}
frontEnd-x.xx/
   INSTALL
   README
   Makefile
   configure
   configure.in
   acconfig.h
   config/
   doc/
   bin/                    (if installed defaultly)
      dumpdep
      hpf2dep
      hpfsc
      java2dep
      unparse
   exampleUse
   src/
      Makefile
      Makefile.in
      basicop/
      xsageop/
      hpf2dep/
      hpfsc/
      java2dep
      tools/
         Makefile
         Makefile.in
         dump/
         unparse/
      include/
   include/                (if installed defaultly)
   testsuite/
   lib/                    (if installed defaultly)
      libbasic.a
      libxsage.a
      libhpf.sl
\end{verbatim}

It has been successfully installed on the following platform.

\begin{itemize}
\item
IBM RS/6000, AIX 3.2.5, 4.1.4, cc/xlC, gcc 2.7.2, lex/yacc, flex 2.7.2/bison
\item
SUN OS 4.1.1,4.1.4, Solaris 2.4, 2.5, gcc 2.7.2, lex, bison
\item
DEC Alpha OSF/1 V3.0, cc/cxx, lex/yacc, bison
\item
HP/9000 HP-UX, gcc 2.7, lex/yacc
\item
PC LINUX 1.2.8, gcc 2.7.2, flex 2.7.2/bison
\item
SGI INDY, IRIX 5.3, gcc 2.7.2, lex/yacc, bison
\end{itemize}

  */

  
  /**@name Functional modules

Generally speaking, there are three functional modules in the package:

\begin{itemize}

\item Language parser. It can convert some language source code
(currently HPF) to our intermediate representation (IR). The IR, as
described in \Ref{Intermediate representation} are different kinds of
structure nodes recording information in the source code, threaded
together by two kinds of link: list link and tree link.

\item Tool box. It contains {\tt dumpdep} and {\tt unparse} two parts,
which will display the intermediate representation generated by
language parser in different ways: either by printing each kind of
nodes along the list or by translating the nodes back to the source
code.

\item A compiler construction library. It provides operations on the
intermediate representation. These include functions for constructing
nodes, analysing information stored in them, and restructuring the
link between them. They are organized as C++ class methods as
described in \Ref{Class library}.

\end{itemize}

From the above, we can see that once defined according to a specific
language, the intermediate representation is the kernel part of the
functional modules. The whole package can be extended based on
this.

First, the IR can be defined to express more kind of programming
languaes, as we will see its structure in the next section. Currently
we are only working on Fortran and Java. Sage++ system used the same
kind of IR for C/C++.

Second, the library functions can also be extended easily. As we will
see, the organization of the library makes it quite easy to add a new
method to a language component, such as a statement or
expression. Different kinds of analyzing and restructuring operations
can be developed as member functions.

  */

  /**@name Intermediate representation

The intermediate representation used in the package is adopted from
Sage++. We give a brief introduction here. 

An abstract syntax tree (AST) and several structure tables constitute
the major part of our intermediate representation for the source
program.

There are two types of nodes in the syntax tree. One is called {\bf
{\em bif node}}, corresponding roughly to statements in the source
program; the other is called {\bf {\em low-level node}}, corresponding
to the expressions in each statement.

First, we only consider the {\em bif} nodes and their relations in the
AST, which is based on program control constructs. The root is called
{\bf {\em global node}} and each node has potentially two distinct
children, the {\bf {\em control children}}. One is called {\bf {\em
true branch}} and the other is called {\bf {\em false branch}}. The
{\em false branch} is mainly used in pure conditional construct such
as {\tt if-then-else}, most {\em bif} node has empty {\em false
branch}.

Each branch may represent a list of statements in the same level, as
can be seen in the following example.

Besides other information, there are two important pieces of
informations contained in each node.  One is called {\bf {\em id}},
which is a unique integer identifier designating the node among
others. The other is called {\bf {\em variant}}, which is a integer
tag telling the nature of the node. Thus, we may use the notation {\em
N[tag]} to refer to a node. 

To describe the control children, we will use an expression of the
form

\begin{verbatim}
      ( N[tag]  (true-branch) (false-branch))
\end{verbatim}

The following simple example in HPF language is to illustrate this
structure.

\begin{verbatim}
      PROGRAM MAIN
      REAL A(100), X
!HPF$ PROCESSORS P(4)
!HPF$ DISTRIBUTE A(BLOCK) ONTO P

      FORALL (i=1:100) A(i) = X + 1

      CALL FOO(A)

      END

      SUBROUTINE FOO(X)
      REAL X(100)
!HPF$ INHERIT X

      IF (X(1).EQ.0) THEN
        X = 1
      ELSE
        X = X + 1
      END IF
      RETURN
      END
\end{verbatim}

The {\em bif} nodes for this program are structured as follows:
\begin{verbatim}
(1[GLOBAL]                 ! 2 components in true-branch
   ((2[PROG_HEDR]          ! 6 components in true-branch
      (3[VAR_DECL]
       4[PROCESSORS_STMT]
       5[DISTRIBUTE_DECL]
       (6[FORALL_STMT]
          (7[ASSIGN_STAT]
           8[CONTROL_END]
          )
          NULL
       )
       9[PROC_STAT]
       10[CONTROL_END]
      )
      NULL
    )
    (11[PROC_HEDR]
       (12[VAR_DECL]
        13[INHERIT_DECL]
        (14[LOGIF_NODE]    ! both branchs are non empty
           (15[ASSIGN_NODE]
            16[CONTROL_END]
           )
           (17[ASSIGN_NODE]
            18[CONTROL_END]
           )
        )
        19[RETURN_STAT]
        20[CONTROL_END]
       )
       NULL
    )
   NULL
)
\end{verbatim}

The integer identifier associated with each node is the key to
accessing the rest of the properties of the node.

Other information contained in a statement node include:

\begin{itemize}

\item the name and line number of the source file containing this
statement,

\item the identifier of the control parent of this node,

\item a symbol reference such as a do loop parameter in Fortran or a
subroutine name in a call statement,

\item two or three expressions associated with the node (see the
following sections for details),

\item the control children of the node.  

\end{itemize}

Among them, expressions give rise to one level of refinement of our
syntax tree. A statement node can have up to three expressions nodes
associated with it, we also call them {\bf {\em low level
expressions}}.

Expression means lists or algebraic expressions of variables,
constants or functions.  For example, a Fortran assignment statement
has a left hand side expression and a right hand side expression.

Besides an integer {\em id}, each expression node has four components.
The first is its {\em variant}. The second is an optional symbol
reference.  The third and fourth components are the {\em left} and
{\em right} operand expression nodes. To describe expressions we can
use the following notation.

\begin{verbatim}
      (N[tag, symbol_id]  left_subtree  right_subtree)
\end{verbatim}

where {\tt N} is the id, {\tt tag} is its variant, {\tt symbol\_id} is
for possible associated symbol table identifier. If both the
left\_tree and right\_tree are empty, we may simply denote the
expression node as {\tt N[tag,symbol\_id]}

Thus, the expression {\tt X+1 } in the above example program takes the form

\begin{verbatim}
      (1[ADD_OP]  2[VAR_REF, X]  3[INT_VAL, 1])
\end{verbatim}

All the symbols in the program have a record in the symbole
table. During the construction of the program IR, these records will
be linked to the bif nodes and the low-level nodes described above. So
when traveling along the parse tree, all information about the symbols
can be obtained.

  */

  //@{

// All lower data structure needed

#ifndef COMPILER_APPLICATION

#define COMPILER_APPLICATION

#include "tags.h"

#include "typedef.h"
#include "bif.h"
#include "ll.h"
#include "symb.h"

#include "dep.h"
#include "sets.h"

  //@}

  
#include "defnodes.h"
#include "vparse.h"

// All maco define in old toolbox
#include "macro.h"

// All the "C" functions from the Rennes toolbox
#include "extcxx_low.h"


  /**@name Package interface

Currently the package is intended mainly for HPF related users. The
possible options are:

\begin{itemize}
\item HPF (or its subset) compiler writers, the primary users.
\item HPF flavored compiler tool builders, they may also use the
package as a basis to start their work.
\item Any language translation developer, who is to take HPF or its
subset as input.
\end{itemize}

Same things may apply to Java language.

The user interface of the package can be devided as two parts. In the
first part, we explain the usage of executable programs in the
package; In the second part, we present the class library based on the
intermediate representation.

  */
  //@{

    /** @name Executable programs

The executable programs constitude the language parser and toolbox in
the system. The data they operates on are files. 

{\bf Language parsers} 

There are two executable programs for HPF language parseing,

\begin{itemize}

\item hpf2dep --- takes as input any HPF program, say {\em
filename.f}, produces {\em filename.dep} as output. {\em filename.dep}
is a file that contains an intermediate representation of the input
HPF program, in terms of an abstract syntax tree (AST) and some tables
(among them the symbol table (ST) is the most important one).

\begin{verbatim}
   Usage: hpfc2dep [-fixed] [-d] F90_file.f
          -d : debugging switch
          -fixed : fixed format
\end{verbatim}

\item hpfsc --- takes as input a file {\em filename.dep}, performs
various semantics check while augmenting the AST to some extend, and
outputs a modified version of {\em filename.dep}.

\begin{verbatim}
   Usage: hpfsc [-vhmcdoiD] HPF_file.dep
          -v : version
          -h : usage
          -m : write .mod files for module .
          -c : compile the source files, but do not link.
          -d : debugging parser
          -o : output filename.
          -i : standard library compilation
          -D : display the module dependence .
\end{verbatim}

\end{itemize}

For java langauage:

\begin{itemize}
\item java2dep --- takes as input a file {\em filename.java}, produces
{\em filename.dep} as output.

\begin{verbatim}
   Usage: java2dep [-d] javafile.java 
          -d : debugging parser 
\end{verbatim}

\end{itemize}

{\bf Toolbox} The toolbox includes two parts, as different ways of
displaying the IR:

\begin{itemize}

\item dumpdep --- takes as input a file {\em filename.dep}, either
from output of hpf2dep or output of hpfsc, produces a readable form of
it to {\tt stdout}.

\begin{verbatim}
  Usage: dumpdep [-d nnn][-m] dep_file
         -d : debugging switch
         -m : generate the readable AST in HTML format
\end{verbatim}

\item unparse --- takes as input a valid file {\em filename.dep},
generate an equivalent HPF program onto {\tt stdout}.

\begin{verbatim}
  Usage: unparse [-debug] [-verbose] [-version] dep_file
         -debug : Print useful debugging information
         -version : Print version number of compiled program
\end{verbatim}

\end{itemize}

    */

    /** @name User library based on the IR

The library of the package provides a set of functions for compiler
writer or language processing programmer to build, analyse and
restructure the source code.

Three kinds of classes are provided:

\begin{itemize}
\item Class corresponding to project and file.
\item Class corresponding to language component.
\item Class used to record attribute for different kinds of language components.
\end{itemize}

We will illustrate the usage of them though an example, and give some
tips about it.

    */
      
    //@{

      /** @name An example of program analyzing

In exampleUse dir, an example for using the class library is
provided. 

A program called {\tt indent} can be built in this dir. You can do so
by just typing {\tt make} in this direcory after configuring the
package.

There are two files in this dir released in the pakcage: {\tt
Makefile.in} and {\tt indent.C}. 

When configuring the package, a {\tt Makefile} is generated. It is
used to make the executable program in this dir. Also it can be used
as a template for further constructing a language processing tool. 

The executable program build is quite simple, it will take as input a
{\em .dep} file generated by the parser and print out the loop structure in
the source program.

      */

      /** @name Class library

      */
          
      //@{
	/* * to be supplied */
    
        //@Include: fromsage++.h

// class library header file
#include "fromsage++.h"


      //@}

      /** @name Suggestions on using the library

to be supplied

      */

    //@}

  //@}

  /** @name Summary

The frontEnd package is used to help building language processing
tools, such as compiler and interpreter. The class library inside the
system provides a clear interface suitable for both library developers
and users. It makes the system easily to be used and extended.

A prototype HPF compiler is being developed based on the frontEnd
system. And the system is also used as education tools in the
compiling technique related courses.

  */

//@}

#endif /* COMPILER_APPLICATION */










