/* #################################################################### */
/* #   Copyright (c) 1996 NPACT Working Group. You may copy,          # */
/* #   distribute, and modify it freely as long as you preserve this  # */
/* #   copyright and permission notice.                               # */
/* #                                                                  # */
/* #   NPACT Working Group is a joint force of                        # */
/* #      NPAC at Syracuse University,                                # */
/* #      PACT at Harbin Institute of Technology, and                 # */
/* #      PACT at Peking University                                   # */
/* #################################################################### */


/* ******************************************************************* */
/*                  pC++/Sage++  Copyright (C) 1993                    */
/*  Indiana University  University of Oregon  University of Rennes     */
/* ******************************************************************* */

#ifndef LIBSAGEXX_H

#define LIBSAGEXX_H 1

#ifndef __GNUC__
# include <string.h>
#else
# pragma interface
#endif

// All the externs (from fromsage++.C) used in fromsage++.h

class SgProject;
class SgFile;
class SgStatement;
class SgExpression;
class SgLabel;
class SgSymbol;
class SgType;
class SgUnaryExp;
class SgClassSymb;
class SgVarDeclStmt;
class SgVarRefExp; /* ajm: I think they should all be here! @$!@ */

extern void **tablebfnd[];
extern void **tablellnd[];
extern void **tabletype[];
extern void **tablesymbol[];
extern void **tablelabel[];
 
extern int numtablebfnd[];
extern int numtablellnd[];
extern int numtabletype[];
extern int numtablesymbol[];
extern int numtablelabel[];
 
extern void **fileTableClass;
extern int allocatedForfileTableClass;
extern void **bfndTableClass;
extern int allocatedForbfndTableClass;
extern void **llndTableClass;
extern int allocatedForllndTableClass;
extern void **typeTableClass;
extern int allocatedFortypeTableClass;
extern void **symbolTableClass;
extern int allocatedForsymbolTableClass;
extern void **labelTableClass;
extern int allocatedForlabelTableClass;
extern SgProject *CurrentProject;

// Prototype definitions  for all the functions in fromsage++.C

void SwitchToFile(int i);
void ReallocatefileTableClass();
void ReallocatebfndTableClass();
void ResetbfndTableClass();
void ReallocatellndTableClass();
void ReallocatesymbolTableClass();
void ReallocatelabelTableClass();
void ReallocatetypeTableClass();
void RemoveFromTableType(void * pt);
void RemoveFromTableSymb(void * pt);
void RemoveFromTableBfnd(void * pt);
void RemoveFromTableFile(void * pt);
void RemoveFromTableLlnd(void * pt);
void RemoveFromTableLabel(void * pt);
void SetMappingInTableForBfnd(PTR_BFND bif, void *pt);
void SetMappingInTableForType(PTR_TYPE type, void *pt);
void SetMappingInTableForSymb(PTR_SYMB symb, void *pt);
void SetMappingInTableForLabel(PTR_LABEL lab, void *pt);
void SetMappingInTableForLlnd(PTR_LLND ll, void *pt);
void SetMappingInTableForFile(PTR_FILE file, void *pt);
SgSymbol *GetMappingInTableForSymbol(PTR_SYMB symb);
SgLabel *GetMappingInTableForLabel(PTR_LABEL lab);
SgStatement *GetMappingInTableForBfnd(PTR_BFND bf);
SgStatement *GetMappingInTableForBfnd(PTR_BFND bf);
SgType *GetMappingInTableForType(PTR_TYPE t);
SgExpression *GetMappingInTableForLlnd(PTR_LLND ll);
SgFile *GetMappingInTableForFile(PTR_FILE file);
SgStatement * BfndMapping(PTR_BFND bif);
SgExpression * LlndMapping(PTR_LLND llin);
SgSymbol * SymbMapping(PTR_SYMB symb);
SgType * TypeMapping(PTR_TYPE ty);
SgLabel * LabelMapping(PTR_LABEL label);


#define SORRY Message("Sorry, not implemented yet",0)

/* includes the attributes data structure */

// 
// Defines the data structure for attributes in sage
// attributes can be used to store any information for any statement, expression, symbol or types nodes
// F. Bodin Indiana July 94.
// 


/** @name the Basic Class
The basic classes are the foundation of \Ref{ the Derived Class} ,which derive 
from the 8 basic classes.  As above,the 8 basic classes are \Ref{SgAttribute}, \Ref{SgProject}, \Ref{SgFile}, \Ref{SgStatement},
\Ref{SgExpression}, \Ref{SgSymbol}, \Ref{SgType} and \Ref{SgLabel}.

*/

//@{
	
/**It is possible to add information to the Sage nodes using attributes. An attribute has two values: 

 1.a type: user supplied types must be positive integers; negative integers are reserved for the system; the reserved types
\begin{verbatim} 
  are:

        #define DEPENDENCE_ATTRIBUTE -1001
        #define INDUCTION_ATTRIBUTE -1002
        #define ACCESS_ATTRIBUTE -1003 
        #define DEPGRAPH_ATTRIBUTE -1004 
        #define USEDLIST_ATTRIBUTE -1005 
        #define DEFINEDLIST_ATTRIBUTE -1006 
        #define NOGARBAGE_ATTRIBUTE -1007 
        #define GARBAGE_ATTRIBUTE -1008 
        #define ANNOTATION_EXPR_ATTRIBUTE -1009 

 2.data: the data field is a pointer to void that can be used to store any information. 
\end{verbatim}
@memo Represents the base class of attributes.
*/
class SgAttribute{
 private:
  /// a label;
  int type; 

  /// the data;
  void *data;  

  /// the size of the data in bytes to allow data to be copied;
  int dataSize; 

  /// to the next attribute of a statements (do that way or not??);
  SgAttribute *next; 

  // link to sage node, allow to go from an attribute to sage stuffs;

  /// indicates if SgStatement, SgExpression, ... ptToSage is pointed to;
  typenode typeNode;   
  
  /// pointer to SgStatement, SgExpression, ... ;
  void *ptToSage; 

  /// the file methods;
  int fileNumber; 

public:
  /**Constructor for SgStatement attribute. The parameter t is the type of the attribute, pt a pointer to the data and size the size of the data (in bytes). This last parameter is useful if no function is given for reading and saving the attributes. The parameter st is the statement the attribute is attached to. When an attribute is created this way, it does not appear in the attributes list of the SgStatement node, but it can be added using the addAttribute(SgAttribute *att) method.
  */ 
  SgAttribute(int t, void *pt, int size, SgStatement &st, int filenum);
  ///Constructor for SgSymbol attribute.
  SgAttribute(int t, void *pt, int size, SgSymbol &st, int filenum);
  ///Constructor for SgExpression attribute.
  SgAttribute(int t, void *pt, int size, SgExpression &st, int filenum);
  ///Constructor for SgType attribute. 
  SgAttribute(int t, void *pt, int size, SgType &st, int filenum);
  ///
  ~SgAttribute();
  ///Get the type of an attribute.
  int getAttributeType();
  ///Set the type of an attribute.
  void setAttributeType(int t);
  ///Get the data field of an attribute. 
  void *getAttributeData();
  ///Set the data field of an attribute.
  void *setAttributeData(void *d);
  ///Get the size of the data field of an attribute. 
  int getAttributeSize();
  ///Set the size of the data field of an attribute.
  void setAttributeSize(int s);
  /**Get the type of the node the attribute is attached to. The type is an enumerate: 

       enum typenode {BIFNODE, LLNODE, SYMBNODE, TYPENODE,...} 

   corresponding respectively to SgStatement, SgExpression, SgSymbol, and SgType. 
  */
  typenode getTypeNode();
  /**Return a pointer to sage node the attribute is attached to. According to getTypeNode(), this value can be casted to the right type. 
  */
  void *getPtToSage(); 
  ///Set the pointer to the Sage node.
  void  setPtToSage(void *sa);
  ///Set the pointer to the Sage node to NULL. 
  void  resetPtToSage();
  ///Set the pointer to the Sage node and the typeNode to be BIFNODE. 
  void  setPtToSage(SgStatement &st);
  ///idem for SgSymbol.
  void  setPtToSage(SgSymbol &st);
  ///idem for SgExpression.
  void  setPtToSage(SgExpression &st);
  ///idem for SgType.
  void  setPtToSage(SgType &st);
  /**Return the SgStatement pointer if it exists and if the attribute is attached to an SgStatement node. Otherwise this function returns NULL.*/
  SgStatement *getStatement();
  ///idem for SgExpression.
  SgExpression *getExpression();
  ///idem for SgSymbol.
  SgSymbol  *getSgSymbol();
  ///idem for SgType.
  SgType  *getType();
  ///returns the fileNumber.
  int getfileNumber();
  ///Get a copy of an attribute. This function does not copy the data field. 
  SgAttribute *copy(); 
  ///The attributes attached to the same node are chained through the next value.
  SgAttribute *getNext();
  ///Set the list chaining of an attribute.
  void setNext(SgAttribute *s);
  ///return the number of SgAttribute nodes linked from this one.
  int listLenght();
  ///return the SgAttribute node numbered "num" from this node.
  SgAttribute *getInlist(int num);
  ///save the private field to the file.
  void save(FILE *file);
  ///
  void save(FILE *file, void (*savefunction)(void *dat,FILE *f));  
};

//
// The ATTRIBUTE TYPE ALREADY USED
//

#define DEPENDENCE_ATTRIBUTE   -1001
#define INDUCTION_ATTRIBUTE    -1002
#define ACCESS_ATTRIBUTE       -1003
#define DEPGRAPH_ATTRIBUTE     -1004
#define USEDLIST_ATTRIBUTE     -1005
#define DEFINEDLIST_ATTRIBUTE  -1006

#define NOGARBAGE_ATTRIBUTE    -1007
#define GARBAGE_ATTRIBUTE      -1008

// store the annotation expression; it is then visible from the
// garbage collection

#define ANNOTATION_EXPR_ATTRIBUTE   -1009

/**************************************************************

Purpose:It contains all the class definitions and the inline
definitions in Sage++. The start of each class and the start of
inlines in each class are easily identifiable.

For example the SgProject class definition starts with class SgProject
(note the 2 spaces between class and SgProject) and the comment line
preceding the inline declarations of SgProject is something like //
SgProject--inlines.

Sections of the include file are within a #ifdef USER #endif. Those
sections are included only in sage++user.h and not in libSage++.C.

Sections of the include file are within a #if 0 #endif.  These refer
to the unimplemented portions of Sage++ library.

***************************************************************/

/**A project is a group of parsed source files (.dep files) making up a program that is to be manipulated by Sage++. Currently, Sage++ can only deal with one project at a time. If multiple files in a project are desired, please see section \Ref{Attributes}. On disk, a project is represented as a file with name ending in .proj, containing a list of names of the component .dep files, one file name per line. A file is one of the parsed source files in a project. Each .dep file contains a parse tree. The root of the tree for a file is called the global node and its immediate children are the top level definitions and functions in the files. The file also contains a symbol table and type table. 
@memo Represent the files in the current project, for all languages.
*/
class  SgProject {
public:
  ///
  SgProject(SgProject &);
  /**Sage++ is initialized by instantiating an SgProject with this constructor.
     'proj\_name' is a string containing the name of the .proj file that 
      corresponds to the program to be manipulated.
  */     
  SgProject(char * proj_name, int argc, char * argv[]);
  
  ///destructor
  ~SgProject();
  
  ///Returns the number of files in the project.
  int numberOfFiles();
  
  ///Returns a reference to the 'ith' file in the project.
  SgFile &file(int i);   
  
  ///Return the name of the 'ith' file in the project.
  char *fileName(int i); 
  
  ///Returns nonzero if every file in the project is a Fortran file. 
  int Fortranlanguage(); 
  
  ///Returns nonzero if at least one file in the project is not a Fortran file.
  int Clanguage();                         
  ///Currently unimplemented.
  void addFile(char * dep_file_name);
  ///Currently unimplemented.
  void deleteFile(SgFile * file);
};

/**This class provides access to the local symbol and type tables,and the top level definition,in the
   dep file it corresponds to. Represents files,for all languages.
*/

class  SgFile{
public:
  ///
  PTR_FILE filept;
  
  /**Provides access to (opens) the existing file, normally created by the parser. */
  SgFile(char* file_name); // the file must exist.
  
  /**This constructor allows the user to create an empty dep file in order to insert statements into it later on */
  SgFile(int Language, char* file_name); // for new empty file objects.
  
  ///destructor							       
  ~SgFile();
  
  ///
  SgFile(SgFile &);
  
  ///Returns CSrc(for C) or ForSrc(for Fortran). 
  int languageType();
  
  ///Saves the(internal representation of the)file as a dep file with filename dep\_file.
  void saveDepFile(char *dep_file);
  
  ///Unparses the (internal representation of the)file and saves the result in the file with file descriptor filedisc
  void unparse(FILE *filedisc);
  
  ///Unparses the (internal representation of the)file to standard output. 
  void unparsestdout();   
  
  ///Returns the statement in the file at which the main program in located,or NULL if the main program is not located in this file.
  SgStatement *mainProgram();
  
  ///Returns the ith function in the file.
  SgStatement *functions(int i);
  
  ///Returns the number of functions in the file.
  int numberOfFunctions();
  
  ///Returns the ith structure in the file.
  SgStatement *getStruct(int i);
  
  ///Returns the number of structures in the file.
  int numberOfStructs();
  
  ///Returns the first statement in the file.
  SgStatement *firstStatement();
  
  ///Returns the first symbol in the file's symbol table.
  SgSymbol *firstSymbol();
  
  ///Returns the first type in the file's type table.
  SgType *firstType();
  
  ///Returns the first expression in the file.
  SgExpression *firstExpression();
  
  ///Returns the first label in the file.
  SgLabel * firstLabel();     
  
  ///Returns the expression in the file with id i.
  SgExpression *SgExpressionWithId(int i);
  
  ///Returns the expression in the file with id i.
  SgStatement *SgStatementWithId( int id);
  
  ///Returns the statement at line 'lineno' in the file.
  SgStatement *SgStatementAtLine(int lineno);
  
  ///Returns the symbol in the file with id 'id'.
  SgSymbol *SgSymbolWithId( int id);
  
  ///Returns the type in the file with id 'id'.
  SgType *SgTypeWithId( int id);

  // for attributes;
 
  /// 
  void saveAttributes(char *file);
  ///
  void saveAttributes(char *file, void  (*savefunction)(void *dat,FILE *f)); 
  ///
  void readAttributes(char *file);
  ///
  void readAttributes(char *file, void * (*readfunction)(FILE *f));
  ///
  int numberOfAttributes();
  ///
  SgAttribute *attribute(int i);
  ///
  int SgFile::expressionGarbageCollection(int deleteExpressionNode, int verbose);
};

// Discuss about control parent, BIF structure etc

/**Files are broken down into statements, 'SgStatement' is the base class for the representation of 
statements. SgStatement has mant derived classes,corresponding to the many different kinds of 
statements in the languages Sage++ can deal with. Each statement has a unique integer idetifier id,
and an integer variant tag which can be used to idetify its type. These are available through 
SgStatement's 'id()' and 'variant()' member functions. SgStatement's 'thebif' member provides access 
to the low level representation of statements, bif nodes. Each statement has a context: it may have 
a lexical predecessor and a lexical successor, and it may be nested within a control block of another
 statement, or in a block structured definition like a C struct. This enclosing statement is called 
the control parent; it defines the structure of the parse tree.
@memo Represents the fundamental class for all statements.
*/

class  SgStatement{
public:

  PTR_BFND thebif;

  ///Creates a SgStatement with variant variant.
  SgStatement(int variant); 
  ///Returns an SgStatement corresponding to the bif node bif.
  SgStatement(PTR_BFND bif);
  ///
  SgStatement(int code, SgLabel *lab, SgSymbol *symb, SgExpression *e1, SgExpression *e2, SgExpression *e3);
  ///
  SgStatement(SgStatement &);
  
  // info about statement

  ///Returns the line number of the statment in the original source file.
  int lineNumber();          

  ///Returns the id of the statement.
  int id();                  

  ///Returns the statement's variant, which can be used to determine what kind of statement it actually is.
  int variant(); 

  /** i = 0,1,2 returns the i-th expression. There can be up to three expressions associated with any statement. For example, the C 'for' loop is of the form 'for(expr\_1;expr\_2;expr\_3)'.*/
  SgExpression *expr(int i); 
  
  ///Returns non-zero if there is a symbol associated with the statement. 

  int hasSymbol();

  /**If there is a symbol associated with the statement, this function returns it. This symbol is used for the name of a subroutine in a subroutine header statement, and, in Fortran only, for the induction variable of a DO loop.
   */

  SgSymbol *symbol();        

  ///Returns the name of the enclosing file.
  char *fileName();
            
  ///Returns non-zero if there is a label associated with the statement.
  int hasLabel();

  ///If the statement has a label, this function returns it.
  SgLabel *label();          

  // Modifying the info

  ///Sets the line number of the statement.
  void setlineNumber(int n); 

  ///Sets the id of the statement.
  void setId(int n);         

  ///Sets the statement's variant.
  void setVariant(int n);    

  ///Set's the statement's ith expression to e.
  void setExpression (int i, SgExpression &e); 

  ///Sets the statement's label. 
  void setLabel(SgLabel &l);

  ///Sets the statement's label.
  void setLabel(SgLabel * l); 

  ///Sets the statement's symbol.
  void setSymbol(SgSymbol &s); 
  
  //@Man: Control structure

  ///Returns the statement's lexical successor statement. 
  SgStatement *lexNext();  

  ///Returns the statement's lexical predecessor statement.
  SgStatement *lexPrev();   

  ///Returns the statement of this statement's control parent. 
  SgStatement *controlParent(); 

  ///Sets this statement's lexical successor statement.
  void setLexNext(SgStatement &s);

  ///Sets this statement's control parent statement.
  void setControlParent(SgStatement &s); 

  // change the control parent of the statement.  
  void setControlParent(SgStatement *s); 
  
  // Access statement using the tree structure

  // Describe BLOB lists here?
 
  ///Returns the number of children in this statement's first list of child statements.
  int numberOfChildrenList1();
  ///Returns the number of children in this statement's second list of child statements.
  int numberOfChildrenList2();
  ///Returns the ith statement in this statement's first list of child statements.
  SgStatement *childList1(int i);
  ///Returns the ith statement in this statement's second list of child statements.
  SgStatement *childList2(int i);
  ///Returns the next statement in this statement's child list.
  SgStatement *nextInChildList();

  /// pass all the statement that is IMPLICIT and use and PARM\_DECL
  SgStatement *firstStmtForDeclaration();   

  ///If the statement is not a leaf of the parse tree, this function returns a pointer to the last declaration in the subtree the statement represent.
  SgStatement *lastDeclaration(); 
  
  /// 
  SgStatement *firstExecutable();
  /**If the statement is not a leaf of the parse tree, this function returns the last executable statement in the subtree the statement defines. This method is not useful in C++.*/ 
  SgStatement *lastExecutable();  
  
  ///If the statement is not a leaf of the parse tree, this function returns a pointer to the last statement in the subtree defines. 
  SgStatement *lastNodeOfStmt();
  
  ///
  SgStatement *nodeBefore();
  ///Inserts statement s before this one.
  void insertStmtBefore(SgStatement &s);
  ///Inserts statement s before this one. Set's s's control parent to cp.
  void insertStmtBefore(SgStatement &s, SgStatement &cp); 
  ///Inserts statement s after this one. 
  void insertStmtAfter(SgStatement &s);
  ///Inserts statement s after this one. Set's s's control parent to cp.
  void insertStmtAfter(SgStatement &s, SgStatement &cp);
  ///Removes this statement from the parse tree, and returns a pointer to the removed statement.
  SgStatement *extractStmt();
  /**If the statement is not a leaf of the parse tree, this function extracts the subtree it defines from the parse tree, and returns a pointer to the first statement in the tree.*/
  SgStatement *extractStmtBody();
  ///Replaces this statement with statement s.
  void replaceWithStmt(SgStatement &s);
  ///
  void simpledeleteStmt();
  ///Removes this statement from the parse tree, and frees the memory associated with it.
  void deleteStmt();
  ///
  SgStatement  &copy (void);
  ///
  SgStatement  *copyPtr (void);
  ///
  SgStatement  &copyOne (void);
  ///
  SgStatement  *copyOnePtr (void);
  ///
  SgStatement  &copyBlock (void);
  ///
  SgStatement  *copyBlockPtr (void);
  ///
  int isIncludedInStmt(SgStatement &s);
  ///
  void replaceSymbByExp(SgSymbol &symb, SgExpression &exp);
  ///
  void replaceSymbBySymb(SgSymbol &symb, SgSymbol &newsymb);
  ///
  void replaceSymbBySymbSameName(SgSymbol &symb, SgSymbol &newsymb);
  ///
  void replaceTypeInStmt(SgType &old, SgType &newtype);

  ///
  char* unparse();

  ///
  void unparsestdout();

  /// unparsing functions
  void sunparse(char *buffer);
 
  ///Preceeding comments are appended to the right of the statement.
  char *comments();     
  ///
  void addComment(char *com);

  /// setComments: set ALL of the node's comments  
  void setComments (char *comments);
  
  /// number of preceeding comments.
  int numberOfComments(); // CAREFUL! 
  
  /// 1 if there are annotations; 0 otherwise
  int hasAnnotations();   
  
  ///
  ~SgStatement();

  //  The following functions are added for compiler to look up and build symbol;

  ///
  SgSymbol * SgStatement::GetSymbolInScope(SgStatement * scope, char * name);
  ///
  SgSymbol * SgStatement::GetSymbolInScope(SgStatement * scope, char * mixstring, char * genname, char * basename); 
  ///
  SgSymbol * SgStatement::GetSymbolInScope(SgStatement * scope, char * mixstring, char * genname, char * basename, int counter);
  ///
  SgSymbol * SgStatement::GetSymbolInScope(SgStatement * scope, char * mixstring, char * genname, char * basename, int c1, int c2);

  ///
  SgSymbol * SgStatement::getSymbolInCurrentScope(char * name);
  ///
  SgSymbol * SgStatement::getSymbolInCurrentScope(char * mixstring, char * genname, char * basename);
  ///
  SgSymbol * SgStatement::getSymbolInCurrentScope(char * mixstring, char * genname, char * basename, int counter);
  ///
  SgSymbol * SgStatement::getSymbolInCurrentScope(char * mixstring, char * genname, char * basename, int c1, int c2);

  //  These function must be removed. Doesn't make sense here.
  int IsSymbolInScope(SgSymbol &symb); // TRUE if symbol is in scope
  int IsSymbolReferenced(SgSymbol &symb);

  /// return where a variable can be declared
  SgStatement *getScopeForDeclare(); 
  
  // FOR ATTRIBUTES 

  ///
  int numberOfAttributes();

  ///
  int numberOfAttributes(int type); //@ManMemo: of a specified type;
  
  ///
  void *attributeValue(int i); 

  ///
  int attributeType(int i); 

  /// only considering one type attribute
  void *attributeValue(int i,int type); 


  /// void * can be NULL;
  void addAttribute(int type, void *a, int size);

  /// void * is NULL;
  void addAttribute(int type); 

  /// no type specifed
  void addAttribute(void *a, int size); 

  ///
  void addAttribute(SgAttribute *att);

  ///
  void *deleteAttribute(int i); 

  ///
  void *deleteAttribute(int i,int type); 
  ///
  void deleteAllAttribute(); 
  ///
  void deleteAllAttribute(int type); 

  ///
  SgAttribute *getAttribute(int i);
  ///
  SgAttribute *getAttribute(int i,int type);

  // FOR DECL_SPECS (friend, inline, extern, static)
  
  /// type should be one of BIT\_EXTERN BIT\_INLINE, \_FRIEND, BIT\_STATIC
  void addDeclSpec(int type);  

  /// resets the decl\_specs field to zero
  void clearDeclSpec();      

  /**returns non-zero if friend modifier set
  returns zero otherwise
  */
  int isFriend();               

  ///
  int isInline();
  ///
  int isExtern();
  ///
  int isStatic();
  
};

/**Many statements contain expressions. Unlike the SgStatement class hierarchy, there is not a subclass for every expression operator type. Basic binary operators such as the standard arithmetic operators are identified only the integer variant returned by the SgExpression member function variant(). The expression nodes that have their own subclass do so because they have a special type of constructor or special fields.  SgExpression is the base class for all the other expression classes. Every expression may have up to two operands (or subexpressions),the left hand side 'lhs()' and the right hand side 'rhs()'.Each expression has a type 'type()',may have a symbol 'symbol()',and has a unique identifier 'id()'.
@memo SgExpression is the base class for all the other expression classes.
*/

class  SgExpression{    
public:
  PTR_LLND thellnd;

  /// generic expression class.

  SgExpression(int variant, SgExpression &lhs, SgExpression &rhs, SgSymbol &s, SgType &type);

  /// 
  SgExpression(int variant, SgExpression *lhs, SgExpression *rhs, SgSymbol *s, SgType *type);

  ///
  SgExpression(int variant, SgExpression *lhs, SgExpression *rhs, SgSymbol *s);

  /// for some node in fortran
  SgExpression(int variant,char *str);

  ///
  SgExpression(int variant);
  ///
  SgExpression(PTR_LLND ll);

  ///
  SgExpression(SgExpression &);
  ///
  ~SgExpression();
  
  ///
  SgExpression *lhs();
  ///
  SgExpression *rhs();
  /**Valid for i = 0, 1. In Sage++ versions 1.7 and earlier i = 1, 2. Starting with Sage++ version 1.9 all arguments that play the role of indices start with zero.*/
  SgExpression *operand(int i);
  ///
  int variant();      
  ///
  SgType *type(); 
  ///
  SgSymbol *symbol();
  ///
  int id();
  ///
  SgExpression *nextInExprTable();
  
  ///
  void setLhs(SgExpression &e);
  ///
  void setLhs(SgExpression *e);
  ///
  void setRhs(SgExpression &e);
  ///
  void setRhs(SgExpression *e);
  ///
  void setSymbol(SgSymbol &s);
  ///
  void setSymbol(SgSymbol *s);
  ///
  void setType(SgType &t);
  ///
  void setType(SgType *t);
  ///
  void setVariant(int v);
  
  ///Returns a deep copy.
  SgExpression &copy();
  ///
  SgExpression *copyPtr();
  ///
  char *unparse(); 
  ///
  void sunparse(char *buffer);
  ///
  void unparsestdout();
  
  ///Returns the most nested expression containing a reference to symbol.
  SgExpression *IsSymbolInExpression(SgSymbol &symbol);
  ///
  void replaceSymbolByExpression(SgSymbol &symbol, SgExpression &expr);
  
  /// get the name refered
  SgSymbol *getVariableRefName();         

  ///
  SgExpression *symbRefs();
  
  ///Returns a list of array reference in the expression.
  SgExpression *arrayRefs(); 
  
  /// 
  int  linearRepresentation(int *coeff, SgSymbol **symb,int *cst, int size);
  
  ///Does a normal form simplify as a linear combination of the symbols in s.
  SgExpression *normalForm(int n, SgSymbol *s);        
  
  ///Finds the multiplicative coeffcient of s in the expression.
  SgExpression *coefficient(SgSymbol &s); 
  
  ///TRUE if the expression can be reduced to an integer value.
  int isInteger();
  ///If isInteger() is true,returns the integer value the expression can be reduced to.
  int valueInteger();

  /// -1 : incorrect
  int valueInteger(int &result); // by zheng.gb
  
  ///
  void replace(SgExpression & e);
  
  ///
  friend SgExpression * operator + ( SgExpression *lhs, SgExpression &rhs); 
  ///
  friend SgExpression * operator * ( SgExpression *lhs, SgExpression &rhs); 
  
  
  ///
  friend SgExpression &operator + ( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &operator - ( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &operator * ( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &operator / ( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &operator % ( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &operator <<( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &operator >>( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &operator < ( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &operator > ( SgExpression &lhs, SgExpression &rhs);
  ///
  friend SgExpression &operator <= ( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &operator >= ( SgExpression &lhs, SgExpression &rhs);
  ///
  friend SgExpression &operator & ( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &operator | ( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &operator &&( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &operator ||( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &operator +=( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &operator &=( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &operator *=( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &operator /=( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &operator %=( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &operator ^=( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &operator <<=( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &operator >>=( SgExpression &lhs, SgExpression &rhs);
  ///
  friend SgExpression &SgAssignOp( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &SgEqOp( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &SgNeqOp( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &SgExprListOp( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &SgRecRefOp( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &SgPointStOp( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &SgScopeOp( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &SgDDotOp( SgExpression &lhs, SgExpression &rhs); 
  ///
  friend SgExpression &SgBitNumbOp( SgExpression &lhs, SgExpression &rhs); 
  
  // ///////////// FOR ATTRIBUTES //////////////////////////

  ///
  int numberOfAttributes();

  /// of a specified type
  int numberOfAttributes(int type); 

  ///
  void *attributeValue(int i); 
  ///
  int  attributeType(int i); 
  /// only considering one type attribute
  void *attributeValue(int i,int type); 

  /// 
  void *deleteAttribute(int i); 

  /// void * can be NULL
  void addAttribute(int type, void *a, int size); //@ManMemo: void * can be NULL;

  /// void * is NULL
  void addAttribute(int type); 

  /// no type specifed
  void addAttribute(void *a, int size); 

  ///
  void addAttribute(SgAttribute *att);
  ///
  SgAttribute *getAttribute(int i);
  ///
  SgAttribute *getAttribute(int i,int type);
};


/**SgSymbol is the base class for the classes representing symbols. As with the other classes there
   is a 'variant()' which idetifies the subclass and an integer identifier 'id()'that is unique for 
   each symbol. The symbols in each file are organized as a list which is accessible from the file
   object as 'SgFile::firstSymbol()' and the following symbol to the current symbol is located with
   'next()'. Every symbol has a character string 'idetifier()' , a type()',which is described in 
   detail below,and a statement called 'scope()' that is the statement in which declaration is 
   scoped.(It is the control parent of the declaration statement). The statement where the variable
   is given by 'declaredInStmt()'. In addtion, there are functions that can be used to generate a 
   copy of the synbol. There are three forms of the copy. The basic copy make a simple copy of the 
   symbol table entry. The level one copy also generate new type information. The level two copy
   also copies the declaration body.
*/

class SgSymbol{
public:
  PTR_SYMB thesymb;

  ///
  SgSymbol(int variant, char *identifier, SgType &t, SgStatement &scope);
  ///
  SgSymbol(int variant, char *identifier, SgType *t, SgStatement *scope);
  ///
  SgSymbol(int variant, char *identifier,  SgStatement &scope);
  ///
  SgSymbol(int variant, char *identifier,  SgStatement *scope);

  ///
  SgSymbol(int variant, char *identifier,   SgType *type, SgStatement *scope, SgSymbol *structsymb, SgSymbol *nextfield );

  ///
  SgSymbol(int variant, char *name);
  ///
  SgSymbol(int variant);
  ///
  SgSymbol(PTR_SYMB symb);
  ///
  SgSymbol(SgSymbol &);
  ///
  ~SgSymbol();
  ///
  int variant();  

  /// unique identifier
  int id();             

  /// the text name for the symbol
  char *identifier();   

  /// the type of the symbol
  SgType *type(); 

  /// the type of the symbol
  void setType(SgType &t);       

  void setType(SgType *t);       

  /// the SgControlStatement where defined
  SgStatement *scope();

  // next symbol reference
  SgSymbol *next();
     
  // next symbol reference in the same scope
  SgSymbol *nextInScope();

  ///set next\_symb field
  void    setNextInScope(SgStatement *s);     

  /**This method is supposed to return a pointer to the statement where the given symbol has been declared. However, in its current implementation, the method is very slow (it does a linear search through the list of statements every time it is called) and unreliable. Often, it is better to write an extra pass through statements and attach the information about the declaration statements to SgSymbols using an SgAttribute so that, at later times, this information can be simply looked up rather than recomputed every time it is needed. 
*/
  SgStatement *declaredInStmt(); 

  ///
  SgSymbol &copy();

  /// copy also parameters
  SgSymbol &copyLevel1(); 

  // copy parameters, body also
  SgSymbol &copyLevel2(); 

  /// special copy to move things from a file
  SgSymbol &copyAcrossFiles(SgStatement &where);

  ///  the Fortran 90 attributes
  int attributes();    

  ///
  void setAttribute(int attribute);
  ///
  void removeAttribute(int attribute);

  /**This method builds a declaration statement for the given symbol and inserts it at the top of the scope given by st.
*/
  void declareTheSymbol(SgStatement &st);
  ///Similar to declareTheSymbol() but takes a second argument that is a parameter list (for function prototypes). 
  void declareTheSymbolWithParamList(SgStatement &st, SgExpression &parlist);
  ///Constructs a declaration expression and returns it.
  SgExpression *makeDeclExpr();
  ///Similar to makeDeclExpr() but takes a second argument that is a parameter list (for function prototypes).
  SgExpression *makeDeclExprWithParamList(SgExpression &parlist);
  ///Constructs a declaration statement and returns it (without inserting it into the file)
  SgVarDeclStmt *makeVarDeclStmt();
  ///Similar to makeVarDeclStmt() but takes a second argument that is a parameter list (for function prototypes) 
  SgVarDeclStmt *makeVarDeclStmtWithParamList(SgExpression &parlist);
 
  ///If the symbol refers to a function, struct, subroutine or class, this returns a pointer to the SgStatement object that is the declaration of the symbol. 
  SgStatement *body(); 

  /// if the symbol is of assumed shape spec
  int isAssumedShapeSymbol();      

  /// if the symbol is of assumed size spec
  int isAssumedSizeSymbol();      

  /// if the symbol is of defered shape array
  int isDeferedShapeSymbol();                       // by zheng.gb

  ///
  SgStatement *declareInStmt(SgStatement *hedr_stmt);


  //  ///////////// FOR ATTRIBUTES //////////////////////////
  ///
  int numberOfAttributes();
  ///
  int numberOfAttributes(int type); 

  ///
  void *attributeValue(int i); 
  ///
  int  attributeType(int i); 
  ///
  void *attributeValue(int i,int type); 
  /// 
  void *deleteAttribute(int i); 

  ///
  void *deleteAttribute(int i,int type); 

  ///
  void addAttribute(int type, void *a, int size); 

  ///
  void addAttribute(int type); 
  ///
  void addAttribute(void *a, int size);
  ///
  void addAttribute(SgAttribute *att);
  ///
  SgAttribute *getAttribute(int i);
  ///
  SgAttribute *getAttribute(int i,int type);
};


/* This code by Andrew Mauer (ajm) */
/* These constants are used by SgType::maskDescriptors() and 
   SgType::getTrueType(). */

const int MASK_NO_DESCRIPTORS = ~0; /* all ones = keep everything */
const int MASK_MOST_DESCRIPTORS =  ( BIT_SIGNED | BIT_UNSIGNED
				     | BIT_LONG | BIT_SHORT
				     | BIT_CONST | BIT_VOLATILE );

const int MASK_ALL_DESCRIPTORS = 0; /* keep nothing */

/**SgType is the base class for all classes that represent types. As with symbols, types in a file form a list that can be accessed with the function SgFile::firstType(). The next() function returns the type node following this. The function copy() returns a copy of the type node. Many types are defined in terms of other types. For example, an array type has a base type that may be a pointer type which has a base type that may be an integer or a class, etc. The baseType() function always access the next type in the chain. 
*/

class SgType{
public:
  ///
  PTR_TYPE thetype;
  ///
  SgType(int variant);
  ///
  SgType(int var, SgExpression *len,SgType *base);
  ///
  SgType(int var, SgSymbol *symb);
  ///
  SgType(int var, SgSymbol *firstfield, SgStatement *structstmt);
  ///
  SgType(PTR_TYPE type);
  ///
  SgType(SgType &);
  ///
  int variant();
  ///
  int id();
  ///
  SgSymbol *symbol();
  ///
  SgType &copy();
  ///
  SgType *copyPtr();
  ///
  SgType *next();
  ///
  int isTheElementType();
  ///
  int equivalentToType(SgType &type);
  ///
  int equivalentToType(SgType *type);
  ///
  SgType *internalBaseType();
  ///
  int hasBaseType();
  ///
  SgType *baseType();
  ///
  SgExpression *length();


  /*  This code by Andrew Mauer (ajm) */
  /**
  This routine strips many descriptive type traits which you are probably
  not interested in cloning for variable declarations, etc.

  Returns the getTrueType of the base type being described IF there
  are no descriptors which are not masked out. The following masks
  can be specified as an optional second argument:
        MASK\_NO\_DESCRIPTORS: Do not mask out anything.
        MASK\_MOST\_DESCRIPTORS: Only leave in: signed, unsigned, short, long,
	                                      const, volatile.
	MASK\_ALL\_DESCRIPTORS: Mask out everything. 

  If you build your own mask, you should make sure that the traits
  you want to set out have their bits UN-set, and the rest should have
  their bits set. The complementation (~) operator is a good one to use.

  See libSage++.h, where the MASK\_\*\_DESCRIPTORS variables are defined.

  */

  SgType *maskDescriptors (int mask);

  /* This code by Andrew Mauer (ajm) */
  /**
  Since Sage stores dereferenced pointers as PTR(-1) -> PTR(1) -> BASE\_TYPE,
  we may need to follow the chain of dereferencing to find the type
  which we expect.

  This code currently assumes that:
  o If you follow the dereferencing pointer (PTR(-1)), you find another
  pointer type or an array type. 

  We do NOT assume that the following situation cannot occur:
      PTR(-1) -> PTR(-1) -> PTR(1) -> PTR(1) -> PTR(-1) -> PTR(1)

  This means there may be more pointers to follow after we come to
  an initial "equilibrium".

  ALGORITHM:

  T\_POINTER:
     [WARNING: No consideration is given to pointers with attributes
     (ls\_flags) set. For instance, a const pointer is treated the same
     as any other pointer.]
     
     1. Return the same type we got if it is not a pointer type or
     the pointer is not a dereferencing pointer type.

     2. Repeat { get next pointer , add its indirection to current total }
     until the current total is 0. We have reached an equilibrium, so
     the next type will not necessarily be a pointer type.

     3. Check the next type for further indirection with another call
     to getTrueType.

  T\_DESCRIPT:
     Returns the result of maskDescriptors called with the given type and mask.

  T\_ARRAY:
     If the array has zero dimensions, we pass over it. This type arose
     for me in the following situation:
          double x[2];
	  x[1] = 0;
     
  T\_DERIVED\_TYPE:
     If we have been told to follow typedefs, get the type of the
     symbol from which this type is derived from, and continue digging.
     Otherwise return this type.


  HITCHES:
     Some programs may dereference a T\_ARRAY as a pointer, so we need
     to be prepared to deal with that.
  */
  SgType *getTrueType (int mask = MASK_MOST_DESCRIPTORS, int follow_typedefs = 0);

  ///
  int numberOfAttributes();
  ///
  int numberOfAttributes(int type);

  ///
  void *attributeValue(int i); 
  ///
  int  attributeType(int i); 
  ///
  void *attributeValue(int i,int type);
  ///
  void *deleteAttribute(int i); 
  ///
  void addAttribute(int type, void *a, int size); 
  ///
  void addAttribute(int type); 
  ///
  void addAttribute(void *a, int size); 
  ///
  void addAttribute(SgAttribute *att);
  ///
  SgAttribute *getAttribute(int i);
  ///
  SgAttribute *getAttribute(int i,int type);
};


//
// SgMakeDeclExp can be called by the user to generate declaration
// expressions from type strings.  it handles all C types.

SgExpression *SgMakeDeclExp(SgSymbol *sym, SgType *t);


/// Represents labels for all languages
class SgLabel{
public:
  ///
  PTR_LABEL thelabel;
  ///
  SgLabel(PTR_LABEL lab);
  ///
  SgLabel(SgLabel &);
  ///
  SgLabel(int i);
  ///
  int id();
  ///
  int labelValue();
  ///
  SgLabel * nextInScope(SgStatement *);
  ///
  int getLastLabelVal();
  ///
  ~SgLabel();
};


// -----------------------------------;

//@}

  /** @name the Derived Class 
the derived classes,which provide operation on the \Ref{Intermediate representation}, inherit from \Ref{the Basic Class}.
  */
//@{


/// a value of one of the base types. variants: INT\_VAL, CHAR\_VAL, FLOAT\_VAL, DOUBLE\_VAL, STRING\_VAL, COMPLEX\_VAL, KEYWORD\_VAL
class  SgValueExp: public SgExpression{
public:
  ///
  SgValueExp(int value);
  ///
  SgValueExp(char char_val);
  ///
  SgValueExp(float float_val);
  ///
  SgValueExp(double double_val);
  ///
  SgValueExp(char *string_val);
  ///
  SgValueExp(double real, double imaginary);
  ///
  SgValueExp(SgValueExp &real, SgValueExp &imaginary);
  ///
  void setValue(int int_val);
  ///
  void setValue(char char_val);
  ///
  void setValue(float float_val);
  ///
  void setValue(double double_val);
  ///
  void setValue(char *string_val);
  ///
  void setValue(double real, double im);
  ///
  void setValue(SgValueExp &real, SgValueExp & im);
  ///
  int intValue();
  ///
  char*  floatValue();
  ///
  char  charValue();
  ///
  char*  doubleValue();
  ///
  char * stringValue();
  ///
  SgExpression *realValue();
  ///
  SgExpression *imaginaryValue();
};



///
class  SgKeywordValExp: public SgExpression{
public:
   ///
   SgKeywordValExp(char *name);
   ///
   char *value();
};


///
class  SgUnaryExp: public  SgExpression{
public:
  ///
  SgUnaryExp(PTR_LLND ll);
  ///
  SgUnaryExp(int variant, SgExpression & e);
  ///
  SgUnaryExp(int variant, int post, SgExpression & e);
  ///
  int post();
  ///
  SgExpression &operand();
};

///
class  SgCastExp: public SgExpression{
public:
  ///
  SgCastExp(PTR_LLND ll);
  ///
  SgCastExp(SgType &t, SgExpression &e);
  ///
  SgCastExp(SgType &t);
  ///
  ~SgCastExp();
};

// delete [size]  expr

/// variant == DELETE\_OP
class SgDeleteExp: public SgExpression{
public:
  ///
  SgDeleteExp(PTR_LLND ll);
  ///
  SgDeleteExp(SgExpression &size, SgExpression &expr);
  ///
  SgDeleteExp(SgExpression &expr);
  ///
  ~SgDeleteExp();
};

// new typename
// new typename (expr)

/// variant == NEW\_OP
class SgNewExp: public SgExpression{
public:
  ///
  SgNewExp(PTR_LLND ll);
  ///
  SgNewExp(SgType &t);
  ///
  SgNewExp(SgType &t, SgExpression &e);
#if 0
  ///
  SgExpression &numberOfArgs();
  ///
  SgExpression &argument(int i);
#endif
  ///
  ~SgNewExp();
};

/// (expr1)? expr2 : expr3, variant == EXPR\_IF 
// functions here can use LlndMapping perhaps.
class SgExprIfExp: public SgExpression{
public:
  ///
  SgExprIfExp(PTR_LLND ll);
  ///
  SgExprIfExp(SgExpression &exp1,SgExpression &exp2, SgExpression &exp3);
  ///
  SgExpression &conditional();
  ///
  SgExpression &trueExp();
  ///
  SgExpression &falseExp();
  ///
  void setConditional(SgExpression &c);
  ///
  void setTrueExp(SgExpression &t);
  ///
  void setFalseExp(SgExpression &f);
  ///
  ~SgExprIfExp();
};

/// function\_name(formal args)  - for function headers and protytpes. variant = FUNCTION\_REF 
class SgFunctionRefExp: public SgExpression{
public:
  SgFunctionRefExp(PTR_LLND ll);
  ///
  SgFunctionRefExp(SgSymbol &fun);
  ///
  ~SgFunctionRefExp();
  ///
  SgSymbol *funName();
  ///
  SgExpression *args();
  ///
  int numberOfArgs();
  ///
  SgExpression *arg(int i);
  ///
  SgExpression * AddArg(char *, SgType &);
  ///
  void addArg(SgExpression &); // this function is added like the one in SgCallStmt;
};

/// function\_name(expr1, expr2, --). variant == FUNC\_CALL
class SgFunctionCallExp: public SgExpression{
public:
  ///
  SgFunctionCallExp(PTR_LLND ll);
  ///
  SgFunctionCallExp(SgSymbol &fun, SgExpression &paramList);
  ///
  SgFunctionCallExp(SgSymbol &fun);
  ///
  ~SgFunctionCallExp();
  ///
  SgSymbol *funName();
  ///
  SgExpression *args();
  ///
  int numberOfArgs();
  ///
  SgExpression *arg(int i);
  ///
  void addArg(SgExpression &arg);
};

/// (functionpointer)(expr1,expr2,expr3). variant == FUNCTION\_OP
class SgFuncPntrExp: public SgExpression{
public:
  ///
  SgFuncPntrExp(PTR_LLND ll);
  ///
  SgFuncPntrExp(SgExpression &ptr);
  ///
  ~SgFuncPntrExp();
  ///
  SgExpression *funExp();
  ///
  void setFunExp(SgExpression &s);
  ///
  int numberOfArgs();
  ///
  SgExpression *arg(int i);
  /// add an argument.
  void addArg(SgExpression &arg);  
  /// add a formal parameter to a pointer to a function prototype or parameter.
  SgExpression* AddArg(SgSymbol *thefunc, char *name, SgType &); // this returns the expression
};

/// variant == EXPR\_LIST, CONTROL\_LIST
class SgExprListExp: public SgExpression{
public:
  ///
  SgExprListExp(PTR_LLND ll);
  ///
  SgExprListExp();
  ///
  SgExprListExp(SgExpression &ptr);
  ///
  ~SgExprListExp();
  ///
  int length();
  ///
  SgExpression *elem(int i);
   ///
  SgExprListExp *next();
  ///
  SgExpression *value();
  ///
  void setValue(SgExpression &ptr);      
  ///
  void append(SgExpression &arg);
  ///
  void insertExpAfter(SgExpression & exp);
  ///
  void linkToEnd(SgExpression &arg);
};

/// Fortran name references, variant == CONST\_REF, TYPE\_REF, INTERFACE\_REF
class  SgRefExp: public SgExpression{
public:
  ///
  SgRefExp(PTR_LLND ll);
  ///
  SgRefExp(int variant, SgSymbol &s);
  ///
  ~SgRefExp();
};

/// a reference to a type in a c++ template argument 
class SgTypeRefExp: public SgExpression{
  public:
  ///
  SgTypeRefExp(SgType &t);
  ///
  ~SgTypeRefExp();
  ///
  SgType *getType();
};

/// scalar variable reference or non-indexed array reference. variant == VAR\_REF
class  SgVarRefExp: public SgExpression{
public:
  ///
  SgVarRefExp (PTR_LLND ll);
  ///
  SgVarRefExp(SgSymbol &s);
  ///
  SgVarRefExp(SgSymbol *s);
  /// if scalar propogation worked
  SgExpression *progatedValue(); 
  ///
  ~SgVarRefExp();
};


/// variant == THIS\_NODE
class  SgThisExp: public SgExpression{
public:
  ///
  SgThisExp (PTR_LLND ll);
  ///
  SgThisExp(SgType &t);
  ///
  ~SgThisExp();
};

/// variant ==DDOT
class SgTripletExp : public SgExpression {
public:
  ///
  SgTripletExp(SgExpression * item0, SgExpression * item1, SgExpression *item2);
  ///
  int numberOfItems();
  ///
  SgExpression * item(int i);
};

/// an array reference. variant == ARRAY\_REF
class  SgArrayRefExp: public SgExpression{
public:
  ///
  SgArrayRefExp(PTR_LLND ll);
  ///
  SgArrayRefExp(SgSymbol &s);
  ///
  SgArrayRefExp(SgSymbol &s, SgExpression &subscripts);
  ///
  SgArrayRefExp(SgSymbol &s, SgExpression &sub1,SgExpression &sub2);
  ///
  SgArrayRefExp(SgSymbol &s, SgExpression &sub1,SgExpression &sub2,SgExpression &sub3);
  ///
  SgArrayRefExp(SgSymbol &s, SgExpression &sub1,SgExpression &sub2,SgExpression &sub3,SgExpression &sub4);
  ///
  ~SgArrayRefExp();
  /// the number of subscripts in reference
  int numberOfSubscripts();  
  ///
  SgExpression *subscripts();
  ///
  SgExpression *subscript(int i);
  ///
  void addSubscript(SgExpression &e);
  ///
  void setSymbol(SgSymbol &s);
  ///
  int isScalarOnDim(int i);
};

/// 
// set NODE\_TYPE.
class SgPntrArrRefExp: public SgExpression{
public:
  ///
  SgPntrArrRefExp(PTR_LLND ll);
  ///
  SgPntrArrRefExp(SgExpression &p);
  ///
  SgPntrArrRefExp(SgExpression &p, SgExpression &subscripts);
  ///
  SgPntrArrRefExp(SgExpression &p, int n, SgExpression &sub1, SgExpression &sub2);
  ///
  SgPntrArrRefExp(SgExpression &p, int n, SgExpression &sub1, SgExpression &sub2, SgExpression &sub3);
  ///
  SgPntrArrRefExp(SgExpression &p, int n, SgExpression &sub1, SgExpression &sub2, SgExpression &sub3, SgExpression &sub4);
  ///
  ~SgPntrArrRefExp();
  /// the number of subscripts in reference
  int dimension(); 
  ///
  SgExpression *subscript(int i);
  ///
  void addSubscript(SgExpression &e);
  ///
  void setPointer(SgExpression &p);
};

/// pointer dereferencing. variant == DEREF\_OP 
class SgPointerDerefExp: public SgExpression{
public:
  ///
  SgPointerDerefExp(PTR_LLND ll);
  ///
  SgPointerDerefExp(SgExpression &pointerExp);
  ///
  ~SgPointerDerefExp();
  ///
  SgExpression *pointerExp();
};

/// a field reference of a structure. variant == RECORD\_REF  
class SgRecordRefExp: public SgExpression{
public:
  SgRecordRefExp(PTR_LLND ll);
  ///
  SgRecordRefExp(SgSymbol &recordName, char *fieldName);
  ///
  SgRecordRefExp(SgExpression &recordExp, char *fieldName);
  ///
  ~SgRecordRefExp();
  ///
  SgSymbol *fieldName();
  ///
  SgSymbol *recordName();
  ///
  SgExpression *record();
};


/// Fortran 90 structure constructor. variant == STRUCTURE\_CONSTRUCTOR
class SgStructConstExp:  public SgExpression{
public:
  ///
  SgStructConstExp(PTR_LLND ll);
  ///
  //further checks on values need to be done.
  SgStructConstExp(SgSymbol &structName, SgExpression &values);
  ///
  SgStructConstExp(SgExpression  &typeRef, SgExpression &values);
  ///
  ~SgStructConstExp();
  ///
  int numberOfArgs();
  ///
  SgExpression *arg(int i);
};

///
class SgConstExp:  public SgExpression{
public:
  ///
  SgConstExp(PTR_LLND ll);
  ///
  SgConstExp(SgExpression &values);
  ///
  ~SgConstExp();
  ///
  int numberOfArgs();
  ///
  SgExpression *arg(int i);
};

/// a vector constant of the form: [ expr1, expr2, expr3]. variant == VECTOR\_CONST
class SgVecConstExp: public SgExpression{
public:
  SgVecConstExp(PTR_LLND ll);
  ///
  SgVecConstExp(SgExpression &expr_list);
  ///
  SgVecConstExp(int n, SgExpression *components);
  ///
  ~SgVecConstExp();
  
  /// the i-th term
  SgExpression *arg(int i); 
  ///
  int numberOfArgs();
  ///
  void setArg(int i, SgExpression &e);
};

/// used for initializations.  form: { expr1,expr2,expr3}, variant == INIT\_LIST 
class SgInitListExp: public SgExpression{
public:
  ///
  SgInitListExp(PTR_LLND ll);
  ///
  SgInitListExp(SgExpression &expr_list);
  ///
  SgInitListExp(int n, SgExpression *components);
  ///
  ~SgInitListExp();
  /// the i-th term
  SgExpression *arg(int i); 
  ///
  int numberOfArgs();
  ///
  void setArg(int i, SgExpression &e);
};

/// used for EQUIVALENCE, NAMELIST and COMMON statements. variant == EQUI\_LIST, NAMELIST\_LIST, COMM\_LIST
class SgObjectListExp: public SgExpression{
public:
  ///
  SgObjectListExp(PTR_LLND ll);
  ///
  SgObjectListExp(int variant, SgSymbol &object, SgExpression &list);
  ///
  SgObjectListExp(int variant,SgExpression &objectRef, SgExpression &list);
  ///
  ~SgObjectListExp();
  ///
  SgSymbol *object();
  ///
  SgExpression *objectRef();
  ///
  int listLength();
  /// i'th symbol
  SgSymbol *symbol(int i);       
  /// i'th symbol
  SgExpression *objectRef(int i);  
};

/** 
Fortran 90 attributes. 
variant ==  PARAMETER\_OP, PUBLIC\_OP, PRIVATE\_OP, ALLOCATABLE\_OP, DIMENSION\_OP, EXTERNAL\_OP, 
IN\_OP, OUT\_OP, INOUT\_OP, INTRINSIC\_OP, POINTER\_OP, OPTIONAL\_OP, SAVE\_OP, TARGET\_OP 
*/

class SgAttributeExp: public SgExpression{
public:
  ///
  SgAttributeExp(PTR_LLND ll);
  ///
  SgAttributeExp(int variant);
  ///
  ~SgAttributeExp();
};

/// Fortran 90 keyword argument. variant == KEYWORD\_ARG
class SgKeywordArgExp: public SgExpression{
public:
  ///
  SgKeywordArgExp(PTR_LLND ll);
  ///
  SgKeywordArgExp(char *argName, SgExpression &exp);
  ///
  ~SgKeywordArgExp();
  ///
  SgSymbol *arg();
  ///
  SgExpression *value();
};

/// Fortran 90 vector subscript expression. variant == DDOT
class SgSubscriptExp: public SgExpression{
public:
  ///
  SgSubscriptExp(PTR_LLND ll);
  ///
  SgSubscriptExp(SgExpression &exp);
  ///
  SgSubscriptExp(SgExpression &lbound, SgExpression &ubound, SgExpression &step);
  ///
  SgSubscriptExp(SgExpression &lbound, SgExpression &ubound);
  ///
  ~SgSubscriptExp();
  // perhaps this function can use LlndMapping
  ///
  SgExpression *lbound();
  ///
  SgExpression *ubound();
  ///
  SgExpression *step();
};

/// variant == INDEX\_OP
// add by zheng.gb
class SgIndexOpExp: public SgExpression {
public:
  ///
  SgIndexOpExp(SgVarRefExp &var, SgSubscriptExp &bound);
  ///
  ~SgIndexOpExp();
};

/// Fortran 90 USE statement ONLY attribute. variant == ONLY\_NODE
class SgUseOnlyExp: public SgExpression{
public:
  ///
  SgUseOnlyExp(PTR_LLND ll);
  ///
  SgUseOnlyExp(SgExpression &onlyList);
  ///
  ~SgUseOnlyExp();
  ///
  SgExpression *onlyList();
};

/// Fortran 90 USE statement renaming, variant == RENAME\_NODE
class SgUseRenameExp: public SgExpression{
public:
  ///
  SgUseRenameExp(PTR_LLND ll);
  ///
  SgUseRenameExp(SgSymbol &newName, SgSymbol &oldName);
  ///
  ~SgUseRenameExp();
   ///
  SgSymbol *newName();
  ///
  SgSymbol *oldName();
  ///
  SgExpression *newNameExp();
  ///
  SgExpression *oldNameExp();
};


/// Fortran default control arguments to Input/Output statements. variant == SPEC\_PAIR
class SgSpecPairExp: public SgExpression{
public:
  ///
  SgSpecPairExp(PTR_LLND ll);
  ///
  SgSpecPairExp(SgExpression &arg, SgExpression &value);
  ///
  SgSpecPairExp(SgExpression &arg);
  ///
  SgSpecPairExp(char *arg, char *value);
  ///
  ~SgSpecPairExp();
  ///
  SgExpression *arg();
  ///
  SgExpression *value();
};

/// Fortran index variable bound instantiation, variant == IOACCESS

// used for do-loop range representation also.
// this form needs to be standardized. 

class SgIOAccessExp: public SgExpression{
public:
  ///
  SgIOAccessExp(PTR_LLND ll);
  // type-checking on bounds needs to be done.
  // Float values are legal in some cases. check manual.

  ///
  SgIOAccessExp(SgSymbol &s, SgExpression lbound, SgExpression ubound, SgExpression step);
  ///
  SgIOAccessExp(SgSymbol &s, SgExpression lbound, SgExpression ubound);
  ///
  ~SgIOAccessExp();
};

/// Fortran index variable bound instantiation. variant == IMPL\_TYPE
class SgImplicitTypeExp: public SgExpression{
public:
  ///
  SgImplicitTypeExp(PTR_LLND ll);
  ///
  SgImplicitTypeExp(SgType &type, SgExpression &rangeList);
  ///
  ~SgImplicitTypeExp();
  ///
  SgType *type();
  ///
  SgExpression *rangeList();
  ///
  char *alphabeticRange();
};

/// Fortran type expression. variant == TYPE\_OP
class SgTypeExp: public SgExpression{
public:
  ///
  SgTypeExp(PTR_LLND ll);
  ///
  SgTypeExp(SgType &type);
  ///
  ~SgTypeExp();
  ///
  SgType *type();
};

/// Fortran index variable bound instantiation. variant == SEQ
class SgSeqExp: public SgExpression{
public:
  ///
  SgSeqExp(PTR_LLND ll);
  ///
  SgSeqExp(SgExpression &exp1, SgExpression &exp2);
  ///
  ~SgSeqExp();
  ///
  SgExpression *front();
  ///
  SgExpression *rear();
};

/// Fortran index variable bound instantiation. variant == LEN\_OP
class SgStringLengthExp: public SgExpression{
public:
  ///
  SgStringLengthExp(PTR_LLND ll);
  ///
  SgStringLengthExp(SgExpression &length);
  ///
  ~SgStringLengthExp();
  ///
  SgExpression *length();
};

/// Fortran default node. variant == DEFAULT
class SgDefaultExp: public SgExpression {
public:
  ///
  SgDefaultExp(PTR_LLND ll);
  ///
  SgDefaultExp();
  ///
  ~SgDefaultExp();
};

/// Fortran label reference. variant == LABEL\_REF 
class SgLabelRefExp: public SgExpression{
public:
  ///
  SgLabelRefExp(PTR_LLND ll);
  ///
  SgLabelRefExp(SgLabel &label);
  ///
  ~SgLabelRefExp();
   ///
  SgLabel *label();
};


// -----------------------------------;


/// Fortran Program block. variant == PROG\_HEDR
class  SgProgHedrStmt: public SgStatement{
public:
  ///
  SgProgHedrStmt(PTR_BFND bif);
  ///
  SgProgHedrStmt(int variant);
  ///
  SgProgHedrStmt(SgSymbol &name, SgStatement &Body);
  ///
  SgProgHedrStmt(SgSymbol &name);
  ///
  SgProgHedrStmt(char *name);
  ///
  SgSymbol &name();
  /// set program name 
  void setName(SgSymbol &symbol); 

  /// the number of functions called
  int numberOfFunctionsCalled();  

  /// the i-th called function
  SgSymbol *calledFunction(int i);

  /// the number of statement funcions;
  int numberOfStmtFunctions();  

  /// the i-th statement function;
  SgStatement *statementFunc(int i); 

  /// the number of entry points;
  int numberOfEntryPoints();     

  /// the i-th entry point;
  SgStatement *entryPoint(int i); 

  /// the number of parameters; 
  int numberOfParameters();      

  /// the i-th parameter
  SgSymbol *parameter(int i); 

  ///
  void insertParameterAfter(SgSymbol & parameter, SgSymbol & added);
  ///
  int numberOfSpecificationStmts();
  ///
  SgStatement *specificationStmt(int i);
  ///
  int numberOfExecutionStmts();
  ///
  SgStatement *executionStmt(int i);
  ///
  int numberOfInternalFunctionsDefined();
  ///
   SgStatement *internalFunction(int i);
  ///
  int numberOfInternalSubroutinesDefined();
  ///
  SgStatement *internalSubroutine(int i);
  ///
  int numberOfInternalSubProgramsDefined();
  ///
  SgStatement *internalSubProgram(int i);
  
#if 0
  /// add a declaration for new variable
  SgSymbol &addVariable(SgType &T, char *name); 

  /// add a new common block
  SgStatement &addCommonBlock(char *blockname, int noOfVars, SgSymbol *Vars); 
#endif        

  ///
  int isSymbolInScope(SgSymbol &symbol);
  ///
  int isSymbolDeclaredHere(SgSymbol &symbol);
  
  ///   ture if a string name is in a scope .
  int isSymbolNameInScope(char *name);             // zheng.gb

  ///
  SgSymbol * firstSymbolInScope();
  
  //  global analysis data

  /// list of used variable access sections
  int numberOfVarsUsed();  

  /// i-th var used section descriptor
  SgExpression *varsUsed(int i); 

  /// list of modifed variable access sections 
  int numberofVarsMod();  

  /// i-th var mod section descriptor
  SgExpression *varsMod(int i);  

  /// 
  ~SgProgHedrStmt();
};

/// Fortran subroutine, variant == PROC\_HEDR
class  SgProcHedrStmt: public SgProgHedrStmt{
public: 
  ///
  SgProcHedrStmt(int variant);
  ///
  SgProcHedrStmt(SgSymbol &name, SgStatement &Body);
  ///
  SgProcHedrStmt(SgSymbol &name);
  ///
  SgProcHedrStmt(char *name);
  ///
  void AddArg(SgExpression &arg); 

  /// returns decl expr created.
  SgExpression * AddArg(char *name, SgType &t);

  /// 
  SgExpression * AddArg(char *name, SgType &t, SgExpression &initializer);

  /// 1 if recursive.
  int isRecursive(); 

  /// the number of entry points other than the main, 0 for C funcs.
  int numberOfEntryPoints(); 

  /// the i-th entry point
  SgStatement *entryPoint(int i); 

  // this is incorrect. Takes only subroutines calls into account.
  // ManMemo: Should be modified to take function calls into account too.

  /// number of calls to this proc.
  int numberOfCalls();          

  /// position of the i-th call. 
  SgStatement *call(int i);

  ///
  ~SgProcHedrStmt();
};      


/// Fortran M process, variant == PROS\_HEDR
class  SgProsHedrStmt: public SgProgHedrStmt{
public:
  ///
  SgProsHedrStmt();
  ///
  SgProsHedrStmt(SgSymbol &name, SgStatement &Body);
  ///
  SgProsHedrStmt(SgSymbol &name);
   ///
  SgProsHedrStmt(char *name);
   ///
  void AddArg(SgExpression &arg);
  /// number of calls to this proc.
  int numberOfCalls();          
  /// position of the i-th call.
  SgStatement *call(int i);     
  ///
  ~SgProsHedrStmt();
};      

/// Fortran and C function, variant == FUNC\_HEDR
class  SgFuncHedrStmt: public SgProcHedrStmt{
public:
  ///
  SgFuncHedrStmt(SgSymbol &name, SgStatement &Body);
  ///
  SgFuncHedrStmt(SgSymbol &name, SgType &type, SgStatement &Body);
  ///
  SgFuncHedrStmt(SgSymbol &name, SgSymbol &resultName, SgType &type, SgStatement &Body);
  ///
  SgFuncHedrStmt(SgSymbol &name);
  ///
  SgFuncHedrStmt(SgSymbol &name, SgExpression *exp);
  ///
  SgFuncHedrStmt(char *name);
  ///
  ~SgFuncHedrStmt();
  
  /// name of result variable.
  SgSymbol *resultName();  
  /// set name of result variable.
  int  setResultName(SgSymbol &symbol); 
  /// type of returned value
  SgType *returnedType(); 
  /// set type of returned value
  void setReturnedType(SgType &type); 
};

class SgClassStmt;

/// a function template or class template
//* in both cases, variant == TEMPLATE_FUNDECL
class SgTemplateStmt: public SgStatement{
public:
  ///
  SgTemplateStmt(SgExpression *arglist);
  /// returns decl expr created. if name == NULL then this is a type reference.
  SgExpression * AddArg(char *name, SgType &t); // returns decl expr created.
  ///
  SgExpression * AddArg(char *name, SgType &t, SgExpression &initializer);
  ///
  int numberOfArgs();
  ///
  SgExpression *arg(int i);
  ///
  SgExpression *argList();
  ///
  void addFunction(SgFuncHedrStmt &theTemplateFunc);
  ///
  void addClass(SgClassStmt &theTemplateClass);
  ///
  SgFuncHedrStmt *isFunction();
  ///
  SgClassStmt *isClass();
};

#if 0
/// Fortran 90 Module statement. variant ==  MODULE\_STMT
class  SgModuleStmt: public SgStatement{
public:
  ///
  SgModuleStmt(SgSymbol &moduleName, SgStatement &body);
  ///
  SgModuleStmt(SgSymbol &moduleName);
  ///
  ~SgModuleStmt();  
  /// module name
  SgSymbol *moduleName();            
  
  /// set module name 
  void setName(SgSymbol &symbol);  

  ///
  int numberOfSpecificationStmts();
  ///
  int numberOfRoutinesDefined();
  ///
  int numberOfFunctionsDefined();
  ///
  int numberOfSubroutinesDefined();
  
  ///
  SgStatement *specificationStmt(int i);
  ///
  SgStatement *routine(int i);
  ///
  SgStatement *function(int i);
  ///
  SgStatement *subroutine(int i);
  
  ///
  int isSymbolInScope(SgSymbol &symbol);
  ///
  int isSymbolDeclaredHere(SgSymbol &symbol);
  
  /// add a declaration for new variable
  SgSymbol &addVariable(SgType &T, char *name); 
                                       
  /// add a new common block
  SgStatement *addCommonBlock(char *blockname, int noOfVars, SgSymbol *Vars);
};

/// Fortran 90 Operator Interface Statement. variant == INTERFACE\_STMT

class  SgInterfaceStmt: public SgStatement{
public:
  ///
  SgInterfaceStmt(SgSymbol &name, SgStatement &body, SgStatement &scope);
  ///
  ~SgInterfaceStmt();
  
  /// interface name if given
  SgSymbol *interfaceName();              
  /// set interface name 
  int setName(SgSymbol &symbol);          
  ///
  int numberOfSpecificationStmts();
  
  ///
  SgStatement *specificationStmt(int i);
  
  ///
  int isSymbolInScope(SgSymbol &symbol);
  ///
  int isSymbolDeclaredHere(SgSymbol &symbol);
};

/// Fortran Block Data statement. variant == BLOCK\_DATA
class  SgBlockDataStmt: public SgStatement{
public:
  ///
  SgBlockDataStmt(SgSymbol &name, SgStatement &body);
  ///
  ~SgBlockDataStmt();
  
  /// block data name if given;
  SgSymbol *name(); 
  /// set block data name 
  int setName(SgSymbol &symbol);         
  ///
  int isSymbolInScope(SgSymbol &symbol);
  ///
  int isSymbolDeclaredHere(SgSymbol &symbol);
};

#endif


/** C++ class statement.   variant == CLASS\_DECL 
   class name : superclass\_list ElementTypeOf collection\_name {
           body
   } variables\_list;
*/

class SgClassStmt: public SgStatement{
public:
  ///
  SgClassStmt(int variant);
  ///
  SgClassStmt(SgSymbol &name);
  ///
  ~SgClassStmt();
  ///
  int numberOfSuperClasses();
  ///
  SgSymbol *name();
  ///
  SgSymbol *superClass(int i);
  ///
  void setSuperClass(int i, SgSymbol &symb);
#if 0
  /// variables in variables\_list
  int numberOfVars();            
  /// i-th variable in variable\_list
  SgExpression variable(int i);  

  /// if an ElementType class.
  SgExpression collectionName(); 
  
  /// body manipulation functions.
  int numberOfPublicVars();
  ///
  int numberOfPrivateVars();
  ///
  int numberOfProtectedVars();
  ///
  SgSymbol *publicVar(int i);
  ///
  SgSymbol *protectedVar(int i);
  ///
  SgSymbol *privateVar(int i);
  ///
  void addPublicVar(SgSymbol &s);
  ///
  void addPrivateVar(SgSymbol &s);
  ///
  void addProtectedVar(SgSymbol &s);
  ///
  int numberOfPublicFuns();
  ///
  int numberOfPrivateFuns();
  ///
  int numberOfProtectedFuns();
  ///
  SgStatement *publicFun(int i);
  ///
  SgStatement *protectedFun(int i);
  ///
  SgStatement *privateFun(int i);
  ///
  void addPublicFun(SgStatement &s);
  ///
  void addPrivateFun(SgStatement &s);
  ///
  void addProtectedFun(SgStatement &s);
#endif
};

/** basic C++ structure.
variant == STRUCT\_DECL
basic C++ structure
struct name {
                   body
             } variables\_list;
 */
class SgStructStmt: public SgClassStmt{
public:
  /// consider like a class.
  SgStructStmt();
  ///
  SgStructStmt(SgSymbol &name);
  ///
  ~SgStructStmt();
  
};

/** basic C++ structure
variant == UNION\_DECL
union name  {
    body
   } variables\_list;
*/

class SgUnionStmt: public SgClassStmt{
public:
  /// consider like a class.
  SgUnionStmt();
  ///
  SgUnionStmt(SgSymbol &name);
  ///
  ~SgUnionStmt();
};

/** basic C++ structure.
 variant == ENUM\_DECL
 enum name  {
     body
    } variables\_list;
*/

class SgEnumStmt: public SgClassStmt{
public:
  /// consider like a class.
  SgEnumStmt();
  ///
  SgEnumStmt(SgSymbol &name);
  ///
  ~SgEnumStmt();
};

/** basic C++ structure.
 variant == COLLECTION\_DECL
 collection name {
                    body
                 } variables\_list;

*/
class SgCollectionStmt: public SgClassStmt{
public:
  ///
  SgCollectionStmt();
  ///
  SgCollectionStmt(SgSymbol &name);
  ///
  ~SgCollectionStmt();
#if 0
  ///
int numberOfElemMethods();
  ///
  SgStatement *elementMethod(int i);
  ///
  void addElementMethod(SgStatement &s);
#endif
  ///
  SgStatement *firstElementMethod();
};

/// in C we have: {  body; } . variant == BASIC\_BLOCK
class SgBasicBlockStmt: public SgStatement{
public:
  ///
  SgBasicBlockStmt();
  ///
  ~SgBasicBlockStmt();
};

// ********************* traditional control Structures ************

/// for Fortran Do and C for(). variant = FOR\_NODE
class  SgForStmt: public SgStatement{
public:
  ///
  SgForStmt(SgSymbol &do_var, SgExpression &start, SgExpression &end,
            SgExpression &step, SgStatement &body);
  ///
  SgForStmt(SgSymbol *do_var, SgExpression *start, SgExpression *end,
            SgExpression *step, SgStatement *body);
  ///
  SgForStmt(SgSymbol &do_var, SgExpression &start, SgExpression &end,
            SgStatement &body);
  ///
  SgForStmt(SgExpression &start, SgExpression &end, SgExpression &step,
            SgStatement &body);
  
  /// the name of the loop (for F90.)
  SgSymbol doName();               

  /// sets the name of the loop(for F90)
  void setDoName(SgSymbol &doName);

  ///
  SgExpression *start();            
  ///
  void setStart(SgExpression &lbound);
  
  ///
  SgExpression *end();
  ///
  void setEnd(SgExpression &ubound);
  
  ///
  SgExpression *step();
  ///
  void setStep(SgExpression &step);
  
  ///
  SgLabel *endOfLoop();

  //SgExpression &bounds();   // bounds are returned as a triplet lb:ub;
  //void setBounds(SgTripletOp &bounds); // bounds are passed as a triplet lb:ub;

  // body is returned with control end statement  still attached. 

  ///
  SgStatement *body();

  /// s is assumed to terminate with a control end statement.
  void set_body(SgStatement &s);
#if 0
  /// new body = s and lex successors.
  int replaceBody(SgStatement &s);
  
  /// 1 if an induction variable can be found.
  int numberOfInductVars();     

  /// i-th induction variable
  SgSymbol *inductionVar(int i);   

  /// range of i-th ind. var.
  SgExpression *indVarRange(int i); 
#endif
  ///
  int isPerfectLoopNest();
  ///
  SgStatement *getNextLoop(); 
  /// returns outer nested loop
  SgStatement *getPreviousLoop(); 
  /// returns innermost nested loop
  SgStatement *getInnermostLoop();
#if 0
  /** TRUE if the bound and step of the loops
   in the loop nest are linear expressions
   and use the index variables of the previous
   loops of the nest.
   */
  int isLinearLoopNest();    
#endif
  /// TRUE if the loop ends with an Enddo
  int isEnddoLoop();           
  
  /// Convert the loop into a Good loop.
  int convertLoop();           
#if 0
  /// TRUE if the body consists only of assignments
  int isAssignLoop();     

  /** TRUE if the body consists only of assigments
               and conditional statements.
   high level program transformations. 
   Most are from SIGMA Toolbox by F.Bodin et al.
   Semantics can be found in the above reference.
   */
  int isAssignIfLoop();       

  ///
  int tiling_p(int i);     
  ///
  int tiling(int i, int tab[]);
  ///
  int stripMining(int i);
  ///
  SgStatement distributeLoop(int i);
  ///
  SgStatement distributeLoopSCC();
  ///
  SgStatement loopFusion(SgForStmt &loop);
  ///
  SgStatement unrollLoop(int i);
  ///
  int interchangeLoops(SgForStmt &loop);
  ///
  int interchangeWithLoop(int i);
  ///
  int normalized();
  ///
  int NormalizeLoop();
  ///
  int vectorize();
  ///
  int vectorizeNest();
  ///
  int ExpandScalar(SgSymbol &symbol, int i);
  ///
  int ScalarForwardSubstitution(SgSymbol &symbol);
  ///
  int pullStatementToFront(SgStatement &s);
  ///
  int pullStatementToEnd(SgStatement &s);
#endif
  ///
  ~SgForStmt();
};

/// variant = FORALL\_CONSTRUCT\_STMT, FORALL\_STMT

// added by zheng.gb
class  SgForallStmt: public SgStatement{
public:
  ///
  SgForallStmt(SgExpression &forall_header_op , SgStatement &body);
  ///
  SgForallStmt(SgExpression &forall_header_op , SgExpression &mask,
	       SgStatement &body);
  
  ///
  ~SgForallStmt();
};

/// for Fortran M ProcessDo statement. variant = PROCESS\_DO\_STAT
class  SgProcessDoStmt: public SgStatement{
public:
  ///
  SgProcessDoStmt(SgSymbol &do_var, SgExpression &start,
		  SgExpression &end, SgExpression &step,
		  SgLabel &endofloop, SgStatement &body);
  ///
  SgProcessDoStmt(SgSymbol &do_var, SgExpression &start,
		  SgExpression &end, SgLabel &endofloop,
		  SgStatement &body);
  ///
  SgProcessDoStmt(SgSymbol &do_var, SgExpression &start,
		  SgExpression &end, SgExpression &step,
		  SgStatement &body);
  ///
  SgProcessDoStmt(SgSymbol &do_var, SgExpression &start,
		  SgExpression &end, SgStatement &body);
  ///
  SgSymbol doName();
  ///
  void setDoName(SgSymbol &doName);
  ///
  SgExpression *start();            
  ///
  SgExpression *end();
  ///
  SgExpression *step();
  ///
  SgLabel *endOfLoop();
  /// body is returned with control end statement still attached. 
  SgStatement *body();
  /// s is assumed to terminate with a control end statement.
  void set_body(SgStatement &s);
  
#if 0
  /// new body = s and lex successors.
  int replaceBody(SgStatement &s);
  
  /// 1 if an induction variable can be found.
  int numberOfInductVars();       

  /// i-th induction variable
  SgSymbol *inductionVar(int i);  

  /// range of i-th ind. var.
  SgExpression *indVarRange(int i); 
#endif

  ///
  int isPerfectLoopNest();
  ///
  SgStatement *getNextLoop(); 
  /// returns outer nested loop
  SgStatement *getPreviousLoop();   

  /// returns innermost nested loop
  SgStatement *getInnermostLoop();  
#if 0
  /** TRUE if the bound and step of the loops
      in the loop nest are linear expressions
      and use the index variables of the previous
      loops of the nest.
  */
  int isLinearLoopNest();
#endif
  /// TRUE if the loop ends with an Enddo
  int isEnddoLoop();            
  /// Convert the loop into a Good loop.
  int convertLoop(); 
#if 0
  /// TRUE if the body consists only of assignments
  int isAssignLoop();        
  /// TRUE if the body consists only of assignments and conditional statements.
  int isAssignIfLoop();    
     
  // high level program transformations. 
  // Most are from SIGMA Toolbox by F.Bodin et al.
  // Semantics can be found in the above reference.
  ///
  int tiling_p(int i);     
  ///
  int tiling(int i, int tab[]);
  ///
  int stripMining(int i);
  ///
  SgStatement distributeLoop(int i);
  ///
  SgStatement distributeLoopSCC();
  ///
  SgStatement loopFusion(SgForStmt &loop);
  ///
  SgStatement unrollLoop(int i);
  ///
  int interchangeLoops(SgForStmt &loop);
  ///
  int interchangeWithLoop(int i);
  ///
  int normalized();
  ///
  int NormalizeLoop();
  ///
  int vectorize();
  ///
  int vectorizeNest();
  ///
  int ExpandScalar(SgSymbol &symbol, int i);
  ///
  int ScalarForwardSubstitution(SgSymbol &symbol);
  ///
  int pullStatementToFront(SgStatement &s);
  ///
  int pullStatementToEnd(SgStatement &s);
#endif
  ///
  ~SgProcessDoStmt();
};

/// for C while()
//* variant = WHILE_NODE
class  SgWhileStmt: public SgStatement{
public:
  ///
  SgWhileStmt(int variant);
  ///
  SgWhileStmt(SgExpression &cond, SgStatement &body);
  /// the while test
  SgExpression *conditional();   
#if 0
  /// 1 if an induction variable can be found.
  int numberOfInductVars();

  /// i-th induction variable
  SgSymbol *inductionVar(int i);  

  /// range of i-th ind. var.
  SgExpression *indVarRange(int i);
#endif
  /// new body = s and lex successors.
  void replaceBody(SgStatement &s); 
  ///
  ~SgWhileStmt();
};


/// variant = DO\_WHILE\_NODE, For Fortran dowhile(  ). and C do{   } while( );
class  SgDoWhileStmt: public SgWhileStmt{
public:
  ///
  SgDoWhileStmt(SgExpression &cond, SgStatement &body);
  ///
  ~SgDoWhileStmt();
};

// forward reference;
class  SgIfStmt;

/// For Fortran logical if  - only one body statement allowed. variant == LOGIF\_NODE
class  SgLogIfStmt: public SgStatement{
public:
  ///
  SgLogIfStmt(int variant);
  ///
  SgLogIfStmt(SgExpression &cond, SgStatement &s);
  /// returns reference to first stmt in the body
  SgStatement *body();
  /// the while test
  SgExpression *conditional();  // 
  // check if the statement s is a single statement. 
  /// new body = s 
  void setBody(SgStatement &s);  
  // this code won't work, since after the addition false
  // clause, it should become SgIfThenElse statement.
  ///  make it into if-then-else
  void addFalseClause(SgStatement &s);
  ///
  SgIfStmt *convertLogicIf();
  ///
  ~SgLogIfStmt();
};

/// For Fortran if then else and C if(). variant == IF\_NODE
class  SgIfStmt: public SgStatement{
public:
  ///
  SgIfStmt(int variant);
  ///
  SgIfStmt(SgExpression &cond, SgStatement &trueBody, SgStatement &falseBody,
	   SgSymbol &construct_name);
  ///
  SgIfStmt(SgExpression &cond, SgStatement &trueBody, SgStatement &falseBody);
  ///
  SgIfStmt(SgExpression &cond, SgStatement &trueBody);
  /// the first stmt in the True clause
  SgStatement *trueBody();      

  // SgBlock is needed? 

  /// i-th stmt in True clause
  SgStatement *trueBody(int i); 

  /// the first stmt in the False
  SgStatement *falseBody();     

  /// i-th statement of the body.
  SgStatement *falseBody(int i);

  /// the while test
  SgExpression *conditional();

  /// 
  SgSymbol *construct_name();
  /// new body=s and lex successors.
  void replaceTrueBody(SgStatement &s);

  /// body=s and lex successors.
  void replaceFalseBody(SgStatement &s);

  ///
  ~SgIfStmt();
};

#if 0

/// For Fortran if then elseif .. elseif ... case.  variant == ELSEIF\_NODE
class  SgIfElseIfStmt: public SgIfStmt {
public:
  ///
  SgIfElseIfStmt(SgExpression &condList, SgStatement &blockList,
                 SgSymbol &constructName);
  ///  the number of conditionals
  int numberOfConditionals();    
  /// block b
  SgStatement *body(int b);  
  /// sets block 
  void setBody(int b);          
  /// the i-th conditional
  SgExpression *conditional(int i);
  /// sets the i-th conditional
  void setConditional(int i);       
  /// 
  void addClause(SgExpression &cond, SgStatement &block);
  /// removes block b and it's conditional
  void removeClause(int b);       
  ///
  ~SgIfElseIfStmt(); 
};

SgIfElseIfStmt::~SgIfElseIfStmt()
{ RemoveFromTableBfnd((void *) this); }; 

#endif

/// For Fortran Arithementic if
//* variant == ARITHIF_NODE
class  SgArithIfStmt: public SgStatement{
public:
  ///
  SgArithIfStmt(int variant);
  ///
  SgArithIfStmt(SgExpression &cond, SgLabel &llabel, SgLabel &elabel, SgLabel &glabel);
  ///
  SgExpression *conditional();
  ///
  void set_conditional(SgExpression &cond);
  /// the <, ==, and > goto labels. in order 0->2.
  SgExpression *label(int i); 
  ///
  void setLabel(SgLabel &label);
  ///
  ~SgArithIfStmt();
};

/// fortran Where stmt. variant == WHERE\_NODE
class  SgWhereStmt: public SgLogIfStmt{
public:
  ///
  SgWhereStmt(SgExpression &cond, SgStatement &body);
  ///
  ~SgWhereStmt();
};

/// fortran Where - Elsewhere stmt. variant == WHERE\_BLOCK\_STMT
class  SgWhereBlockStmt: public SgIfStmt{
public:
  ///
  SgWhereBlockStmt(SgExpression &cond, SgStatement &trueBody, SgStatement &falseBody);
  ///
  ~SgWhereBlockStmt();
};


/// Fortran Case and C switch(). variant == SWITCH\_NODE
class  SgSwitchStmt: public SgStatement{
public:
  ///
  SgSwitchStmt(SgExpression &selector, SgStatement &caseOptionList,
	       SgSymbol &constructName);
  ///
  ~SgSwitchStmt();
  /// the switch selector
  SgExpression *selector(); 
  void setSelector(SgExpression &cond);
  /// the number of cases
  int numberOfCaseOptions();      
  /// i-th case block
  SgStatement *caseOption(int i);  
  ///
  void addCaseOption(SgStatement &caseOption);
#if 0
  ///
  void  deleteCaseOption(int i);
#endif
};

/// Fortran case option statement. variant == CASE\_NODE
class SgCaseOptionStmt: public SgStatement{
public:
  ///
  SgCaseOptionStmt(SgExpression &caseRangeList, SgStatement &body, 
		   SgSymbol &constructName);
  ///
  ~SgCaseOptionStmt();
  
  ///
  SgExpression *caseRangeList();
  ///
  void setCaseRangeList(SgExpression &caseRangeList);
  ///
  SgExpression *caseRange(int i);
  ///
  void setCaseRange(int i, SgExpression &caseRange);
  ///
  SgStatement *body();
  ///
  void setBody(SgStatement &body);
};


/// this is really a non-control, non-declaration stmt. no special functions here.
class  SgExecutableStatement: public  SgStatement{
public: 
   SgExecutableStatement(int variant);
};

/// Fortran assignment Statment. variant == ASSIGN\_STAT
class SgAssignStmt: public SgExecutableStatement{
public:
  ///
  SgAssignStmt(int variant);
  ///
  SgAssignStmt(SgExpression &lhs, SgExpression &rhs);
  /// the left hand side
  SgExpression *lhs();  
  /// the right hand side
  SgExpression *rhs(); 
  /// replace lhs with e
  void replaceLhs(SgExpression &e); 
  /// replace rhs with e
  void replaceRhs(SgExpression &e);
#if 0
  ///
  SgExpression *varReferenced();
  ///
  SgExpression *varUsed();
  ///
  SgExpression *varDefined();
#endif
};


/// C non-control expression Statment. variant == EXPR\_STMT\_NODE
class  SgCExpStmt: public SgExecutableStatement{
public:
  ///
  SgCExpStmt(SgExpression &exp);
  ///
  SgCExpStmt(SgExpression &lhs, SgExpression &rhs);
  /// the expression
  SgExpression *expr();
  /// replace exp with e
  void replaceExpression(SgExpression &e);
  ///
  ~SgCExpStmt();
};

/// Fortran  pointer assignment statement. variant == POINTER\_ASSIGN\_STAT
class  SgPointerAssignStmt: public SgAssignStmt{
public:
  ///
  SgPointerAssignStmt(SgExpression lhs, SgExpression rhs);
  ///
  ~SgPointerAssignStmt();
};

/** Fortran heap space allocation and deallocation statements
 variant == ALLOCATE\_STMT or DEALLOCATE\_STMT 
 heap and nullify statements can be sub-classes of list executable statement class.
 */
class  SgHeapStmt: public SgExecutableStatement{
public:
  ///
  SgHeapStmt(int variant, SgExpression &allocationList, SgExpression &statVariable);
  ///
  ~SgHeapStmt();
  ///
  SgExpression *allocationList();
  ///
  void setAllocationList(SgExpression &allocationList);
  ///
  SgExpression *statVariable();
  ///
  void setStatVariable(SgExpression &statVar);
};

/// Fortran pointer initialization statement. variant == NULLIFY\_STMT 
class  SgNullifyStmt: public SgExecutableStatement{
public:
  ///
  SgNullifyStmt(SgExpression &objectList);
  ///
  ~SgNullifyStmt();
  ///
  SgExpression *nullifyList();
  ///
  void setNullifyList(SgExpression &nullifyList);
};

/// variant == CONT\_STAT in Fortran and variant == CONTINUE\_NODE in C
class  SgContinueStmt: public SgExecutableStatement{
public:
  ///
  SgContinueStmt();
  ///
  ~SgContinueStmt();
};

/// the end of a basic block. variant == CONTROL\_END 
class  SgControlEndStmt: public SgExecutableStatement{
public:
  ///
  SgControlEndStmt(int variant);
  ///
  SgControlEndStmt();
  ///
  ~SgControlEndStmt();
};


/// the end of a basic block. variant == BREAK\_NODE 
class  SgBreakStmt: public SgExecutableStatement{
public:
  ///
  SgBreakStmt();
  ///
  ~SgBreakStmt();
};

/// the fortran 90 cycle statement. variant == CYCLE\_STMT
class  SgCycleStmt: public SgExecutableStatement{
public:
  ///
  SgCycleStmt(SgSymbol &symbol);
  /// the name of the loop to cycle
  SgSymbol *constructName(); 
  ///
  void setConstructName(SgSymbol &constructName);
  ///
  ~SgCycleStmt();
};

/// the return (expr) node variant == RETURN\_NODE/RETURN\_STAT
class  SgReturnStmt: public SgExecutableStatement{
public:
  ///
  SgReturnStmt(SgExpression &returnValue);
  ///
  SgReturnStmt();
  ///
  SgExpression *returnValue();
  ///
  void setReturnValue(SgExpression &retVal);
  ///
  ~SgReturnStmt();
};


/// the fortran 90 exit statement. variant == EXIT\_STMT
class  SgExitStmt: public SgControlEndStmt{
public:
  ///
  SgExitStmt(SgSymbol &construct_name);
  ///
  ~SgExitStmt();
  /// the name of the loop to cycle
  SgSymbol *constructName();  
  ///
  void setConstructName(SgSymbol &constructName);
};

/// the fortran or C goto. variant == GOTO\_NODE
class  SgGotoStmt: public SgExecutableStatement{
public:
  ///
  SgGotoStmt(SgLabel &label);
  ///
  SgLabel *branchLabel();
#if 0
  ///
  SgStatement *target(); //the statement we go to
#endif
  ///
  ~SgGotoStmt();
};


/// the fortran statements containg a list of labels
class SgLabelListStmt: public SgExecutableStatement{
public:
  ///
  SgLabelListStmt(int variant);
  ///
  int numberOfTargets();
  ///
  SgExpression *labelList();
  ///
  void setLabelList(SgExpression &labelList);
#if 0
  ///
  SgStatement *target(int i); //the statement we go to
#endif
};


/// the fortran. variant == ASSGOTO\_NODE
class  SgAssignedGotoStmt: public SgLabelListStmt{
public:
  ///
  SgAssignedGotoStmt(SgSymbol &symbol, SgExpression &labelList);
  ///
  SgSymbol *symbol();
  ///
  void setSymbol(SgSymbol &symb);
  ///
  ~SgAssignedGotoStmt();
};


/// the fortran goto. variant == COMGOTO\_NODE
class  SgComputedGotoStmt: public SgLabelListStmt{
public:
  ///
  SgComputedGotoStmt(SgExpression &expr, SgLabel &label);
  ///
  void addLabel(SgLabel &label);
  ///
  SgExpression *exp();
  ///
  void setExp(SgExpression &exp);
  ///
  ~SgComputedGotoStmt();
};

/// the fortran stop. variant == STOP\_STAT
class  SgStopOrPauseStmt: public SgExecutableStatement{
public:
  ///
  SgStopOrPauseStmt(int variant, SgExpression *expr);
  ///
  SgExpression *exp();
  ///
  void setExp(SgExpression &exp);
  ///
  ~SgStopOrPauseStmt();
};

/// the fortran call. variant == PROC\_STAT
class  SgCallStmt: public SgExecutableStatement{
public:
  ///
  SgCallStmt(SgSymbol &name, SgExpression &args);
  ///
  SgCallStmt(SgSymbol &name);
  /// name of subroutine being called
  SgSymbol *name();    
  /// the number of arguement expressions
  int numberOfArgs(); 
  ///
  void  addArg(SgExpression &arg);
  /// the i-th argument expression
  SgExpression *arg(int i); 
  ///
  ~SgCallStmt();
  
#if 0
  /// global analysis functions
  int numberOfVarsUsed();
  /// i-th region description
  SgExpression *varsUsed(int i);  
  int numberOfVarsMod();
  /// i-th region description
  SgExpression *varsMod(int i); 
#endif
};  


/// the Fortran M process call
//* variant == PROS_STAT
class  SgProsCallStmt: public SgExecutableStatement{
public:
  ///
  SgProsCallStmt(SgSymbol &name, SgExprListExp &args);
  ///
  SgProsCallStmt(SgSymbol &name);
  /// name of process being called
  SgSymbol *name();   
  /// name of process being called
  int numberOfArgs(); 
  ///
  void addArg(SgExpression &arg);
  ///
  SgExprListExp *args();
  /// the i-th argument expression
  SgExpression *arg(int i); 
  ///
  ~SgProsCallStmt();
};
 
 
/// the Fortran M process call with location
//* variant == PROS_STAT_LCTN
class  SgProsCallLctn: public SgExecutableStatement{
public:
  ///
  SgProsCallLctn(SgSymbol &name, SgExprListExp &args, SgExprListExp &lctn);
  ///
  SgProsCallLctn(SgSymbol &name, SgExprListExp &lctn);
  /// name of process being called
  SgSymbol *name();    
  /// the number of arguement expressions
  int numberOfArgs(); 
  ///
  void  addArg(SgExpression &arg);
  ///
  SgExprListExp *args();
  /// the i-th argument expression
  SgExpression *arg(int i);
  ///
  SgExpression *location();
  ///
  ~SgProsCallLctn();
};
 

///  the Fortran M process call with submachine
//* variant == PROS_STAT_SUBM
class  SgProsCallSubm: public SgExecutableStatement{
public:
  ///
  SgProsCallSubm(SgSymbol &name, SgExprListExp &args, SgExprListExp &subm);
  ///
  SgProsCallSubm(SgSymbol &name, SgExprListExp &subm);
  /// name of process being called
  SgSymbol *name();    
  /// the number of arguement expressions
  int numberOfArgs(); 
  ///
  void  addArg(SgExpression &arg);
  ///
  SgExprListExp *args();
  ///  the i-th argument expression
  SgExpression *arg(int i); 
  ///
  SgExpression *submachine();
  ///
  ~SgProsCallSubm();
};
 

/// the Fortran M processes statement
//* variant == PROCESSES_STAT
class  SgProcessesStmt: public SgStatement{
public:
  ///
  SgProcessesStmt();
  ///
  ~SgProcessesStmt();
};
 

/// the Fortran M endprocesses statement 
//* variant == PROCESSES_END
class  SgEndProcessesStmt: public SgStatement{
public:
  ///
  SgEndProcessesStmt();
  ///
  ~SgEndProcessesStmt();
};


/// variant == PORT\_TYPE\_OP, INPORT\_TYPE\_OP, or OUTPORT\_TYPE\_OP
class SgPortTypeExp: public SgExpression{
public:
  ///
  SgPortTypeExp(SgType &type);
  ///
  SgPortTypeExp(SgType &type, SgExpression &ref);
  ///
  SgPortTypeExp(int variant, SgExpression &porttype);
  ///
  ~SgPortTypeExp();
  ///
  SgType *type();
  ///
  int numberOfRef();
  /// return a ref or a port type
  SgExpression *ref(); 
  ///
  SgPortTypeExp *next();
};
 
 
/// the Fortran M inport statement
//* variant == INPORT_DECL
class SgInportStmt: public SgStatement
{
public:
  ///
  SgInportStmt(SgExprListExp &name);
  ///
  SgInportStmt(SgExprListExp &name, SgPortTypeExp &porttype);
  ///
  ~SgInportStmt();
  ///
  void addname(SgExpression &name);
  ///
  int numberOfNames();
  ///
  SgExprListExp *names();
  ///
  SgExpression *name(int i);
  ///
  void addporttype(SgExpression &porttype);
  ///
  int numberOfPortTypes();
  ///
  SgPortTypeExp *porttypes();
  ///
  SgPortTypeExp *porttype(int i);
};

 
/// the Fortran M outport statement. variant == OUTPORT\_DECL
class SgOutportStmt: public SgStatement{
public:
  ///
  SgOutportStmt(SgExprListExp &name);
  ///
  SgOutportStmt(SgExprListExp &name, SgPortTypeExp &porttype);
  ///
  ~SgOutportStmt();
  ///
  void addname(SgExpression &name);
  ///
  int numberOfNames();
  ///
  SgExprListExp *names();
  ///
  SgExpression *name(int i);
  ///
  void addporttype(SgExpression &porttype);
  ///
  int numberOfPortTypes();
  ///
  SgPortTypeExp *porttypes();
  ///
  SgPortTypeExp *porttype(int i);
};

 
/// the Fortran M channel statement. variant == CHANNEL\_STAT
class SgChannelStmt: public SgStatement{
public:
  ///
  SgChannelStmt(SgExpression &outport, SgExpression &inport);
  ///
  SgChannelStmt(SgExpression &outport, SgExpression &inport,
		SgExpression &io_or_err);
  ///
  SgChannelStmt(SgExpression &outport, SgExpression &inport,
		SgExpression &iostore, SgExpression &errlabel);
  ///
  ~SgChannelStmt();
   ///
  SgExpression *outport();
  ///
  SgExpression *inport();
  ///
  SgExpression *ioStore();
   ///
  SgExpression *errLabel();
};
 

/// the Fortran M merger statement. variant == MERGER\_STAT
class SgMergerStmt: public SgStatement{
public:
  ///
  SgMergerStmt(SgExpression &outport, SgExpression &inport);
  ///
  SgMergerStmt(SgExpression &outport, SgExpression &inport,
	       SgExpression &io_or_err);
  ///
  SgMergerStmt(SgExpression &outport, SgExpression &inport,
	       SgExpression &iostore, SgExpression &errlabel);
  ///
  ~SgMergerStmt();
  ///
  void addOutport(SgExpression &outport);
  /// can't add it before outports
  void addIoStore(SgExpression &iostore); 
  /// can't add it before iostore
  void addErrLabel(SgExpression &errlabel); 
  ///
  int numberOfOutports();
  ///
  SgExpression *outport(int i);
  ///
  SgExpression *inport();
  ///
  SgExpression *ioStore();
  ///
  SgExpression *errLabel();
};


/// the Fortran M moveport statement
//* variant == MOVE_PORT
class SgMoveportStmt: public SgStatement{
public:
  ///
  SgMoveportStmt(SgExpression &fromport, SgExpression &toport);
  ///
  SgMoveportStmt(SgExpression &fromport, SgExpression &toport,
		 SgExpression &io_or_err);
  ///
  SgMoveportStmt(SgExpression &fromport, SgExpression &toport,
		 SgExpression &iostore, SgExpression &errlabel);
  ///
  ~SgMoveportStmt();
  ///
  SgExpression *fromport();
  ///
  SgExpression *toport();
  ///
  SgExpression *ioStore();
  ///
  SgExpression *errLabel();
};
 

/// the Fortran M send statement
//* variant == SEND_STAT
class SgSendStmt: public SgStatement{
public:
  ///
  SgSendStmt(SgExpression &control, SgExprListExp &argument);
  ///
  SgSendStmt(SgExpression &outport, SgExprListExp &argument, SgExpression &io_or_err);
  ///
  SgSendStmt(SgExpression &outport, SgExprListExp &argument, SgExpression &iostore, SgExpression &errlabel);
  ///
  ~SgSendStmt();
  ///
  void addOutport(SgExpression &outport);
  /// can't add it before outports
  void addIoStore(SgExpression &iostore); 
  /// add it before iostore
  void addErrLabel(SgExpression &errlabel);
  ///
  void addArgument(SgExpression &argument);
  ///
  int numberOfOutports();
  ///
  int numberOfArguments();
  ///
  SgExpression *controls();
  ///
  SgExpression *outport(int i);
  ///
  SgExprListExp *arguments();
   ///
  SgExpression *argument(int i);
  ///
  SgExpression *ioStore();
  ///
  SgExpression *errLabel();
};

/// the Fortran M receive statement
//* variant == RECEIVE_STAT
class SgReceiveStmt: public SgStatement{
public:
  ///
  SgReceiveStmt(SgExpression &control, SgExprListExp &argument);
  ///
  SgReceiveStmt(SgExpression &inport, SgExprListExp &argument,
		SgExpression &e1);
  ///
  SgReceiveStmt(SgExpression &inport, SgExprListExp &argument,
		SgExpression &e1, SgExpression &e2);
  ///
  SgReceiveStmt(SgExpression &inport, SgExprListExp &argument,
		SgExpression &e1, SgExpression &e2, SgExpression &e3);
  ///
  ~SgReceiveStmt();
  ///
  void addInport(SgExpression &inport);
  /// can't add it before inports
  void addIoStore(SgExpression &iostore);
  /// add it before iostore
  void addErrLabel(SgExpression &errlabel);
  /// add it before errlabel
  void addEndLabel(SgExpression &endlabel);
  /// 
  void addArgument(SgExpression &argument);
  ///
  int numberOfInports();
  ///
  int numberOfArguments();
  ///
  SgExpression *controls();
  ///
  SgExpression *inport(int i);
  ///
  SgExprListExp *arguments();
  ///
  SgExpression *argument(int i);
  ///
  SgExpression *ioStore();
  ///
  SgExpression *errLabel();
  ///
  SgExpression *endLabel();
};



/// the Fortran M endchannel statement
//* variant == ENDCHANNEL_STAT
class SgEndchannelStmt: public SgStatement{
public:
  ///
  SgEndchannelStmt(SgExpression &outport);
  ///
  SgEndchannelStmt(SgExpression &outport, SgExpression &io_or_err);
  ///
  SgEndchannelStmt(SgExpression &outport, SgExpression &iostore,
                          SgExpression &errlabel);
  ///
  ~SgEndchannelStmt();
  ///
  void addOutport(SgExpression &outport);
  /// can't add it before outports
  void addIoStore(SgExpression &iostore);
  /// can't add it before iostore
  void addErrLabel(SgExpression &errlabel);
  int numberOfOutports();
  ///
  SgExpression *controls();
  ///
  SgExpression *outport(int i);
  ///
  SgExpression *ioStore();
  ///
  SgExpression *errLabel();
};


/// the Fortran M probe statement
//* variant == PROBE_STAT
class SgProbeStmt: public SgStatement{
public:
  ///
  SgProbeStmt(SgExpression &inport);
  ///
  SgProbeStmt(SgExpression &inport, SgExpression &e1);
  ///
  SgProbeStmt(SgExpression &inport, SgExpression &e1,
	      SgExpression &e2);
  ///
  SgProbeStmt(SgExpression &inport, SgExpression &e1,
	      SgExpression &e2, SgExpression &e3);
  ///
  ~SgProbeStmt();
  ///
  void addInport(SgExpression &inport);
  /// can't add before inports
  void addIoStore(SgExpression &iostore);
  /// can't add before iostore
  void addErrLabel(SgExpression &errlabel);
  /// can't add before errlabel
  void addEmptyStore(SgExpression &endlabel);
  ///
  int numberOfInports();
  ///
  SgExpression *controls();
  ///
  SgExpression *inport(int i);
  ///
  SgExpression *ioStore();
  ///
  SgExpression *errLabel();
  ///
  SgExpression *emptyStore();
};


/// variant == PROCESSORS\_REF
class SgProcessorsRefExp: public SgExpression{
public:
  ///
  SgProcessorsRefExp(PTR_LLND ll);
  ///
  SgProcessorsRefExp();
  ///
  SgProcessorsRefExp(SgExpression &subscripts);
  ///
  SgProcessorsRefExp(SgExpression &sub1,SgExpression &sub2);
 
  ///
  SgProcessorsRefExp(SgExpression &sub1,SgExpression &sub2,
		     SgExpression &sub3);
  
  ///
  SgProcessorsRefExp(SgExpression &sub1,SgExpression &sub2,
		     SgExpression &sub3,SgExpression &sub4);
  ///
  ~SgProcessorsRefExp();
  /// the number of subscripts in reference
  int numberOfSubscripts();  
  ///
  SgExpression *subscripts();
  ///
  SgExpression *subscript(int i);
  ///
  void addSubscript(SgExpression &e);
};

/** parent of INPORT\_NAME, OUTPORT\_NAME, FROMPORT\_NAME, TOPORT\_NAME
          IOSTAT\_STORE, EMPTY\_STORE, ERR\_LABEL, END\_LABEL
*/

class SgControlExp: public SgExpression{
public:
  ///
  SgControlExp(int variant);
  ///
  ~SgControlExp();
  ///
  SgExpression *exp();
};


/// variant == INPORT\_NAME
class SgInportExp: public SgControlExp{
public:
  ///
  SgInportExp(SgExprListExp &exp);
  ///
  ~SgInportExp();
};


/// variant == OUTPORT\_NAME
class SgOutportExp: public SgControlExp{
public:
  ///
  SgOutportExp(SgExprListExp &exp);
  ///
  ~SgOutportExp();
};


/// variant == FROMPORT\_NAME
class SgFromportExp: public SgControlExp{
public:
  ///
  SgFromportExp(SgExprListExp &exp);
  ///
  ~SgFromportExp();
};


///
//* variant == TOPORT_NAME
class SgToportExp: public SgControlExp{
public:
  ///
  SgToportExp(SgExprListExp &exp);
  ///
  ~SgToportExp();
};


/// variant == IOSTAT\_STORE
class SgIO_statStoreExp: public SgControlExp{
public:
  ///
  SgIO_statStoreExp(SgExprListExp &exp);
  ///
  ~SgIO_statStoreExp();
};


/// variant == EMPTY\_STORE
class SgEmptyStoreExp: public SgControlExp{
public:
  ///
  SgEmptyStoreExp(SgExprListExp &exp);
  ///
  ~SgEmptyStoreExp();
};


/// variant == ERR\_LABEL
class SgErrLabelExp: public SgControlExp{  
public:
  ///
  SgErrLabelExp(SgExprListExp &exp);
  ///
  ~SgErrLabelExp();
};

/// variant == END\_LABEL
class SgEndLabelExp: public SgControlExp{
public:
  ///
  SgEndLabelExp(SgExprListExp &exp);
  ///
  ~SgEndLabelExp();
};

/// variant == DATA\_IMPL\_DO
class SgDataImpliedDoExp: public SgExpression{
public:
  ///
  SgDataImpliedDoExp(SgExprListExp &dlist, SgSymbol &iname,
		     SgExprListExp &ilist);
  ///
  ~SgDataImpliedDoExp();
  ///
  void addDataelt(SgExpression &data);
  ///
  void addIconexpr(SgExpression &icon);
  ///
  SgSymbol *iname();
  ///
  int numberOfDataelt();
  ///
  SgExprListExp *dataelts();

  /**
     only the first 3 elements in the
     iconexpr list are useful. They represent
     the initial value, the limit, and the 
     increment of the implied do expression
     respectively 
  */
  SgExprListExp *iconexprs(); 
  ///
  SgExpression *dataelt(int i);
  ///
  SgExpression *init();
  ///
  SgExpression *limit();
  ///
  SgExpression *increment();
};
 

/// variant == DATA\_ELT
class SgDataEltExp: public SgExpression{
public:
  ///
  SgDataEltExp(SgExpression &dataimplieddo);
  ///
  SgDataEltExp(SgSymbol &name, SgExpression &datasubs,
	       SgExpression &datarange);
  ///
  ~SgDataEltExp();
  ///
  SgExpression *dataimplieddo();
  ///
  SgSymbol *name();
  ///
  SgExpression *datasubs();
  ///
  SgExpression *datarange();
};


/// variant == DATA\_SUBS
class SgDataSubsExp: public SgExpression{
public:
  ///
  SgDataSubsExp(SgExprListExp &iconexprlist);
  ///
  ~SgDataSubsExp();
  ///
  SgExprListExp *iconexprlist();
};

/// variant == DATA\_RANGE
class SgDataRangeExp: public SgExpression{
public:
  ///
  SgDataRangeExp(SgExpression &iconexpr1, SgExpression &iconexpr2);
  ///
  ~SgDataRangeExp();
  ///
  SgExpression *iconexpr1();
  ///
  SgExpression *iconexpr2();
};


/// variant == ICON\_EXPR
class SgIconExprExp: public SgExpression{
public:
  ///
  SgIconExprExp(SgExpression &expr);
  ///
  ~SgIconExprExp();
  ///
  SgExpression *expr();
};


/// variant == IMPLIED\_DO;
class SgImpliedDoExp: public SgExpression {
public:
  ///
  SgImpliedDoExp();
  ///
  SgSymbol * index();
  ///
  int numberOfTriplet();
  ///
  SgExpression * triplet(int);
  ///
  SgExpression * list();
  ///
  ~SgImpliedDoExp();
};

/// fortran input/output and their control statements. abstract class
class SgIOStmt: public SgExecutableStatement{
public:
   SgIOStmt(int variant);
};

/// fortran input and output statements. variant = READ\_STAT, WRITE\_STATE, PRINT\_STAT
class SgInputOutputStmt: public SgIOStmt{
public:
  ///
  SgInputOutputStmt(int variant, SgExpression &specList, SgExpression &itemList);
  ///
  SgExpression *specList();
  ///
  void setSpecList(SgExpression &specList);
  ///
   SgExpression *itemList();
  ///
  void setItemList(SgExpression &itemList);
  ///
  ~SgInputOutputStmt();
};

/** fortran input/output control and editing statements
 variant = OPEN\_STAT, CLOSE\_STAT, INQUIRE\_STAT, BACKSPACE\_STAT,
 REWIND\_STAT, ENDFILE\_STAT, FORMAT\_STAT
*/

class SgIOControlStmt: public SgExecutableStatement{
public:
  ///
  SgIOControlStmt(int variant, SgExpression &controlSpecifierList);
  ///
  SgExpression *controlSpecList();
  ///
  void setControlSpecList(SgExpression &controlSpecList);
  ///
  ~SgIOControlStmt();
};

// ******************** Declaration Nodes ***************************

/// Declaration class. abstract class
class  SgDeclarationStatement: public SgStatement{
public:
  ///
  SgDeclarationStatement(int variant);
  ///
  ~SgDeclarationStatement();
  
  ///
  SgExpression *varList();
  ///
  int numberOfVars();
  ///
  SgExpression *var(int i);
  ///
  void deleteVar(int i);
  ///
  void deleteTheVar(SgExpression &var);
  ///
  void addVar(SgExpression &exp);
};

/// Declaration Statement. variant == VAR\_DECL
class  SgVarDeclStmt: public SgDeclarationStatement{
public:
  ///
  // varRefValList is a list of low-level nodes of
  // variants VAR_REFs or ARRAY_REFs or ASSIGN_OPs
  SgVarDeclStmt(SgExpression &varRefValList, SgExpression &attributeList, SgType &type);
  ///
  SgVarDeclStmt(SgExpression &varRefValList, SgType &type);
  ///
  SgVarDeclStmt(SgExpression &varRefValList);
  ///
  ~SgVarDeclStmt();
  /// the type; 
  SgType *type();  

  /** the number of F90 attributes;
   the attributes are: PARAMETER\_OP | PUBLIC\_OP |
      PRIVATE\_OP | ALLOCATABLE\_OP | EXTERNAL\_OP |
      OPTIONAL\_OP | POINTER\_OP | SAVE\_OP TARGET\_OP
  */
  int numberOfAttributes(); 

#if 0
  ///
  SgExpression &attribute(int i);
#endif
  /// the number of variables declared;        
  int numberOfSymbols(); 
  ///
  SgSymbol *symbol(int i);
  
  /// 
  void deleteSymbol(int i); //i start from 0

  ///
  void deleteTheSymbol(SgSymbol &symbol);

  /// the initial value ofthe i-th variable

  SgExpression *initialValue(int i);  

  // The complete ASSGN_OP
  SgExpression *completeInitialValue(int i); 

  // expression of the initial value 
  // *BW* from M. Golden
  
  /** sets the initial value ofthe i-th variable
   an alternative way to initialize variables. The low-level node (VAR\_REF or ARRAY\_REF) is
   replaced by a ASSIGN\_OP low-level node.
  */

  void setInitialValue(int i, SgExpression &initVal); 

  /// removes initial value of the i-th declaration 
  void clearInitialValue(int i);  
};

/// the Fortran M Intent Statement. variant == INTENT\_STMT
class  SgIntentStmt: public SgDeclarationStatement{
public:
  ///
  SgIntentStmt(SgExpression &varRefValList, SgExpression &attribute);
  ///
  ~SgIntentStmt();
  ///  the number of arguement expressions
  int numberOfArgs(); 
  ///
  void  addArg(SgExpression &arg);
  ///
  SgExpression *args();
  ///  the i-th argument expression
  SgExpression *arg(int i); 
  ///
  SgExpression *attribute();
};


/** Declaration Statement.
 variant == OPTIONAL\_STMT, SAVE\_STMT, PUBLIC\_STMT,
 PRIVATE\_STMT, EXTERNAL\_STAT, INTRINSIC\_STAT, DIM\_STAT, 
 ALLOCATABLE\_STAT, POINTER\_STAT, TARGET\_STAT, MODULE\_PROC\_STMT,
 PROCESSORS\_STAT (for Fortran M processors statement)
*/

class  SgVarListDeclStmt: public SgDeclarationStatement{
public:
  ///
  SgVarListDeclStmt(int variant, SgExpression &symbolRefList);
  ///
  SgVarListDeclStmt(int variant, SgSymbol &symbolList, SgStatement &scope);
  
  ///
  ~SgVarListDeclStmt();
  
  ///
  int numberOfSymbols();
  ///
  SgSymbol *symbol(int i);
  ///
  void appendSymbol(SgSymbol &symbol);
  ///
  void deleteSymbol(int i);
  ///
  void deleteTheSymbol(SgSymbol &symbol);
};


/// Fortran 90 structure declaration statement. variant == STRUCT\_DECL
class SgStructureDeclStmt: public SgDeclarationStatement{
public:
  ///
  SgStructureDeclStmt(SgSymbol &name, SgExpression &attributes, SgStatement &body);
  ///
  ~SgStructureDeclStmt();
  
#if 0
  ///
  int isPrivate();
  ///
  int isPublic();
  ///
  int isSequence();
#endif
};

/** Declaration statement.
 variant == NAMELIST\_STAT, EQUI\_STAT, COMM\_STAT,
            and  PROS\_COMM for Fortran M

 These statements have the format of a list of variable lists. For example,
 EQUIVALENCE (A, C, D), (B, G, F), ....
*/

class SgNestedVarListDeclStmt: public SgDeclarationStatement{
public:

  ///
  // varList must be of low-level variant appropriate to variant. For example,
  // if the variant is COMM_STAT, listOfVarList must be of variant COMM_LIST.
  SgNestedVarListDeclStmt(int variant, SgExpression &listOfVarList);

  /// 
  ~SgNestedVarListDeclStmt();
  
  ///
  SgExpression *lists();
  ///
  int numberOfLists();
  ///
  SgExpression *list(int i);

#if 0
  ///
  SgExpression *leadingVar(int i);
#endif

  ///
  void addList(SgExpression &list);
  ///
  void  addVarToList(SgExpression &varRef);
  ///
  void deleteList(int i);
  ///
  void deleteTheList(SgExpression &list);
  ///
  void deleteVarInList(int i, SgExpression &varRef);
  ///
  void deleteVarInTheList(SgExpression &list, SgExpression &varRef);
};

/// Fortran constants declaration statement. variant = PARAM\_DECL
class SgParameterStmt: public SgDeclarationStatement{      
public:
  ///
  SgParameterStmt(SgExpression &constants, SgExpression &values);
  ///
  SgParameterStmt(SgExpression &constantsWithValues);
  ///
  ~SgParameterStmt();
  
  /// the number of constants declared
  int numberOfConstants();    
  
  /// the i-th variable
  SgSymbol *constant(int i); 

  /// the value of i-th variable  
  SgExpression *value(int i); 
  
  ///
  void addConstant(SgSymbol &constant, SgExpression &value);
  ///
  void deleteConstant(int i);
  ///
  void deleteTheConstant(SgSymbol &constant);
};

/// Fortran implicit type declaration statement. variant = IMPL\_DECL
class SgImplicitStmt: public SgDeclarationStatement{      
public:
  ///
  SgImplicitStmt(SgExpression &implicitLists);
  ///
  ~SgImplicitStmt();  
  /// the number of implicit types declared;
  int numberOfImplicitTypes();  
  /// the i-th implicit type
  SgType *implicitType(int i); 
  ///
  SgExpression *implicitRangeList(int i) ;
  ///
  void  appendImplicitNode(SgExpression &impNode);

#if 0
  ///
  void  addImplicitType(SgType Type, char alphabet[]);
  ///
  int deleteImplicitItem(int i);
  ///
  int deleteTheImplicitItem(SgExpression &implicitItem);
#endif

};

#if 0

/// Fortran 90 module usuage statement. variant = USE\_STMT
class SgUseStmt: public SgDeclarationStatement{
public:
  /// renameList must be a list of low-level nodes of variant RENAME\_NODE
  SgUseStmt(SgSymbol &moduleName, SgExpression &renameList, SgStatement &scope);
  ///
  ~SgUseStmt();  
  ///
  int isOnly();
  ///
  SgSymbol *moduleName();
  ///
  void setModuleName(SgSymbol &moduleName);
  ///
  int numberOfRenames();
  ///
  SgExpression *renameNode(int i);
  ///
  void  addRename(SgSymbol &localName, SgSymbol &useName);
  ///
  void  addRenameNode(SgExpression &renameNode);
  ///
  void  deleteRenameNode(int i);
  ///
  void deleteTheRenameNode(SgExpression &renameNode);
};

/// Fortran statement function declaration. variant == STMTFN\_DECL
class  SgStmtFunctionStmt: public SgDeclarationStatement{
public:        
  ///
  SgStmtFunctionStmt(SgSymbol &name, SgExpression &args, SgStatement Body);
  ///
  ~SgStmtFunctionStmt();
  ///
  SgSymbol *name();
  ///
  void setName(SgSymbol &name);
  ///
  SgType *type();
  /// the number of parameters
  int numberOfParameters();       
  ///  the i-th parameter
  SgSymbol *parameter(int i);
};      

/// Fortran 90 simple miscellaneous statements. variant == CONTAINS\_STMT, PRIVATE\_STMT, SEQUENCE\_STMT
class  SgMiscellStmt: public SgDeclarationStatement{
public:
  ///
  SgMiscellStmt(int variant);
  ///
  ~SgMiscellStmt();
};      

#endif

/// for HPF, INHERIT\_DECL 
class  SgInheritStmt : public SgStatement {
public:
  ///
  SgInheritStmt();
  ///
  int numberOfVars();
  ///
  SgExpression * var(int);

};


// More stuffs for types and symbols

/** a variable. variant = VARIABLE\_NAME */

class SgVariableSymb: public SgSymbol {
public:
  ///
  SgVariableSymb(char *identifier, SgType &t, SgStatement &scope);
  ///
  SgVariableSymb(char *identifier, SgType *t, SgStatement *scope);
  ///
  SgVariableSymb(char *identifier, SgType &t);
  ///
  SgVariableSymb(char *identifier,  SgStatement &scope);
   ///
  SgVariableSymb(char *identifier,  SgStatement *scope);
   ///
  SgVariableSymb(char *identifier);
  ///
  ~SgVariableSymb();
  
  /// This function allocates and returns a new variable reference expression to this symbol. 
  /* (ajm) */
  SgVarRefExp *varRef (void);

#if 0
  ///
  int isAttributeSet(int attribute);
  ///
  void setAttribute(int attribute);
  
  /// number of uses.
  int numberOfUses();            
  ///  statement where i-th use occurs
  SgStatement  *useStmt(int i); 
  /// expression where i-th use occurs
  SgExpression *useExpr(int i);  
  ///
  int numberOfDefs();
#endif 

};

/// a symbol for a constant object. variant == CONST\_NAME 
class SgConstantSymb: public SgSymbol{
public:
  ///
  SgConstantSymb(char *identifier, SgStatement &scope, 
                 SgExpression &value);
  ///
  ~SgConstantSymb();
  ///
  SgExpression *constantValue();
  /// set const\_value    
  void setConstantValue(SgExpression *value); //zheng.gb
};


/// a subroutine, function or main program. variant == PROGRAM\_NAME, PROCEDURE\_NAME, or FUNCTION\_NAME
class SgFunctionSymb: public SgSymbol{
public:
  ///
  SgFunctionSymb(int variant);
  ///
  SgFunctionSymb(int variant, char *identifier, SgType &t, 
                 SgStatement &scope);
  ///
  ~SgFunctionSymb();
  ///
  void addParameter(int, SgSymbol &parameters);
  ///
  void insertParameter(int position, SgSymbol &symb);
  ///
  int numberOfParameters();
  ///
  SgSymbol *parameter(int i); 
  ///
  SgSymbol *result();
  ///
  void  setResult(SgSymbol &symbol);
#if 0
  ///
  int isRecursive();
  ///
  int setRecursive();
#endif
  SgStatement *function_hedr();
};


/** a member function for a class or struct or collection.
 variant = MEMBER\_FUNC
 may be either MEMB\_PRIVATE, MEMB\_PUBLIC,
 MEMP\_METHOELEM or MEMB\_PROTECTED
*/
class SgMemberFuncSymb: public SgFunctionSymb{
public:
  ///
  SgMemberFuncSymb(char *identifier, SgType &t, SgStatement &cla,
		   int status);
  ///
  ~SgMemberFuncSymb();
#if 0
  ///
  int status();       
  ///
  int isVirtual();       // 1 if virtual.
#endif
  ///
  int isMethodOfElement();
  ///
  SgSymbol *className();
  ///
  void setClassName(SgSymbol &symb);
};

/// a field in an enum or in a struct. variant == ENUM\_NAME or FIELD\_NAME
class SgFieldSymb: public SgSymbol{
public:
  ///  
  // no check is made to see if the field "identifier"
  // already exists in the structure. 
  SgFieldSymb(char *identifier, SgType &t, SgSymbol &structureName);
  ///
  ~SgFieldSymb();
  /// position in the structure
  int offset(); 
  /// parent structure 
  SgSymbol *structureName(); 
  ///
  SgSymbol *nextField();
  ///
  int isMethodOfElement();
#if 0
  ///
  int isPrivate();
  ///
  int isSequence();
  ///
  void  setPrivate();
  ///
  void setSequence();
#endif
};

/** the class, union, struct and collection type.
 variant == CLASS\_NAME, UNION\_NAME, STRUCT\_NAME or COLLECTION\_NAME
*/

class SgClassSymb: public SgSymbol{
public:
  SgClassSymb(int variant, char *name, SgStatement &scope);
  ///
  ~SgClassSymb();
  ///
  int numberOfFields();
  ///
  SgSymbol *field(int i);
};

#if 0

/// a C typedef.  the type() function returns the base type. variant == TYPE\_NAME
class SgTypeSymb: public SgSymbol{
public:
  ///
  SgTypeSymb(char *name, SgType &baseType);
  ///
  SgType &baseType();
  ///
  ~SgTypeSymb();
};

#endif

/// a C label name. variant == LABEL\_NAME
class SgLabelSymb: public SgSymbol{
public:
  ///
  SgLabelSymb(char *name);
  ///
  ~SgLabelSymb();
};

/// a Fortran label variable for an assigned goto stmt. variant == LABEL\_NAME
class SgLabelVarSymb: public SgSymbol{
public:
  ///
  SgLabelVarSymb(char *name, SgStatement &scope);
  ///
  ~SgLabelVarSymb();
};

/// for fortran external statement. variant == ROUTINE\_NAME
class SgExternalSymb: public SgSymbol{
public:
  ///
  SgExternalSymb(char *name, SgStatement &scope);
  ///
  ~SgExternalSymb();
};

/// for fortran statement with construct names. variant == CONSTRUCT\_NAME
class SgConstructSymb: public SgSymbol{
public:
  ///
  SgConstructSymb(char *name, SgStatement &scope);
  ///
  ~SgConstructSymb();
};

// A lot of work needs to be done on this class.

/// for fortran interface statement. variant == INTERFACE\_NAME
class SgInterfaceSymb: public SgSymbol{
public:
  ///
  SgInterfaceSymb(char *name, SgStatement &scope);
  ///
  ~SgInterfaceSymb();
};

// A lot of work needs to be done on this class.
/// for fortran module statement
//* variant == MODULE_NAME
class SgModuleSymb: public SgSymbol{
public:
  ///
  SgModuleSymb(char *name);
  ///
  ~SgModuleSymb();
};

// ********************* Types *******************************

/// A new array type is generated for each array. variant == T\_ARRAY
class SgArrayType: public SgType{
public:
  ///
  SgArrayType(SgType &base_type);

  ///
  SgArrayType(SgType &base_type,int dim, SgExpression &ranges);

  ///
  int dimension();
  ///
  SgExpression *sizeInDim(int i);
  ///
  void addDimension(SgExpression *e);
  ///
  SgExpression * getDimList();
  ///
  SgType * baseType();
  ///
  void setBaseType(SgType &bt);
  ///
  void addRange(SgExpression &e);
  ///  get the array size if caculable , -1  failed 
  int  getSize();
  ~SgArrayType();
};


#if 0

/**
  a C struct or Fortran Record, a C++ class, a C Union and a C Enum
  and a pC++ collection.  note: derived classes are another type.
  this type is very simple.  it only contains the standard type
  info from SgType and a pointer to the class declaration stmt
  and a pointer to the symbol that is the first field in the struct.
  variant == T\_STRUCT, T\_ENUM, T\_CLASS, T\_ENUM, T\_COLLECTION
*/

class SgClassType: public SgType{
public:

  /** why is struct\_decl needed. No appropriate field found.
   assumes that first\_field has been declared as
   FIELD\_NAME and the remaining fields have been stringed to it.
  */
  SgClassType(int variant, char *name, SgStatement &struct_decl, int num_fields,
              SgSymbol &first_field);
  ///
  SgStatement &structureDecl();
  ///
  SgSymbol *firstFieldSymb();
  ///
  SgSymbol *fieldSymb(int i);
  ///
  ~SgClassType();
};

#endif


/** A pointer type contains only one hany bit of information: the base type.
 can also have a modifier like BIT\_CONST BIT\_GLOBAL. see SgDescriptType.
 variant == T\_POINTER
*/
class SgPointerType: public SgType{
public:
  ///
  SgPointerType(SgType &base_type);
  ///
  SgPointerType(SgType *base_type);
  ///
  SgType *baseType();
  ///
  int indirection();
  ///
  void setIndirection(int);
  ///
  int modifierFlag();
  ///
  void setModifierFlag(int flag); 
  ///
  void setBaseType(SgType &baseType);
  ///
   ~SgPointerType();
};


/// Function Types have a returned value type. variant == T\_FUNCTION
class SgFunctionType: public SgType{
public:
  ///
  SgFunctionType(SgType &return_val_type);
  ///
  SgType *returnedValue();
  ///
  void changeReturnedValue(SgType &rv);
  ///
  ~SgFunctionType();
};


/// A reference (\&xx in c+=) type contains only one hany bit of information: the base type. variant == T\_REFERENCE
class SgReferenceType: public SgType{
public:
  ///
  SgReferenceType(SgType &base_type);
  ///
  SgType *baseType();
  ///
  void setBaseType(SgType &baseType);
  ///
  ~SgReferenceType();
  ///
  int modifierFlag();
  ///
  void setModifierFlag(int flag); 
};

/** for example:  typedef int integer;  go to the symbol table
 for the base type and Id.
 variant == T\_DERIVED\_TYPE
*/
class SgDerivedType: public SgType{
public:
  ///
  SgDerivedType(SgSymbol &type_name);
  ///
  SgSymbol * typeName();
  ///
  ~SgDerivedType();
};

/// variant == T\_DERIVED\_CLASS
class SgDerivedClassType: public SgType{
  // for example:  typedef int integer;  go to the symbol table
  // for the base type and Id.

public:
  ///
  SgDerivedClassType(SgSymbol &type_name);
  ///
  SgSymbol *typeName();
  ///
  ~SgDerivedClassType();
};

/// this is the type for a template object: T\_DERIVED\_TEMPLATE
class SgDerivedTemplateType: public SgType{
public:
  ///
  SgDerivedTemplateType(SgExpression *arg_vals, SgSymbol *classname);
  ///
  SgExpression *argList();
  ///
  void addArg(SgExpression *arg);
  ///
  int numberOfArgs();
  ///
  SgExpression  *arg(int i);
  ///
  void setName(SgSymbol &s);
  /// the name of the template class.
  SgSymbol *typeName(); 
};


/** for example in C: long volatile int x; 
 long and volatile are modifiers and there is a descriptor
 type whose base type is the real type of x.
 the modifier is an integer with bits set if the modifier
 holds.
 the bits are:
 BIT\_SYN, BIT\_SHARED, BIT\_PRIVATE, BIT\_FUTURE, BIT\_VIRTUAL, 
 BIT\_, BIT\_UNSIGNED, BIT\_SIGNED, BIT\_LONG, BIT\_SHORT,
 BIT\_VOLATILE, BIT\_CONST, BIT\_TYPEDEF, BIT\_EXTERN, BIT\_AUTO,
 BIT\_STATIC, BIT\_REGISTER, BIT\_FRIEND, BIT\_GLOBAL, and more.

 variant = T\_DESCRIPT
 */
class SgDescriptType: public SgType{
public:
  ///
  SgDescriptType(SgType &base_type, int bit_flag);
  ///
  int modifierFlag();
  ///
  void setModifierFlag(int flag);
  ///
  ~SgDescriptType();
};

/**
 for example:
 Collection DistributedArray {body1} ;
 class object {body2} ;
 DistributedArray<object>  X;
 X is of type with variant = T\_DERIVED\_COLLECTION
*/
class SgDerivedCollectionType: public SgType{
public:
  ///
  SgDerivedCollectionType(SgSymbol &s, SgType &t);
  ///
  SgType *elementClass();
  ///
  void setElementClass(SgType &ty);
  ///
  SgSymbol *collectionName();
  ///
   SgStatement *createCollectionWithElemType();
  ///
  ~SgDerivedCollectionType();
};

//  following are added by liyu

/// for variant=DISTRIBUTE\_OP
class SgDistributeAttrExp : public SgExpression
{
  ///
  inline SgDistributeAttrExp(PTR_LLND ll);
};

/// for variant=ALIGN\_ATT
class SgAlignAttrExp : public SgExpression
{
  ///
  inline SgAlignAttrExp(PTR_LLND ll);
};

/// for variant=BLOCK\_OP,CYCLIC\_OP,COLON\_OP,STAR\_OP
class SgDataDistExp : public SgExpression
{
  ///
  inline SgDataDistExp(PTR_LLND ll);
};

//@}

// end of liyu addition

/** @name the Global Functions 
these global functions are useful when reconstructing the source program.
*/
//@{
 
/**This function is used to create a pointer pointing to some object of Class SgType whose atomic type of public fields is INTEGER */
SgType *SgTypeInt();
/**This function is used to create a pointer pointing to some object of Class SgType whose atomic type of public fields is CHARACTER */
SgType *SgTypeChar();
/**This function is used to create a pointer pointing to some object of class SgType whose atomic type of public fields is FLOAT   */

SgType *SgTypeFloat();
/**This function is used to create a pointer pointing to some object of Class SgType whose atomic type of public fields is DOUBLE */
   
SgType *SgTypeDouble();
/**This function is used to create a pointer pointing to some object of Class SgType whose atomic type of public fields is VOID */
SgType *SgTypeVoid();
/**This function is used to create a pointer pointing to some object of Class SgType whose atomic type of public fields is BOOLEAN */

SgType *SgTypeBool();
/**This function is used to create a pointer pointing to some object of Class SgType whose atomic type of public fields is DEFAULT */
SgType *SgTypeDefault();
/**This function is used to create a pointer pointing to some object of Class SgType whose atomic type of public fields is  PROCESSOR */
SgType *SgTypeProcessors();
/**This function is used to create a pointer pointing to some object of Class SgType whose atomic type of public fields is TEMPLATE */
SgType *SgTypeTemplates();

///This function is used to create an object of Class SgUnaryExp which represents a unary opration "*expr".                                         
SgUnaryExp & SgDerefOp(SgExpression &e);
///This function is used to create an object of Class SgUnaryExp which represents a unary opration  "\&expr".
SgUnaryExp & SgAddrOp(SgExpression &e);
///This function is used to create an object of Class SgUnaryExp which represents a unary opration  "-expr".

SgUnaryExp & SgUMinusOp(SgExpression &e);
///This function is used to create an object of Class SgUnaryExp which represents a unary opration  "+expr".

SgUnaryExp & SgUPlusOp(SgExpression &e);
///This function is used to create an object of Class SgUnaryExp which represents a unary opration "++expr".
SgUnaryExp & SgPrePlusPlusOp(SgExpression &e);
///This function is used to create an object of Class SgUnaryExp which represents a unary opration  "--expr".
SgUnaryExp & SgPreMinusMinusOp(SgExpression &e);
///This function is used to create an object of Class SgUnaryExp which represents a unary opration  "expr++".

SgUnaryExp & SgPostPlusPlusOp(SgExpression &e);
///This function is used to create an object of Class SgUnaryExp which represents a unary opration  "expr--".
SgUnaryExp & SgPostMinusMinusOp(SgExpression &e);
///This function is used to create an object of Class SgUnaryExp which represents a unary opration  "~expr".

SgUnaryExp & SgBitCompfOp(SgExpression &e);
///This function is used to create an object of Class SgUnaryExp which represents a unary opration  "!expr".

SgUnaryExp & SgNotOp(SgExpression &e);
///This function is used to create an object of Class SgUnaryExp which represents a unary opration  "sizeof(expr)".

 
SgUnaryExp & SgSizeOfOp(SgExpression &e);


///This function is used to create an object of Class SgUnaryExp which represents some unary expression of type "code". 
SgUnaryExp & makeAnUnaryExpression(int code,PTR_LLND ll1);


///This function is used to judge whether the type of object "*pt" is among INT\_VAL, CHAR\_VAL, FLOAT\_VAL, DOUBLE\_VAL, STRING\_VAL, COMPLEX\_VAL and KEYWORD\_VAL.
SgValueExp *             isSgValueExp(SgExpression *pt);
///This function is used to judge whether the type of object "*pt" is KEYWORD\_VAL.
SgKeywordValExp *        isSgKeywordValExp(SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the unary expression such as DEREF\_OP, ADDRESS\_OP, SIZE\_OP, MINUS\_OP, UNARY\_ADD\_OP, PLUSPLUS\_OP, MINUSMINUS\_OP, BIT\_COMPLEMENT\_OP and  NOT\_OP.  
SgUnaryExp *             isSgUnaryExp(SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "CAST\_OP".
SgCastExp *              isSgCastExp(SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "DELETE\_OP".
SgDeleteExp *            isSgDeleteExp(SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "NEW\_OP".

SgNewExp *               isSgNewExp(SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "EXPR\_IF".
SgExprIfExp *            isSgExprIfExp(SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "FUNC\_CALL".
SgFunctionCallExp *      isSgFunctionCallExp(SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "FUNCTION\_OP".
SgFuncPntrExp *          isSgFuncPntrExp(SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "EXPR\_LIST".
SgExprListExp *          isSgExprListExp(SgExpression *pt);
///This function is used to judge whether the object "*pt" represents some expression among "CONST\_REF", "TYPE\_REF" and "INTERFACE\_REF".
SgRefExp *               isSgRefExp (SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "VAR\_REF".
SgVarRefExp *            isSgVarRefExp (SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "THIS\_NODE".
SgThisExp *              isSgThisExp (SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "ARRAY\_REF".
SgArrayRefExp *          isSgArrayRefExp (SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "ARRAY\_OP".
SgPntrArrRefExp *        isSgPntrArrRefExp(SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "DEREF\_OP".
SgPointerDerefExp *      isSgPointerDerefExp (SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "RECORD\_REF". 
SgRecordRefExp *         isSgRecordRefExp (SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "STRUCTURE\_CONSTRUCTOR".
SgStructConstExp*        isSgStructConstExp (SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "CONSTRUCTOR\_REF".
SgConstExp*              isSgConstExp (SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "VECTOR\_CONST".
SgVecConstExp *          isSgVecConstExp (SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "INIT\_LIST".
SgInitListExp *          isSgInitListExp (SgExpression *pt);
///This function is used to judge whether the object "*pt" represents some expression among "EQUI\_LIST", "NAMELIST\_LIST" and "COMM\_LIST".
SgObjectListExp *        isSgObjectListExp (SgExpression *pt);
///This function is used to judge whether the object "*pt" represents some expression among "PARAMETER\_OP", "PUBLIC\_OP", "PRIVATE\_OP", "ALLOCATABLE\_OP", "DIMENSION\_OP", "EXTERNAL\_OP", "IN\_OP", "OUT\_OP", "INOUT\_OP", "INTRINSIC\_OP", "POINTER\_OP", "OPTIONAL\_OP", "SAVE\_OP" and "TARGET\_OP".
SgAttributeExp *         isSgAttributeExp (SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "KEYWORD\_ARG".
SgKeywordArgExp *        isSgKeywordArgExp (SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "DDOT".
SgSubscriptExp*          isSgSubscriptExp (SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "ONLY\_NODE".
SgUseOnlyExp *           isSgUseOnlyExp (SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "RENAME\_NODE".
SgUseRenameExp *         isSgUseRenameExp (SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "SPEC\_PAIR".
SgSpecPairExp *          isSgSpecPairExp (SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "IOACCESS".
SgIOAccessExp *          isSgIOAccessExp (SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "IMPL\_TYPE".
SgImplicitTypeExp *      isSgImplicitTypeExp (SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "TYPE\_OP".
SgTypeExp *              isSgTypeExp (SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "SEQ".
SgSeqExp *               isSgSeqExp (SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "LEN\_OP".
SgStringLengthExp *      isSgStringLengthExp (SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "DEFAULT 
SgDefaultExp *           isSgDefaultExp (SgExpression *pt);
///This function is used to judge whether the object "*pt" represents the expression "LABEL\_REF".
SgLabelRefExp *          isSgLabelRefExp (SgExpression *pt);
///This function is used to judge whether the object "*pt" represents some statement among "PROC\_HEDR", "FUNC\_HEDR" and "PROG\_HEDR".
SgProgHedrStmt *         isSgProgHedrStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents some statement among "FUNC\_HEDR" and "PROC\_HEDR".
SgProcHedrStmt *         isSgProcHedrStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "FUNC\_HEDR".
SgFuncHedrStmt *         isSgFuncHedrStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents some statement among "CLASS\_DECL", "TECLASS\_DECL", "STRUCT\_DECL", "UNION\_DECL", "ENUM\_DECL" and "COLLECTION\_DECL".
SgClassStmt *            isSgClassStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "STRUCT\_DECL".
SgStructStmt *           isSgStructStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "UNION\_DECL".
SgUnionStmt *            isSgUnionStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "ENUM\_DECL".
SgEnumStmt *             isSgEnumStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "COLLECTION\_DECL".
SgCollectionStmt *       isSgCollectionStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "BASIC\_BLOCK".
SgBasicBlockStmt *       isSgBasicBlockStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "FOR\_NODE".
SgForStmt *              isSgForStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "WHILE\_NODE".
SgWhileStmt *            isSgWhileStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "DO\_WHILE\_NODE".
SgDoWhileStmt *          isSgDoWhileStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "LOGIF\_NODE".
SgLogIfStmt *            isSgLogIfStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "IF\_NODE".
SgIfStmt *               isSgIfStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "ARITHIF\_NODE".
SgArithIfStmt *          isSgArithIfStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "WHERE\_NODE".
SgWhereStmt *            isSgWhereStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "WHERE\_BLOCK\_STMT".
SgWhereBlockStmt *       isSgWhereBlockStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statememt "SWITCH\_NODE".
SgSwitchStmt *           isSgSwitchStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "CASE\_NODE".
SgCaseOptionStmt *       isSgCaseOptionStmt (SgStatement *pt);
///This function is used to judge whether the object "* represents executable statement.
SgExecutableStatement *  isSgExecutableStatement (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statememt "ASSIGN\_STAT".
SgAssignStmt *           isSgAssignStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "EXPR\_STMT\_NODE".
SgCExpStmt *             isSgCExpStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "POINTER\_ASSIGN\_STAT".  

SgPointerAssignStmt *    isSgPointerAssignStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents some statement among "ALLOCATE\_STMT" and "DEALLOCATE\_STMT".
SgHeapStmt *             isSgHeapStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "NULLIFY\_STMT".
SgNullifyStmt *          isSgNullifyStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "CONT\_STAT".
SgContinueStmt *         isSgContinueStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "CONTROL\_END".
SgControlEndStmt *       isSgControlEndStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "BREAK\_NODE".
SgBreakStmt *            isSgBreakStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "CYCLE\_STMT".
SgCycleStmt *            isSgCycleStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents some statement among "RETURN\_NODE" and "RETURN\_STAT".
SgReturnStmt *           isSgReturnStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "EXIT\_STMT".
SgExitStmt *             isSgExitStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "GOTO\_NODE".
SgGotoStmt *             isSgGotoStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents some statement among "COMGOTO\_NODE" and "ASSGOTO\_NODE".
SgLabelListStmt *        isSgLabelListStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "ASSGOTO\_NODE".
SgAssignedGotoStmt *     isSgAssignedGotoStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "COMGOTO\_NODE".
SgComputedGotoStmt *     isSgComputedGotoStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "STOP\_STAT".
SgStopOrPauseStmt *      isSgStopOrPauseStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "PROC\_STAT".
SgCallStmt*              isSgCallStmt (SgStatement *pt);
/// Fortran M 
SgProsHedrStmt *	 isSgProsHedrStmt (SgStatement *pt);
/// Fortran M 
SgProcessDoStmt *        isSgProcessDoStmt (SgStatement *pt);
/// Fortran M 
SgProsCallStmt*          isSgProsCallStmt (SgStatement *pt); 
/// Fortran M 
SgProsCallLctn*          isSgProsCallLctn (SgStatement *pt); 
/// Fortran M 
SgProsCallSubm*          isSgProsCallSubm (SgStatement *pt); 
/// Fortran M 
SgInportStmt *           isSgInportStmt (SgStatement *pt); 
/// Fortran M 
SgOutportStmt *          isSgOutportStmt (SgStatement *pt);
/// Fortran M 
SgIntentStmt *		 isSgIntentStmt (SgStatement *pt); 
/// Fortran M 
SgChannelStmt *          isSgChannelStmt (SgStatement *pt);
/// Fortran M 
SgMergerStmt *           isSgMergerStmt (SgStatement *pt); 
/// Fortran M 
SgMoveportStmt *         isSgMoveportStmt (SgStatement *pt);
/// Fortran M 
SgSendStmt *             isSgSendStmt (SgStatement *pt); 
/// Fortran M 
SgReceiveStmt *          isSgReceiveStmt (SgStatement *pt);
/// Fortran M 
SgEndchannelStmt *       isSgEndchannelStmt (SgStatement *pt); 
/// Fortran M 
SgProbeStmt *            isSgProbeStmt (SgStatement *pt); 
/// Fortran M 
SgProcessorsRefExp *	 isSgProcessorsRefExp(SgExpression *pt);
/// Fortran M 
SgPortTypeExp *          isSgPortTypeExp (SgExpression *pt); 
/// Fortran M 
SgInportExp *            isSgInportExp (SgExpression *pt); 
/// Fortran M 
SgOutportExp *           isSgOutportExp (SgExpression *pt);
/// Fortran M 
SgFromportExp *          isSgFromportExp (SgExpression *pt); 
/// Fortran M 
SgToportExp *            isSgToportExp (SgExpression *pt); 
/// Fortran M 
SgIO_statStoreExp *      isSgIO_statStoreExp (SgExpression *pt);
/// Fortran M 
SgEmptyStoreExp *        isSgEmptyStoreExp (SgExpression *pt);
/// Fortran M 
SgErrLabelExp *          isSgErrLabelExp (SgExpression *pt);
/// Fortran M 
SgEndLabelExp *          isSgEndLabelExp (SgExpression *pt);
/// Fortran M 
SgDataImpliedDoExp *     isSgDataImpliedDoExp (SgExpression *pt);
/// Fortran M 
SgDataEltExp *           isSgDataEltExp (SgExpression *pt); 
/// Fortran M 
SgDataSubsExp *          isSgDataSubsExp (SgExpression *pt); 
/// Fortran M 
SgDataRangeExp *         isSgDataRangeExp (SgExpression *pt); 
/// Fortran M 
SgIconExprExp *          isSgIconExprExp (SgExpression *pt); 
///This function is used to judge whether the object "*pt" represents the input/output statement .
SgIOStmt *               isSgIOStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents some statement among "READ\_STAT", "WRITE\_STAT" and "PRINT\_STAT".  
SgInputOutputStmt *      isSgInputOutputStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents some statement among "CLOSE\_STAT", "INQUIRE\_STAT", "BACKSPACE\_STAT", "REWIND\_STAT", "ENDFILE\_STAT" and "FORMAT\_STAT".
SgIOControlStmt *        isSgIOControlStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents some statement among "VAR\_DECL", "ENUM\_DECL", "STRUCT\_DECL", "CLASS\_DECL", "TECLASS\_DECL" and "COLLECTION\_DECL".
SgDeclarationStatement * isSgDeclarationStatement (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "VAR\_DECL".
SgVarDeclStmt *          isSgVarDeclStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents some statement among "INTENT\_STMT", "OPTIONAL\_STMT", "SAVE\_DECL", "PUBLIC\_STMT", "PRIVATE\_STMT", "EXTERN\_STAT", "INTRIN\_STAT", "DIM\_STAT", "ALLOCATABLE\_STMT", "POINTER\_STMT", "TARGET\_STMT", "MODULE\_PROC\_STMT" and "PROCESSORS\_STAT".
SgVarListDeclStmt *      isSgVarListDeclStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "STRUCT\_DECL".
SgStructureDeclStmt *    isSgStructureDeclStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents some statement among "NAMELIST\_STAT", "EQUI\_STAT", "PROS\_COMM" and "COMM\_STAT".
SgNestedVarListDeclStmt* isSgNestedVarListDeclStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "PARAM\_DECL".
SgParameterStmt *        isSgParameterStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt" represents the statement "IMPL\_DECL".
SgImplicitStmt *         isSgImplicitStmt (SgStatement *pt);
///This function is used to judge whether the object "*pt:" represents the symbol "VARIABLE\_NAME".
SgVariableSymb *         isSgVariableSymb (SgSymbol *pt);
///This function is used to judge whether the object "*pt" represents the symbol "CONST\_NAME".
SgConstantSymb *         isSgConstantSymb (SgSymbol *pt);
///This function is used to judge whether the object "*pt" represents some symbol "PROGRAM\_NAME", "PROCEDURE\_NAME", "FUNCTION\_NAME" and "MEMBER\_FUNC".
SgFunctionSymb *         isSgFunctionSymb (SgSymbol *pt);
///This function is used to judge whether the object "*pt" represents the symbol "MEMBER\_FUNC".
SgMemberFuncSymb *       isSgMemberFuncSymb (SgSymbol *pt);
///This function is used to judge whether the object "*pt" represents some symbol among "ENUM\_NAME" and "FIELD\_NAME".
SgFieldSymb *            isSgFieldSymb (SgSymbol *pt);
///This function is used to judge whether the object "*pt" represents some symbol among "CLASS\_NAME", "TECLASS\_NAME", "UNION\_NAME", "STRUCT\_NAME" and "COLLECTION\_NAME".
SgClassSymb *            isSgClassSymb (SgSymbol *pt);
///This function is used to judge whether the object "*pt" represents the symbol "LABEL\_NAME".
SgLabelSymb *            isSgLabelSymb (SgSymbol *pt);
///This function is used to judge whether the object "*pt" represents the symbol "LABEL\_NAME".
SgLabelVarSymb *         isSgLabelVarSymb (SgSymbol *pt);
///This function is used to judge whether the object "*pt" represents the symbol "ROUTINE\_NAME".
SgExternalSymb *         isSgExternalSymb (SgSymbol *pt);
///This function is used to judge whether the object "*pt" represents the symbol "CONSTRUCT\_NAME".
SgConstructSymb *        isSgConstructSymb (SgSymbol *pt);
///This function is used to judge whether the object "*pt" represents the symbol "INTERFACE\_NAME".
SgInterfaceSymb *        isSgInterfaceSymb (SgSymbol *pt);
///This function is used to judge whether the object "*pt" represents the symbol "MODULE\_NAME".
SgModuleSymb *           isSgModuleSymb (SgSymbol *pt);
///This function is used to judge whether the object "*pt" represents the type "T\_ARRAY".
SgArrayType *            isSgArrayType (SgType *pt);
///This function is used to judge whether the object "*pt" represents the type "T\_POINTER".
SgPointerType *          isSgPointerType (SgType *pt);
///This function is used to judge whether the object "*pt" represents the type "T\_FUNCTION".

SgFunctionType *         isSgFunctionType (SgType *pt);
///This function is used to judge whether the object "*pt" represents the type "T\_REFERENCE".
SgReferenceType *        isSgReferenceType (SgType *pt);
///This function is used to judge whether the object "*pt" represents the type "T\_DERIVED\_TYPE".
SgDerivedType *          isSgDerivedType (SgType *pt);
///This function is used to judge whether the object "*pt" represents the type "T\_DERIVED\_CLASS".
SgDerivedClassType *     isSgDerivedClassType (SgType *pt);
///This function is used to judge whether the object "*pt" represents the type "T\_DESCRIPT".
SgDescriptType *         isSgDescriptType (SgType *pt);
///This function is used to judge whether the object "*pt" represents the type "T\_DERIVED\_COLLECTION".
SgDerivedCollectionType* isSgDerivedCollectionType (SgType *pt);

//@}

#endif /* ndef LIBSAGEXX_H */
