%start comment

%{

/*  From hpf2dep */

#ifdef HAVE_CONFIG_H
#include <config.h>                     /* config header file */
#endif

#ifdef STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include <stdio.h>
#include "keywords.h"
#   include "typedef.h"
#   include "bif.h"
#   include "ll.h"
#   include "symb.h"

#include "defnodes.h"

/***********/

#include "javaparser.tab.h"

extern  int  lineNum_g,lineNum_l;
extern  int  debugFlag_g;
extern  FILE *inf_ptr;

char *copys();

extern int parloop;   /* PARLOOP */

%}

Digit                  [0-9]
Letter                 [a-zA-Z_$]
Identifier             {Letter}(({Digit}|{Letter})*)

Longsuffix             [lL]
OctalConst             0([0-7])+{Longsuffix}?
HexConst               0x([0-9a-fA-F])*{Longsuffix}?
DecConst               0|([1-9]{Digit}*){Longsuffix}?

DigitString            {Digit}+
SignedDigitString      [+-]?{DigitString}
Significand            (({DigitString}"."{DigitString}?)|("."{DigitString}))

Floatsuffix            [f|d|F|D]

RealLiteralConstant    ({Significand}([Ee]{SignedDigitString})?|{DigitString}[eE]{SignedDigitString}){Floatsuffix}?

DoubleLiteralConstant  ({Significand}([Dd]{SignedDigitString})?|{DigitString}[dD]{SignedDigitString}){Floatsuffix}?

Whitespace 	            [\t\f ]+

Unicodeesc              \\u[0-9a-fA-F]{4}

%%

%{
   yyin=inf_ptr;
                    /* zheng.gb 1996.5.29 */

#ifdef YYTEXT_IS_POINTER
#ifdef IS_UNSIGNED_CHAR
   yytext = (unsigned char *) calloc(1,(unsigned)0x400);
#else
   yytext = (char *) calloc(1,(unsigned)0x400);
#endif
#endif

%}

{Whitespace} {
   ; 
}


"//".*\n { 
   if(debugFlag_g == 1)
      printf("(comment//,)\n"); 
   lineNum_g++;
   lineNum_l++;
}

"/*" { 
   if(debugFlag_g == 1)
      printf("(comment/* begin,)");
   BEGIN comment; 
}

<comment>[^*\n]* {
   if(debugFlag_g == 1)
      printf("(comment... %s)",yytext);
}

<comment>"*"[^*/\n]* {
   if(debugFlag_g == 1)
      printf("(comment... %s)",yytext);
}

<comment>\n {
   lineNum_g++;
   lineNum_l++;
}

<comment>"*/" { 
   int i;
   if(debugFlag_g == 1)
      printf("(comment*/end,)\n");
   for(i=0;yytext[i]!='\0';i++)
      if (yytext[i] == '\n')
        {
         lineNum_g++;
         lineNum_l++;
         }
   BEGIN 0; 
}

<comment><<EOF>> { 
   /* javar_out(0, "Unexpected <EOF> in comment"); */
   BEGIN 0;
   printf(" Unexpected <EOF> in comment \n" );
}

"\n" { 
   lineNum_g++;
   lineNum_l++;
}

{Unicodeesc} { 
  /* javar_out(0, "%s : Sorry, unicode escapes "
			          "are not supported\n", yytext); */
   if(debugFlag_g==1)
         printf("(Unicode),");
   return XXX_;
} 


"=="   return EQ_;
"<="   return LE_;
">="   return GE_;
"!="   return NE_;
"&&"   return AND_;
"||"   return OR_;
"++"   return PP_;
"--"   return MM_;

"<<"   return LL_;
">>"   return RR_;
">>>"  return RRR_;

"+="   return PIS_; 
"-="   return MINIS_;
"*="   return MULIS_;
"/="   return DIS_;
"&="   return ANDIS_; 
"|="   return ORIS_;
"^="   return UIS_;
"%="   return IDIVIS_;
"<<="  return LLIS_;
">>="  return RRIS_;
">>>=" return RRRIS_; 

"null" { 
   /* yylval.expr = make_nul_lit(); */
   return NULL_LITERAL;
}

"true"  {
   return BOOL_TRUE_LITERAL;
} 
"false" { 
   return BOOL_FALSE_LITERAL;
}

\"([^"]|"\"\"")*\"  {
   char *temp=copys(yytext);
   int i;
   for(i=0;i<strlen(yytext)-2;i++)
      temp[i]=yytext[i+1];
   temp[i]='\0';
   yylval.charp=copys(temp);
   return(STRING_LITERAL);
}

\'([^']|"''")*\'  {
   char *temp=copys(yytext);
   int i;
   for(i=0;i<strlen(yytext)-2;i++)
      temp[i]=yytext[i+1];
   temp[i]='\0';
   yylval.charp=copys(temp);
   return(CHAR_LITERAL); 
}

{DecConst}  {
   yylval.charp=copys(yytext);
   return(INT_DEC_LITERAL);
} 
{OctalConst} {
   yylval.charp=copys(yytext);
   return(INT_OCT_LITERAL);
}
{HexConst}  {
   yylval.charp=copys(yytext);
   return(INT_HEX_LITERAL);
}

{RealLiteralConstant}  {
   yylval.charp=copys(yytext);
   return(REAL_LITERAL);
}
 
{DoubleLiteralConstant} {
   yylval.charp=copys(yytext);
   return(DOUBLE_LITERAL);
}

{Identifier} {
   return(keynameDeal_m(yytext));
}


.             { return yytext[0]; }


%%

#ifndef yywrap
yywrap()
{ 
   return 1; 
}
#endif

/*   int searchKey_m(char name[40],char tab[47][15],int limit)  */
int searchKey_m(name,tab,limit)
char name[40];
char tab[47][15];
{
   int i=0;
   int flag=0;

   do
    {
        if(strcmp(name,tab[i])==0)
            flag=1;
        else
             i++;
    }while (!((flag==1)||(i==limit))) ;
    return(i);
}

/*   int keynameDeal_m(char name[40])   */
int keynameDeal_m(name)
char name[40];
{
   int i;
   int returnValue_m;
   char chara;

   yylval.charp=copys( name );

   if((i=searchKey_m(name,KEY_WORD_G,KEY_WORD_LIMIT_G))<KEY_WORD_LIMIT_G)
      {
       if(debugFlag_g==1)
            printf("(keyword: %s  %d)\n",name,i);
       returnValue_m=ABSTRACT+i; 
       return(returnValue_m);
       }
   else
      {
       if(debugFlag_g==1)
             printf("(It's a identifier_name: %s)\n",name);
       return(IDENTIFIER);
       }
}
