//  ----------------------------------
//  Darryl Brown
//  University of Oregon pC++/Sage++
//
//  baseClasses.C  -  module for basic classes used by 
//       breakpoint modules.
//
//  
//  ----------------------------------;

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>

#ifdef STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include "baseClasses.h"

//--------------------------------------------------//  ;
//   HERE ARE SOME HELPFUL ROUTINES.;
//--------------------------------------------------//  ;

/*  str_save  -- accepts a string, and mallocs space for it, or;
 *    returns an error if malloc fails...;
 */
char * brk_stringSave(char * str)
{
  char * tmp;
  if ((tmp = new char[ strlen(str)+1] ) == NULL) 
    {
      perror("new failed\n");
    }
  strcpy (tmp, str);
  return tmp;
}
//--------------------------------------------------//  ;
/*   strsame compares two strings returning 1 for same, 0 for not
 *           same.
 */
int brk_strsame(char *s1, char *s2) {
  return (strcmp(s1, s2) == 0);
}

//--------------------------------------------------
void brk_printtabs(int tabs) {
  for (int i=0; i < tabs; i++) {
    printf("\t");
  }
}


//--------------------------------------------------
void brk_printtabs(int tabs, FILE *fptr) {
  for (int i=0; i < tabs; i++) {
    fprintf(fptr, "\t");
  }
}




//--------------------------------------------------//  
//   HERE ARE THE CLASSES
//--------------------------------------------------//  
 brk_ptrNode::brk_ptrNode () {
   next = prev = NULL;
   node = NULL;
 }

 brk_ptrNode::brk_ptrNode (brk_basePtr *b) {
   next = prev = NULL;
   node = b;
 }

 brk_ptrNode::compare(brk_basePtr *h) {
   brk_ptrNode *p;
   p = (brk_ptrNode *) h;
   return node->compare( p->node );
 }


//--------------------------------------------------//  ;
//--------------------------------------------------//  ;

 brk_linkedList::brk_linkedList () {
   start = current = end = NULL;
   length = 0;
 }

brk_basePtr * brk_linkedList::insert(int i, brk_basePtr * p) {
  
  //  insert hierachy h into this list at index i;
  
  brk_ptrNode *tmp;
  int j = 0;
  brk_ptrNode *h = new brk_ptrNode(p);
  
  for (tmp = start; 
       tmp && j < i; 
       j++, tmp = tmp->next);
  
  
  if (tmp) {
    h->prev = tmp->prev;
    h->next = tmp;
    if (tmp->prev) {
      tmp->prev->next = h;
    } else {
      start = h;
    }
    length--;
    return tmp->node;
  } else {
    if (j == i) {
      // it is the end of the list...
	h->prev = end;
      h->next = NULL;
      end->next = h;
      end = h;
    }
    return NULL;
  }
}

void brk_linkedList::push (brk_basePtr *h) {
  
  // push brk_basePtr i onto front of list;
  brk_ptrNode *i = new brk_ptrNode(h);
  
  if (length == 0) {
    end = start = i;
    current = end->next = end->prev = NULL;
    length = 1;
  } else {
    i->next = start;
    i->prev = NULL;
    start->prev = i;
    start = i;
    length++;
  }
}

void brk_linkedList::pushLast (brk_basePtr *h) {
  
  // push listmember i onto end of list;
  brk_ptrNode *i = new brk_ptrNode(h);
  
  fflush(stdout);
  if (length == 0) {
    end = start = i;
    current = end->next = end->prev = NULL;
    length = 1;
  } else {
    i->next = NULL;
    i->prev = end;
    end->next = i;
    end = i;
    length++;
  }
}

brk_basePtr * brk_linkedList::remove (int i) {
  
  // remove and return member at index i;
  
  brk_ptrNode *tmp;
  int j = 0;
  
  for (tmp = start; 
       tmp && j < i; 
       j++, tmp = tmp->next);
  
  if (tmp) {
    if (tmp->prev) {
      tmp->prev->next = tmp->next;
    } else {
      start = tmp->next;
    }
    if (tmp->next) {
      tmp->next->prev = tmp->prev;
    } else {
      end = tmp->prev;
    }
    length--;
    return tmp->node;
  } else {
    return NULL;
  }
}

brk_basePtr * brk_linkedList::getIth (int i) {
  
  // return value at index i, without removing it;
  
  brk_ptrNode *tmp;
  int j = 0;
  
  for (tmp = start; 
       tmp && j < i; 
       j++, tmp = tmp->next);
  
  if (tmp) {
    return tmp->node;
  } else {
    return NULL;
  }
}

brk_basePtr *brk_linkedList::pop () {
  
  // remove and return member at front of list;
  
  brk_ptrNode *tmp;
  if (length == 0) {
    return NULL;
  } else {
    tmp = start;
    start = tmp->next;
    if (start) {
      start->prev = NULL;
    } else {
      end = NULL;
    }
    length--;
    return tmp->node;
  }
}

brk_basePtr * brk_linkedList::popLast () {
  
  // remove and return member at end of list;
  
  brk_ptrNode *tmp;
  if (length == 0) {
    return NULL;
  } else {
    tmp = end;
    end = tmp->prev;
    if (end) {
      end->next = NULL;
    } else {
      start = NULL;
    }
    length--;
    return tmp->node;
  }
}

brk_basePtr * brk_linkedList::searchList () {
  current = start;
  if (current) {
    return current->node;
  } else {
    return NULL;
  }
}

brk_ptrNode * brk_linkedList::findMember(brk_basePtr *node) {
  brk_basePtr *h;

  h = searchList();
  while (h) {
    if (h->compare(node) == 0) {
      return current;
    }
    h = nextItem();
  }
  return NULL;
}
    

brk_basePtr *brk_linkedList::nextItem () {
  
  //  return the item pointed to by current, 
  //  and increment current;
   
  if (current && (current=current->next)) {
    return current->node;
  } else {
    return NULL;
  }
}

void brk_linkedList::printIth(int i) {
  // send print message to each item in list;
  getIth(i)->print();
}

void brk_linkedList::print(int tNum) {
  print(tNum, stdout);
}

void brk_linkedList::print(int tNum, FILE *fptr) {
  // send print message to each item in list;
  brk_basePtr *tmp;
  
  tmp = searchList();
  
  while (tmp) {
    tmp->print(tNum, fptr);
    tmp = nextItem();
  }
  fflush(stdout);
}


void brk_linkedList::printToBuf(int tNum, char *buf) {
  // send print message to each item in list;
  brk_basePtr *tmp;
  char tbuf[100];
  *buf = '\0';

  tmp = searchList();
  while (tmp) {
    tmp->printToBuf(tNum, tbuf);
    strcat(buf, tbuf);
    tmp = nextItem();
  }
  fflush(stdout);
}

//  prints whole list;
void brk_linkedList::print() {
  print(0, stdout);
}


//  routine to pass into qsort routine..;
//  this routine used by sort to call the compare functions;
//  for two brk_linkedList elements;
int  qcompare(const void *bp1, const void *bp2) {
  brk_basePtr **p1, **p2;
  p1 = (brk_basePtr **) bp1;
  p2 = (brk_basePtr **) bp2;
  if ((*p1)->userCompare == NULL) {
    return (*p1)->compare((*p2));
  } else {
    return (*(*p1)->userCompare)(*p1, *p2);
  }
}

//  sorts list, but uses user provided comparison function;
//  The user provided function should return an int, and ;
//  take as arguments two type of (brk_basePtr *) (or a type;
//  derived from brk_basePtr *);				      
void brk_linkedList::sort(int 
		      (* compareFunc)
		      (brk_basePtr *, brk_basePtr *)) {
  brk_basePtr *b ;
  
  // if 0 or 1 elements, no need to do anything.;
  if (length < 2) {
    return;
  }
  
  b = searchList();
  while (current) {
    b->userCompare = compareFunc;
    b = nextItem();
  }
  
  // sorting;
  sort();
  
  b = searchList();
  while (current) {
    b->userCompare = NULL;
    b = nextItem();
  }
}

//  sorts the relation list;
void brk_linkedList::sort() {
  brk_basePtr **ptrVector, **p1;
  
  // if 0 or 1 elements, no need to do anything.;
  if (length < 2) {
    return;
  }
  // create vector of pointers to the nodes that will be compared.;
  p1 = ptrVector = new brk_basePtr * [length];
  searchList();
  while (current) {
    *p1 = current->node;
    p1++;
    nextItem();
  }
  
  // sort the nodes in the vector;
  qsort(ptrVector, length, sizeof (brk_basePtr *), &qcompare);
  
  // put the nodes back in the list;
  p1 = ptrVector;
  searchList();
  while (current) {
    current->node =  *p1;
    p1++;
    nextItem();
  }
  delete [] ptrVector;
}

void brk_linkedList::swap(brk_ptrNode *l, brk_ptrNode *r) {
  brk_basePtr *tmpr, *tmpl;
  
  tmpr = r->node;
  tmpl = l->node;
  insert(memberNum(l), tmpr);
  remove(memberNum(l));
  insert(memberNum(r), tmpl);
  remove(memberNum(r));
  
}

int brk_linkedList::memberNum(brk_ptrNode *m) {
  brk_ptrNode *p;
  int i;
  for (i = 0, p = start; p != m && i < length; i++, p = p->next) ;
  if (i == length) return -1;
  else return i;
}

  

// ---------------------------------------------------;
// ---------------------------------------------------;

brk_basePtr::brk_basePtr() {
  userCompare = NULL;
}

void brk_basePtr::print() {
  return;
}

void brk_basePtr::printToBuf(int, char *) {
  return;
}

void brk_basePtr::printAll(int) {
  return;
}

void brk_basePtr::printAll() {
  return;
}

#if 0
void brk_basePtr::printAll(int, FILE *) {
  return;
}

void brk_basePtr::printAll(FILE *) {
  return;
}
#endif

void brk_basePtr::print(int) {
  return;
}

void brk_basePtr::print(int, FILE *) {
  return;
}

int brk_basePtr::compare(brk_basePtr *) {
  return -1;
}


// ---------------------------------------------------;
// ---------------------------------------------------;

