/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

#include <stdio.h>

#ifndef __GNUC__
# include <stdlib.h>
#else
extern "C" void abort(void);
extern "C" void exit(int status);
# pragma implementation
#endif

#include "typedef.h"
#include "bif.h"
#include "ll.h"
#include "symb.h"

#include "defnodes.h"
#include "db.h"
#include "vparse.h"

#include "macro.h"

#include "vpc.h"
//#include "f90.h"

#include "extcxx_low.h"
extern int number_of_ll_node;

#undef USER

//
// define for having the debugging
//
//define  DEBUGLIB 1
#define  MAX_FILES 1000
//
//
// Array to keep track of table for a file
//
//

#define ALLOCATECHUNK 1000

#include "mapping.h"

void **tablebfnd[MAX_FILES];
void **tablellnd[MAX_FILES];
void **tabletype[MAX_FILES];
void **tablesymbol[MAX_FILES];
void **tablelabel[MAX_FILES];

int numtablebfnd[MAX_FILES];
int numtablellnd[MAX_FILES];
int numtabletype[MAX_FILES];
int numtablesymbol[MAX_FILES];
int numtablelabel[MAX_FILES];



int CurrentFileNumber = 0;

//
// Table for making link between the nodes and the classes
// Take the id and return a pointer
//

void **fileTableClass;
int allocatedForfileTableClass;
void **bfndTableClass;
int allocatedForbfndTableClass;
void **llndTableClass;
int allocatedForllndTableClass;
void **typeTableClass;
int allocatedFortypeTableClass;
void **symbolTableClass;
int allocatedForsymbolTableClass;
void **labelTableClass;
int allocatedForlabelTableClass;


//
// Some definition for this module
//

#define SORRY Message("Sorry, not implemented yet",0)

class SgProject;
class SgFile;
class SgStatement;
class SgExpression;
class SgLabel;
class SgSymbol;
class SgType;
class SgUnaryExp;
class SgClassSymb;
class SgVarDeclStmt; 


//
// Set of function to care about the table management
//

void InitializeTable()
{
  int i;
  for (i=0 ; i < MAX_FILES; i++)
    {
      tablebfnd[i] = NULL;
      tablellnd[i]= NULL;
      tabletype[i]= NULL;
      tablesymbol[i]= NULL;
      tablelabel[i]= NULL;

      numtablebfnd[i]= 0;
      numtablellnd[i]= 0;
      numtabletype[i]= 0;
      numtablesymbol[i]= 0;
      numtablelabel[i]= 0;
	
      // FOR ATTRIBUTES;
      tablebfndAttribute[i] = NULL;
      tablellndAttribute[i]= NULL;
      tabletypeAttribute[i]= NULL;
      tablesymbolAttribute[i]= NULL;
      tablelabelAttribute[i]= NULL;

      numtablebfndAttribute[i]= 0;
      numtablellndAttribute[i]= 0;
      numtabletypeAttribute[i]= 0;
      numtablesymbolAttribute[i]= 0;
      numtablelabelAttribute[i]= 0;
    }

  
  fileTableClass =  NULL;
  bfndTableClass =  NULL;
  llndTableClass  = NULL;
  typeTableClass  = NULL;
  symbolTableClass  = NULL;
  labelTableClass  =  NULL;
  allocatedForfileTableClass = 0;
  allocatedForbfndTableClass = 0;
  allocatedForllndTableClass = 0;
  allocatedFortypeTableClass = 0;
  allocatedForsymbolTableClass =0;
  allocatedForlabelTableClass = 0;

  // FOR ATTRIBUTES;
  fileTableAttribute =  NULL;
  bfndTableAttribute =  NULL;
  llndTableAttribute  = NULL;
  typeTableAttribute  = NULL;
  symbolTableAttribute  = NULL;
  labelTableAttribute  =  NULL;
  allocatedForfileTableAttribute = 0;
  allocatedForbfndTableAttribute = 0;
  allocatedForllndTableAttribute = 0;
  allocatedFortypeTableAttribute = 0;
  allocatedForsymbolTableAttribute =0;
  allocatedForlabelTableAttribute = 0;
}


void SwitchToFile(int i)
{
  if (i >= MAX_FILES)
    {
      Message("Too many files",0);
      exit(1);
    }
  
 tablebfnd[CurrentFileNumber] = bfndTableClass;
 tablellnd[CurrentFileNumber]= llndTableClass;
 tabletype[CurrentFileNumber]= typeTableClass;
 tablesymbol[CurrentFileNumber]= symbolTableClass;
 tablelabel[CurrentFileNumber]= labelTableClass;

 numtablebfnd[CurrentFileNumber]= allocatedForbfndTableClass;
 numtablellnd[CurrentFileNumber]= allocatedForllndTableClass;
 numtabletype[CurrentFileNumber]= allocatedFortypeTableClass;
 numtablesymbol[CurrentFileNumber]= allocatedForsymbolTableClass;
 numtablelabel[CurrentFileNumber]= allocatedForlabelTableClass;

 bfndTableClass =  tablebfnd[i];
 llndTableClass  = tablellnd[i];
 typeTableClass  = tabletype[i];
 symbolTableClass  = tablesymbol[i];
 labelTableClass  =  tablelabel[i];

 allocatedForbfndTableClass =  numtablebfnd[i];
 allocatedForllndTableClass = numtablellnd[i];
 allocatedFortypeTableClass = numtabletype[i];
 allocatedForsymbolTableClass =numtablesymbol[i];
 allocatedForlabelTableClass = numtablelabel[i];


  // FOR ATTRIBUTES
 tablebfndAttribute[CurrentFileNumber] = bfndTableAttribute;
 tablellndAttribute[CurrentFileNumber]= llndTableAttribute;
 tabletypeAttribute[CurrentFileNumber]= typeTableAttribute;
 tablesymbolAttribute[CurrentFileNumber]= symbolTableAttribute;
 tablelabelAttribute[CurrentFileNumber]= labelTableAttribute;

 numtablebfndAttribute[CurrentFileNumber]= allocatedForbfndTableAttribute;
 numtablellndAttribute[CurrentFileNumber]= allocatedForllndTableAttribute;
 numtabletypeAttribute[CurrentFileNumber]= allocatedFortypeTableAttribute;
 numtablesymbolAttribute[CurrentFileNumber]= allocatedForsymbolTableAttribute;
 numtablelabelAttribute[CurrentFileNumber]= allocatedForlabelTableAttribute;

 bfndTableAttribute =  tablebfndAttribute[i];
 llndTableAttribute  = tablellndAttribute[i];
 typeTableAttribute  = tabletypeAttribute[i];
 symbolTableAttribute  = tablesymbolAttribute[i];
 labelTableAttribute  =  tablelabelAttribute[i];

 allocatedForbfndTableAttribute =  numtablebfndAttribute[i];
 allocatedForllndTableAttribute = numtablellndAttribute[i];
 allocatedFortypeTableAttribute = numtabletypeAttribute[i];
 allocatedForsymbolTableAttribute =numtablesymbolAttribute[i];
 allocatedForlabelTableAttribute = numtablelabelAttribute[i];
 CurrentFileNumber = i;

}

// add a chunk to the size 
void ReallocatefileTableClass()
{
  int i;
  void **pt;
  
  pt  =  new void *[allocatedForfileTableClass + ALLOCATECHUNK];
  for (i=0; i<allocatedForfileTableClass + ALLOCATECHUNK; i++)
    pt[i] = NULL;
  
  for (i=0 ; i < allocatedForfileTableClass; i++)
    {
      pt[i] = fileTableClass[i];
    }
  if (allocatedForfileTableClass)
    delete fileTableClass;
  fileTableClass = pt;
  allocatedForfileTableClass = allocatedForfileTableClass + ALLOCATECHUNK;
}


void ReallocatebfndTableClass()
{
  int i;
  void **pt;
  
  pt  =  new void *[allocatedForbfndTableClass + ALLOCATECHUNK];
  for (i=0; i<allocatedForbfndTableClass + ALLOCATECHUNK; i++)
    pt[i] = NULL;
  for (i=0 ; i < allocatedForbfndTableClass; i++)
    {
      pt[i] = bfndTableClass[i];
    }
  if (allocatedForbfndTableClass)
    delete bfndTableClass;
  bfndTableClass = pt;
  allocatedForbfndTableClass = allocatedForbfndTableClass + ALLOCATECHUNK;
}


void ResetbfndTableClass()
{
  int i;
  
  for (i=0 ; i < allocatedForbfndTableClass; i++)
    {
//      delete bfndTableClass[i];
      bfndTableClass[i] = NULL;
    }
}

void ReallocatellndTableClass()
{
  int i;
  void **pt;
  
  pt  =  new void *[allocatedForllndTableClass + ALLOCATECHUNK];
  for (i=0; i<allocatedForllndTableClass + ALLOCATECHUNK; i++)
    pt[i] = NULL;
  for (i=0 ; i < allocatedForllndTableClass; i++)
    {
      pt[i] = llndTableClass[i];
    }
  if (allocatedForllndTableClass)
    delete llndTableClass;
  llndTableClass = pt;
  allocatedForllndTableClass = allocatedForllndTableClass + ALLOCATECHUNK;
}

void ReallocatesymbolTableClass()
{
  int i;
  void **pt;
  
  pt  =  new void *[allocatedForsymbolTableClass + ALLOCATECHUNK];
  for (i=0; i<allocatedForsymbolTableClass + ALLOCATECHUNK; i++)
    pt[i] = NULL;
  for (i=0 ; i < allocatedForsymbolTableClass; i++)
    {
      pt[i] = symbolTableClass[i];
    }
  if (allocatedForsymbolTableClass)
    delete symbolTableClass;
  symbolTableClass = pt;
  allocatedForsymbolTableClass = allocatedForsymbolTableClass + ALLOCATECHUNK;
}


void ReallocatelabelTableClass()
{
  int i;
  void **pt;
  
  pt  =  new void *[allocatedForlabelTableClass + ALLOCATECHUNK];
  for (i=0; i<allocatedForlabelTableClass + ALLOCATECHUNK; i++)
    pt[i] = NULL;
  for (i=0 ; i < allocatedForlabelTableClass; i++)
    {
      pt[i] = labelTableClass[i];
    }
  if (allocatedForlabelTableClass)
    delete labelTableClass;
  labelTableClass = pt;
  allocatedForlabelTableClass = allocatedForlabelTableClass + ALLOCATECHUNK;
}


void ReallocatetypeTableClass()
{
  int i;
  void **pt;
  
  pt  =  new void *[allocatedFortypeTableClass + ALLOCATECHUNK];
  for (i=0; i<allocatedFortypeTableClass + ALLOCATECHUNK; i++)
    pt[i] = NULL;
  for (i=0 ; i < allocatedFortypeTableClass; i++)
    {
      pt[i] = typeTableClass[i];
    }
  if (allocatedFortypeTableClass)
    delete typeTableClass;
  typeTableClass = pt;
  allocatedFortypeTableClass = allocatedFortypeTableClass + ALLOCATECHUNK;
}


void RemoveFromTableType(void * pt)
{
  int i;
  for (i=0 ; i < allocatedFortypeTableClass; i++)
    {
      if (typeTableClass[i] == pt)
        {
          typeTableClass[i] = NULL;
          return;
        }
    }
}


void RemoveFromTableSymb(void * pt)
{
  int i;
  for (i=0 ; i < allocatedForsymbolTableClass; i++)
    {
      if (symbolTableClass[i] == pt)
        {
          symbolTableClass[i] = NULL;
          return;
        }
    }
}



void RemoveFromTableBfnd(void * pt)
{
  int i;
  for (i=0 ; i < allocatedForbfndTableClass; i++)
    {
      if (bfndTableClass[i] == pt)
        {
          bfndTableClass[i] = NULL;
          return;
        }
    }
}


void RemoveFromTableFile(void * pt)
{
  int i;
  for (i=0 ; i < allocatedForfileTableClass; i++)
    {
      if (fileTableClass[i] == pt)
        {
          fileTableClass[i] = NULL;
          return;
        }
    }
}

// forward, to be defined later in this file;
void RemoveFromTableLlnd(void * pt); 


void RemoveFromTableLabel(void * pt)
{
  int i;
  for (i=0 ; i < allocatedForlabelTableClass; i++)
    {
      if (labelTableClass[i] == pt)
        {
          labelTableClass[i] = NULL;
          return;
        }
    }
}



void 
SetMappingInTableForBfnd(PTR_BFND bif, void *pt)
{
  if (!bif)
    return ;
  while (allocatedForbfndTableClass <= BIF_ID(bif))
    {
      ReallocatebfndTableClass();
    }
  bfndTableClass[BIF_ID(bif)] = pt;
}


void 
SetMappingInTableForType(PTR_TYPE type, void *pt)
{
  if (!type)
    return ;
  while (allocatedFortypeTableClass <= TYPE_ID(type))
    {
      ReallocatetypeTableClass();
    }
  typeTableClass[TYPE_ID(type)] = pt;
}


void 
SetMappingInTableForSymb(PTR_SYMB symb, void *pt)
{
  if (!symb)
    return ;
  while (allocatedForsymbolTableClass <= SYMB_ID(symb))
    {
      ReallocatesymbolTableClass();
    }
  symbolTableClass[SYMB_ID(symb)] = pt;
}

void 
SetMappingInTableForLabel(PTR_LABEL lab, void *pt)
{
  if (!lab)
    return ;
  while (allocatedForlabelTableClass <= LABEL_ID(lab))
    {
      ReallocatelabelTableClass();
    }
  labelTableClass[SYMB_ID(lab)] = pt;
}



void 
SetMappingInTableForLlnd(PTR_LLND ll, void *pt)
{
  if (!ll)
    return ;
  while (allocatedForllndTableClass <= NODE_ID(ll))
    {
      ReallocatellndTableClass();
    }
  llndTableClass[NODE_ID(ll)] = pt;
}


void 
SetMappingInTableForFile(PTR_FILE file, void *pt)
{
  int id;
  if (!file)
    return ;
  id = GetFileNum(FILE_FILENAME(file));
  while (allocatedForfileTableClass <= id)
    {
      ReallocatefileTableClass();
    }
  fileTableClass[id] = pt;
}


SgSymbol *
GetMappingInTableForSymbol(PTR_SYMB symb)
{
  int id;
  if (!symb)
    return NULL;
  id = SYMB_ID(symb);
  if (allocatedForsymbolTableClass <= id)    
    {
      return NULL;
    }
  return (SgSymbol *) symbolTableClass[id];
}



SgLabel *
GetMappingInTableForLabel(PTR_LABEL lab)
{
  int id;
  if (!lab)
    return NULL;
  id = LABEL_ID(lab);
  if (allocatedForlabelTableClass <= id)    
    {
      return NULL;
    }
  return (SgLabel *) labelTableClass[id];
}


SgStatement *
GetMappingInTableForBfnd(PTR_BFND bf)
{
  int id;
  if (!bf)
    return NULL;
  id = BIF_ID(bf);
  if (allocatedForbfndTableClass <= id)    
    {
      return NULL;
    }
  return (SgStatement *) bfndTableClass[id];
}


SgType *
GetMappingInTableForType(PTR_TYPE t)
{
  int id;
  if (!t)
    return NULL;
  id = TYPE_ID(t);
  if (allocatedFortypeTableClass <= id)    
    {
      return NULL;
    }
  return (SgType *) typeTableClass[id];
}


SgExpression *
GetMappingInTableForLlnd(PTR_LLND ll)
{
  int id;
  if (!ll)
    return NULL;
  id = NODE_ID(ll);
  if (allocatedForllndTableClass <= id)    
    {
      return NULL;
    }
  return (SgExpression *) llndTableClass[id];
}


SgFile *
GetMappingInTableForFile(PTR_FILE file)
{
  int id;
  if (!file)
    return NULL;
  id = GetFileNum(FILE_FILENAME(file));
  if (allocatedForfileTableClass <= id)
    {
      return NULL;
    }
  return (SgFile *) fileTableClass[id];
}


