/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>

#include "tags.h"

#include "typedef.h"
#include "bif.h"
#include "ll.h"
#include "symb.h"

#include "dep.h"
#include "sets.h"

#include "defnodes.h"
#include "vparse.h"

// All maco define in old toolbox
#include "macro.h"
// All the "C" functions from the Rennes toolbox
#include "extcxx_low.h"
// class library header file
#include "fromsage++.h"

#include "sage++callgraph.h"

#ifdef STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif
 
// this is the source code for sage++callgraph 

int SgCallGraph::numberOfFunctionsInGraph(){ return num_funs_in_table; }

SgSymbol *SgCallGraph::firstFunction()
{
  return (currentFun = FunListHead)?currentFun->s:(SgSymbol *)NULL;
}

SgSymbol *SgCallGraph::nextFunction()
{
  if(currentFun)
    return (currentFun = currentFun->next)?currentFun->s:(SgSymbol *)NULL;
  else
    return (SgSymbol *)NULL;
}

SgStatement *SgCallGraph::functionBody(SgSymbol *fname)
{
  SgCallGraphFunRec *f;  
  return (f=locateFunctionInTable(fname))?f->body:(SgStatement *)NULL;
}

int SgCallGraph::functionId(SgSymbol *fname)
{
  SgCallGraphFunRec *f;  
  return (f=locateFunctionInTable(fname))?f->id:-1;
}

SgSymbol *SgCallGraph::function(int i){    // i-th function in table (0 = first)
	    SgCallGraphFunRec *f=FunListHead;
	    static SgCallGraphFunRec *lastf=FunListHead;
	    static int first_time=1;
	    if(first_time){
              first_time=0;
	      fprintf(stderr, "SgCallGraph:: Using obsolete function:\n        --    SgSymbol *function(int)\n");
	    }
            if(i < num_funs_in_table){
	      if(i >= lastf->id)i-=(f=lastf)->id;
	      for(;i>0;i--)f = f->next;
	      return f->s;
	    }
	    else return NULL;
}

SgStatement *SgCallGraph::functionBody(int i){    // i-th function in table (0 = first)
	    SgCallGraphFunRec *f=FunListHead;
	    static SgCallGraphFunRec *lastf=FunListHead;
	    static first_time=1;
	    if(first_time){
              first_time=0;
	      fprintf(stderr, "SgCallGraph:: Using obsolete function:\n        --    SgStatement *functionBody(int)\n");
	    }
            if(i < num_funs_in_table){
	      if(i >= lastf->id)i-=(f=lastf)->id;
	      for(;i>0;i--)f = f->next;
	      return f->body;
	    }
	    else return NULL;
}

int SgCallGraph::numberOfCallSites(SgSymbol *funname){ // number of call sites for funname
	SgCallGraphFunRec *f;
        if(f = locateFunctionInTable(funname)) return f->Num_Call_Sites;
	else return 0;
}

SgStatement *SgCallGraph::firstCallSiteStmt(SgSymbol *fname){ // stmt of first call of fname
	SgCallGraphFunRec *f;
	if((f = locateFunctionInTable(fname)) && (f->Num_Call_Sites>0))
          return (f->currentCallSite = f->callSites->next)->stmt;
	else return NULL;
}

SgStatement *SgCallGraph::nextCallSiteStmt(SgSymbol *fname){ // stmt of next call of fname
	SgCallGraphFunRec *f;
        if((f = locateFunctionInTable(fname)) 
	   && f->currentCallSite
	   && (f->currentCallSite!=f->callSites))
          return (f->currentCallSite = f->currentCallSite->next)->stmt;
	else return NULL;
}

							     
SgExpression *SgCallGraph::firstCallSiteExpr(SgSymbol *fname){ // expression of first call
	SgCallGraphFunRec *f;
        if((f = locateFunctionInTable(fname)) && (f->Num_Call_Sites>0))
          return (f->currentCallExpr = f->callSites->next)->expr;
	else return NULL;
}

SgExpression *SgCallGraph::nextCallSiteExpr(SgSymbol *fname){ // expression of next call
	SgCallGraphFunRec *f;
        if((f = locateFunctionInTable(fname)) 
	   && f->currentCallExpr
	   && (f->currentCallExpr!=f->callSites))
          return (f->currentCallExpr = f->currentCallExpr->next)->expr;
	else return NULL;
}

SgSymbol *SgCallGraph::firstCalledFunction(SgSymbol *fname){ // first function called in fname
	SgCallGraphFunRec *f;
        if((f = locateFunctionInTable(fname)) && (f->Num_Call_List>0))
          return (f->currentFunCall = f->callList->next)->fr->s;
	else return NULL;
}

SgSymbol *SgCallGraph::nextCalledFunction(SgSymbol *fname){ // next function called in fname
	SgCallGraphFunRec *f;
        if((f = locateFunctionInTable(fname)) 
	   && f->currentFunCall
	   && (f->currentFunCall!=f->callList))
          return (f->currentFunCall = f->currentFunCall->next)->fr->s;
	else return NULL;
}

SgStatement *SgCallGraph::SgCalledFunctionStmt(SgSymbol *fname){ //get statement of current called function
  SgCallGraphFunRec *f;
  if((f = locateFunctionInTable(fname)) 
    && f->currentFunCall)
    return (f->currentFunCall)->stmt;
  else return NULL;
}

SgExpression *SgCallGraph::SgCalledFunctionExpr(SgSymbol *fname){ //get expression of current called function
  SgCallGraphFunRec *f;
  if((f = locateFunctionInTable(fname)) 
    && f->currentFunCall)
    return (f->currentFunCall)->expr;
  else return NULL;
}

SgStatement *SgCallGraph::callSiteStmt(SgSymbol *fname, int i){ // stmt of i-th call of funname
	SgCallSiteList *cs;
	SgCallGraphFunRec *f;

        static int first_time=1;
	if(first_time){
	  first_time=0;
	  fprintf(stderr, "SgCallGraph:: Using obsolete function:\n        --    SgStatement *callSiteStmt(SgSymbol *, int)\n");
	}
	if((f = locateFunctionInTable(fname)) && (i < f->Num_Call_Sites)){
	  cs = f->callSites->next;
	  for(;i>0;i--)cs = cs->next;
	  return cs->stmt;
	}
	else return NULL;
/*
        if((f = locateFunctionInTable(fname)) && (i < f->Num_Call_Sites)) return f->callsite[i];
	else return NULL;
*/
}
	
SgExpression *SgCallGraph::callSiteExpr(SgSymbol *fname, int i){ // expression of i-th call
	SgCallSiteList *cs;
	SgCallGraphFunRec *f;

        static int first_time=1;
	if(first_time){
	  first_time=0;
	  fprintf(stderr, "SgCallGraph:: Using obsolete function:\n        --    SgExpression *callSiteExpr(SgSymbol *, int)\n");
	}
	if((f = locateFunctionInTable(fname)) && (i < f->Num_Call_Sites)){
	  cs = f->callSites->next;
	  for(;i>0;i--)cs = cs->next;
	  return cs->expr;
	}
	else return NULL;

/*
        if((f = locateFunctionInTable(fname)) && (i < f->Num_Call_Sites)) return f->callexpr[i];
	else return NULL;
*/
}
	
int SgCallGraph::numberOfFunsCalledFrom(SgSymbol *funname){ // how many call sites in funname
	SgCallGraphFunRec *f;
        if(f = locateFunctionInTable(funname)) return f->Num_Call_List;
	else return 0;
}
	
SgSymbol *SgCallGraph::calledFunction(SgSymbol *funname, int i){ // i-th function called in funname
	SgCallGraphFunRecList *cf;
	SgCallGraphFunRec *f;

        static int first_time=1;
	if(first_time){
	  first_time=0;
	  fprintf(stderr, "SgCallGraph:: Using obsolete function:\n        --    SgSymbol *calledFunction(SgSymbol *, int)\n");
	}
	if((f = locateFunctionInTable(funname)) && (i < f->Num_Call_List)){
	  cf = f->callList->next;
	  for(;i>0;i--)cf = cf->next;
	  return cf->fr->s;
	}
	else return NULL;

/*
        if((f = locateFunctionInTable(funname)) && (i < f->Num_Call_List)) 
                  return f->callList[i]->s;
	else return NULL;
*/
}
int SgCallGraph::isACollectionFunc(SgSymbol* funname){  // 1 if funname is a method of a collection.
	SgCallGraphFunRec *f;
        if((f = locateFunctionInTable(funname))&& (f->type != SGMOE_FUN)) 
		return f->isCollection;
	else return 0;
}
int SgCallGraph::isAMethodOfElement(SgSymbol* funname){  // 1 if funname is a method of a method of an elemnt.
	SgCallGraphFunRec *f;
        if((f = locateFunctionInTable(funname)) && (f->type == SGMOE_FUN))
               return 1;
	else return 0;
}
	
int SgCallGraph::isCalledInSeq(SgSymbol* funname){  // 1 if funname is called in a sequential sect.
	SgCallGraphFunRec *f;
        if(f = locateFunctionInTable(funname)) return f->calledInSeq;
	else return 0;
}
	
int SgCallGraph::isCalledInPar(SgSymbol* funname){  // 1 if funname is called in parallel code
	SgCallGraphFunRec *f;
        if(f = locateFunctionInTable(funname)) return f->calledInPar;
	else return 0;
}
	


void SgCallGraph::init(){
	int i;

	num_funs_in_table = 0;
	currentFun = FunListHead = NULL;
	for(i = 0; i < SGMAX_HASH; i++){
		hash_table[i] = 0;
		}
}
			

void SgCallGraph::printFunctionEntry(SgSymbol *s){

   SgCallSiteList *cs;
   SgCallGraphFunRecList *cf;
   SgCallGraphFunRec *t = locateFunctionInTable(s);
   
   if(!t){
       printf("function %s not found in table\n", s->identifier());
       return;
   }

   printf("---------------------------------------------------------\n");
   printf("[%d] name: %s\n",t->id,t->s->identifier());
   switch(t->type){
	case SGMOE_FUN: 
		printf("Method Of ELement.");
		if(t->className) 
			printf(" Collection: %s\n", t->className->identifier());
		else printf("\n");
		break;
	case SGMOC_FUN:
                if(t->isCollection) printf("Method of Collection\n");
		else printf("Class member function\n");
		if(t->className) 
			printf(" Class: %s\n", t->className->identifier());
		else printf("\n");
		break;
	default:
		printf("ordinary function\n");
	}
   if(t->calledInPar) printf("invoked in parallel!\n");
   else if(t->calledInSeq) printf("part of main thread \n");
   if(t->calledInPar && t->calledInSeq) printf("WARNING: both types!!\n");

   printf("call sites:");
   if(t->Num_Call_Sites == 0) printf("none\n");
   for(cs = t->callSites; cs; cs=cs->next) cs->stmt->unparsestdout();
   printf("which calls:");
   if(t->Num_Call_List == 0) printf(" nothing\n");
   else printf("\n");
   for(cf = t->callList; cf; cf=cf->next) 
                    printf("          %s\n", cf->fr->s->identifier());
   printf("\n");
}

void SgCallGraph::printTableEntry(int i){
  SgSymbol *s = function(i);
  static first_time=1;
  if(!first_time){
    first_time=0;
    fprintf(stderr, "SgCallGraph:: Using obsolete function:\n        --    void printTableEntry(int)\n");
  }
  printFunctionEntry(s);
}

SgCallGraphFunRec * SgCallGraph::locateFunctionInTable(SgSymbol *s){
    unsigned long int hi = hashSymbol(s);
    SgCallGraphFunRec *f;
 
    if(f = hash_table[hi]){
      while(f->s != s)
	if(!(f = f->next_hash))
	  return NULL;
      return f;
    }
    return NULL;
}

void SgCallGraph::insertInHashTable(SgSymbol *s, SgCallGraphFunRec *f){
    unsigned long int hi = hashSymbol(s);
    SgCallGraphFunRec *f2;

    f->next_hash = NULL; 
    if(!(f2 = hash_table[hi]))
      hash_table[hi] = f;
    else{
      while(f2->next_hash){
	f2 = f2->next_hash;
      }
      f2->next_hash = f;
    }
}

unsigned long int SgCallGraph::hashSymbol(SgSymbol *s){
  return (((unsigned long int)(void *)s) % SGMAX_HASH);
}



//
// true if s is  same class as c or a super class of c.
//
int isSameClassOrSuperClass(SgSymbol *c, SgSymbol *s){
   SgStatement *body;
   SgClassStmt *t;
   int i;
   if(!c || !s) return 0;
   if(c == s) return 1;
   if((body = c->body()) && (t = isSgClassStmt(body))){
	for(i = 0; i < t->numberOfSuperClasses(); i++){
		if(isSameClassOrSuperClass(t->superClass(i), s)){
			 return 1;
			}
		}
	}
    return 0;
}
	
SgCallGraphFunRec * SgCallGraph::lookForFunctionOpForClass(SgSymbol *c) {
    if(!c) return NULL;;
    for(SgCallGraphFunRec *f=FunListHead; f; f=f->next){
	if (isSameClassOrSuperClass(c,f->className) &&
            (!strcmp(f-> s->identifier(),"()")))
          return f;
	}
    return NULL;
}
//
// the following function takes a function call reference (statment, expr) to
// a function t and updates the table with the reference info, i.e. who called
// whom.
//

void SgCallGraph::updateFunctionTableConnections(SgCallGraphFunRec *t, SgStatement *stat, SgExpression *e)
  {
   SgStatement *st;
   SgCallGraphFunRec *t2 = NULL;   
	if(t){
              SgCallSiteList *cs = new SgCallSiteList;
              cs->stmt = stat;
              cs->expr = e;
	      if(t->Num_Call_Sites++){
		cs->next = t->callSites->next;
		t->callSites->next = cs;
		t->callSites = cs;
	      }
	      else
		t->callSites = cs->next = cs;
              st = stat->controlParent();
              while(st && (st->variant() != FUNC_HEDR) && 
                              (st->variant() != GLOBAL))
		   st = st->controlParent();
              if(st && st->variant() == FUNC_HEDR){
                    SgFuncHedrStmt *h = isSgFuncHedrStmt(st);
                    t2 = locateFunctionInTable(&(h->name()));
                    }
	      if(!t2){
                    printf("error: could not locate call:");
                    stat->unparsestdout();
                    }
               else{
		 SgCallGraphFunRecList *fr = new SgCallGraphFunRecList;
		 fr->fr = t;
		 fr->stmt = stat;
		 fr->expr = e;
		 if(t2->Num_Call_List++){
		   fr->next = t2->callList->next;
		   t2->callList->next = fr;
		   t2->callList = fr;
		 }
		 else
		   t2->callList = fr->next = fr;
	       }
	    }
         else { /* printf("unknown function in expression:");
         	e->unparsestdout(); printf("\n"); */
		}
 }

//
// the following function builds a new, simplified type for a given
// type generated by the compiler.  Its main function is to cancel
// out the occurrences of adjacent ref and deref operators.
// note: this must be used with care.
// it strips off attributes like const, global, sync, etc.
//

SgType *makeReducedType(SgType *t){
   int done, sign;
   SgPointerType *p;
   SgDerivedType *d;
   SgDescriptType *des;
   SgArrayType *ar;
   SgSymbol *s;
     if(!t) return NULL;
     done = 0;
     sign = 0;
     while(!done){
	if(des = isSgDescriptType(t)){
	    //printf("descript found \n");
            if(des->hasBaseType() && des->baseType()) t = des->baseType();
	    else done = 1;
	    }
        else if(d = isSgDerivedType(t)){
	    // printf("derived found \n");
	    if((s = d->typeName()) && s->type()) t = s->type();
	    else done = 1;
	    }
        else if(ar = isSgArrayType(t)){
	    // vprintf("array found \n");
            if(t->baseType()){
		 t = t->baseType();
		}
            else done = 1;                        
 	    }
	else if(p = isSgPointerType(t)){
	    //printf("ptr found \n");
	    if(p->baseType()){
		sign += p->indirection();
	        t = p->baseType();
		}
	    else done = 1;
	    }
	else done = 1;
	}
     // now add back balance of pointer refs.
     //  printf("sign = %d\n", sign);
     while(sign){
	p = new SgPointerType(*t);
	if(sign < 0){
        	p->setIndirection(-1);
		sign++;
		}
	else{
		p->setIndirection(1);
		sign--;
		}
	t = p;
      }
     return t;
}
//
// to find collection references we need to look at types
// to do this we need to reduce the type expression.
// there may be several layers of dereferences of pointer types
// and typedefs and type modifiers that hide the true type.  
// this functions tries to reduce these to the true type
// of the object.
//

SgType *findTrueType(SgExpression *e){
	SgType *t;
        SgPointerType *p, *p2;
        SgDerivedType *d;
        SgDescriptType *des;
        SgSymbol *s;
        int done = 0;
     if(!e) return NULL;
     if(!(t = e->type())) return NULL;
     while(!done){
	if(des = isSgDescriptType(t)){
            if(des->hasBaseType() && des->baseType()) t = des->baseType();
	    else done = 1;
	    }
        else if(d = isSgDerivedType(t)){
	    if((s = d->typeName()) && s->type()) t = s->type();
	    else done = 1;
	    }
        /*  need to fix baseType for array
        else if(ar = isSgArrayType(t)){
	    if((p2 = isSgPointerType(ar->baseType())) &&
	       (1 + p2->indirection() == 0)) t = p2->baseType();
            else done = 1;                        
 	    }
         */
	else if(p = isSgPointerType(t)){
	    if((p2 = isSgPointerType(p->baseType())) &&
	       (p->indirection() + p2->indirection() == 0)) t = p2->baseType();
	    else if((d = isSgDerivedType(p->baseType())) && 
		    (s = d->typeName()) &&
		    (p2 = isSgPointerType(s->type())) &&
		    (p->indirection() + p2->indirection() == 0) )
                                                           t = p2->baseType();
	    else done = 1;
	    }
	else done = 1;
	}
     return t;
}
	
// 
// this is a utility function that searches an expression in a statement
// for calls to functions.  when they are found they are recored
// in the table
//
void SgCallGraph::findFunctionCalls(SgStatement *stat, SgExpression *e){
   SgFunctionCallExp *f;
   SgFuncPntrExp *fop;
   if(!e) return;
   if(f = isSgFunctionCallExp(e)){
                SgSymbol *s = f->funName();
		SgCallGraphFunRec *t = locateFunctionInTable(s);
                updateFunctionTableConnections(t, stat, e);
          }
   else if(fop = isSgFuncPntrExp(e)){
       // of the form (*funptr)(args) or Coll(args)->...
       // look for a collection type.
       SgExpression *funptr = fop->funExp();
       SgExpression *ref;
       SgType *type;
       SgDerivedCollectionType *colref;
       while(funptr->variant() == EXPR_LIST) funptr = funptr->lhs();
       type = findTrueType(funptr);
       // while(funptr->variant() == DEREF_OP) funptr = funptr->lhs();
       ref = funptr;
       if(ref && type)
            if(colref = isSgDerivedCollectionType(type)){
                   // have a collection and a () ref.
		  /*
		   printf("collection refernce: "); ref->unparsestdout();
                   printf(" with  type:%s <",  colref->collectionName()->identifier());
                   if(elemtype = isSgDerivedType(colref->elementClass())){
			SgSymbol *elemsym = elemtype->typeName();
                        printf("%s >\n", elemsym->identifier());
			}
                   else printf(" ?? >\n");
		   */
		   SgCallGraphFunRec *t = lookForFunctionOpForClass(colref->collectionName());
                   updateFunctionTableConnections(t, stat, fop);
                   }
        }
     
       
  if(e->lhs())findFunctionCalls(stat, e->lhs());
  if(e->rhs())findFunctionCalls(stat, e->rhs());
}

//
// computClosure traverses the graph and looks at who calls whom.
//

void SgCallGraph::computeClosures(){
        int changed, j;
        SgCallGraphFunRec *theMain = NULL;
        SgCallGraphFunRec *f;
	for(f=FunListHead; f; f=f->next){
		 f->flag = 0;
		 if(strcmp(f->s->identifier(),"main") == 0) theMain = f;
		 else if(strcmp(f->s->identifier(),"Processor_Main") == 0) 
			theMain = f;
		}
        //
        // find the functions that are called as part of the main (sequential) thread
        //
	if(!theMain) Message("no main() found in file");
	else{
		changed = 1;
		theMain->calledInSeq = 1;
		while(changed){
			changed = 0;
			for(f=FunListHead; f; f = f->next){
				if(!f->flag && f->calledInSeq){
				        SgCallGraphFunRecList *fl=f->callList;
					for(j = 0; j < f->Num_Call_List; j++,fl=fl->next)
					   if((fl->fr->type != SGMOE_FUN) && 
                                              (fl->fr->isCollection == 0))
						  fl->fr->calledInSeq = 1;
					changed = 1;
					f->flag = 1;
					}
				}
			}
		}
        //
        // find the functions that are called in the parallel regions.
        //
	for(f=FunListHead; f; f=f->next){
		f->flag = 0;
		if((f->type == SGMOE_FUN) || f->isCollection) f->calledInPar = 1;
		}
	changed = 1;
	while(changed){
		changed = 0;
		for(f=FunListHead; f; f=f->next){
			if(!f->flag && f->calledInPar){
			        SgCallGraphFunRecList *fl=f->callList;
				for(j = 0; j < f->Num_Call_List; j++,fl=fl->next)
				  fl->fr->calledInPar = 1;
				changed = 1;
				f->flag = 1;
				}
			}
		}
}
	

void SgCallGraph::GenCallTree(SgFile *f){
   SgSymbol *s = f->firstSymbol();
   SgStatement *st = f->firstStatement();
   SgFunctionSymb *fname;
   SgMemberFuncSymb *memfname;
   SgCallGraphFunRec *frec;
   SgCallGraphFunRec *tail=NULL;

   init();
   for(; s; s = s->next()){
      if(memfname = isSgMemberFuncSymb(s)){

             if(!(frec = new SgCallGraphFunRec)){
			Message("Out of memory in sgCallGraph\n",0);
			exit(1);
			}
	     frec->id = num_funs_in_table++;
	     if(!tail)
               FunListHead = frec;
             else
	       tail->next = frec;
	     frec->next = NULL;
	     tail = frec;
	     insertInHashTable(s, frec);

	     if(memfname->isMethodOfElement())
             	frec->type = SGMOE_FUN;
	     else frec->type = SGMOC_FUN;
             frec->body = memfname->body();
             frec->s = s;
             frec->Num_Call_Sites = 0;
             frec->Num_Call_List = 0;
	     frec->callList = NULL;
	     frec->callSites = NULL;
	     frec->currentCallSite = NULL;
	     frec->currentCallExpr = NULL;
	     frec->currentFunCall = NULL;
	     frec->isCollection = 0;
	     frec->calledInPar = 0;
	     frec->calledInSeq = 0;
             frec->className = memfname->className();
             if(memfname->className() && (
                 (memfname->className()->variant() == COLLECTION_NAME) ||
                 (memfname->className()->variant() == TECLASS_NAME)))
		frec->isCollection = 1;
           }
      else if(fname = isSgFunctionSymb(s)){
             if(!(frec = new SgCallGraphFunRec)){
			Message("Out of memory in sgCallGraph\n",0);
			exit(1);
			}
             frec = new SgCallGraphFunRec;
	     frec->id = num_funs_in_table++;
	     if(!tail)
               FunListHead = frec;
             else
	       tail->next = frec;
	     frec->next = NULL;
	     tail = frec;
	     insertInHashTable(s, frec);

             frec->type = SGNORMAL_FUN;
             frec->body = fname->body();
             frec->s = s;
             frec->Num_Call_Sites = 0;
             frec->Num_Call_List = 0;
	     frec->callList = NULL;
	     frec->callSites = NULL;
	     frec->currentCallSite = NULL;
	     frec->currentCallExpr = NULL;
	     frec->currentFunCall = NULL;
	     frec->isCollection = 0;
	     frec->calledInPar = 0;
	     frec->calledInSeq = 0;
	     frec->className = NULL;
          }
      }
  
    for( ; st; st = st->lexNext()){
             findFunctionCalls(st, st->expr(0));   
             findFunctionCalls(st, st->expr(1));   
             findFunctionCalls(st, st->expr(2));
	     }

    // computeClosures();
}   
               
/******************** end of class SgCallGraph **********************/
