
//  ----------------------------------
//  Darryl Brown
//  University of Oregon pC++/Sage++
//  
//  SgClassHierarchy.C  - the implementation of the class classHierarchy.
//  
//  a class(es) for inspecting the class hierarchy
//  of a sage++ project.
//  
//  ----------------------------------;

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>

#ifdef STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

// All lower data structure needed

#include "tags.h"

#include "typedef.h"
#include "bif.h"
#include "ll.h"
#include "symb.h"

#include "dep.h"
#include "sets.h"

#include "defnodes.h"
#include "vparse.h"

// All maco define in old toolbox
#include "macro.h"
// All the "C" functions from the Rennes toolbox
#include "extcxx_low.h"
// class library header file
#include "fromsage++.h"

#include "baseClasses.h"

#include "sage++classhierarchy.h"


#define DEBUG 0	

// ----------------------------------
char * typeToStr (int ty) {
  
  switch (ty) {
  case T_ENUM: {
    return "enum";
    /*break;*/
  }
  case T_CLASS: {
    return "class";
    /*break;*/
  }
  case T_COLLECTION: {
    return "Collection";
    /*break;*/
  }
  case T_ARRAY: {
    return "array";
    /*break;*/
  }
  case T_STRUCT: {
    return "struct";
    /*break;*/
  }
  case T_INT: {
    return "int";
    /*break;*/
  }
  case T_FLOAT: {
    return "float";
    /*break;*/
  }
  case T_DOUBLE: {
    return "double";
    /*break;*/
  }
  case T_CHAR: {
    return "char";
    /*break;*/
  }
  case T_STRING: {
    return "string";
    /*break;*/
  }
  case T_BOOL: {
    return "bool";
    /*break;*/
  }
  case T_COMPLEX: {
    return "complex";
    /*break;*/
  }
  case T_VOID: {
    return "void";
    /*break;*/
  }
  case T_POINTER: {
    return "pointer";
    /*break;*/
  }
  case T_UNION: {
    return "union";
    /*break;*/
  }
  case T_FUNCTION: {
    return "function";
    /*break;*/
  }
  case T_DERIVED_CLASS: {
    return "derived class";
    /*break;*/
  }
  case T_DERIVED_TYPE: {
    return "derived type";
    /*break;*/
  }
  case T_DERIVED_COLLECTION: {
    return "derived Collection";
    /*break;*/
  }
  case T_REFERENCE: {
    return "reference";
    /*break;*/
  }
  case T_MEMBER_POINTER: {
    return "member pointer";
    /*break;*/
  }
  case T_DESCRIPT: {
    return "descript";
    /*break;*/
  }
  case T_UNKNOWN: {
    return "unknown";
    /*break;*/
  }
  case DEFAULT: {
    return "default";
    /*break;*/
  }
  }
  char tmp[20];
  sprintf(tmp, "\n WIERD variant Type %d\n", ty);
  printf("returning %s\n", tmp);
  return tmp;
}


// ----------------------------------
int strToType(char *s) {
  
  if (brk_strsame(s, "class")) {
    return T_CLASS;
  }  else if (brk_strsame(s, "Collection ")) {
    return T_COLLECTION;
  } else if (brk_strsame(s, "struct ")) {
    return T_STRUCT;
  } else if (brk_strsame(s, "derived class ")) {
    return T_DERIVED_CLASS;
  }  else if (brk_strsame(s, "derived Collection ")) {
    return T_DERIVED_COLLECTION;
  } else if (brk_strsame(s, "derived type ")) {
    return T_DERIVED_TYPE;
  } else if (brk_strsame(s, "union ")) {
    return T_UNION;
  } else if (brk_strsame(s, "int ")) {
    return T_INT;
  } else if (brk_strsame(s, "long ")) {
    return T_INT;
  } else if (brk_strsame(s, "short ")) {
    return T_INT;
  } else if (brk_strsame(s, "char ")) {
    return T_CHAR;
  } else if (brk_strsame(s, "float ")) {
    return T_FLOAT;
  } else if (brk_strsame(s, "double ")) {
    return T_DOUBLE;
  } else if (brk_strsame(s, "pointer ")) {
    return T_POINTER;
  } else if (brk_strsame(s, "default ")) {
    return DEFAULT;
  } else if (brk_strsame(s, "array ")) {
    return T_ARRAY;
  } else if (brk_strsame(s, "void ")) {
    return T_VOID;
  } else if (brk_strsame(s, "reference ")) {
    return T_REFERENCE;
  } else if (brk_strsame(s, "unknown ")) {
    return T_UNKNOWN;
  } else if (brk_strsame(s, "descriptor ")) {
    printf("----descriptor---\n");
    return T_DESCRIPT;
  } else if (brk_strsame(s, "function ")) {
    return T_FUNCTION;
  } else if (brk_strsame(s, "string ")) {
    return T_STRING;
  } else {
    return T_UNKNOWN;
  }
}

//--------------------------------------------------//  
//   HERE ARE THE CLASSES
//--------------------------------------------------//  


//--------------------------------------------------//  
 relationList::relationList () {
   start = current = end = NULL;
   length = 0;
 }

 void relationList::printAll(int tNum) {
  // send print message to each item in list;
  brk_ptrNode *tmp;
  
  tmp = start;
  while (tmp) {
    tmp->node->printAll(tNum);
    tmp = tmp->next;
  }
  fflush(stdout);
}

 void relationList::printAll() {
   printAll(0);
}


//--------------------------------------------------//  
// class hierarchy functions

 classHierarchy::classHierarchy(char * cn) {
   
   allocateLists();

   // store the class name.;
   className = brk_stringSave (cn);
   
 }

 classHierarchy::classHierarchy(void) {
   allocateLists();
 }

 classHierarchy::classHierarchy(SgSymbol * cs) {
   
   allocateLists();
   
   // store the class symbol, ;
   classSymbol = cs;
   
   // and class name, ;
   className = brk_stringSave (cs->identifier());
   
   // and declarative statement;
   declaration = (SgClassStmt *) cs->body();
   
 }

 void classHierarchy::allocateLists() {
   children = new relationList();
   parents = new relationList();
   classMembers = new relationList();
 }   

 classHierarchy::classHierarchy(SgClassStmt * clSt) {

   allocateLists();

   // store the statement declaration...;
   declaration = clSt;

   // store the class symbol ;
   classSymbol = clSt->name();
   
   // and class name;
   className = brk_stringSave(classSymbol->identifier());
 }

void classHierarchy::determineMembers(SgFile *aFile) {
  // declaration must be assigned before this function can;
  // be accomplished;
  SgSymbol  *curSym;
  SgMemberFuncSymb *clSym;
  SgClassStmt *clSt;
  SgStatement *st;
  SgFuncHedrStmt *fst;
  SgVarDeclStmt *vrSt;
  classHierarchy *tmp;
  classMember *tmpCv;
  memberType mt;
  SgFieldSymb *flSym;
  int flag;
  SgType *memTy;

  // loop over Symbol table of file.;
  
  curSym  = aFile->firstSymbol(); 
  while (curSym) {
    
    if (clSym = isSgMemberFuncSymb(curSym)) {
      
      tmp = findClassHierarchy(clSym->className()->identifier());
      if (tmp) {
	
	fst = (SgFuncHedrStmt *) clSym->body();

	// determine if it is private/public/protected...;
	flag = clSym->attributes();

	if (flag & PRIVATE_FIELD) {
	  mt = PRIVATE_FUNC;
	}
	else if (flag & PUBLIC_FIELD) {
	  if (flag & ELEMENT_FIELD)
	    mt = ELEMENT_FUNC;
	  else
	    mt = PUBLIC_FUNC;
	}
	else if (flag & PROTECTED_FIELD) {
	  mt = PROTECTED_FUNC;
	}
	else {
	  mt = UNKNOWN_FUNC;
	}

	tmpCv = new classMember(curSym, mt);

	// determine type ;
	if (fst) {
	  memTy = fst->returnedType();
	  tmpCv->typeOf = typeToStr(memTy->variant());
	} else {
	  tmpCv->typeOf = "unknown";
	} 

	clSt = tmp->declaration;
	// need to discriminate here between what is public, private, etc;
	
	tmp->classMembers->pushLast(tmpCv);
	
      } else {
#       if DEBUG
	printf("ERROR: found a member function with no corresponding class def.\n");
#       endif
      }
      
    } else {
	    
      if (flSym = isSgFieldSymb(curSym)) {

	st = NULL;

#if DEBUG
	if ((vrSt = (SgVarDeclStmt *) curSym->declaredInStmt()) &&
	    (st = vrSt->controlParent())) {
	} else {
	  printf("%s has no declared in stmt\n", curSym->identifier());
        }
#endif

	// parent controler is not a struct, it is a class or collection...;
	if (
	    // statement of declaration..;
	    ((vrSt = (SgVarDeclStmt *) curSym->declaredInStmt()) &&
	    (st = vrSt->controlParent()) &&
	     // now, control parent of the declarative statement;
	    // check to see if correct variants.
	    ((st->variant() == CLASS_DECL) ||
	     (st->variant() == COLLECTION_DECL)))) {
	  
	  tmp = findClassHierarchy(st->symbol()->identifier());
	  
	  if (tmp) {
	    
	    // determine if it is private/public/protected...;
	    flag = curSym->attributes();
	    
	    if (flag & PRIVATE_FIELD) {
	      mt = PRIVATE_VAR;
	    }
	    else if (flag & PUBLIC_FIELD) {
	      if (flag & ELEMENT_FIELD)
	        mt = ELEMENT_VAR;
	      else
	        mt = PUBLIC_VAR;
	    }
	    else if (flag & PROTECTED_FIELD) {
	      mt = PROTECTED_VAR;
	    }
	    else {
	      mt = UNKNOWN_VAR;
	    }
	    
	    tmpCv = new classMember(curSym, mt);
	    
	    // determine type..;
	    memTy = curSym->type();
	    tmpCv->typeOf = typeToStr(memTy->variant());
	    
	    // need to discriminate here between what is public, private, etc;
	    tmp->classMembers->pushLast(tmpCv);
	    
	  } else {
#           if DEBUG
	    printf("ERROR:found a member field with no corresponding class def %s.\n",
		   curSym->identifier());
	    printf("declared In %s", curSym->declaredInStmt()->symbol()->identifier());
	    printf("controlParent: %s\n", st->symbol()->identifier());
#           endif
	  }
	}
      }
    }
    
    // get next symbol in table.;
    curSym = curSym->next();
  }
  
}
       
void classHierarchy::printAllCollections(int tabs) {
  classHierarchy *tmp;

  if (isCollection()) {
    print(tabs);
  }
  // if at least one child, surround with brackets...;
  if (children->start) {
    printf("  { \n");
  } else {
    printf("\n");
  }
  
  fflush(stdout);
  
  // print children;
  tmp = (classHierarchy *) children->searchList();
  while (tmp) {
    tmp->printAllCollections((tabs+1));
    tmp = (classHierarchy *) children->nextItem();
  }
  
  // if at least one child, surround with brackets...;
  if (children->start) {
    for (int i=0; i<tabs; i++) {
      printf("\t");
    }
    printf("}\n");
  }
  
}

 void classHierarchy::printAll(int tabs) {
  print(tabs);
  
  // if at least one chile, surround with brackets...;
  if (children->start) {
    printf("  { \n");
  } else {
    printf("\n");
  }
  
  fflush(stdout);
  
  // print children;
  children->printAll((tabs + 1));
  
  // if at least one chile, surround with brackets...;
  if (children->start) {
    for (int i=0; i<tabs; i++) {
      printf("\t");
    }
    printf("}\n");
  }
  
}

 void classHierarchy::print(int tabs) {
  for (int i=0; i<tabs; i++) {
    printf("\t");
  }
  printf("%s", className);
}

 void classHierarchy::printMemberType(memberType mt) {
  classMember *tmp;

  tmp = (classMember *) classMembers->searchList();
  while (tmp) {
    if (tmp->typeVariant == mt) {
      tmp->print();
    }
    tmp = (classMember *) classMembers->nextItem();
  }
}

 void classHierarchy::print() {
  print(0);
}

//  prints whole list;
 void classHierarchy::printAll() {
  printAll(0);
}

int classHierarchy::numParents() {
  return parents->length;
}

int classHierarchy::numChildren() {
  return children->length;
}

classHierarchy *classHierarchy::findMember(brk_basePtr *bp) {
  classHierarchy *tmp;
  brk_ptrNode *p;
  
  if (p = children->findMember(bp)) {
    return (classHierarchy *) (p->node);
  } 
  tmp = (classHierarchy *) children->searchList();
  while (tmp) {
    if (p = tmp->children->findMember(bp)) {
      return (classHierarchy *) (p->node);
    }
    tmp = (classHierarchy *) children->nextItem();
  }
  return NULL;
}

// -----------------------------------------------------------------;
// (recursively) finds where (or if) a new class fits in this classHierarchy;
// using the new classs superclass (cl).;

 classHierarchy * classHierarchy::findClassHierarchy(char *cl) {
  brk_ptrNode *cur;
  classHierarchy *tmp;
  
  // initialize searching utilities for list;
  children->searchList();
  
  // make certain there is at least one item in list;
  if (children->current) {
    
    // retrieve pointer to classHierarchy from current list member;
    while (cur = children->current) {
      // check to see if this is the classHierarchy of the designated super class;
      tmp = (classHierarchy *) cur->node;
      if (brk_strsame(cl, tmp->className)) {
	return tmp;
      } else {
	// check depth first down this classHierarchy ...;
	tmp = (classHierarchy *) cur->node;
	if (tmp = tmp->findClassHierarchy(cl)) {
	  return tmp;
	}
      }
      children->nextItem();
    }
  }
  return NULL;
  
}

// -----------------------------------------------------------------;
// stores the class declared in clSt at the top level of this classHierarchy ;

classHierarchy * classHierarchy::pushOnTop(SgClassStmt *clSt) {
  
  classHierarchy *tmp;
  if (parents->length > 0) {
    // there is (at least) one parent, so look up the classHierarchy. ;
    tmp = (classHierarchy *) parents->start->node;
    return tmp->pushOnTop(clSt);
  } else {
    // must be at the top if there are no parents;
    children->pushLast(new classHierarchy(clSt));
    return this;
  }
  
}

// -----------------------------------------------------------------;
void classHierarchy::sort(int (* compareFunc) (brk_basePtr *, brk_basePtr *)) {
  // sorts the list, elements must have compare function.,;
  classHierarchy *tmp;
  
  // sort this instances children..;
  children->sort(compareFunc);
  
  // sort rest of descendants..;
  tmp = (classHierarchy *) children->searchList();
  while (tmp) {
    tmp->sort(compareFunc);
    tmp = (classHierarchy *) children->nextItem();
  }
}   
   
// -----------------------------------------------------------------;
 void classHierarchy::sort() {
  // sorts the list, elements must have compare function.,;
  classHierarchy *tmp;
  
  // sort this instances children..;
  children->sort();
  
  // sort rest of descendants..;
  tmp = (classHierarchy *) children->searchList();
  while (tmp) {
    tmp->sort();
    tmp = (classHierarchy *) children->nextItem();
  }
}

// -----------------------------------------------------------------;
// stores a new class in the global hierarchies;
// structure using the statement that the new;
// class is declared in (clSt).;

 classHierarchy * classHierarchy::storeInClassHierarchy (SgClassStmt *clSt) {
  
  char *cl, *superName;
  SgSymbol *super;
  int i, num;
  classHierarchy *parent;
  classHierarchy *h;
  
  // name of class.;
  cl = clSt->name()->identifier();
  

  h = new classHierarchy(clSt);
  if (findMember(h)) {
    delete(h);
    return NULL;
  }
  // check for superclasses.;
  if (clSt->numberOfSuperClasses() == 0) {
    
    // if no superclasses, it is a top level class,  ;
    // so find the top level of this classHierarchy and  ;
    // put it in that class.  ;
    
    return pushOnTop(clSt);
  } else {
    
    // it has a superclass, so look for where this class;
    // fits in terms of other previously defined classes;
    // in the classHierarchy.;
    num = clSt->numberOfSuperClasses();
    for (i = 0; i < num; i++) {
      super = clSt->superClass(i);
      superName = super->identifier();
      
      parent = findClassHierarchy(superName);
      // make certain we did find a place for the new class;
      // i.e. parent is not NULL;
      if (parent) {
	// create new hierarchy...;
	h = new classHierarchy(clSt);
	
	// store new hierarchy in the superclass children list...;
	parent->children->pushLast(h);
	
	// store the parent in the new hierarchies parents list...;
	h->parents->pushLast(parent);
	
      }
    }
    return parent;
  }
}

// -----------------------------------------------------------------;
//  finds and prints out all classes in a file.;

void classHierarchy::determineClassHierarchy(SgFile *aFile) {
  SgStatement *curSt;
  SgClassStmt *clSt;
  
  // loop through statements....;
  
  curSt = aFile->firstStatement();
  while (curSt) {
    
    //  checks to see if symbol is a class,;
    //  and prints is out if it is.    	;
    if ((clSt = isSgClassStmt(curSt)) &&
        (clSt->lastNodeOfStmt() != clSt) &&  // no forward declaration
	((clSt->variant() == CLASS_DECL) ||
	 (clSt->variant() == COLLECTION_DECL))) {
      
      // stores values in classHierarchy
	storeInClassHierarchy(clSt);
    }
    // get next type in table.;
    curSt = curSt->lexNext();
  }
  determineMembers(aFile);
}

int classHierarchy::numberOfParents (void) {
	return parents->length;
}

int classHierarchy::numberOfChildren (void) {
	return children->length;
}

int classHierarchy::numberOfDescendants (void) {
  int sum = 0;
  classHierarchy *tmp;

  children->searchList();
  while (children->current) {
    tmp = (classHierarchy *) children->current->node;
    sum = sum + 1 + tmp->numberOfDescendants();
    children->nextItem();
  }
  return sum;
}

int classHierarchy::isCollection() {
  
  // check to see if declaration statement was ever defined...;
  if (declaration) {
    // then check the variant type...;
    if (declaration->variant() == COLLECTION_DECL) {
      return 1;
    }  else  {
      // it is not a collection...;
      return 0;
    }
  } else {
    // we dont have enough information, so assume it is not a collection...;
    return 0;
  }
}

char * classHierarchy::fileName() {
  // check to see if declaration statement was ever defined...;
  if (declaration) {
    // then check the variant type...;
    return declaration->fileName();
  } else {
    return NULL;
  }
}

int classHierarchy::lineNumber() {
  // check to see if declaration statement was ever defined...;
  if (declaration) {
    // then check the variant type...;
    return declaration->lineNumber();
  } else {
    return -1;
  }
}

int classHierarchy::compare(brk_basePtr *h) {
  classHierarchy *ch;
  char *s1, *s2, *tmp;
  char dif;

  ch = (classHierarchy *) h;
  // convert all from upper to lower case;
  s1 = brk_stringSave(className);
  s2 = brk_stringSave(ch->className);
  dif = 'a' - 'A';

  tmp = s1;
  while (*tmp) {
    if ((*tmp > 'A')  && (*tmp < 'Z')) {
      *tmp = *tmp + dif;
    }
    tmp++;
  }

  tmp = s2;
  while (*tmp) {
    if ((*tmp > 'A') && (*tmp < 'Z')) {
      *tmp = *tmp + dif;
    }
    tmp++;
  }

  // check to see if declaration statement was ever defined...;
  return strcmp(s1, s2);
}

//--------------------------------------------------//  
// class member functions

 classMember::classMember (SgSymbol *sym, memberType mt) {
   typeVariant = mt;
   symbol = sym;
   name = brk_stringSave(sym->identifier());
   memType = sym->type();
   declaration = sym->declaredInStmt();
 }

 classMember::classMember(SgStatement *decl, memberType mt) {
   typeVariant = mt;
   symbol = decl->symbol();
   declaration = decl;
   name = declaration->symbol()->identifier();
 }

void classMember::print() {
  printf("%s (%s)\n", name, typeOf);
}


void classMember::print(int tabs) {
  for(int i = 0; i < tabs; i++) { 
    printf("\t");
  }
  printf("%s\n", name);
}
