/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

#include <stdio.h>

#ifndef __GNUC__
# include <stdlib.h>
#else
extern "C" void abort(void);
extern "C" void exit(int status);
# pragma implementation
#endif

#include "tags.h"
#include "typedef.h"
#include "bif.h"
#include "ll.h"
#include "symb.h"
#include "dep.h"
#include "sets.h"
#include "defnodes.h"
#include "db.h"
#include "vparse.h"

#include "macro.h"

#include "vpc.h"
//#include "f90.h"

#include "extcxx_low.h"
extern int number_of_ll_node;

#undef USER


//
// define for having the debugging
//
//define  DEBUGLIB 1
#define  MAX_FILES 1000
//
//
// Array to keep track of table for a file
//
//

#define ALLOCATECHUNK 1000

#define SORRY Message("Sorry, not implemented yet",0)

class SgProject;
class SgFile;
class SgStatement;
class SgExpression;
class SgLabel;
class SgSymbol;
class SgType;
class SgUnaryExp;
class SgClassSymb;
class SgVarDeclStmt; 
class SgAttribute;

extern void SwitchToFile(int);
extern void RemoveFromTableFile(void *);
extern void RemoveFromTableBfnd(void *);
extern void RemoveFromTableLlnd(void *);
extern void RemoveFromTableSymb(void *);
extern void RemoveFromTableLabel(void *);
extern void RemoveFromTableType(void *);

extern void ResetbfndTableClass();

extern SgFile * GetMappingInTableForFile(PTR_FILE);

extern SgStatement * GetMappingInTableForBfnd(PTR_BFND);
extern SgSymbol * GetMappingInTableForSymbol(PTR_SYMB);

extern SgExpression * GetMappingInTableForLlnd(PTR_LLND);
extern SgType * GetMappingInTableForType(PTR_TYPE);
extern SgLabel * GetMappingInTableForLabel(PTR_LABEL);

extern SgAttribute * GetMappingInTableForBfndAttribute(PTR_BFND);
extern SgAttribute * GetMappingInTableForSymbolAttribute(PTR_SYMB);
extern SgAttribute * GetMappingInTableForTypeAttribute(PTR_TYPE);
extern SgAttribute * GetMappingInTableForLlndAttribute(PTR_LLND);

extern void SetMappingInTableForFile(PTR_FILE, void *);
extern void SetMappingInTableForBfnd(PTR_BFND, void *);
extern void SetMappingInTableForLlnd(PTR_LLND, void *);
extern void SetMappingInTableForSymb(PTR_SYMB, void *);
extern void SetMappingInTableForType(PTR_TYPE, void *);
extern void SetMappingInTableForLabel(PTR_LABEL, void *);

extern void SetMappingInTableForBfndAttribute(PTR_BFND, SgAttribute *);
extern void SetMappingInTableForTypeAttribute(PTR_TYPE, SgAttribute *);
extern void SetMappingInTableForSymbolAttribute(PTR_SYMB, SgAttribute *);
extern void SetMappingInTableForLlndAttribute(PTR_LLND, SgAttribute *);
extern void SetMappingInTableForBfndAttribute(PTR_BFND, SgAttribute *);

extern int CurrentFileNumber;

extern void **fileTableClass;
extern int allocatedForfileTableClass;
extern void **bfndTableClass;
extern int allocatedForbfndTableClass;
extern void **llndTableClass;
extern int allocatedForllndTableClass;
extern void **typeTableClass;
extern int allocatedFortypeTableClass;
extern void **symbolTableClass;
extern int allocatedForsymbolTableClass;
extern void **labelTableClass;
extern int allocatedForlabelTableClass;

extern SgAttribute **fileTableAttribute;
extern int allocatedForfileTableAttribute;
extern SgAttribute **bfndTableAttribute;
extern int allocatedForbfndTableAttribute;
extern SgAttribute **llndTableAttribute;
extern int allocatedForllndTableAttribute;
extern SgAttribute **typeTableAttribute;
extern int allocatedFortypeTableAttribute;
extern SgAttribute **symbolTableAttribute;
extern int allocatedForsymbolTableAttribute;
extern SgAttribute **labelTableAttribute;
extern int allocatedForlabelTableAttribute;
extern "C" void ll_tree_exp(PTR_LLND, int);

//Fortran and C++ Structures
//
// There several families of classes here.
//   Projects-    which correspond to a collection of parsed
//                source files.
//   Files -      which corresponds to an individual source file
//   Statements-  Fortran or C statements
//   Expressions- Fortran or C expression trees.
//   Symbols-     Symbol Table entries.
//   Types-       Each symbol has a type which lives in a type table.
//   Labels-      Statement labels in fortran or C
//   Dependences- Data Dependence Class 
//
//  naming convention: Classnames begin with Sg (for Sage)
//  class functions begin with a lower case and have first letters
//  of words in Caps likeThisWord.
//  
//  In general functions return references when ever possible.  
//
//
// ************* Project and File Types ******************
// the sage fortran 90 and c++ parsers generate files with
// a .dep extension.  A project is a file with a .proj extension
// that consists of a list of .dep files that make the basis
// of the project.  The following describes the
// basic mechanisms to access and modify the structures
// The class hierarch is as follows:
//
//SgProject     = the class representing multi source file projects
//
//SgFile        = the basic source file object.
//   - SgFortranFile  = the subclass for Fortran sources
//   - SgCFile        = the subclass for C files.
//
// ******************************************************************

// As you can see, some statements are specifically Fortran and
// some apply only to C and C++.
//

// the generic statement class has functions to access or modify any
// property of a given statement.

// forward ref

extern SgStatement * BfndMapping(PTR_BFND bif);
extern SgExpression * LlndMapping(PTR_LLND llin);
extern SgSymbol * SymbMapping(PTR_SYMB symb);
extern SgType * TypeMapping(PTR_TYPE ty);
extern SgLabel * LabelMapping(PTR_LABEL label);

#include "fromsage++.h"

extern SgProject *CurrentProject;

// SgProject--inlines

 SgProject::~SgProject()
{}
 SgProject::SgProject(SgProject &)
{ 
 Message("SgProject copy constructor not allowed",0);
}

 int SgProject::numberOfFiles()
{ return LibnumberOfFiles(); }

 char *SgProject::fileName(int i)
{
  PTR_FILE file;
  char * x;

  file = GetFileWithNum(i);
  SetCurrentFileTo(file);
  SwitchToFile(GetFileNumWithPt(file));
  if (!file)
    x = NULL;
  else
    x = FILE_FILENAME(file);
  return x;
}    

 int SgProject::Fortranlanguage()
{ return LibFortranlanguage(); }

 int SgProject::Clanguage()
{ return LibClanguage(); }


// SgFile--inlines

 SgFile::~SgFile()
{  RemoveFromTableFile((void *) this); }

SgFile::SgFile(SgFile &f)
{
  filept = f.filept;  
  Message("SgFile: copy constructor not allowed",0); 
}

 int SgFile::languageType()
{  return FILE_LANGUAGE(filept); }

 void SgFile::saveDepFile(char *dep_file)
{
  SetCurrentFileTo(filept);
  SwitchToFile(GetFileNumWithPt(filept));
  LibsaveDepFile(dep_file);
// id may have change all the bifnode class are deleted....
  ResetbfndTableClass();
}

 void SgFile::unparse(FILE *filedisc)
{
  SetCurrentFileTo(filept);
  SwitchToFile(GetFileNumWithPt(filept));
  UnparseProgram(filedisc);
}
     
 void SgFile::unparsestdout()
{
  SetCurrentFileTo(filept);
  SwitchToFile(GetFileNumWithPt(filept));
  UnparseProgram(stdout);
}


 SgStatement *SgFile::mainProgram()
{
  SetCurrentFileTo(filept);
  SwitchToFile(GetFileNumWithPt(filept));
  return BfndMapping(getMainProgram());
}

 int SgFile::numberOfFunctions()
{
  SetCurrentFileTo(filept);
  SwitchToFile(GetFileNumWithPt(filept));
  return getNumberOfFunction();
}

 int SgFile::numberOfStructs()
{
  SetCurrentFileTo(filept);
  SwitchToFile(GetFileNumWithPt(filept));
  return getNumberOfStruct();
}

 SgStatement *SgFile::firstStatement()
{
  SetCurrentFileTo(filept);
  SwitchToFile(GetFileNumWithPt(filept));
  return BfndMapping(getFirstStmt());
}

 SgSymbol *SgFile::firstSymbol()
{
  SetCurrentFileTo(filept);
  SwitchToFile(GetFileNumWithPt(filept));
  return SymbMapping(PROJ_FIRST_SYMB ());
}

 SgExpression *SgFile::firstExpression()
{
  SetCurrentFileTo(filept);
  SwitchToFile(GetFileNumWithPt(filept));
  return LlndMapping(PROJ_FIRST_LLND ());
}

 SgType *SgFile::firstType()
{
  SetCurrentFileTo(filept);
  SwitchToFile(GetFileNumWithPt(filept));
  return TypeMapping(PROJ_FIRST_TYPE ());
}


 SgExpression *SgFile::SgExpressionWithId(int i)
{ return LlndMapping(Get_ll_with_id (i));}

 SgLabel *SgFile::firstLabel()
{
  SetCurrentFileTo(filept);
  SwitchToFile(GetFileNumWithPt(filept));
  return LabelMapping(PROJ_FIRST_LABEL());
}

 SgStatement *SgFile::SgStatementWithId( int id)
{ return BfndMapping(Get_bif_with_id (id)); }
        
 SgStatement *SgFile::SgStatementAtLine(int lineno)
{ return BfndMapping(rec_num_near_search(lineno));}

 SgSymbol *SgFile::SgSymbolWithId( int id)
{ return SymbMapping(Get_Symb_with_id (id)); }

 SgType *SgFile::SgTypeWithId( int id)
{ return TypeMapping(Get_type_with_id (id)); }



// SgStatement--inlines

 SgStatement::SgStatement(SgStatement &s)
{
  Message("SgStatement: copy constructor not allowed",0);
  thebif = s.thebif; 
}

 int SgStatement::lineNumber()
{ return BIF_LINE(thebif); }

 int  SgStatement::id()
{ return BIF_ID(thebif);}

 int  SgStatement::variant()
{ return BIF_CODE(thebif); }

// inline functions should contain single return
// hence int x is needed.
 int SgStatement::hasSymbol()
{
  int x;

  if (BIF_SYMB(thebif))
    x = TRUE;
  else 
    x = FALSE;

  return x;
}

 SgSymbol * SgStatement::symbol()
{ return SymbMapping(BIF_SYMB(thebif)); }

 char * SgStatement::fileName()
{ return BIF_FILE_NAME(thebif)->name; }

 int SgStatement::hasLabel()
{
  int x;
  if (BIF_LABEL(thebif))
    x = TRUE;
  else
    x = FALSE;
  return x;
}

 void SgStatement::setlineNumber(int n)
{ BIF_LINE(thebif) = n; }

 void SgStatement::setId(int)
{ Message("Id cannot be changed",BIF_LINE(thebif)); }

 void SgStatement::setVariant(int n)
{ BIF_CODE(thebif) = n; }

 void  SgStatement::setLabel(SgLabel &l) 
{ BIF_LABEL(thebif) = l.thelabel; }

void  SgStatement::setLabel(SgLabel * l)
{
  if (l!=NULL)
    BIF_LABEL(thebif) = l->thelabel;
  else
    BIF_LABEL(thebif) = NULL;
}

 void  SgStatement::setSymbol(SgSymbol &s)
{ BIF_SYMB(thebif) =  s.thesymb; }


 SgStatement * SgStatement::lexNext()
{ return BfndMapping(BIF_NEXT(thebif)); }

 SgStatement * SgStatement::lexPrev()
{ return BfndMapping(getNodeBefore(thebif)); }


 SgStatement * SgStatement::controlParent()
{ 
    if (this->variant() != GLOBAL)
    {
        return BfndMapping(BIF_CP(thebif));
    }
    else
    {
        return 0;
    }
}

 int SgStatement::numberOfChildrenList1()
{ return (blobListLength(BIF_BLOB1(thebif)));}

 int SgStatement::numberOfChildrenList2()
{ return (blobListLength(BIF_BLOB2(thebif))); }

 SgStatement * SgStatement::childList1(int i)
{ return BfndMapping(childfInBlobList(BIF_BLOB1(thebif),i)); }

 SgStatement * SgStatement::childList2(int i)
{ return BfndMapping(childfInBlobList(BIF_BLOB2(thebif),i)); }


 void  SgStatement::setLexNext(SgStatement &s)
{ BIF_NEXT(thebif) = s.thebif; }

SgStatement * SgStatement::lastDeclaration()
{ 
  if(BIF_BLOB1(thebif))
    return BfndMapping(LibLastDeclaration(childfInBlobList(BIF_BLOB1(thebif),0))); 
  else
    return BfndMapping(LibLastDeclaration(childfInBlobList(BIF_BLOB1(BIF_CP(thebif)),0)));
}

SgStatement * SgStatement::firstExecutable()
{ 
  if(BIF_BLOB1(thebif))
    return BfndMapping(LibFirstExecutable(childfInBlobList(BIF_BLOB1(thebif),0))); 
  else
    return BfndMapping(LibFirstExecutable(childfInBlobList(BIF_BLOB1(BIF_CP(thebif)),0))); 
    
}

 SgStatement * SgStatement::lastExecutable()
{
  PTR_BFND last;
  last = getLastNodeOfStmt(thebif);
  last = getNodeBefore(last);
  return BfndMapping(last);
}

 SgStatement *SgStatement::lastNodeOfStmt()
{ return BfndMapping(getLastNodeOfStmt(thebif)); }

 SgStatement *SgStatement::nodeBefore()
{ return BfndMapping(getNodeBefore(thebif)); }

 void SgStatement::insertStmtBefore(SgStatement &s)
{ insertBfndBeforeIn(s.thebif,thebif,NULL); }

 void SgStatement::insertStmtBefore(SgStatement &s,SgStatement &cp )
{ insertBfndBeforeIn(s.thebif,thebif,cp.thebif); }


 SgStatement * SgStatement::extractStmt()
{ return BfndMapping(LibextractStmt(thebif)); }

 SgStatement *SgStatement::extractStmtBody()
{ return BfndMapping(LibextractStmtBody(thebif)); }

 void  SgStatement::replaceWithStmt(SgStatement &s)
{ LibreplaceWithStmt(thebif,s.thebif); }

 void  SgStatement::simpledeleteStmt()
{ LibdeleteStmt(thebif); }

 int SgStatement::isIncludedInStmt(SgStatement &s)
{return isInStmt(thebif, s.thebif);}

 SgStatement & SgStatement::copy()
{ return *BfndMapping(duplicateStmtsNoExtract(thebif)); }

 SgStatement * SgStatement::copyPtr()
{ return BfndMapping(duplicateStmtsNoExtract(thebif)); }

 SgStatement * SgStatement::copyOnePtr()
{
     SgStatement *new_stmt = BfndMapping(duplicateOneStmt(thebif));

     /* Hackery to make sure the control parent propagates correctly.
	Unfortunately, the copy function itself it badly broken. */

     new_stmt->setControlParent (this->controlParent());

     return new_stmt;
}

 SgStatement & SgStatement::copyOne()
{
     return *copyOnePtr();
}

 SgStatement * SgStatement::copyBlockPtr()
{ return BfndMapping(duplicateStmtsBlock(thebif)); }
  
 SgStatement & SgStatement::copyBlock()
{ return *copyBlockPtr(); }

 void SgStatement::replaceSymbByExp(SgSymbol &symb, SgExpression &exp)
{
  LibreplaceSymbByExpInStmts(thebif, getLastNodeOfStmt(thebif), symb.thesymb, exp.thellnd);
}

 void SgStatement::replaceSymbBySymb(SgSymbol &symb,SgSymbol &newsymb )
{
  replaceSymbInStmts(thebif, getLastNodeOfStmt(thebif), symb.thesymb, newsymb.thesymb);
}

 void SgStatement::replaceSymbBySymbSameName(SgSymbol &symb,SgSymbol &newsymb)
{
  replaceSymbInStmtsSameName(thebif, getLastNodeOfStmt(thebif), symb.thesymb, newsymb.thesymb);
}

 void SgStatement::replaceTypeInStmt(SgType &old, SgType &newtype)
{// do redundant work by should be ok go twice in member function 
  if (BIF_SYMB(thebif))
    replaceTypeUsedInStmt(BIF_SYMB(thebif),thebif,old.thetype,newtype.thetype);
  else
    replaceTypeUsedInStmt(NULL,thebif,old.thetype,newtype.thetype);
}

 void

SgStatement::setComments(char *comments)
{
     LibSetAllComments (thebif, comments);
}

 SgStatement *SgStatement::getScopeForDeclare()
{
  return BfndMapping(LibGetScopeForDeclare(thebif));
}

 void SgStatement::unparsestdout()
{
  UnparseBif(thebif);
}

 char * SgStatement::comments()
{
  char *x;

  if (BIF_CMNT(thebif))
    x = CMNT_STRING(BIF_CMNT(thebif));
  else
    x = NULL;

  return x;

}

 SgStatement::~SgStatement()
{
   RemoveFromTableBfnd((void *) this);
}

 void SgStatement::addDeclSpec(int type)
{
  BIF_DECL_SPECS(thebif) = BIF_DECL_SPECS(thebif) | type;
}  

 void SgStatement::clearDeclSpec()
{
  BIF_DECL_SPECS(thebif) = 0;
}        

 int SgStatement::isFriend()
{
  return (BIF_DECL_SPECS(thebif) & BIT_FRIEND);
}

 int SgStatement::isInline()
{
  return (BIF_DECL_SPECS(thebif) & BIT_INLINE);
}

 int SgStatement::isExtern()
{
  return (BIF_DECL_SPECS(thebif) & BIT_EXTERN);
}

 int SgStatement::isStatic()
{
  return (BIF_DECL_SPECS(thebif) & BIT_STATIC);
}




// SgExpression--inlines

 SgExpression::SgExpression(SgExpression &e)
{
 Message("SgExpression: copy constructor not allowed",0);
 thellnd = e.thellnd;
}

 SgExpression::~SgExpression()
{ RemoveFromTableLlnd((void *) this); }
        
 SgExpression *SgExpression::lhs()
{ return LlndMapping(NODE_OPERAND0(thellnd)); }

 SgExpression *SgExpression::rhs()
{  return LlndMapping(NODE_OPERAND1(thellnd)); }

 SgExpression *SgExpression::nextInExprTable()
{  return LlndMapping(NODE_NEXT(thellnd)); }

 int SgExpression::variant()
{  return NODE_CODE(thellnd); }

 SgType *SgExpression::type()
{ return TypeMapping(NODE_TYPE(thellnd)); }

 int SgExpression::id()
{ return NODE_ID(thellnd); }

 void SgExpression::setLhs(SgExpression &e)
{ NODE_OPERAND0(thellnd) = e.thellnd; }

 void SgExpression::setLhs(SgExpression *e)
{ NODE_OPERAND0(thellnd) = (e == 0) ? 0 : e->thellnd; }

 void SgExpression::setRhs(SgExpression &e)
{ NODE_OPERAND1(thellnd) = e.thellnd; }

 void SgExpression::setRhs(SgExpression *e)
{ NODE_OPERAND1(thellnd) = ( e == 0 ) ? 0 : e->thellnd; }

 void SgExpression::setSymbol(SgSymbol &s)
{ NODE_SYMB(thellnd) = s.thesymb; }

 void SgExpression::setSymbol(SgSymbol *s)
{ NODE_SYMB(thellnd) = ( s == 0 ) ? 0 : s->thesymb; }

 void SgExpression::setType(SgType &t)
{ NODE_TYPE(thellnd) = t.thetype; }

 void SgExpression::setType(SgType *t)
{ NODE_TYPE(thellnd) = (t == 0) ? 0 : t->thetype; }

 void SgExpression::setVariant(int v)
{  
  Message("Variant of a low level node node should not be change",0);
  NODE_CODE(thellnd) = v; 
}

 SgExpression *SgExpression::copyPtr()
{  return LlndMapping(copyLlNode(thellnd)); }

 SgExpression &SgExpression::copy()
{  return *copyPtr(); }

 SgExpression *SgExpression::IsSymbolInExpression(SgSymbol &symbol)
{ return LlndMapping(LibIsSymbolInExpression(thellnd, symbol.thesymb)); }

 void SgExpression::replaceSymbolByExpression(SgSymbol &symbol, SgExpression &expr)
{ LibreplaceSymbByExp(thellnd, symbol.thesymb, expr.thellnd); }

 SgExpression *SgExpression::arrayRefs()
{ return LlndMapping(LibarrayRefs(thellnd)); }

 SgExpression *SgExpression::symbRefs()
{ return LlndMapping(LibsymbRefs(thellnd,NULL));}

 void SgExpression::unparsestdout()
{ UnparseLLND(thellnd); }


// SgSymbol--inlines

 SgSymbol::~SgSymbol()
{  RemoveFromTableSymb((void *) this); }

 SgSymbol::SgSymbol(SgSymbol &s)
{
 thesymb = s.thesymb;
 Message("SgSymbol: no copy constructor allowed",0);
}


 int SgSymbol::variant()
{ return SYMB_CODE(thesymb); }

 int SgSymbol::id()
{ return SYMB_ID(thesymb); }

 char *SgSymbol::identifier()
{ return SYMB_IDENT(thesymb); }

 SgType *SgSymbol::type()
{ return TypeMapping(SYMB_TYPE(thesymb)); }


 void SgSymbol::setType(SgType &t)
{ SYMB_TYPE(thesymb) = t.thetype; }

 void SgSymbol::setType(SgType *t)
{ SYMB_TYPE(thesymb) = (t == 0) ? 0 : t->thetype; }

 SgStatement *SgSymbol::scope()
{ return BfndMapping(SYMB_SCOPE(thesymb)); }

 SgSymbol *SgSymbol::next()
{ return SymbMapping(SYMB_NEXT(thesymb));}

 SgSymbol *SgSymbol::nextInScope()
{ return SymbMapping(thesymb->next_symb);}

 SgSymbol &SgSymbol::copy() 
{ return *SymbMapping(duplicateSymbol(thesymb)); }

 SgSymbol &SgSymbol::copyLevel1() 
{ return *SymbMapping(duplicateSymbolLevel1(thesymb)); }

 SgSymbol &SgSymbol::copyLevel2() 
{ return *SymbMapping(duplicateSymbolLevel2(thesymb)); }

 SgSymbol &SgSymbol::copyAcrossFiles(SgStatement &where) 
{ 
  resetDoVarForSymb();
  return *SymbMapping(duplicateSymbolAcrossFiles(thesymb,where.thebif)); 
}

 void SgSymbol::declareTheSymbolWithParamList
            (SgStatement &st, SgExpression &parlist)
{ declareAVarWPar(thesymb, parlist.thellnd, st.thebif); }


 SgExpression *SgSymbol::makeDeclExprWithParamList
           (SgExpression &parlist)
{ return LlndMapping(makeDeclExpWPar(thesymb, parlist.thellnd));} 

// SgType--inlines

 SgType::SgType(PTR_TYPE type)
{
  thetype = type;
  SetMappingInTableForType(thetype, (void *) this);
}

 SgType::SgType(SgType &t)
{
  thetype = t.thetype;
  Message("SgType: no copy constructor allowed",0); 
}

 int SgType::variant()
{ return TYPE_CODE(thetype); }

 int SgType::id()
{ return TYPE_ID(thetype); }

 SgSymbol *SgType::symbol()
{ return SymbMapping(TYPE_SYMB_DERIVE(thetype));}

 SgType *SgType::copyPtr()
{ return TypeMapping(duplicateType(thetype));}

 SgType &SgType::copy()
{ return *copyPtr(); }

 SgType *SgType::next()
{ return TypeMapping(TYPE_NEXT(thetype)); }

 int SgType::isTheElementType()
{ return isElementType(thetype);}

 int SgType::equivalentToType(SgType &type)
{ return isTypeEquivalent(thetype, type.thetype);}

 int SgType::equivalentToType(SgType *type)
{
  if ( type == 0 )
    return 0;
  else
    return isTypeEquivalent(thetype, type->thetype);
}


 SgType *SgType::internalBaseType()
{
  PTR_TYPE ty;
  ty = lookForInternalBasetype(thetype);
  return TypeMapping(ty);
}

 int SgType::hasBaseType()
{
  return hasTypeBaseType(TYPE_CODE(thetype));
}

 SgType *SgType::baseType()
{
  SgType * x;
  if (hasTypeBaseType(TYPE_CODE(thetype)))
    x = TypeMapping(TYPE_BASE(thetype));
  else
    x = NULL;

  return x;
}

 SgExpression *SgType::length()
{ return LlndMapping(TYPE_RANGES(thetype)); }

// SgLabel--inlines

 SgLabel::SgLabel(SgLabel & lab){
  Message("SgLabel: copy constructor not allowed",0);
  thelabel = lab.thelabel;
}

 SgLabel::SgLabel(PTR_LABEL lab)
{
  thelabel = lab;
  SetMappingInTableForLabel(thelabel, (void *) this);
}

 SgLabel::SgLabel(int i)
{
  thelabel = (PTR_LABEL) newNode (LABEL_KIND);
  LABEL_STMTNO(thelabel) = i;
  SetMappingInTableForLabel(thelabel, (void *) this);
}
 int SgLabel::id()
{ return LABEL_ID(thelabel); }

int SgLabel::labelValue()
{
  return LABEL_STMTNO(thelabel);
}

SgLabel * SgLabel::nextInScope(SgStatement * s)
{
  PTR_LABEL lab=thelabel;
  while (LABEL_NEXT(lab)!=NULL) {
    if (LABEL_SCOPE(LABEL_NEXT(lab))==s->thebif)
      return LabelMapping(LABEL_NEXT(thelabel));
    lab=LABEL_NEXT(lab);
  }  
  return NULL;
}

 int SgLabel::getLastLabelVal()
{ return getLastLabelId();}

  SgLabel::~SgLabel()
{  RemoveFromTableLabel((void *) this); }


// SgValueExp--inlines

 SgValueExp::SgValueExp(int value):SgExpression(INT_VAL)
{
  NODE_TYPE(thellnd) =  GetAtomicType(T_INT);
  NODE_INT_CST_LOW (thellnd) = value;
}

 SgValueExp::SgValueExp(char char_val):SgExpression( CHAR_VAL)
{
  NODE_TYPE(thellnd) = GetAtomicType(T_CHAR);
  NODE_CHAR_CST(thellnd) = char_val;
}

 SgValueExp::SgValueExp(float float_val):SgExpression(FLOAT_VAL)
{
  char tmp[100]; // No doubles longer than 100 digits;
  sprintf (tmp,"%e",float_val);
  NODE_STR(thellnd) = (char*) xmalloc ((strlen(tmp) + 1)*sizeof(char));
  strcpy(NODE_STR(thellnd),tmp);
  NODE_TYPE(thellnd) = GetAtomicType(T_FLOAT);
  
}

 SgValueExp::SgValueExp(double double_val):SgExpression(DOUBLE_VAL)
{
  char tmp[100]; // No doubles longer than 100 digits ;
  sprintf (tmp,"%e",double_val);
  NODE_STR(thellnd) = (char*) xmalloc ((strlen(tmp) + 1)*sizeof(char));
  strcpy(NODE_STR(thellnd),tmp);
  NODE_TYPE(thellnd) = GetAtomicType(T_DOUBLE);
}

 SgValueExp::SgValueExp(char *string_val):SgExpression(STRING_VAL)
{
  NODE_TYPE(thellnd) = GetAtomicType(T_STRING);
  NODE_STRING_POINTER(thellnd) = string_val;
}

 SgValueExp::SgValueExp(double real, double imaginary):SgExpression(COMPLEX_VAL)
{
  NODE_TYPE(thellnd) = GetAtomicType(T_COMPLEX);
  NODE_OPERAND0(thellnd) = SgValueExp(real).thellnd;
  NODE_OPERAND1(thellnd) = SgValueExp(imaginary).thellnd;
}

 SgValueExp::SgValueExp(SgValueExp &real, SgValueExp &imaginary):SgExpression(COMPLEX_VAL)
{
  NODE_TYPE(thellnd) = GetAtomicType(T_COMPLEX);
  NODE_OPERAND0(thellnd) = real.thellnd;
  NODE_OPERAND1(thellnd) = imaginary.thellnd;
}

// are these setValue functions really needed?
// the user can simply say, SgValueExp(3.0) and
// get the same functionality, in most cases.
// Moreover, the code is wrong. The NODE_ CODE field
// must be checked.
  void SgValueExp::setValue(int int_val)
{
  NODE_INT_CST_LOW (thellnd) = int_val;
}

  void SgValueExp::setValue(char char_val)
{
  NODE_CHAR_CST(thellnd) = char_val;
}

  void SgValueExp::setValue(float float_val)
{
  char tmp[100]; // No doubles longer than 100 digits ;
  sprintf (tmp,"%e",float_val);
  if (!NODE_STR(thellnd))
    NODE_STR(thellnd) = (char*) xmalloc ((strlen(tmp) + 1)*sizeof(char));
  strcpy(NODE_STR(thellnd),tmp);
}

  void SgValueExp::setValue(double double_val)
{
  char tmp[100]; // No doubles longer than 100 digits ;
  sprintf (tmp,"%e",double_val);
  if (!NODE_STR(thellnd))
    NODE_STR(thellnd) = (char*) xmalloc ((strlen(tmp) + 1)*sizeof(char));
  strcpy(NODE_STR(thellnd),tmp);
}

  void SgValueExp::setValue(char *string_val)
{
  NODE_STRING_POINTER(thellnd) = string_val;
}

  void SgValueExp::setValue(double real, double im)
{
  NODE_OPERAND0(thellnd) = SgValueExp(real).thellnd;
  NODE_OPERAND1(thellnd) = SgValueExp(im).thellnd;
}

  void SgValueExp::setValue(SgValueExp &real, SgValueExp & im)
{
  NODE_OPERAND0(thellnd) = real.thellnd;
  NODE_OPERAND1(thellnd) = im.thellnd;
}

 int SgValueExp::intValue()
{
  int x;
  if (NODE_CODE(thellnd) != INT_VAL)
    {
      Message("message initValue not understood");
      x = 0;
    }
  else
    x = NODE_INT_CST_LOW (thellnd);
  return x;
}

 char* SgValueExp::floatValue()
{
  char*  x;

  if (NODE_CODE(thellnd) != FLOAT_VAL)
    {
      Message("message floatValue not understood");
      x = NULL;
    }
  else 
    x = NODE_FLOAT_CST(thellnd);

  return x;
}

 char SgValueExp::charValue()
{
  char x;

  if (NODE_CODE(thellnd) != CHAR_VAL)
    {
      Message("message charValue not understood");
      x = 0;
    }
  else
    x = NODE_CHAR_CST(thellnd);

  return x;
}

 char*  SgValueExp::doubleValue()
{
  char*  x;

  if (NODE_CODE(thellnd) != DOUBLE_VAL)
    {
      Message("message doubleValue not understood");
      x = NULL;
    }
  else
    x = NODE_DOUBLE_CST(thellnd);

  return x;
}

 char * SgValueExp::stringValue()
{
  char *x;

  if (NODE_CODE(thellnd) != STRING_VAL)
    {
      Message("message stringValue not understood");
      x = NULL;
    }
  else
    x = NODE_STRING_POINTER(thellnd);

  return x;
}

 SgExpression * SgValueExp:: realValue()
{
  SgExpression *x;

  if (NODE_CODE(thellnd) != COMPLEX_VAL)
    {
      Message("message realValue not understood");
      x = NULL;
    }
  else 
    x = LlndMapping(NODE_OPERAND0(thellnd));

  return x;
}

 SgExpression * SgValueExp::imaginaryValue()
{
  SgExpression *x;

  if (NODE_CODE(thellnd) != COMPLEX_VAL)
    {
      Message("message imaginaryValue not understood");
      x = NULL;
    }
  else
    x = LlndMapping(NODE_OPERAND1(thellnd));

  return x;
}



// SgKeywordValExp--inlines
  SgKeywordValExp::SgKeywordValExp(char *name):SgExpression(KEYWORD_VAL)
{ NODE_STRING_POINTER(thellnd) = name; }

 char * SgKeywordValExp::value()
{ return NODE_STRING_POINTER(thellnd); }


// SgUnaryExp--inlines

// In the code below, no type checking has been done. 
// Some of the parser code may be modified to do the type-checking.
// For example, SgUnaryExp(ADDRESS_OP, 2) should not
// be detected.
// the standard unary expressons
// variant:DEREF_OP      * expr
// variant:ADDRESS_OP    & expr
// variant:MINUS_OP      - expr
// variant:UNARY_ADD_OP  + expr
// variant:PLUSPLUS_OP   ++lhd  or rhs++
// variant:MINUSMINUS_OP --lhs  or rhs--
// variant:BIT_COMPLEMENT_OP  ~ expr
// variant:NOT_OP        ! expr
// variant:SIZE_OP       sizeof( expr)

 SgUnaryExp::SgUnaryExp(PTR_LLND ll):SgExpression(ll)
{}
 SgUnaryExp::SgUnaryExp(int variant, SgExpression & e):SgExpression(variant)
{
  NODE_OPERAND0(thellnd) = e.thellnd;
}

 SgUnaryExp::SgUnaryExp(int variant, int post, SgExpression &e):SgExpression(variant)
{  // post =1 rhs++
  if (post)
    NODE_OPERAND1(thellnd) = e.thellnd;
  else
    NODE_OPERAND0(thellnd) = e.thellnd;
}

  int SgUnaryExp::post() // returns TRUE if a post inc or dec op.
{ if (NODE_OPERAND1(thellnd)) return TRUE; else return FALSE;}


// SgCastExp--inlines

 SgCastExp::SgCastExp(PTR_LLND ll):SgExpression(ll)
{}

 SgCastExp::SgCastExp(SgType &t, SgExpression &e):SgExpression(CAST_OP)
{
  NODE_TYPE(thellnd) = t.thetype;
  NODE_OPERAND0(thellnd) = e.thellnd;
  // an experiment to fix the bernd bug.
  NODE_OPERAND1(thellnd) = (SgMakeDeclExp(NULL, &t))->thellnd;
}

 SgCastExp::SgCastExp(SgType &t):SgExpression(CAST_OP)
{ NODE_TYPE(thellnd) = t.thetype; }

 SgCastExp::~SgCastExp(){RemoveFromTableLlnd((void *) this);}


// SgDeleteExp--inlines

 SgDeleteExp::SgDeleteExp(PTR_LLND ll):SgExpression(ll)
{}

 SgDeleteExp::SgDeleteExp(SgExpression &size,SgExpression &expr):SgExpression(DELETE_OP)
{
  NODE_OPERAND0(thellnd) = expr.thellnd;
  NODE_OPERAND1(thellnd) = size.thellnd;
}

 SgDeleteExp::SgDeleteExp( SgExpression &expr):SgExpression(DELETE_OP)
{
  NODE_OPERAND0(thellnd) = expr.thellnd;
}

 SgDeleteExp::~SgDeleteExp()
{ RemoveFromTableLlnd((void *) this); }



// SgNewExp--inlines


 SgNewExp::SgNewExp(PTR_LLND ll):SgExpression(ll)
{}
 
 SgNewExp::SgNewExp(SgType &t):SgExpression(NEW_OP)
{
  SgCastExp *pt;
   pt =  new SgCastExp(t);
   NODE_OPERAND0(thellnd) = pt->thellnd;
}
 
 SgNewExp::SgNewExp(SgType &t, SgExpression &e):SgExpression(NEW_OP)
{
  SgCastExp *pt;
  pt =  new SgCastExp(t);
  NODE_OPERAND0(thellnd) = pt->thellnd;
  NODE_OPERAND1(thellnd) = e.thellnd;
}

 SgNewExp::~SgNewExp()
{ RemoveFromTableLlnd((void *) this); }


// SgExprIfExp--inlines

 SgExprIfExp::SgExprIfExp(PTR_LLND ll): SgExpression(ll)
{}

 SgExprIfExp::SgExprIfExp(SgExpression &exp1, 
                                SgExpression &exp2, 
                                SgExpression &exp3):SgExpression(EXPR_IF)
{
  NODE_OPERAND0(thellnd)= exp1.thellnd;
  NODE_OPERAND1(thellnd)= newExpr(EXPR_IF_BODY,NODE_TYPE(exp2.thellnd),exp2.thellnd,exp3.thellnd);
}

 void SgExprIfExp::setConditional(SgExpression &c)
{
  NODE_OPERAND0(thellnd) = c.thellnd;
}

// SgFunctionRefExp--inlines
 SgFunctionRefExp::SgFunctionRefExp(PTR_LLND ll):SgExpression(ll)
{}
 SgFunctionRefExp::SgFunctionRefExp(SgSymbol &fun):SgExpression(FUNC_CALL)
{
  NODE_SYMB (thellnd) = fun.thesymb;
}
 SgFunctionRefExp::~SgFunctionRefExp()
{ RemoveFromTableLlnd((void *) this); }

 SgSymbol *SgFunctionRefExp::funName()
{ return SymbMapping(NODE_SYMB(thellnd)); }

 SgExpression * SgFunctionRefExp::args()
{ return LlndMapping(NODE_OPERAND0(thellnd)); }

  int SgFunctionRefExp::numberOfArgs()
{ return exprListLength(NODE_OPERAND0(thellnd)); }

 SgExpression * SgFunctionRefExp::arg(int i)
{ return LlndMapping(getPositionInExprList(NODE_OPERAND0(thellnd),i));  }

 void SgFunctionRefExp::addArg(SgExpression &arg)
{
  struct llnd * tail;
  tail = thellnd;
  if (NODE_OPERAND0(tail)==NULL) {
    NODE_OPERAND0(tail)=newExpr(EXPR_LIST,NULL, arg.thellnd, NULL);
  }
  else {
    tail=NODE_OPERAND0(tail);
    while (NODE_OPERAND1(tail)!=NULL)
      tail=NODE_OPERAND1(tail);
    NODE_OPERAND1(tail)=newExpr(EXPR_LIST,NULL, arg.thellnd, NULL);
  };
}

// SgFunctionCallExp--inlines

 SgFunctionCallExp::SgFunctionCallExp(PTR_LLND ll):SgExpression(ll)
{}

 SgFunctionCallExp::SgFunctionCallExp(SgSymbol &fun, SgExpression &paramList):SgExpression(FUNC_CALL)
{
  NODE_SYMB (thellnd) = fun.thesymb;
  NODE_OPERAND0(thellnd) = paramList.thellnd;
}

 SgFunctionCallExp::SgFunctionCallExp(SgSymbol &fun):SgExpression(FUNC_CALL)
{
  NODE_SYMB (thellnd) = fun.thesymb;
}
 SgFunctionCallExp::~SgFunctionCallExp()
{ RemoveFromTableLlnd((void *) this); }

 SgSymbol *SgFunctionCallExp::funName()
{ return SymbMapping(NODE_SYMB(thellnd)); }

 SgExpression * SgFunctionCallExp::args()
{ return LlndMapping(NODE_OPERAND0(thellnd)); }

  int SgFunctionCallExp::numberOfArgs()
{ return exprListLength(NODE_OPERAND0(thellnd)); }

 SgExpression * SgFunctionCallExp::arg(int i)
{ return LlndMapping(getPositionInExprList(NODE_OPERAND0(thellnd),i));  }

 void SgFunctionCallExp::addArg(SgExpression &arg)
{ NODE_OPERAND0(thellnd) = addToExprList(NODE_OPERAND0(thellnd),arg.thellnd); }



// SgFuncPntrExp--inlines

 SgFuncPntrExp::SgFuncPntrExp(PTR_LLND ll):SgExpression(ll)
{}

 SgFuncPntrExp::SgFuncPntrExp(SgExpression &ptr):SgExpression(FUNCTION_OP)
{ NODE_OPERAND0(thellnd) = ptr.thellnd; }

 SgFuncPntrExp::~SgFuncPntrExp(){RemoveFromTableLlnd((void *) this);}

 SgExpression * SgFuncPntrExp::funExp()
{  return LlndMapping(NODE_OPERAND0(thellnd)); }

  void SgFuncPntrExp::setFunExp(SgExpression &s)
{ NODE_OPERAND0(thellnd) = s.thellnd; }

 int SgFuncPntrExp::numberOfArgs()
{ return exprListLength(NODE_OPERAND1(thellnd)); }

  SgExpression * SgFuncPntrExp::arg(int i)
{ return LlndMapping(getPositionInExprList(NODE_OPERAND1(thellnd),i)); }

 void SgFuncPntrExp::addArg(SgExpression &arg)
{ NODE_OPERAND1(thellnd) = addToExprList(NODE_OPERAND1(thellnd),arg.thellnd);}
          


// SgExprListExp--inlines

 SgExprListExp::SgExprListExp(PTR_LLND ll):SgExpression(ll)
{}

 SgExprListExp::SgExprListExp():SgExpression(EXPR_LIST)
{}

 SgExprListExp::SgExprListExp(SgExpression &ptr):SgExpression(EXPR_LIST)
{ NODE_OPERAND0(thellnd) = ptr.thellnd; }

 SgExprListExp::~SgExprListExp(){RemoveFromTableLlnd((void *) this);}

  int SgExprListExp::length()
{ return exprListLength(thellnd); }

  SgExpression * SgExprListExp::elem(int i)
{ return LlndMapping(getPositionInExprList(thellnd,i)); }

 SgExprListExp * SgExprListExp::next()
{ return (SgExprListExp *) LlndMapping(NODE_OPERAND1(thellnd)); }

 SgExpression * SgExprListExp::value()
{ return LlndMapping(NODE_OPERAND0(thellnd)); }

 void SgExprListExp::setValue(SgExpression &ptr)
{ NODE_OPERAND0(thellnd) = ptr.thellnd; }

 void SgExprListExp::append(SgExpression &arg)
{ thellnd = addToList(thellnd,arg.thellnd); }

void SgExprListExp::insertExpAfter(SgExpression & exp)
{
  PTR_LLND temp = NODE_OPERAND1(thellnd);
  NODE_OPERAND1(thellnd) = exp.thellnd;
  NODE_OPERAND1(exp.thellnd)=temp;
}

// SgRefExp--inlines
 SgRefExp::SgRefExp(PTR_LLND ll):SgExpression(ll)
{}

 SgRefExp::SgRefExp(int variant, SgSymbol &s):SgExpression(variant)
{
  NODE_SYMB(thellnd) = s.thesymb;
  NODE_TYPE(thellnd) = SYMB_TYPE(s.thesymb);
}

 SgRefExp::~SgRefExp()
{ RemoveFromTableLlnd((void *) this); }

// SgTypeRefExp -- inlines

 SgTypeRefExp::SgTypeRefExp(SgType &t): SgExpression(TYPE_REF){
     NODE_TYPE(thellnd) = t.thetype;
}

 SgType * SgTypeRefExp::getType(){
     return TypeMapping(NODE_TYPE(thellnd));
}

 SgTypeRefExp::~SgTypeRefExp()
{ RemoveFromTableLlnd((void *) this); }

// SgVarRefExp--inlines

 SgVarRefExp::SgVarRefExp (PTR_LLND ll):SgExpression(ll)
{}

 SgVarRefExp::SgVarRefExp(SgSymbol &s):SgExpression(VAR_REF)
{
  NODE_TYPE(thellnd) = SYMB_TYPE(s.thesymb);
  NODE_SYMB(thellnd) = s.thesymb;
}
 SgVarRefExp::SgVarRefExp(SgSymbol *s):SgExpression(VAR_REF)
{
  if(s){
  NODE_TYPE(thellnd) = SYMB_TYPE(s->thesymb);
  NODE_SYMB(thellnd) = s->thesymb;
  }
}

 SgVarRefExp::~SgVarRefExp()
{ RemoveFromTableLlnd((void *) this); }


// SgThisExp--inlines

 SgThisExp::SgThisExp (PTR_LLND ll):SgExpression(ll)
{}

 SgThisExp::SgThisExp(SgType &t):SgExpression(THIS_NODE)
{ NODE_TYPE(thellnd) = t.thetype; }

 SgThisExp::~SgThisExp()
{ RemoveFromTableLlnd((void *) this); }

// SgTripletExp;

SgTripletExp::SgTripletExp(SgExpression * item0, SgExpression * item1, SgExpression *item2):SgExpression(DDOT)
{}

int SgTripletExp::numberOfItems()
{
  if ((this->operand(1)==NULL)&&(this->operand(2)==NULL))
    return (0);
  if (this->operand(2)==NULL)
    return(2);
  else 
    return(3);
}

SgExpression * SgTripletExp::item(int i)
{
  if (i==0)
    return (this->operand(1)->operand(1));
  if (i==1)
    return (this->operand(1)->operand(2));
  if (i==2)
    return (this->operand(2));
  return this;
}

// SgArrayRefExp--inlines

 SgArrayRefExp::SgArrayRefExp(PTR_LLND ll):SgExpression(ll)
{}

 SgArrayRefExp::SgArrayRefExp(SgSymbol &s):SgExpression(ARRAY_REF)
{
  PTR_SYMB symb;
  
  symb = s.thesymb;
// We don't do the check.
//  if (!arraySymbol(symb))
//    Message("Attempt to create an array ref with a symbol not of type array",0);
  NODE_SYMB(thellnd) = symb;
  NODE_TYPE(thellnd) = lookForInternalBasetype(SYMB_TYPE(symb));
}

 SgArrayRefExp::SgArrayRefExp(SgSymbol &s, SgExpression &subscripts):SgExpression(ARRAY_REF)
{
  PTR_SYMB symb;
  
  symb = s.thesymb;
  if (!arraySymbol(symb))
    Message("Attempt to create an array ref with a symbol not of type array",0);
  
  NODE_SYMB(thellnd) = symb;
  if(NODE_CODE(subscripts.thellnd) == EXPR_LIST)
     NODE_OPERAND0(thellnd) = subscripts.thellnd;
  else
     NODE_OPERAND0(thellnd) =  addToExprList(NODE_OPERAND0(thellnd),subscripts.thellnd);
  NODE_TYPE(thellnd) = lookForInternalBasetype(SYMB_TYPE(symb));
}

 SgArrayRefExp::SgArrayRefExp(SgSymbol &s, SgExpression &sub1,SgExpression &sub2):SgExpression(ARRAY_REF)
{
  PTR_SYMB symb;
  
  symb = s.thesymb;
  
  if (!arraySymbol(symb))
    Message("Attempt to create an array ref with a symbol not of type array",0);
  NODE_SYMB(thellnd) = symb;
  NODE_OPERAND0(thellnd) =  addToExprList(NODE_OPERAND0(thellnd),sub1.thellnd);
  NODE_OPERAND0(thellnd) =  addToExprList(NODE_OPERAND0(thellnd),sub2.thellnd);
  NODE_TYPE(thellnd) = lookForInternalBasetype(SYMB_TYPE(symb));
}


 SgArrayRefExp::SgArrayRefExp(SgSymbol &s, SgExpression &sub1,SgExpression &sub2,SgExpression &sub3):SgExpression(ARRAY_REF)
{
  PTR_SYMB symb;
  
  symb = s.thesymb;
  
  if (!arraySymbol(symb))
    Message("Attempt to create an array ref with a symbol not of type array",0);
  NODE_SYMB(thellnd) = symb;
  NODE_OPERAND0(thellnd) =  addToExprList(NODE_OPERAND0(thellnd),sub1.thellnd);
  NODE_OPERAND0(thellnd) =  addToExprList(NODE_OPERAND0(thellnd),sub2.thellnd);
  NODE_OPERAND0(thellnd) =  addToExprList(NODE_OPERAND0(thellnd),sub3.thellnd);
  NODE_TYPE(thellnd) = lookForInternalBasetype(SYMB_TYPE(symb));
}

 SgArrayRefExp::SgArrayRefExp(SgSymbol &s, SgExpression &sub1,SgExpression &sub2,SgExpression &sub3,SgExpression &sub4):SgExpression(ARRAY_REF)
{
  PTR_SYMB symb;
  
  symb = s.thesymb;
  
  if (!arraySymbol(symb))
    Message("Attempt to create an array ref with a symbol not of type array",0);
  NODE_SYMB(thellnd) = symb;
  NODE_OPERAND0(thellnd) =  addToExprList(NODE_OPERAND0(thellnd),sub1.thellnd);
  NODE_OPERAND0(thellnd) =  addToExprList(NODE_OPERAND0(thellnd),sub2.thellnd);
  NODE_OPERAND0(thellnd) =  addToExprList(NODE_OPERAND0(thellnd),sub3.thellnd);
  NODE_OPERAND0(thellnd) =  addToExprList(NODE_OPERAND0(thellnd),sub4.thellnd);
  NODE_TYPE(thellnd) = lookForInternalBasetype(SYMB_TYPE(symb));
}

 SgArrayRefExp::  ~SgArrayRefExp()
{ RemoveFromTableLlnd((void *) this); }

// the number of subscripts in reference
 int SgArrayRefExp::numberOfSubscripts()
{ return exprListLength(NODE_OPERAND0(thellnd));}

 SgExpression * SgArrayRefExp::subscripts()
{ return LlndMapping(NODE_OPERAND0(thellnd)); }

 SgExpression * SgArrayRefExp::subscript(int i)
{
  PTR_LLND ll = NULL;
  ll = getPositionInExprList(NODE_OPERAND0(thellnd),i);
  return LlndMapping(ll);
}

 void SgArrayRefExp::addSubscript(SgExpression &e)
{ NODE_OPERAND0(thellnd) =  addToExprList(NODE_OPERAND0(thellnd),e.thellnd);}

 void SgArrayRefExp::setSymbol(SgSymbol &s)
{ NODE_SYMB(thellnd) = s.thesymb;}

int SgArrayRefExp::isScalarOnDim(int i)
{
  PTR_LLND ll = NULL;
  ll = getPositionInExprList(NODE_OPERAND0(thellnd),i);
  
  if (NODE_CODE(ll)!=DDOT)
    return (1);
  else 
    return (0);
}

// SgProcessorsRefExp--inlines

 SgProcessorsRefExp::SgProcessorsRefExp(PTR_LLND ll):SgExpression(ll)
{}

 SgProcessorsRefExp::SgProcessorsRefExp():SgExpression(PROCESSORS_REF)
{
  SgSymbol *symb;
  
  symb = new SgSymbol(VARIABLE_NAME, "_PROCESSORS");
  NODE_SYMB(thellnd) = symb->thesymb;
  NODE_TYPE(thellnd) = lookForInternalBasetype(SYMB_TYPE(symb->thesymb));
}

 SgProcessorsRefExp::SgProcessorsRefExp(SgExpression &subscripts):SgExpression(PROCESSORS_REF)
{
  SgSymbol *symb;
  
  symb = new SgSymbol(VARIABLE_NAME, "_PROCESSORS");
  NODE_SYMB(thellnd) = symb->thesymb;
  NODE_OPERAND0(thellnd) =  addToExprList(NODE_OPERAND0(thellnd),subscripts.thellnd);
  NODE_TYPE(thellnd) = lookForInternalBasetype(SYMB_TYPE(symb->thesymb));
}

 SgProcessorsRefExp::SgProcessorsRefExp(SgExpression &sub1,SgExpression &sub2):SgExpression(PROCESSORS_REF)
{
  SgSymbol *symb;
  
  symb = new SgSymbol(VARIABLE_NAME, "_PROCESSORS");
  NODE_SYMB(thellnd) = symb->thesymb;
  NODE_OPERAND0(thellnd) =  addToExprList(NODE_OPERAND0(thellnd),sub1.thellnd);
  NODE_OPERAND0(thellnd) =  addToExprList(NODE_OPERAND0(thellnd),sub2.thellnd);
  NODE_TYPE(thellnd) = lookForInternalBasetype(SYMB_TYPE(symb->thesymb));
}


 SgProcessorsRefExp::SgProcessorsRefExp(SgExpression &sub1,SgExpression &sub2,SgExpression &sub3):SgExpression(PROCESSORS_REF)
{
  SgSymbol *symb;
  
  symb = new SgSymbol(VARIABLE_NAME, "_PROCESSORS");
  NODE_SYMB(thellnd) = symb->thesymb;
  NODE_OPERAND0(thellnd) =  addToExprList(NODE_OPERAND0(thellnd),sub1.thellnd);
  NODE_OPERAND0(thellnd) =  addToExprList(NODE_OPERAND0(thellnd),sub2.thellnd);
  NODE_OPERAND0(thellnd) =  addToExprList(NODE_OPERAND0(thellnd),sub3.thellnd);
  NODE_TYPE(thellnd) = lookForInternalBasetype(SYMB_TYPE(symb->thesymb));
}

 SgProcessorsRefExp::SgProcessorsRefExp(SgExpression &sub1,SgExpression &sub2,SgExpression &sub3,SgExpression &sub4):SgExpression(PROCESSORS_REF)
{
  SgSymbol *symb;
  
  symb = new SgSymbol(VARIABLE_NAME, "_PROCESSORS");
  NODE_SYMB(thellnd) = symb->thesymb;
  NODE_OPERAND0(thellnd) =  addToExprList(NODE_OPERAND0(thellnd),sub1.thellnd);
  NODE_OPERAND0(thellnd) =  addToExprList(NODE_OPERAND0(thellnd),sub2.thellnd);
  NODE_OPERAND0(thellnd) =  addToExprList(NODE_OPERAND0(thellnd),sub3.thellnd);
  NODE_OPERAND0(thellnd) =  addToExprList(NODE_OPERAND0(thellnd),sub4.thellnd);
  NODE_TYPE(thellnd) = lookForInternalBasetype(SYMB_TYPE(symb->thesymb));
}

 SgProcessorsRefExp::  ~SgProcessorsRefExp()
{ RemoveFromTableLlnd((void *) this); }

// the number of subscripts in reference
 int SgProcessorsRefExp::numberOfSubscripts()
{ return exprListLength(NODE_OPERAND0(thellnd));}

 SgExpression * SgProcessorsRefExp::subscripts()
{ return LlndMapping(NODE_OPERAND0(thellnd)); }

 SgExpression * SgProcessorsRefExp::subscript(int i)
{
  PTR_LLND ll = NULL;
  ll = getPositionInExprList(NODE_OPERAND0(thellnd),i);
  return LlndMapping(ll);
}

 void SgProcessorsRefExp::addSubscript(SgExpression &e)
{ NODE_OPERAND0(thellnd) =  addToExprList(NODE_OPERAND0(thellnd),e.thellnd);}



// SgPntrArrRefExp--inlines

 SgPntrArrRefExp::SgPntrArrRefExp(PTR_LLND ll):SgExpression(ll)
{}

 SgPntrArrRefExp::SgPntrArrRefExp(SgExpression &p):SgExpression(ARRAY_OP)
{ NODE_OPERAND0(thellnd) = p.thellnd; }

 SgPntrArrRefExp::SgPntrArrRefExp(SgExpression &p, SgExpression &subscripts):SgExpression(ARRAY_OP)
{
  NODE_OPERAND0(thellnd) = p.thellnd;
  NODE_OPERAND1(thellnd) =  addToExprList(NODE_OPERAND1(thellnd),subscripts.thellnd);
}

 SgPntrArrRefExp::SgPntrArrRefExp(SgExpression &p, int, SgExpression &sub1, SgExpression &sub2):SgExpression(ARRAY_OP)
{
  NODE_OPERAND0(thellnd) = p.thellnd;
  NODE_OPERAND1(thellnd) =  addToExprList(NODE_OPERAND1(thellnd),sub1.thellnd);
  NODE_OPERAND1(thellnd) =  addToExprList(NODE_OPERAND1(thellnd),sub2.thellnd);
}

 SgPntrArrRefExp::SgPntrArrRefExp(SgExpression &p, int, SgExpression &sub1, SgExpression &sub2, SgExpression &sub3):SgExpression(ARRAY_OP)
{
  NODE_OPERAND0(thellnd) = p.thellnd;
  NODE_OPERAND1(thellnd) =  addToExprList(NODE_OPERAND1(thellnd),sub1.thellnd);
  NODE_OPERAND1(thellnd) =  addToExprList(NODE_OPERAND1(thellnd),sub2.thellnd);
  NODE_OPERAND1(thellnd) =  addToExprList(NODE_OPERAND1(thellnd),sub3.thellnd);
}

 SgPntrArrRefExp::SgPntrArrRefExp(SgExpression &p, int, SgExpression &sub1, SgExpression &sub2, SgExpression &sub3, SgExpression &sub4):SgExpression(ARRAY_OP)
{
  NODE_OPERAND0(thellnd) = p.thellnd;
  NODE_OPERAND1(thellnd) =  addToExprList(NODE_OPERAND1(thellnd),sub1.thellnd);
  NODE_OPERAND1(thellnd) =  addToExprList(NODE_OPERAND1(thellnd),sub2.thellnd);
  NODE_OPERAND1(thellnd) =  addToExprList(NODE_OPERAND1(thellnd),sub3.thellnd);
  NODE_OPERAND1(thellnd) =  addToExprList(NODE_OPERAND1(thellnd),sub4.thellnd);
}

 SgPntrArrRefExp::~SgPntrArrRefExp()
{ RemoveFromTableLlnd((void *) this); }

 int SgPntrArrRefExp::dimension()
{ return exprListLength(NODE_OPERAND1(thellnd)); }

 SgExpression *SgPntrArrRefExp::subscript(int i)
{ return LlndMapping(getPositionInExprList(NODE_OPERAND1(thellnd),i)); }

 void SgPntrArrRefExp::addSubscript(SgExpression &e)
{ NODE_OPERAND1(thellnd) =  addToExprList(NODE_OPERAND1(thellnd),e.thellnd); }

 void SgPntrArrRefExp::setPointer(SgExpression &p)
{ NODE_OPERAND0(thellnd) = p.thellnd; }


// SgPointerDerefExp--inlines

 SgPointerDerefExp::SgPointerDerefExp(PTR_LLND ll):SgExpression(ll)
{}

 SgPointerDerefExp::SgPointerDerefExp(SgExpression &pointerExp):SgExpression(DEREF_OP)
{
  PTR_TYPE expType;
  
  expType = NODE_TYPE(pointerExp.thellnd);
  if (!pointerType(expType))
    Message("Attempt to create SgPointerDerefExp with non pointer type",0);

  NODE_OPERAND0(thellnd) = pointerExp.thellnd;
  NODE_TYPE(thellnd) = lookForInternalBasetype(expType);
}

 SgPointerDerefExp::~SgPointerDerefExp()
{ RemoveFromTableLlnd((void *) this);}


 SgExpression * SgPointerDerefExp::pointerExp()
{ return LlndMapping(NODE_OPERAND0(thellnd)); }


// SgRecprdRefExp--inlines

 SgRecordRefExp::SgRecordRefExp(PTR_LLND ll):SgExpression(ll)
{}

 SgRecordRefExp::SgRecordRefExp(SgSymbol &recordName, char *fieldName):SgExpression(RECORD_REF)
{ 
  PTR_SYMB recordSym, fieldSym;
  
  recordSym = recordName.thesymb;

  if ((fieldSym = getFieldOfStructWithName(fieldName, SYMB_TYPE(recordSym))) == SMNULL)
    Message("No such field",0);

  NODE_OPERAND0(thellnd) = newExpr(VAR_REF,SYMB_TYPE(recordName.thesymb), recordName.thesymb);
  NODE_OPERAND1(thellnd) = newExpr(VAR_REF,SYMB_TYPE(fieldSym), fieldSym);
  NODE_TYPE(thellnd) = SYMB_TYPE(fieldSym);
}

 SgRecordRefExp::SgRecordRefExp(SgExpression &recordExp, char *fieldName):SgExpression(RECORD_REF)
{
  PTR_SYMB  fieldSym;

            
  if ((fieldSym = getFieldOfStructWithName(fieldName, NODE_TYPE(recordExp.thellnd))) == SMNULL)
    Message("No such field",0);
  
  NODE_OPERAND0(thellnd) = recordExp.thellnd;
  NODE_OPERAND1(thellnd) = newExpr(VAR_REF,SYMB_TYPE(fieldSym),fieldSym);
  NODE_TYPE(thellnd) = SYMB_TYPE(fieldSym);
}

 SgRecordRefExp::~SgRecordRefExp(){RemoveFromTableLlnd((void *) this);}

 SgSymbol * SgRecordRefExp::fieldName()
{ return SymbMapping(NODE_SYMB(NODE_OPERAND1(thellnd))); }

 SgSymbol * SgRecordRefExp::recordName()
{
  SgSymbol *x;

  if (NODE_CODE(NODE_OPERAND0(thellnd)) != VAR_REF)
    x = NULL;
  else 
    x = SymbMapping(NODE_SYMB(NODE_OPERAND0(thellnd)));
  
  return x;
}

 SgExpression * SgRecordRefExp::record()
{ return LlndMapping(NODE_OPERAND0(thellnd)); }



// SgStructConstExp--inlines

 SgStructConstExp::SgStructConstExp(PTR_LLND ll):SgExpression(ll)
{}

 SgStructConstExp::SgStructConstExp(SgSymbol &structName, SgExpression &values):SgExpression(STRUCTURE_CONSTRUCTOR)
{
  NODE_OPERAND0(thellnd) = newExpr(TYPE_REF,SYMB_TYPE(structName.thesymb),structName.thesymb);
  NODE_OPERAND1(thellnd) = values.thellnd;
  NODE_TYPE(thellnd) = SYMB_TYPE(structName.thesymb);
}

 SgStructConstExp::SgStructConstExp(SgExpression  &typeRef, SgExpression &values):SgExpression(STRUCTURE_CONSTRUCTOR)
{
  NODE_OPERAND0(thellnd) = typeRef.thellnd;
  NODE_OPERAND1(thellnd) = values.thellnd;
  NODE_TYPE(thellnd) = NODE_TYPE(typeRef.thellnd);
}

 SgStructConstExp::~SgStructConstExp()
{ RemoveFromTableLlnd((void *) this); }

 int SgStructConstExp::numberOfArgs()
{ return exprListLength(NODE_OPERAND1(thellnd)); }
 
 SgExpression * SgStructConstExp::arg(int i)
{ return LlndMapping(getPositionInExprList(NODE_OPERAND1(thellnd),i)); }


// SgConstExp--inlines

  SgConstExp::SgConstExp(PTR_LLND ll):SgExpression(ll)
{}

// NODE_ TYPE needs to be filled here.
// type-checking of values needs to be done.
  SgConstExp::SgConstExp(SgExpression &values):SgExpression(CONSTRUCTOR_REF)
{
  NODE_OPERAND0(thellnd) = values.thellnd;
}

 SgConstExp::~SgConstExp(){RemoveFromTableLlnd((void *) this);}

 int SgConstExp::numberOfArgs()
{ return exprListLength(NODE_OPERAND1(thellnd)); }

 SgExpression * SgConstExp::arg(int i)
{ return LlndMapping(getPositionInExprList(NODE_OPERAND1(thellnd),i)); }



// SgVecConstExp--inlines

 SgVecConstExp::SgVecConstExp(PTR_LLND ll):SgExpression(ll)
{}

#ifdef NOT_YET_IMPLEMENTED
 SgVecConstExp::SgVecConstExp(SgExpression &expr_list):SgExpression(VECTOR_CONST)
{ SORRY; }
#endif

#ifdef NOT_YET_IMPLEMENTED
 SgVecConstExp::SgVecConstExp(int n, SgExpression *components):SgExpression(VECTOR_CONST)
{ SORRY; }
#endif

 SgVecConstExp::~SgVecConstExp()
{ RemoveFromTableLlnd((void *) this); }
        
#ifdef NOT_YET_IMPLEMENTED
 SgExpression * SgVecConstExp::arg(int i)
{ 
  SORRY;
  return (SgExpression *) NULL;
}
#endif

#ifdef NOT_YET_IMPLEMENTED
 int SgVecConstExp::numberOfArgs()
{
  SORRY;
  return 0;
}
#endif

#ifdef NOT_YET_IMPLEMENTED
 void SgVecConstExp::setArg(int i, SgExpression &e)
{
  SORRY;
}
#endif



// SgInitListExp--inlines

 SgInitListExp::SgInitListExp(PTR_LLND ll):SgExpression(ll)
{}

 SgInitListExp::SgInitListExp(SgExpression &expr_list):SgExpression(INIT_LIST)
{
  NODE_OPERAND0(thellnd)=expr_list.thellnd;
  NODE_TYPE(thellnd)=NODE_TYPE(expr_list.thellnd);
}

#ifdef NOT_YET_IMPLEMENTED
 SgInitListExp::SgInitListExp(int n, SgExpression *components):SgExpression(INIT_LIST)
{
  SORRY;
}
#endif

 SgInitListExp::~SgInitListExp()
{ RemoveFromTableLlnd((void *) this); }

        
#ifdef NOT_YET_IMPLEMENTED
 SgExpression * SgInitListExp::arg(int i)
{
  SORRY;
  return (SgExpression *) NULL;
}
#endif

#ifdef NOT_YET_IMPLEMENTED
 int SgInitListExp::numberOfArgs()
{
  SORRY;
  return 0;
}
#endif

#ifdef NOT_YET_IMPLEMENTED
 void SgInitListExp::setArg(int i, SgExpression &e)
{
  SORRY;
}
#endif


// SgObjectListExp--inlines

 SgObjectListExp::SgObjectListExp(PTR_LLND ll):SgExpression(ll)
{}

 SgObjectListExp::SgObjectListExp(int variant, SgSymbol &object, SgExpression &list):SgExpression(variant)
{
#ifdef AJM_SUGGESTS

// This is not what is expected in a COMMON block.
//  NODE_OPERAND0(thellnd) = newExpr(VAR_REF, SYMB_TYPE(object.thesymb), object.thesymb);
  NODE_SYMB(thellnd) = object.thesymb;
  NODE_OPERAND0(thellnd) = list.thellnd;

#else /* Original */

  NODE_OPERAND0(thellnd) = newExpr(VAR_REF, SYMB_TYPE(object.thesymb), object.thesymb);
  NODE_OPERAND1(thellnd) = list.thellnd;

#endif
}

 SgObjectListExp::SgObjectListExp(int variant,SgExpression &objectRef, SgExpression &list):SgExpression(variant)
{
#ifdef AJM_SUGGESTS
// Not what a common block wants.
//  NODE_OPERAND0(thellnd) = objectRef.thellnd;
  NODE_SYMB(thellnd)=objectRef.symbol()->thesymb;
  NODE_OPERAND0(thellnd) = list.thellnd;
#else
  NODE_OPERAND0(thellnd) = objectRef.thellnd;
  NODE_OPERAND1(thellnd) = list.thellnd;
#endif
}

 SgObjectListExp::~SgObjectListExp()
{ RemoveFromTableLlnd((void *) this); }
        
 SgSymbol * SgObjectListExp::object()
{ return SymbMapping(NODE_SYMB(NODE_OPERAND0(thellnd))); }

 SgExpression * SgObjectListExp::objectRef()
{ return LlndMapping(NODE_OPERAND0(thellnd)); }

 int SgObjectListExp::listLength()
{ return exprListLength(NODE_OPERAND1(thellnd)); }

 SgSymbol * SgObjectListExp::symbol(int i)         //i'th symbol
{ 
  PTR_LLND pt;
  SgSymbol *x;

  if ( (pt = getPositionInExprList(NODE_OPERAND1(thellnd),i)) &&
       NODE_OPERAND0(pt))
    x = SymbMapping(NODE_SYMB(NODE_OPERAND0(pt))); 
  else
    x = NULL;
  return x;
}

 SgExpression * SgObjectListExp::objectRef(int i)  //i'th symbol
{ 
  PTR_LLND pt;
  SgExpression *x;

  if ( (pt = getPositionInExprList(NODE_OPERAND1(thellnd),i)) &&
       NODE_OPERAND0(pt))
    x = LlndMapping(NODE_OPERAND0(pt)); 
  else
    x = NULL;

  return x;
}


// SgAttributeExp--inlines
 SgAttributeExp::SgAttributeExp(PTR_LLND ll):SgExpression(ll)
{}

 SgAttributeExp::SgAttributeExp(int variant):SgExpression(variant)
{}

 SgAttributeExp::~SgAttributeExp()
{ RemoveFromTableLlnd((void *) this); }


// SgKeywordArgExp--inlines

 SgKeywordArgExp::SgKeywordArgExp(PTR_LLND ll):SgExpression(ll)
{}

 SgKeywordArgExp::SgKeywordArgExp(char *argName, SgExpression &exp):SgExpression(KEYWORD_ARG)
{ 
  NODE_OPERAND0(thellnd) = exp.thellnd;
  NODE_OPERAND1(thellnd) = SgKeywordValExp(argName).thellnd;
  NODE_TYPE(thellnd) = NODE_TYPE(exp.thellnd);
}

 SgKeywordArgExp::~SgKeywordArgExp()
{ RemoveFromTableLlnd((void *) this); }
  
 SgSymbol * SgKeywordArgExp::arg()
{ return SymbMapping(NODE_SYMB(thellnd)); }

 SgExpression * SgKeywordArgExp::value()
{ return LlndMapping(NODE_OPERAND0(thellnd)); }


// SgSubscriptExp--inlines

 SgSubscriptExp::SgSubscriptExp(PTR_LLND ll):SgExpression(ll)
{}

// add by zheng.gb
 SgSubscriptExp::SgSubscriptExp(SgExpression &exp):SgExpression(exp.thellnd)
{}

 SgSubscriptExp::SgSubscriptExp(SgExpression &lbound, SgExpression &ubound, SgExpression &step):SgExpression(DDOT)
{
  PTR_LLND lb, ub, inc;
  
  lb = lbound.thellnd; ub = ubound.thellnd; inc = step.thellnd;
// modified by zheng.gb
  if (!isIntegerType(lb->type) && !isIntegerType(ub->type) && !isIntegerType(inc->type))
    Message("Non integer type for SgSubscriptExp",0);
  
/*               modified by zheng.gb
  NODE_OPERAND0(thellnd) = lbound.thellnd;
  NODE_OPERAND1(thellnd) = newExpr(DDOT,NULL,ubound.thellnd, step.thellnd);
*/
  NODE_OPERAND0(thellnd)=newExpr(RANGE_OP,NULL,lbound.thellnd, ubound.thellnd);
  NODE_OPERAND1(thellnd)=step.thellnd;
}

 SgSubscriptExp::SgSubscriptExp(SgExpression &lbound, SgExpression &ubound):SgExpression(DDOT)
{
  PTR_LLND lb, ub;
  
  lb = lbound.thellnd; ub = ubound.thellnd;
// modified by zheng.gb
  if (!isIntegerType(lb->type) && !isIntegerType(ub->type))
    Message("Non integer type for SgSubscriptExp",0);
  
//  NODE_OPERAND0(thellnd) = lbound.thellnd;
//  NODE_OPERAND1(thellnd) =  ubound.thellnd;

  NODE_OPERAND0(thellnd)=newExpr(RANGE_OP,NULL,lbound.thellnd, ubound.thellnd);
}

 SgSubscriptExp:: ~SgSubscriptExp()
{ RemoveFromTableLlnd((void *) this);}

// SgIndexOpExp--inlines                      add by zheng.gb

 SgIndexOpExp::SgIndexOpExp(SgVarRefExp &var, SgSubscriptExp &bound):SgExpression(INDEX_OP)
{
  NODE_OPERAND0(thellnd) = var.thellnd;
  NODE_OPERAND1(thellnd) = bound.thellnd;
}

 SgIndexOpExp:: ~SgIndexOpExp()
{ RemoveFromTableLlnd((void *) this);}

// SgUseOnlyExp--inlines

 SgUseOnlyExp::SgUseOnlyExp(PTR_LLND ll):SgExpression(ll)
{}

 SgUseOnlyExp::SgUseOnlyExp(SgExpression &onlyList):SgExpression(ONLY_NODE)
{ NODE_OPERAND0(thellnd) = onlyList.thellnd; }

 SgUseOnlyExp::~SgUseOnlyExp()
{ RemoveFromTableLlnd((void *) this); }

 SgExpression * SgUseOnlyExp::onlyList()
{ return LlndMapping(NODE_OPERAND0(thellnd)); }


 SgUseRenameExp::SgUseRenameExp(PTR_LLND ll):SgExpression(ll)
{}

#ifdef NOT_YET_IMPLEMENTED
 SgUseRenameExp::SgUseRenameExp(SgSymbol &newName, SgSymbol &oldName):SgExpression( RENAME_NODE)
{ SORRY; }
#endif

 SgUseRenameExp::~SgUseRenameExp()
{ RemoveFromTableLlnd((void *) this); }


#ifdef NOT_YET_IMPLEMENTED
 SgSymbol *SgUseRenameExp::newName()
{
  SORRY;
  return (SgSymbol *) NULL;
}
#endif

#ifdef NOT_YET_IMPLEMENTED
 SgSymbol *SgUseRenameExp::oldName()
{
  SORRY;
  return (SgSymbol *) NULL;
}
#endif

#ifdef NOT_YET_IMPLEMENTED
 SgExpression * SgUseRenameExp::newNameExp()
{
  SORRY;
  return (SgExpression *) NULL;
}
#endif

#ifdef NOT_YET_IMPLEMENTED
 SgExpression * SgUseRenameExp::oldNameExp()
{
  SORRY;
  return (SgExpression *) NULL;
}
#endif


// SgSpecPairExp--inlines

 SgSpecPairExp::SgSpecPairExp(PTR_LLND ll):SgExpression(ll)
{}

 SgSpecPairExp::SgSpecPairExp(SgExpression &arg, SgExpression &value):SgExpression(SPEC_PAIR)
{
  NODE_OPERAND0(thellnd) = arg.thellnd;
  NODE_OPERAND1(thellnd) = value.thellnd;
}

 SgSpecPairExp::SgSpecPairExp(SgExpression &arg):SgExpression(SPEC_PAIR)
{ NODE_OPERAND0(thellnd) = arg.thellnd; }

 SgSpecPairExp::SgSpecPairExp(char *arg, char *):SgExpression(SPEC_PAIR)
{
  NODE_OPERAND0(thellnd) = SgKeywordValExp(arg).thellnd;
  NODE_OPERAND1(thellnd) = SgKeywordValExp(arg).thellnd;
}

 SgSpecPairExp::~SgSpecPairExp()
{ RemoveFromTableLlnd((void *) this); }

 SgExpression *SgSpecPairExp::arg()
{ return LlndMapping(NODE_OPERAND0(thellnd)); }

 SgExpression * SgSpecPairExp::value()
{ return LlndMapping(NODE_OPERAND1(thellnd)); }


// SgIOAccessExp--inlines

 SgIOAccessExp::SgIOAccessExp(PTR_LLND ll):SgExpression(ll)
{}

// type-checking on bounds needs to be done.
// Float values are legal in some cases. check manual.
 SgIOAccessExp::SgIOAccessExp(SgSymbol &s, SgExpression lbound, SgExpression ubound, SgExpression step):SgExpression(IOACCESS)
{ 
  NODE_SYMB(thellnd) = s.thesymb;
  NODE_OPERAND0(thellnd) = newExpr(SEQ,NULL, newExpr(DDOT,NULL, lbound.thellnd, ubound.thellnd), step.thellnd);
}

 SgIOAccessExp::SgIOAccessExp(SgSymbol &s, SgExpression lbound, SgExpression ubound):SgExpression(IOACCESS)
{
  NODE_SYMB(thellnd) = s.thesymb;
  NODE_OPERAND0(thellnd) = newExpr(SEQ,NULL, newExpr(DDOT,NULL, lbound.thellnd, ubound.thellnd), NULL);
}

 SgIOAccessExp::~SgIOAccessExp()
{ RemoveFromTableLlnd((void *) this); }


// SgImplicitTypExp--inlines

 SgImplicitTypeExp::SgImplicitTypeExp(PTR_LLND ll):SgExpression(ll)
{}

 SgImplicitTypeExp::SgImplicitTypeExp(SgType &type, SgExpression &rangeList):SgExpression(IMPL_TYPE)
{
  NODE_TYPE(thellnd) = type.thetype;
  NODE_OPERAND0(thellnd) = rangeList.thellnd;
}

 SgImplicitTypeExp::~SgImplicitTypeExp()
{ RemoveFromTableLlnd((void *) this);}

 SgType * SgImplicitTypeExp::type()
{ return TypeMapping(NODE_TYPE(thellnd)); }

 SgExpression * SgImplicitTypeExp::rangeList()
{ return LlndMapping(NODE_OPERAND0(thellnd)); }

#ifdef NOT_YET_IMPLEMENTED
 char * SgImplicitTypeExp::alphabeticRange()
{
  SORRY;
  return (char *) NULL;
}
#endif


// SgTypeExp--inlines

 SgTypeExp::SgTypeExp(PTR_LLND ll):SgExpression(ll)
{}

 SgTypeExp::SgTypeExp(SgType &type):SgExpression(TYPE_OP)
{ NODE_TYPE(thellnd) = type.thetype; }

 SgTypeExp::~SgTypeExp()
{ RemoveFromTableLlnd((void *) this);}

 SgType * SgTypeExp::type()
{ return TypeMapping( NODE_TYPE(thellnd)); }


// SgSeqExp--inlines

 SgSeqExp::SgSeqExp(PTR_LLND ll):SgExpression(ll)
{}

 SgSeqExp::SgSeqExp(SgExpression &exp1, SgExpression &exp2):SgExpression(SEQ)
{
  NODE_OPERAND0(thellnd) = exp1.thellnd;
  NODE_OPERAND1(thellnd) = exp2.thellnd;
}

 SgSeqExp::~SgSeqExp()
{ RemoveFromTableLlnd((void *) this);}

 SgExpression * SgSeqExp::front()
{ return LlndMapping(NODE_OPERAND0(thellnd)); }

 SgExpression * SgSeqExp::rear()
{ return LlndMapping(NODE_OPERAND1(thellnd)); }



// SgStringLengthExp--inlines

 SgStringLengthExp::SgStringLengthExp(PTR_LLND ll):SgExpression(ll)
{}

 SgStringLengthExp::SgStringLengthExp(SgExpression &length):SgExpression(LEN_OP)
{ NODE_OPERAND0(thellnd) = length.thellnd; }

 SgStringLengthExp::~SgStringLengthExp()
{ RemoveFromTableLlnd((void *) this);}

 SgExpression * SgStringLengthExp::length()
{ return LlndMapping(NODE_OPERAND0(thellnd)); }



// SgDefaultExp--inlines

 SgDefaultExp::SgDefaultExp(PTR_LLND ll):SgExpression(ll)
{}

 SgDefaultExp::SgDefaultExp():SgExpression(DEFAULT) 
{}

 SgDefaultExp::~SgDefaultExp()
{ RemoveFromTableLlnd((void *) this); }


// SgLabelRefExp--inlines

 SgLabelRefExp::SgLabelRefExp(PTR_LLND ll):SgExpression(ll)
{}

 SgLabelRefExp::SgLabelRefExp(SgLabel &label):SgExpression(LABEL_REF)
{ NODE_LABEL(thellnd) = label.thelabel; }

 SgLabelRefExp::~SgLabelRefExp()
{ RemoveFromTableLlnd((void *) this); }

 SgLabel * SgLabelRefExp::label()
{ return LabelMapping(NODE_LABEL(thellnd)); }


// SgProgHedrStmt--inlines


 SgProgHedrStmt::SgProgHedrStmt(PTR_BFND bif):SgStatement(bif)
{}

 SgProgHedrStmt::SgProgHedrStmt(int variant):SgStatement(variant)
{ addControlEndToStmt(thebif); }

 SgProgHedrStmt::SgProgHedrStmt(SgSymbol &name, SgStatement &Body):SgStatement(PROG_HEDR)
{
  BIF_SYMB(thebif) = name.thesymb;
  insertBfndListIn(Body.thebif,thebif,thebif);
  addControlEndToStmt(thebif);
}

 SgProgHedrStmt::SgProgHedrStmt(SgSymbol &name):SgStatement(PROG_HEDR)
{
  BIF_SYMB(thebif) = name.thesymb;
  addControlEndToStmt(thebif);
}

 SgProgHedrStmt::SgProgHedrStmt(char *name):SgStatement(PROG_HEDR)
{
  SgSymbol *proc;
  proc = new SgSymbol(PROGRAM_NAME, name); 
  SYMB_SCOPE(proc->thesymb) = PROJ_FIRST_BIF();
  SYMB_TYPE(proc->thesymb) = GetAtomicType(DEFAULT);
  BIF_SYMB(thebif) = proc->thesymb;
  addControlEndToStmt(thebif);
}

 SgSymbol & SgProgHedrStmt::name()
{
  PTR_SYMB symb;
  SgSymbol *pt = NULL;
  symb = BIF_SYMB(thebif);
  if (!symb)
      Message("The bif has no symbol",0);
  else 
    {
      pt = GetMappingInTableForSymbol(symb);
      if (!pt)
        pt = new SgSymbol(symb);      
    }
  return *pt;
}     
   
 void SgProgHedrStmt::setName(SgSymbol &symbol)
{ BIF_SYMB(thebif) = symbol.thesymb; }

#ifdef NOT_YET_IMPLEMENTED
 int SgProgHedrStmt::numberOfFunctionsCalled()
{
  SORRY;
  return 0;
}
#endif

#ifdef NOT_YET_IMPLEMENTED
 SgSymbol * SgProgHedrStmt::calledFunction(int i)
{
  SORRY;
  return (SgSymbol *) NULL;
}
#endif

 int SgProgHedrStmt::numberOfStmtFunctions()
{ return countInStmtNode1(thebif, STMTFN_STAT); }

 SgStatement * SgProgHedrStmt::statementFunc(int i)
{ return BfndMapping(GetcountInStmtNode1(thebif, STMTFN_STAT, i)); }

 int SgProgHedrStmt::numberOfEntryPoints()
{ return countInStmtNode1(thebif, ENTRY_STAT); }

 SgStatement * SgProgHedrStmt::entryPoint(int i)
{ return BfndMapping(GetcountInStmtNode1(thebif, ENTRY_STAT, i)); }

 int SgProgHedrStmt::numberOfParameters()
{ return lenghtOfParamList(BIF_SYMB(thebif)); }

 SgSymbol * SgProgHedrStmt::parameter(int i)
{
  PTR_SYMB symb;
  symb = GetThParam(BIF_SYMB(thebif),i);
  return SymbMapping(symb);
}

void SgProgHedrStmt::insertParameterAfter(SgSymbol & parameter, SgSymbol & added)
{
  PTR_SYMB temp = SYMB_NEXT_DECL(parameter.thesymb);
  SYMB_NEXT_DECL(parameter.thesymb) = added.thesymb;
  SYMB_NEXT_DECL(added.thesymb) = temp;
  // not care control parent here;
  
}
        
#ifdef NOT_YET_IMPLEMENTED
 int SgProgHedrStmt::numberOfSpecificationStmts()
{
  SORRY;
  return 0;
}
#endif

#ifdef NOT_YET_IMPLEMENTED
 int SgProgHedrStmt::numberOfExecutionStmts()
{
  SORRY;
  return 0;
}
#endif

#ifdef NOT_YET_IMPLEMENTED
 SgStatement * SgProgHedrStmt::specificationStmt(int i)
{
  SORRY;
  return (SgStatement *) NULL;
}
#endif

#ifdef NOT_YET_IMPLEMENTED
 SgStatement * SgProgHedrStmt::executionStmt(int i)
{
  SORRY;
  return (SgStatement *) NULL;
}
#endif
 
 int SgProgHedrStmt::numberOfInternalFunctionsDefined()
{ return countInStmtNode1(thebif, FUNC_HEDR); }

 int SgProgHedrStmt::numberOfInternalSubroutinesDefined()
{ return countInStmtNode1(thebif, PROC_HEDR); }

 int SgProgHedrStmt::numberOfInternalSubProgramsDefined()
{
  return (countInStmtNode1(thebif, FUNC_HEDR) +
          countInStmtNode1(thebif, PROC_HEDR)) ;
}

#ifdef NOT_YET_IMPLEMENTED
 SgStatement * SgProgHedrStmt::internalSubProgram(int i)
{
  SORRY;
  return (SgStatement *) NULL;
}
#endif

 SgStatement * SgProgHedrStmt::internalFunction(int i)
{ return BfndMapping(GetcountInStmtNode1(thebif, FUNC_HEDR, i)); }

 SgStatement * SgProgHedrStmt::internalSubroutine(int i)
{ return BfndMapping(GetcountInStmtNode1(thebif, PROC_HEDR, i)); }


#ifdef NOT_YET_IMPLEMENTED
SgSymbol &addVariable(SgType &T, char *name)
{
  SORRY;
}
#endif

#ifdef NOT_YET_IMPLEMENTED
//add a declaration for new variable
SgStatement &addCommonBlock(char *blockname, int noOfVars,
                            SgSymbol *Vars)
{
  SORRY;
}
#endif
        
#ifdef NOT_YET_IMPLEMENTED
 int SgProgHedrStmt::isSymbolInScope(SgSymbol &symbol)
{
  SORRY;
  return 0;
}
#endif

#ifdef NOT_YET_IMPLEMENTED
 int SgProgHedrStmt::isSymbolDeclaredHere(SgSymbol &symbol)
{
  SORRY;
  return 0;
}
#endif

SgSymbol * SgProgHedrStmt::firstSymbolInScope()
{
  SgSymbol * symb = this->symbol();
   
  while ((symb->next()!=NULL)) {
    symb=symb->next();
    if (symb->scope()==this)
      break;
  }

  return symb;
}

// global analysis data

#ifdef NOT_YET_IMPLEMENTED
 int SgProgHedrStmt::numberOfVarsUsed()
{
  SORRY;
  return 0;
}
#endif

#ifdef NOT_YET_IMPLEMENTED
 SgExpression * SgProgHedrStmt::varsUsed(int i)
{
  SORRY;
  return (SgExpression *) NULL;
}
#endif

#ifdef NOT_YET_IMPLEMENTED
 int SgProgHedrStmt::numberofVarsMod()
{
  SORRY;
  return 0;
}
#endif

#ifdef NOT_YET_IMPLEMENTED
 SgExpression *varsMod(int i)
{
  SORRY;
  return (SgExpression *) NULL;
}
#endif

 SgProgHedrStmt::~SgProgHedrStmt()
{ RemoveFromTableBfnd((void *) this); }



// SgProcHedrStmt--inlines

 SgProcHedrStmt::SgProcHedrStmt(int variant):SgProgHedrStmt(variant)
{ }

 SgProcHedrStmt::SgProcHedrStmt(SgSymbol &name, SgStatement &Body):SgProgHedrStmt(PROC_HEDR)
{
  BIF_SYMB(thebif) = name.thesymb;
    if(LibClanguage()){
        printf("SgProcHedrStmt: not a valid C construct. use FuncHedr\n");
	}
  name.thesymb->entry.proc_decl.proc_hedr = thebif;
  insertBfndListIn(Body.thebif,thebif,thebif);
}

 SgProcHedrStmt::SgProcHedrStmt(SgSymbol &name):SgProgHedrStmt(PROC_HEDR)
{ BIF_SYMB(thebif) = name.thesymb;  
  name.thesymb->entry.proc_decl.proc_hedr = thebif;
    if(LibClanguage()){
        printf("SgProcHedrStmt: not a valid C construct. use FuncHedr\n");
	}
}

 SgProcHedrStmt::SgProcHedrStmt(char *name):SgProgHedrStmt(PROC_HEDR)
{
  SgSymbol *proc;
  proc = new SgSymbol(PROCEDURE_NAME, name); 
  SYMB_SCOPE(proc->thesymb) = PROJ_FIRST_BIF();
  SYMB_TYPE(proc->thesymb) = GetAtomicType(DEFAULT);
  BIF_SYMB(thebif) = proc->thesymb;
  proc->thesymb->entry.proc_decl.proc_hedr = thebif;
    if(LibClanguage()){
        printf("SgProcHedrStmt: not a valid C construct. use FuncHedr\n");
	}
        
}

 void SgProcHedrStmt::AddArg(SgExpression &arg)
{
  PTR_SYMB symb;
  PTR_LLND ll;

  if(LibFortranlanguage())
 	BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif),arg.thellnd);
  else{ 
        ll = BIF_LL1(thebif);
        ll = NODE_OPERAND0(ll);
	NODE_OPERAND0(ll) = addToExprList(NODE_OPERAND0(ll),arg.thellnd);
      }
  ll = giveLlSymbInDeclList(arg.thellnd);
  if (ll && (symb= NODE_SYMB(ll)))
    {
      appendSymbToArgList(BIF_SYMB(thebif),symb); 
      SYMB_SCOPE(symb) = thebif;
      if(LibFortranlanguage())
            declareAVar(symb,thebif);
    } 
  else
    Message("bad symbol in SgProcHedrStmt::AddArg",0);
}


#ifdef NOT_YET_IMPLEMENTED
 int SgProcHedrStmt::isRecursive()  // 1 if recursive.
{ 
  SORRY;
  return 0;
  //return isAttributeSet(BIF_SYMB(thebif), RECURSIVE_BIT); 
}
#endif

 int SgProcHedrStmt::numberOfEntryPoints()
{ return countInStmtNode1(thebif,ENTRY_STAT); }

 SgStatement * SgProcHedrStmt::entryPoint(int i)
{ return BfndMapping(GetcountInStmtNode1(thebif,ENTRY_STAT,i)); }

// this is incorrect. Takes only subroutines calls into account.
// Should be modified to take function calls into account too.
 int SgProcHedrStmt::numberOfCalls()
{ return countInStmtNode1(thebif,PROC_STAT); }

 SgStatement * SgProcHedrStmt::call(int i)
{ return BfndMapping(GetcountInStmtNode1(thebif,PROC_STAT,i)); }

 SgProcHedrStmt::~SgProcHedrStmt()
{ RemoveFromTableBfnd((void *) this); }



// SgProsHedrStmt--inlines

 SgProsHedrStmt::SgProsHedrStmt():SgProgHedrStmt(PROS_HEDR)
{}

 SgProsHedrStmt::SgProsHedrStmt(SgSymbol &name, SgStatement &Body)
                      :SgProgHedrStmt(PROS_HEDR)
{
  BIF_SYMB(thebif) = name.thesymb;
  insertBfndListIn(Body.thebif,thebif,thebif);
}

 SgProsHedrStmt::SgProsHedrStmt(SgSymbol &name):SgProgHedrStmt(PROS_HEDR)
{ BIF_SYMB(thebif) = name.thesymb; }

 SgProsHedrStmt::SgProsHedrStmt(char *name):SgProgHedrStmt(PROS_HEDR)
{
  SgSymbol *pros;
  pros = new SgSymbol(PROCESS_NAME, name); 
  SYMB_SCOPE(pros->thesymb) = PROJ_FIRST_BIF();
  SYMB_TYPE(pros->thesymb) = GetAtomicType(DEFAULT);
  BIF_SYMB(thebif) = pros->thesymb;
}

 void SgProsHedrStmt::AddArg(SgExpression &arg)
{
  PTR_SYMB symb;
  PTR_LLND ll;

  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif),arg.thellnd);
  ll = giveLlSymbInDeclList(arg.thellnd);
  if (ll && (symb= NODE_SYMB(ll)))
    {
      appendSymbToArgList(BIF_SYMB(thebif),symb); 
      SYMB_SCOPE(symb) = thebif;
      declareAVar(symb,thebif);
    } 
  else
    Message("Pb in SgProsHedrStmt::AddArg",0);
}

 int SgProsHedrStmt::numberOfCalls()
{ return countInStmtNode1(thebif,PROS_STAT); }

 SgStatement * SgProsHedrStmt::call(int i)
{ return BfndMapping(GetcountInStmtNode1(thebif,PROS_STAT,i)); }

 SgProsHedrStmt::~SgProsHedrStmt()
{ RemoveFromTableBfnd((void *) this); }



// SgFuncHedrStmt--inlines
 SgFuncHedrStmt::SgFuncHedrStmt(SgSymbol &name, SgStatement &Body): 
   SgProcHedrStmt(FUNC_HEDR)
{
  BIF_SYMB(thebif) = name.thesymb;
  if(LibClanguage()){
	SgExpression *fref = new SgExpression(FUNCTION_REF);
	fref->setSymbol(name);
  	BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif),fref->thellnd);
	}
  SYMB_FUNC_HEDR(name.thesymb) = thebif;
  insertBfndListIn(Body.thebif,thebif,thebif);
}

 SgFuncHedrStmt::SgFuncHedrStmt(SgSymbol &name, SgType &type, SgStatement &Body): SgProcHedrStmt(FUNC_HEDR)
{
  BIF_SYMB(thebif) = name.thesymb;
  if(LibClanguage()){
	SgExpression *fref = new SgExpression(FUNCTION_REF);
	fref->setSymbol(name);
  	BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif),fref->thellnd);
	}
  SYMB_TYPE(BIF_SYMB(thebif)) = type.thetype;
  SYMB_FUNC_HEDR(name.thesymb) = thebif;
  insertBfndListIn(Body.thebif,thebif,thebif);
}

 SgFuncHedrStmt::SgFuncHedrStmt(SgSymbol &name, SgSymbol &resultName, 
                  SgType &type, SgStatement &Body): SgProcHedrStmt(FUNC_HEDR)
{
  BIF_SYMB(thebif) = name.thesymb;
  if(LibClanguage()){
	SgExpression *fref = new SgExpression(FUNCTION_REF);
	fref->setSymbol(name);
  	BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif),fref->thellnd);
	}
  SYMB_TYPE(BIF_SYMB(thebif)) = type.thetype;
  SYMB_DECLARED_NAME(BIF_SYMB(thebif)) = resultName.thesymb;
  SYMB_FUNC_HEDR(name.thesymb) = thebif;
  insertBfndListIn(Body.thebif,thebif,thebif);
}

 SgFuncHedrStmt::SgFuncHedrStmt(SgSymbol &name): SgProcHedrStmt(FUNC_HEDR)
{ BIF_SYMB(thebif) = name.thesymb; 
  if(LibClanguage()){
	SgExpression *fref = new SgExpression(FUNCTION_REF);
	fref->setSymbol(name);
  	BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif),fref->thellnd);
	}
}

 SgFuncHedrStmt::SgFuncHedrStmt(SgSymbol &name, SgExpression *exp): SgProcHedrStmt(FUNC_HEDR)
{ 
  BIF_SYMB(thebif) = name.thesymb; 
  if (exp)
    BIF_LL1(thebif) = exp->thellnd;
  SYMB_FUNC_HEDR(name.thesymb) = thebif;
}

 SgFuncHedrStmt::SgFuncHedrStmt(char *name): SgProcHedrStmt(FUNC_HEDR)
{
  SgSymbol *proc;
  proc = new SgSymbol(FUNCTION_NAME, name); 
  if(LibClanguage()){
	SgExpression *fref = new SgExpression(FUNCTION_REF);
	fref->setSymbol(*proc);
  	BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif),fref->thellnd);
	}
  SYMB_SCOPE(proc->thesymb) = PROJ_FIRST_BIF();
  SYMB_TYPE(proc->thesymb) = GetAtomicType(T_INT);
  SYMB_FUNC_HEDR(proc->thesymb) = thebif;
  BIF_SYMB(thebif) = proc->thesymb;
}

 SgFuncHedrStmt::~SgFuncHedrStmt()
{ RemoveFromTableBfnd((void *) this); } 

 SgType * SgFuncHedrStmt::returnedType()
{
  PTR_TYPE ty = NULL;
  if (BIF_SYMB(thebif))
    ty = SYMB_TYPE(BIF_SYMB(thebif));
  return TypeMapping(ty);
}

 void SgFuncHedrStmt::setReturnedType(SgType &type)
{
  if (BIF_SYMB(thebif))
    SYMB_TYPE(BIF_SYMB(thebif)) = type.thetype;
}

 SgSymbol * SgFuncHedrStmt::resultName()  // name of result variable.
{
  SgSymbol *x;

  if (BIF_SYMB(thebif) == SMNULL)
    x = NULL;
  else
    x = SymbMapping(SYMB_DECLARED_NAME(BIF_SYMB(thebif)));

  return x;
}

// Use Message to flag error and type it void?
 int  SgFuncHedrStmt::setResultName(SgSymbol &symbol)// set name of result variable.
{
  int x;
  if (BIF_SYMB(thebif) == SMNULL)
    x = 0;
  else {
    SYMB_DECLARED_NAME(BIF_SYMB(thebif)) = symbol.thesymb;
    x = 1;
  }
  return x;
}



// SgClassStmt--inlines

 SgClassStmt::SgClassStmt(int variant):SgStatement(variant)
{}

 SgClassStmt::SgClassStmt(SgSymbol &name):SgStatement(CLASS_DECL)
{ BIF_SYMB(thebif) = name.thesymb; }

 SgClassStmt::~SgClassStmt()
{ RemoveFromTableBfnd((void *) this); }

 int SgClassStmt::numberOfSuperClasses()
{ return exprListLength(BIF_LL2(thebif)); }

 SgSymbol * SgClassStmt::name()
{ return SymbMapping(BIF_SYMB(thebif)); }

 SgSymbol * SgClassStmt::superClass(int i)
{
  PTR_LLND pt;
  SgSymbol *x;

  pt = getPositionInExprList(BIF_LL2(thebif),i);
  pt = giveLlSymbInDeclList(pt);
  if (pt)
    x = SymbMapping(NODE_SYMB(pt));
  else
    x = SymbMapping(NULL);

  return x;
}

 void SgClassStmt::setSuperClass(int i, SgSymbol &symb)
{
  PTR_LLND pt;
  
  if (!BIF_LL2(thebif))
    {
      BIF_LL2(thebif) = addToExprList(BIF_LL2(thebif),newExpr(VAR_REF,NULL,symb.thesymb));
    } 
  else
    {
      pt = getPositionInExprList(BIF_LL2(thebif),i);
      pt = giveLlSymbInDeclList(pt);
      if (pt)
        NODE_SYMB(pt) = symb.thesymb;
      else
        BIF_LL2(thebif) = addToExprList(BIF_LL2(thebif),newExpr(VAR_REF,NULL,symb.thesymb));
    }
}


// SgStructStmt--inlines

 SgStructStmt::SgStructStmt():SgClassStmt(STRUCT_DECL)
{}

 SgStructStmt::SgStructStmt(SgSymbol &name):SgClassStmt(name)
{
  BIF_SYMB(thebif) = name.thesymb;
  BIF_CODE(thebif) = STRUCT_DECL;
}

 SgStructStmt::~SgStructStmt()
{ RemoveFromTableBfnd((void *) this); }



// SgUnionStmt--inlines
// consider like a class.
 SgUnionStmt::SgUnionStmt():SgClassStmt(UNION_DECL)
{}

 SgUnionStmt::SgUnionStmt(SgSymbol &name):SgClassStmt(name)
{
  BIF_SYMB(thebif) = name.thesymb;
  BIF_CODE(thebif) = UNION_DECL;
}

 SgUnionStmt::~SgUnionStmt()
{ RemoveFromTableBfnd((void *) this); }



// SgEnumStmt--inlines
// consider like a class.
 SgEnumStmt::SgEnumStmt():SgClassStmt(ENUM_DECL)
{}

 SgEnumStmt::SgEnumStmt(SgSymbol &name):SgClassStmt(name)
{
  BIF_SYMB(thebif) = name.thesymb;
  BIF_CODE(thebif) = ENUM_DECL;
}

 SgEnumStmt::~SgEnumStmt()
{ RemoveFromTableBfnd((void *) this); }



// SgCollectionStmt--inlines

 SgCollectionStmt::SgCollectionStmt():SgClassStmt(COLLECTION_DECL)
{}

 SgCollectionStmt::SgCollectionStmt(SgSymbol &name):SgClassStmt(name)
{ BIF_CODE(thebif) = COLLECTION_DECL; }

 SgCollectionStmt::~SgCollectionStmt()
{ RemoveFromTableBfnd((void *) this); }

 SgStatement * SgCollectionStmt::firstElementMethod()
{ return BfndMapping(LibfirstElementMethod(thebif)); }


// SgBasicBlockStmt--inlines
 SgBasicBlockStmt::SgBasicBlockStmt(): SgStatement(BASIC_BLOCK)
{}

 SgBasicBlockStmt::~SgBasicBlockStmt()
{ RemoveFromTableBfnd((void *) this); }



// SgForStmt--inlines
 SgForStmt::SgForStmt(SgSymbol &do_var, SgExpression &start, SgExpression &end,
                            SgExpression &step, SgStatement &body):SgStatement(FOR_NODE)
{
  if (CurrentProject->Fortranlanguage())
    {
      BIF_SYMB(thebif) = do_var.thesymb;
      BIF_LL1(thebif) = newExpr(DDOT,NODE_TYPE(end.thellnd),start.thellnd,end.thellnd);
      BIF_LL2(thebif) = step.thellnd;

      SgStatement * enddo= new SgStatement(ENDDO_STMT);
      insertBfndListIn(enddo->thebif,thebif,thebif);

      insertBfndListIn(body.thebif,thebif,thebif);

      addControlEndToStmt(thebif);
    } else
      {
        SORRY;
      }
}

 SgForStmt::SgForStmt(SgSymbol *do_var, SgExpression *start, SgExpression *end,
            SgExpression *step, SgStatement *body):SgStatement(FOR_NODE)
{
  if (CurrentProject->Fortranlanguage())
    {
      if (do_var)
        BIF_SYMB(thebif) = do_var->thesymb;
      if (start && end)
        BIF_LL1(thebif) = newExpr(DDOT,NODE_TYPE(end->thellnd),start->thellnd,end->thellnd);
      if (step)
        BIF_LL2(thebif) = step->thellnd;
      if (body)
        insertBfndListIn(body->thebif,thebif,thebif);
      addControlEndToStmt(thebif);
    } else
      {
        SORRY;
      }
}

 SgForStmt::SgForStmt(SgSymbol &do_var, SgExpression &start, SgExpression &end
                            , SgStatement &body):SgStatement(FOR_NODE)
{
  if (CurrentProject->Fortranlanguage())
    {
      BIF_SYMB(thebif) = do_var.thesymb;
      BIF_LL1(thebif) = newExpr(DDOT,NODE_TYPE(end.thellnd),start.thellnd,end.thellnd);
      BIF_LL2(thebif) = NULL;
      insertBfndListIn(body.thebif,thebif,thebif);
      addControlEndToStmt(thebif);
    } else
      {
        SORRY;
      }
}
// For C Statement;
 SgForStmt::SgForStmt(SgExpression &start, SgExpression &end,
                            SgExpression &step, SgStatement &body):SgStatement(FOR_NODE)
{
  if (CurrentProject->Fortranlanguage())
    {
      BIF_LL1(thebif) = newExpr(DDOT,NODE_TYPE(end.thellnd),start.thellnd,end.thellnd);
      BIF_LL2(thebif) = step.thellnd;
      insertBfndListIn(body.thebif,thebif,thebif);
      addControlEndToStmt(thebif);
    } else
      {
        BIF_LL1(thebif) = start.thellnd;
        BIF_LL2(thebif) = end.thellnd;
        BIF_LL3(thebif) = step.thellnd;
        insertBfndListIn(body.thebif,thebif,thebif);
        addControlEndToStmt(thebif);
      }
}

 void SgForStmt::setDoName(SgSymbol &doName)
{ BIF_SYMB(thebif) = doName.thesymb; }  // sets the name of the loop (for F90.)

 SgSymbol SgForStmt::doName()
{ SgSymbol result(BIF_SYMB(thebif));   // modified by zheng.gb
  return result;
}    // the name of the loop (for F90.)

 SgExpression * SgForStmt::start()
{       
  SgExpression *x;

  if (CurrentProject->Fortranlanguage())
    {
      if ((BIF_LL1(thebif) != LLNULL) &&
          (NODE_CODE(BIF_LL1(thebif)) == DDOT))
         x = LlndMapping(NODE_OPERAND0(BIF_LL1(thebif)));
      else {
        x = NULL;
        SORRY;
      }
    } 
  else
      x = LlndMapping(BIF_LL1(thebif));

  return x;
}

 void SgForStmt::setStart(SgExpression &lbound)
{

  if (CurrentProject->Fortranlanguage())
    {
      if ((BIF_LL1(thebif) != LLNULL) &&
          (NODE_CODE(BIF_LL1(thebif)) == DDOT))
        {
          NODE_OPERAND0(BIF_LL1(thebif)) =  lbound.thellnd;
        }
      else
        {
          BIF_LL1(thebif) = newExpr(DDOT,NODE_TYPE(lbound.thellnd),lbound.thellnd,NULL);
        }
    }
  else 
    {
      BIF_LL1(thebif) = lbound.thellnd;
    }
}

 SgExpression * SgForStmt::end()
{
  SgExpression *x;

  if (CurrentProject->Fortranlanguage())
    {
      if ((BIF_LL1(thebif) != LLNULL) &&
          (NODE_CODE(BIF_LL1(thebif)) == DDOT))
        x = LlndMapping(NODE_OPERAND1(BIF_LL1(thebif)));
      else {
        x = NULL;
        SORRY;
      }
    } 
  else  /* BW, change contributed by Michael Golden */
    {
      if (BIF_LL2(thebif) == LLNULL)
        x = NULL;
      else
        x = LlndMapping(BIF_LL2(thebif));
    }
  return x;
}

 void SgForStmt::setEnd(SgExpression &ubound)
{
  if (CurrentProject->Fortranlanguage())
    {
      if ((BIF_LL1(thebif) != LLNULL) &&
          (NODE_CODE(BIF_LL1(thebif)) == DDOT))
        NODE_OPERAND1(BIF_LL1(thebif)) =  ubound.thellnd;
      else
        {
          BIF_LL1(thebif) = newExpr(DDOT,NODE_TYPE(ubound.thellnd),NULL,ubound.thellnd);
        }
    }
  else 
    {
      BIF_LL2(thebif) = ubound.thellnd;
    }
}


 SgLabel * SgForStmt::endOfLoop()
    { return LabelMapping(BIF_LABEL_USE(thebif)); }

 SgExpression * SgForStmt::step()
{
  SgExpression *x;
  if (CurrentProject->Fortranlanguage())
    {
      x = LlndMapping(BIF_LL2(thebif));
    } 
  else  /* BW, change contributed by Michael Golden */
    {
      if (BIF_LL3(thebif) == LLNULL)
        x =  NULL;
      else
        x = LlndMapping(BIF_LL3(thebif));
    }

  return x;
}

 void SgForStmt::setStep(SgExpression &step)
{
  if (CurrentProject->Fortranlanguage())
    {
      BIF_LL2(thebif) = step.thellnd;
    }
  else
    {
      BIF_LL3(thebif) = step.thellnd;
    }
}

 SgStatement * SgForStmt::body()
{
  PTR_BFND bif =NULL;
  
  if (BIF_BLOB1(thebif)) 
    bif = BLOB_VALUE(BIF_BLOB1(thebif));
            
  return BfndMapping(bif);
}

// s is assumed to terminate with a
//   control end statement.
 void SgForStmt::set_body(SgStatement &s)
{
  BIF_BLOB1(thebif) = NULL;
  insertBfndListIn(s.thebif,thebif,thebif);
}

// False if the loop is not a prefect nest
// else returns size of the loop nest

 int SgForStmt::isPerfectLoopNest()
{ return LibperfectlyNested (thebif); }

// returns inner nested loop
 SgStatement * SgForStmt::getNextLoop()
{ return BfndMapping(LibgetNextNestedLoop (thebif)); }

// returns outer nested loop
 SgStatement * SgForStmt::getPreviousLoop()
{ return BfndMapping(LibgetPreviousNestedLoop (thebif));  }

// returns innermost nested loop
 SgStatement * SgForStmt::getInnermostLoop()
{ return BfndMapping(LibgetInnermostLoop (thebif)); }

// TRUE if the loop ends with an Enddo
 int SgForStmt::isEnddoLoop()
{ return LibisEnddoLoop (thebif); }

// Convert the loop into a Good loop.
 int SgForStmt::convertLoop()
{ return convertToEnddoLoop (thebif); }

 SgForStmt::~SgForStmt()
{ RemoveFromTableBfnd((void *) this);}

// SgForallStmt--inlines                      add by zheng.gb
 SgForallStmt::SgForallStmt(SgExpression &forall_header_op ,
            SgStatement &body):SgStatement(FORALL_STMT)
{
      BIF_LL1(thebif) = forall_header_op.thellnd;
      //            add control_end
      insertBfndListIn(body.thebif,thebif,thebif);
      addControlEndToStmt(thebif);
}

 SgForallStmt::SgForallStmt(SgExpression &forall_header_op , SgExpression &mask,
            SgStatement &body):SgStatement(FORALL_STMT)
{
      BIF_LL1(thebif) = forall_header_op.thellnd;
      BIF_LL2(thebif) = mask.thellnd;
      //            add control_end
      insertBfndListIn(body.thebif,thebif,thebif);
      addControlEndToStmt(thebif);
}

 SgForallStmt::~SgForallStmt()
{ RemoveFromTableBfnd((void *) this);}

// SgProcessDoStmt--inlines
 SgProcessDoStmt::SgProcessDoStmt(SgSymbol &do_var, SgExpression &start,
                                        SgExpression &end, SgExpression &step,
                                        SgLabel &endofloop, SgStatement &body)
                       :SgStatement(PROCESS_DO_STAT)
{
  if (CurrentProject->Fortranlanguage())
    {
      BIF_SYMB(thebif) = do_var.thesymb;
      BIF_LL1(thebif) = newExpr(DDOT,NODE_TYPE(start.thellnd),start.thellnd,end.thellnd);
      BIF_LL2(thebif) = step.thellnd;
      BIF_LABEL_USE(thebif) = endofloop.thelabel;
      insertBfndListIn(body.thebif,thebif,thebif);
      addControlEndToStmt(thebif);
    } else
      {
        SORRY;
      }
}

 SgProcessDoStmt::SgProcessDoStmt(SgSymbol &do_var, SgExpression &start,
                                        SgExpression &end, SgLabel &endofloop,
                                        SgStatement &body)
                       :SgStatement(PROCESS_DO_STAT)
{
  if (CurrentProject->Fortranlanguage())
    {
      BIF_SYMB(thebif) = do_var.thesymb;
      BIF_LL1(thebif) = newExpr(DDOT,NODE_TYPE(start.thellnd),start.thellnd,end.
thellnd);
      BIF_LABEL_USE(thebif) = endofloop.thelabel;
      insertBfndListIn(body.thebif,thebif,thebif);
      addControlEndToStmt(thebif);
    } else
      {
        SORRY;
      }
}

 SgProcessDoStmt::SgProcessDoStmt(SgSymbol &do_var, SgExpression &start,
                                        SgExpression &end, SgExpression &step,
                                        SgStatement &body)
                       :SgStatement(PROCESS_DO_STAT)
{
  if (CurrentProject->Fortranlanguage())
    {
      BIF_SYMB(thebif) = do_var.thesymb;
      BIF_LL1(thebif) = newExpr(DDOT,NODE_TYPE(start.thellnd),start.thellnd,end.
thellnd);
      BIF_LL2(thebif) = step.thellnd;
      insertBfndListIn(body.thebif,thebif,thebif);
      addControlEndToStmt(thebif);
    } else
      {
        SORRY;
      }
}

 SgProcessDoStmt::SgProcessDoStmt(SgSymbol &do_var, SgExpression &start,
                                        SgExpression &end, SgStatement &body)
                       :SgStatement(PROCESS_DO_STAT)
{
  if (CurrentProject->Fortranlanguage())
    {
      BIF_SYMB(thebif) = do_var.thesymb;
      BIF_LL1(thebif) = newExpr(DDOT,NODE_TYPE(start.thellnd),start.thellnd,end.
thellnd);
      insertBfndListIn(body.thebif,thebif,thebif);
      addControlEndToStmt(thebif);
    } else
      {
        SORRY;
      }
}


 void SgProcessDoStmt::setDoName(SgSymbol &doName)
{ BIF_SYMB(thebif) = doName.thesymb; }


 SgSymbol SgProcessDoStmt::doName() 
{ SgSymbol result(BIF_SYMB(thebif));          // modified by zheng.gb
  return  result;
}
 

 SgExpression * SgProcessDoStmt::start()
{       
  SgExpression *x;

  if (CurrentProject->Fortranlanguage())
    {
      if ((BIF_LL1(thebif) != LLNULL) &&
          (NODE_CODE(BIF_LL1(thebif)) == DDOT))
         x = LlndMapping(NODE_OPERAND0(BIF_LL1(thebif)));
      else {
        x = NULL;
        SORRY;
      }
    } 
  else {
    x = NULL;
    SORRY;
  }

  return x;
}

 SgExpression * SgProcessDoStmt::end()
{
  SgExpression *x;

  if (CurrentProject->Fortranlanguage())
    {
      if ((BIF_LL1(thebif) != LLNULL) &&
          (NODE_CODE(BIF_LL1(thebif)) == DDOT))
        x = LlndMapping(NODE_OPERAND1(BIF_LL1(thebif)));
      else {
        x = NULL;
        SORRY;
      }
    } 
  else {
    x = NULL;
    SORRY;
    }

  return x;
}

 SgExpression * SgProcessDoStmt::step()
{
  SgExpression *x;
  if (CurrentProject->Fortranlanguage())
    {
      x = LlndMapping(BIF_LL2(thebif));
    } 
  else {
    x = NULL;
    SORRY;
    };

  return x;
}

 SgLabel * SgProcessDoStmt::endOfLoop()
{ return LabelMapping(BIF_LABEL_USE(thebif)); }

 SgStatement * SgProcessDoStmt::body()
{
  PTR_BFND bif =NULL;
  
  if (BIF_BLOB1(thebif)) 
    bif = BLOB_VALUE(BIF_BLOB1(thebif));
            
  return BfndMapping(bif);
}

// s is assumed to terminate with a
//   control end statement.
 void SgProcessDoStmt::set_body(SgStatement &s)
{
  BIF_BLOB1(thebif) = NULL;
  insertBfndListIn(s.thebif,thebif,thebif);
}

// False if the loop is not a prefect nest
// else returns size of the loop nest

 int SgProcessDoStmt::isPerfectLoopNest()
{ return LibperfectlyNested (thebif); }

// returns inner nested loop
 SgStatement * SgProcessDoStmt::getNextLoop()
{ return BfndMapping(LibgetNextNestedLoop (thebif)); }

// returns outer nested loop
 SgStatement * SgProcessDoStmt::getPreviousLoop()
{ return BfndMapping(LibgetPreviousNestedLoop (thebif));  }

// returns innermost nested loop
 SgStatement * SgProcessDoStmt::getInnermostLoop()
{ return BfndMapping(LibgetInnermostLoop (thebif)); }

// TRUE if the loop ends with an Enddo
 int SgProcessDoStmt::isEnddoLoop()
{ return LibisEnddoLoop (thebif); }

// Convert the loop into a Good loop.
 int SgProcessDoStmt::convertLoop()
{ return convertToEnddoLoop (thebif); }

 SgProcessDoStmt::~SgProcessDoStmt()
{ RemoveFromTableBfnd((void *) this);}



// SgWhileStmt--inlines

 SgWhileStmt::SgWhileStmt(int variant):SgStatement(variant)
{}

 SgWhileStmt::SgWhileStmt(SgExpression &cond, SgStatement &body):SgStatement(WHILE_NODE)
{
  BIF_LL1(thebif) = cond.thellnd;
  insertBfndListIn(body.thebif,thebif,thebif);
  addControlEndToStmt(thebif);
}

// the while test
 SgExpression * SgWhileStmt::conditional()
{ return LlndMapping(BIF_LL1(thebif)); }

 void SgWhileStmt::replaceBody(SgStatement &s)
{
  BIF_BLOB1(thebif) = NULL;
  insertBfndListIn(s.thebif,thebif,thebif);
  addControlEndToStmt(thebif);
}



 
 SgWhileStmt::~SgWhileStmt()
{ RemoveFromTableBfnd((void *) this); }


// SgDoWhileStmt--inlines

 SgDoWhileStmt::SgDoWhileStmt(SgExpression &cond, SgStatement &body): SgWhileStmt(DO_WHILE_NODE)
{
  BIF_LL1(thebif) = cond.thellnd;
  insertBfndListIn(body.thebif,thebif,thebif);
  addControlEndToStmt(thebif);
}

 SgDoWhileStmt::~SgDoWhileStmt()
{ RemoveFromTableBfnd((void *) this); }


// SgLofIfStmt--inlines

 SgLogIfStmt::SgLogIfStmt(int variant):SgStatement(variant)
{}

 SgLogIfStmt::SgLogIfStmt(SgExpression &cond, SgStatement &s):SgStatement(LOGIF_NODE)
{
  BIF_LL1(thebif) = cond.thellnd;
  insertBfndListIn(s.thebif,thebif,thebif);
  addControlEndToStmt(thebif);
}

 SgStatement * SgLogIfStmt::body()
{
  PTR_BFND bif =NULL;
  if (BIF_BLOB1(thebif))
    bif = BLOB_VALUE(BIF_BLOB1(thebif));
  return BfndMapping(bif);
}

 SgExpression * SgLogIfStmt::conditional()
{  return LlndMapping(BIF_LL1(thebif)); }  // the while test

// check if the statement s is a single statement. 
 void SgLogIfStmt::setBody(SgStatement &s)
{
  BIF_BLOB1(thebif) = NULL;
  insertBfndListIn(s.thebif,thebif,thebif);
}

// this code won't work, since after the addition false
//   clause, it should become SgIfThenElse statement.
 void SgLogIfStmt::addFalseClause(SgStatement &s)
{
  appendBfndListToList2(s.thebif,thebif);
  addControlEndToList2(thebif);
}

//need a forward definition;
SgIfStmt * isSgIfStmt (SgStatement *pt);

 SgIfStmt *SgLogIfStmt::convertLogicIf()
{
  LibconvertLogicIf(thebif);
  return isSgIfStmt(this);
}

 SgLogIfStmt::~SgLogIfStmt()
{ RemoveFromTableBfnd((void *) this); }


// SgIfStmt--inlines

 SgIfStmt::SgIfStmt(int variant): SgStatement(variant)
{}

 SgIfStmt::SgIfStmt(SgExpression &cond, SgStatement &trueBody, SgStatement &falseBody, SgSymbol &construct_name):SgStatement(IF_NODE)
{
  BIF_LL1(thebif) = cond.thellnd;
  BIF_SYMB(thebif) = construct_name.thesymb;
  insertBfndListIn(trueBody.thebif,thebif,thebif);
  appendBfndListToList2(falseBody.thebif,thebif);
  addControlEndToStmt(thebif);
}

 SgIfStmt::SgIfStmt(SgExpression &cond, SgStatement &trueBody, SgStatement &falseBody):SgStatement(IF_NODE)
{
  BIF_LL1(thebif) = cond.thellnd;
  insertBfndListIn(trueBody.thebif,thebif,thebif);
  appendBfndListToList2(falseBody.thebif,thebif);
  addControlEndToStmt(thebif);
}

 SgIfStmt::SgIfStmt(SgExpression &cond, SgStatement &trueBody):SgStatement(IF_NODE)
{
  BIF_LL1(thebif) = cond.thellnd;
  insertBfndListIn(trueBody.thebif,thebif,thebif);
  addControlEndToStmt(thebif);
}

// the first stmt in the True clause
 SgStatement * SgIfStmt::trueBody()
{
  PTR_BFND bif = NULL;
  if (BIF_BLOB1(thebif))
    bif = BLOB_VALUE(BIF_BLOB1(thebif));
  return BfndMapping(bif);
}

// SgBlock is needed? 
// i-th stmt in True clause
 SgStatement * SgIfStmt::trueBody(int i)
{
  PTR_BFND bif =NULL;
  if (BIF_BLOB1(thebif))
    bif = BLOB_VALUE(BIF_BLOB1(thebif));
  return BfndMapping(getStatementNumber(bif,i));
}

// the first stmt in the False
 SgStatement * SgIfStmt::falseBody()
{
  PTR_BFND bif = NULL;
  if (BIF_BLOB2(thebif))
    bif = BLOB_VALUE(BIF_BLOB2(thebif));
  return BfndMapping(bif);
}

// i-th statement of the body.
 SgStatement * SgIfStmt::falseBody(int i)
{
  PTR_BFND bif =NULL;
  if (BIF_BLOB2(thebif))
    bif = BLOB_VALUE(BIF_BLOB2(thebif));
  return BfndMapping(getStatementNumber(bif,i));
}

// the while test
 SgExpression * SgIfStmt::conditional()
{ return LlndMapping(BIF_LL1(thebif)); }

 SgSymbol * SgIfStmt::construct_name()
{ return SymbMapping(BIF_SYMB(thebif)); }

// new body=s and lex successors.
 void SgIfStmt::replaceTrueBody(SgStatement &s)
{
  BIF_BLOB1(thebif) = NULL;
  insertBfndListIn(s.thebif,thebif,thebif);
}

// new body=s and lex successors.
 void SgIfStmt::replaceFalseBody(SgStatement &s)
{
  BIF_BLOB2(thebif) = NULL;
  appendBfndListToList2(s.thebif,thebif);
  addControlEndToList2(thebif);
}

 SgIfStmt::~SgIfStmt()
{ RemoveFromTableBfnd((void *) this); }


// SgArithIfStmt--inlines

 SgArithIfStmt::SgArithIfStmt(int variant):SgStatement(variant)
{}

 SgArithIfStmt::SgArithIfStmt(SgExpression &cond, SgLabel &llabel, SgLabel &elabel, SgLabel &glabel):SgStatement(ARITHIF_NODE)
{
  BIF_LL1(thebif) = cond.thellnd;
  BIF_LL2(thebif) = addLabelRefToExprList(BIF_LL2(thebif),llabel.thelabel);
  BIF_LL2(thebif) = addLabelRefToExprList(BIF_LL2(thebif),elabel.thelabel);
  BIF_LL2(thebif) = addLabelRefToExprList(BIF_LL2(thebif),glabel.thelabel);
}

 SgExpression * SgArithIfStmt::conditional()
{ return LlndMapping(BIF_LL1(thebif)); }

 void SgArithIfStmt::set_conditional(SgExpression &cond)
{ BIF_LL1(thebif) = cond.thellnd; }

// the <, ==, and > goto labels. in order 0->2.
 SgExpression * SgArithIfStmt::label(int i)
{ return LlndMapping(getPositionInExprList(BIF_LL1(thebif),i)); }

#ifdef NOT_YET_IMPLEMENTED
 void SgArithIfStmt::setLabel(SgLabel &label)
{
  BIF_LL3(thebif) = addLabelRefToExprList(BIF_LL3(thebif) , label.thelabel);
  SORRY;
}
#endif

 SgArithIfStmt::~SgArithIfStmt()
{ RemoveFromTableBfnd((void *) this); }


// SgWhereStmt--inlines

 SgWhereStmt::SgWhereStmt(SgExpression &cond, SgStatement &body):SgLogIfStmt(WHERE_NODE)
{
  BIF_LL1(thebif) = cond.thellnd;
  insertBfndListIn(body.thebif,thebif,thebif);
  addControlEndToStmt(thebif);
}

 SgWhereStmt::~SgWhereStmt()
{ RemoveFromTableBfnd((void *) this); }


// SgWhereBlockStmt--inlines

 SgWhereBlockStmt::SgWhereBlockStmt(SgExpression &cond, SgStatement &trueBody, SgStatement &falseBody):SgIfStmt(WHERE_BLOCK_STMT)
{
  BIF_LL1(thebif) = cond.thellnd;
  insertBfndListIn(trueBody.thebif,thebif,thebif);
  appendBfndListToList2(falseBody.thebif,thebif);
  // appendBfndListToList2 does not update BIF_ NEXT...
    addControlEndToList2(thebif);
}

 SgWhereBlockStmt::~SgWhereBlockStmt()
{ RemoveFromTableBfnd((void *) this); }


// SgSwitchStmt--inlines

 SgSwitchStmt::SgSwitchStmt(SgExpression &selector, SgStatement &caseOptionList,
                                  SgSymbol &constructName):SgStatement(SWITCH_NODE)
{
  BIF_SYMB(thebif) = constructName.thesymb;
  BIF_LL1(thebif) = selector.thellnd;
  insertBfndListIn(caseOptionList.thebif,thebif,thebif);
}

 SgSwitchStmt::~SgSwitchStmt()
{ RemoveFromTableBfnd((void *) this); }

 SgExpression * SgSwitchStmt::selector()
{ return LlndMapping(BIF_LL1(thebif)); }

 void  SgSwitchStmt::setSelector(SgExpression &cond)
{ BIF_LL1(thebif) = cond.thellnd; }

// the number of cases
 int SgSwitchStmt::numberOfCaseOptions()
{ return countInStmtNode1(thebif,CASE_NODE); }

// i-th case block
 SgStatement * SgSwitchStmt::caseOption(int i)
{ return BfndMapping(GetcountInStmtNode1(thebif,CASE_NODE,i)); }


 void SgSwitchStmt::addCaseOption(SgStatement &caseOption)
{ insertBfndListIn(caseOption.thebif,thebif,thebif); }

#if 0 
// extractBifSectionBetween not defined
 void  SgSwitchStmt::deleteCaseOption(int i)
{
  PTR_BFND pt;
  if ( pt = GetcountInStmtNode1(thebif,CASE_NODE,i))
    extractBifSectionBetween(pt,getLastNodeOfStmt(pt));
}
#endif


// SgCaseOptionStmt--inlines


 SgCaseOptionStmt::SgCaseOptionStmt(SgExpression &caseRangeList, SgStatement &body, 
                                          SgSymbol &constructName):SgStatement(CASE_NODE)
{
  BIF_SYMB(thebif) = constructName.thesymb;
  BIF_LL1(thebif) = caseRangeList.thellnd;
  insertBfndListIn(body.thebif,thebif,thebif);
  addControlEndToStmt(thebif);
}

 SgCaseOptionStmt::~SgCaseOptionStmt()
{ RemoveFromTableBfnd((void *) this);}

 SgExpression * SgCaseOptionStmt::caseRangeList()
{ return LlndMapping(BIF_LL1(thebif)); }

 void SgCaseOptionStmt::setCaseRangeList(SgExpression &caseRangeList)
{ BIF_LL1(thebif) = caseRangeList.thellnd; }

 SgExpression * SgCaseOptionStmt::caseRange(int i)
{ return LlndMapping(getPositionInExprList(BIF_LL1(thebif),i));}

 void SgCaseOptionStmt::setCaseRange(int, SgExpression &caseRange)
{ 
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif),caseRange.thellnd); 
}

 SgStatement * SgCaseOptionStmt::body()
{
  PTR_BFND bif =NULL;
  if (BIF_BLOB1(thebif))
    bif = BLOB_VALUE(BIF_BLOB1(thebif));
  return BfndMapping(bif);
}

 void SgCaseOptionStmt::setBody(SgStatement &body)
{
  BIF_BLOB1(thebif) = NULL;
  insertBfndListIn(body.thebif,thebif,thebif);
}


// ******************** Leaf Executable Nodes ***********************

// SgExecutableStatement--inlines

 SgExecutableStatement::SgExecutableStatement(int variant):SgStatement(variant)
{}

// SgAssignStmt--inlines

 SgAssignStmt::SgAssignStmt(int variant):SgExecutableStatement(variant)
{}
 SgAssignStmt::SgAssignStmt(SgExpression &lhs, SgExpression &rhs):SgExecutableStatement(ASSIGN_STAT)
{
  BIF_LL1(thebif) = lhs.thellnd;
  BIF_LL2(thebif) = rhs.thellnd;
}

 SgExpression * SgAssignStmt::lhs()
{ return LlndMapping(BIF_LL1(thebif)); }

// the right hand side
 SgExpression * SgAssignStmt::rhs()
{ return LlndMapping(BIF_LL2(thebif)); }

// replace lhs with e
 void SgAssignStmt::replaceLhs(SgExpression &e)
{ BIF_LL1(thebif) = e.thellnd; }

// replace rhs with e
 void SgAssignStmt::replaceRhs(SgExpression &e)
{ BIF_LL2(thebif) = e.thellnd; }


// SgCExpStmt--inlines
 SgCExpStmt::SgCExpStmt(SgExpression &exp):SgExecutableStatement(EXPR_STMT_NODE)
{ BIF_LL1(thebif) = exp.thellnd; }

 SgCExpStmt::SgCExpStmt(SgExpression &lhs, SgExpression &rhs):SgExecutableStatement(EXPR_STMT_NODE)
{ BIF_LL1(thebif) =addToExprList(BIF_LL1(thebif),newExpr(ASSGN_OP,NULL,lhs.thellnd,rhs.thellnd)); }

// the expression
 SgExpression *SgCExpStmt::expr()
{ return LlndMapping(BIF_LL1(thebif)); }

// replace exp with e
 void SgCExpStmt::replaceExpression(SgExpression &e)
{ BIF_LL1(thebif) = e.thellnd; }

 SgCExpStmt::~SgCExpStmt()
{ RemoveFromTableBfnd((void *) this); }


// SgPointerAssignStmt--inlines

 SgPointerAssignStmt::SgPointerAssignStmt(SgExpression lhs, SgExpression rhs):SgAssignStmt(POINTER_ASSIGN_STAT)
{
  BIF_LL1(thebif) = lhs.thellnd;
  BIF_LL2(thebif) = rhs.thellnd;
}

 SgPointerAssignStmt::~SgPointerAssignStmt()
{ RemoveFromTableBfnd((void *) this); }



// SgHeapStmt--inlines

 SgHeapStmt::SgHeapStmt(int variant, SgExpression &allocationList, SgExpression &statVariable):SgExecutableStatement(variant)
{
  BIF_LL1(thebif) = allocationList.thellnd;
  BIF_LL2(thebif) = statVariable.thellnd;
}

 SgHeapStmt::~SgHeapStmt()
{ RemoveFromTableBfnd((void *) this); }

 SgExpression * SgHeapStmt::allocationList()
{ return LlndMapping(BIF_LL1(thebif)); }

 void SgHeapStmt::setAllocationList(SgExpression &allocationList)
{ BIF_LL1(thebif) = allocationList.thellnd;}

 SgExpression * SgHeapStmt::statVariable()
{ return LlndMapping(BIF_LL2(thebif)); }

 void SgHeapStmt::setStatVariable(SgExpression &statVar)
{ BIF_LL2(thebif) = statVar.thellnd; }


// SgNullifyStmt--inlines

 SgNullifyStmt::SgNullifyStmt(SgExpression &objectList):SgExecutableStatement(NULLIFY_STMT)
{ BIF_LL1(thebif) = objectList.thellnd; }

 SgNullifyStmt::~SgNullifyStmt()
{ RemoveFromTableBfnd((void *) this); }

 SgExpression * SgNullifyStmt::nullifyList()
{ return LlndMapping(BIF_LL1(thebif)); }

 void SgNullifyStmt::setNullifyList(SgExpression &nullifyList)
{ BIF_LL1(thebif) = nullifyList.thellnd; }


// SgContinueStmt--inlines

 SgContinueStmt::SgContinueStmt():SgExecutableStatement(CONT_STAT)
{}
 SgContinueStmt::~SgContinueStmt()
{ RemoveFromTableBfnd((void *) this); }


// SgControlEndStmt--inlines

 SgControlEndStmt::SgControlEndStmt(int variant):SgExecutableStatement(variant)
{}

 SgControlEndStmt::SgControlEndStmt():SgExecutableStatement(CONTROL_END)
{}

 SgControlEndStmt::~SgControlEndStmt()
{ RemoveFromTableBfnd((void *) this); }


// SgBreakStmt--inlines

 SgBreakStmt::SgBreakStmt():SgExecutableStatement(BREAK_NODE)
{}

 SgBreakStmt::~SgBreakStmt()
{ RemoveFromTableBfnd((void *) this); }


// SgCycleStmt--inlines


 SgCycleStmt::SgCycleStmt(SgSymbol &symbol):SgExecutableStatement(CYCLE_STMT)
{ BIF_SYMB(thebif) = symbol.thesymb; }

// the name of the loop to cycle
 SgSymbol * SgCycleStmt::constructName()
{ return SymbMapping(BIF_SYMB(thebif)); }

 void SgCycleStmt::setConstructName(SgSymbol &constructName)
{ BIF_SYMB(thebif) = constructName.thesymb; }

 SgCycleStmt::~SgCycleStmt()
{ RemoveFromTableBfnd((void *) this); }


 SgExpression * SgReturnStmt::returnValue()
{ return LlndMapping(BIF_LL1(thebif)); }

 void SgReturnStmt::setReturnValue(SgExpression &retVal)
{ BIF_LL1(thebif) = retVal.thellnd; }

 SgReturnStmt::~SgReturnStmt()
{ RemoveFromTableBfnd((void *) this); }


// SgExitStmt--inlines

 SgExitStmt::SgExitStmt(SgSymbol &construct_name):SgControlEndStmt(EXIT_STMT)
{ BIF_SYMB(thebif) = construct_name.thesymb; }

 SgExitStmt::~SgExitStmt()
{ RemoveFromTableBfnd((void *) this); }

 SgSymbol * SgExitStmt::constructName()
{ return SymbMapping(BIF_SYMB(thebif)); }  // the name of the loop to cycle

 void SgExitStmt::setConstructName(SgSymbol &constructName)
{ BIF_SYMB(thebif) = constructName.thesymb; }



// SgGotoStmt--inlines
 SgGotoStmt::SgGotoStmt(SgLabel &label):SgExecutableStatement(GOTO_NODE)
{ BIF_LL3(thebif) = SgLabelRefExp(label).thellnd; }
/* Tried to fix a bug reported by anl's people.
   The following line is the original code.
{ BIF_LABEL(thebif) = label.thelabel; }
*/


 SgLabel * SgGotoStmt::branchLabel()
{ SgLabelRefExp *e =  (SgLabelRefExp *) LlndMapping(BIF_LL3(thebif));
  return (e)?e->label(): (SgLabel *)NULL;
 }
	

 SgGotoStmt::~SgGotoStmt(){RemoveFromTableBfnd((void *) this);}


// SgLabelListStmt--inlines

 SgLabelListStmt::SgLabelListStmt(int variant):SgExecutableStatement(variant)
{}

 int SgLabelListStmt::numberOfTargets()
{ return exprListLength(BIF_LL1(thebif)); }

 SgExpression * SgLabelListStmt::labelList()
{ return LlndMapping(BIF_LL1(thebif)); }

 void SgLabelListStmt::setLabelList(SgExpression &labelList)
{ BIF_LL1(thebif) = labelList.thellnd; }



// SgAssignedGotoStmt--inlines

 SgAssignedGotoStmt::SgAssignedGotoStmt(SgSymbol &symbol, SgExpression &labelList):SgLabelListStmt(ASSGOTO_NODE)
{
  BIF_SYMB(thebif) = symbol.thesymb;
  BIF_LL1(thebif) = labelList.thellnd;
}

 SgSymbol * SgAssignedGotoStmt::symbol()
{ return SymbMapping(BIF_SYMB(thebif)); }

 void SgAssignedGotoStmt::setSymbol(SgSymbol &symb)
{ BIF_SYMB(thebif) = symb.thesymb; }

 SgAssignedGotoStmt::~SgAssignedGotoStmt()
{ RemoveFromTableBfnd((void *) this); }



// SgComputedGotoStmt--inlines

 SgComputedGotoStmt::SgComputedGotoStmt(SgExpression &expr, SgLabel &label):SgLabelListStmt(COMGOTO_NODE)
{
  BIF_LL1(thebif) = addLabelRefToExprList(BIF_LL1(thebif),label.thelabel);
  BIF_LL2(thebif) = expr.thellnd;
}

 void SgComputedGotoStmt::addLabel(SgLabel &label)
{
  BIF_LL1(thebif) = addLabelRefToExprList(BIF_LL1(thebif),label.thelabel); 
}

 SgExpression * SgComputedGotoStmt::exp()
{ return LlndMapping(BIF_LL2(thebif)); }

 void SgComputedGotoStmt::setExp(SgExpression &exp)
{ BIF_LL2(thebif) = exp.thellnd; }

 SgComputedGotoStmt::~SgComputedGotoStmt()
{ RemoveFromTableBfnd((void *) this); }


// SgStopOrPauseStmt--inlines

 SgStopOrPauseStmt::SgStopOrPauseStmt(int variant, SgExpression *expr):SgExecutableStatement(variant)
{ 
if (expr)
  BIF_LL1(thebif) = expr->thellnd;
 }

 SgExpression * SgStopOrPauseStmt::exp()
{ return LlndMapping(BIF_LL1(thebif)); }

 void SgStopOrPauseStmt::setExp(SgExpression &exp)
{ BIF_LL1(thebif) = exp.thellnd; }

 SgStopOrPauseStmt::~SgStopOrPauseStmt()
{ RemoveFromTableBfnd((void *) this); }


// SgCallStmt--inlines

 SgCallStmt::SgCallStmt(SgSymbol &name, SgExpression &args):SgExecutableStatement(PROC_STAT)
{
  BIF_SYMB(thebif) = name.thesymb;
  BIF_LL1(thebif) = args.thellnd;
}

 SgCallStmt::SgCallStmt(SgSymbol &name):SgExecutableStatement(PROC_STAT)
{ 
  SgExpression * expr = new SgExpression (PROC_CALL, NULL, NULL, &name);
  BIF_LL1(thebif) = expr->thellnd; 
  //BIF_SYMB(thebif) = name.thesymb; 
}

// name of subroutine being called
 SgSymbol * SgCallStmt::name()
{ return SymbMapping(NODE_SYMB(BIF_LL1(thebif))); }

// the number of arguement expressions
 int SgCallStmt::numberOfArgs()
{ return exprListLength(NODE_OPERAND0(BIF_LL1(thebif))); }

 void SgCallStmt::addArg(SgExpression &arg)
{
  int debug=0;
  //printf("The ll1 id is %d\n", this->expr(0)->id());
  struct llnd * tail;
  tail = BIF_LL1(thebif);
  if (NODE_OPERAND0(tail)==NULL) {
    NODE_OPERAND0(tail)=newExpr(EXPR_LIST, NULL, arg.thellnd, NULL);
  }
  else {
    tail=NODE_OPERAND0(tail);
    while (NODE_OPERAND1(tail)!=NULL)
      tail=NODE_OPERAND1(tail);
    NODE_OPERAND1(tail)=newExpr(EXPR_LIST, NULL, arg.thellnd, NULL);
  };
}

/*
void SgCallStmt::addArg(SgExpression &arg)
{
  SgExpression * proccall=this->expr(0);

  SgExpression * newarg = new SgExpression (EXPR_LIST, &arg, NULL, NULL);

  if (proccall->operand(1)==NULL)
    proccall->setLhs(newarg);
  else {
    SgExpression * tail = proccall->operand(2);

    while(tail->operand(2)!=NULL)
      tail=tail->operand(2);
    tail->setRhs(newarg);

  }
}
*/

// the i-th argument expression
 SgExpression * SgCallStmt::arg(int i)
{ return LlndMapping(getPositionInExprList(NODE_OPERAND0(BIF_LL1(thebif)), i)); }

 SgCallStmt::~SgCallStmt()
{ RemoveFromTableBfnd((void *) this); }



// SgProsCallStmt--inlines
 
 SgProsCallStmt::SgProsCallStmt(SgSymbol &name, SgExprListExp &args):SgExecutableStatement(PROS_STAT)
{
  BIF_SYMB(thebif) = name.thesymb;
  BIF_LL1(thebif) = args.thellnd;
}
 
 SgProsCallStmt::SgProsCallStmt(SgSymbol &name):SgExecutableStatement(PROS_STAT)
{ BIF_SYMB(thebif) = name.thesymb; }

// name of process being called
 SgSymbol * SgProsCallStmt::name()
{ return SymbMapping(BIF_SYMB(thebif)); }
 
// the number of arguement expressions
 int SgProsCallStmt::numberOfArgs()
{ return exprListLength(BIF_LL1(thebif)); }
 
 void SgProsCallStmt::addArg(SgExpression &arg)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif),arg.thellnd); }

 SgExprListExp *SgProsCallStmt::args()
{ return (SgExprListExp *) LlndMapping(BIF_LL1(thebif)); }
 
// the i-th argument expression
 SgExpression * SgProsCallStmt::arg(int i)
{ return LlndMapping(getPositionInExprList(BIF_LL1(thebif), i)); }
 
 SgProsCallStmt::~SgProsCallStmt()
{ RemoveFromTableBfnd((void *) this); }
 
 
 
// SgProsCallLctn--inlines
 
 SgProsCallLctn::SgProsCallLctn(SgSymbol &name, SgExprListExp &args,
                                      SgExprListExp &lctn)
                      :SgExecutableStatement(PROS_STAT_LCTN)
{
  BIF_SYMB(thebif) = name.thesymb;
  BIF_LL1(thebif) = args.thellnd;
  BIF_LL2(thebif) = lctn.thellnd;
}
 
 SgProsCallLctn::SgProsCallLctn(SgSymbol &name, SgExprListExp &lctn)
                      :SgExecutableStatement(PROS_STAT_LCTN)
{
   BIF_SYMB(thebif) = name.thesymb;
   BIF_LL2(thebif) = lctn.thellnd;
}
 
// name of process being called
 SgSymbol * SgProsCallLctn::name()
{ return SymbMapping(BIF_SYMB(thebif)); }
 
// the number of arguement expressions
 int SgProsCallLctn::numberOfArgs()
{ return exprListLength(BIF_LL1(thebif)); }
 
 void SgProsCallLctn::addArg(SgExpression &arg)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif),arg.thellnd); }
 
 SgExprListExp *SgProsCallLctn::args()
{ return (SgExprListExp *) LlndMapping(BIF_LL1(thebif)); }
 
// the i-th argument expression
 SgExpression * SgProsCallLctn::arg(int i)
{ return LlndMapping(getPositionInExprList(BIF_LL1(thebif), i)); }
 
 SgExpression * SgProsCallLctn::location()
{ return LlndMapping(BIF_LL2(thebif)); }

 SgProsCallLctn::~SgProsCallLctn()
{ RemoveFromTableBfnd((void *) this); }
 
 

// SgProsCallSubm--inlines
 
 SgProsCallSubm::SgProsCallSubm(SgSymbol &name, SgExprListExp &args,
                                      SgExprListExp &subm)
                      :SgExecutableStatement(PROS_STAT_SUBM)
{
  BIF_SYMB(thebif) = name.thesymb;
  BIF_LL1(thebif) = args.thellnd;
  BIF_LL2(thebif) = subm.thellnd;
}
 
 SgProsCallSubm::SgProsCallSubm(SgSymbol &name, SgExprListExp &subm)
                      :SgExecutableStatement(PROS_STAT_SUBM)
{
   BIF_SYMB(thebif) = name.thesymb;
   BIF_LL2(thebif) = subm.thellnd;
}
 
// name of process being called
 SgSymbol * SgProsCallSubm::name()
{ return SymbMapping(BIF_SYMB(thebif)); }
 
// the number of arguement expressions
 int SgProsCallSubm::numberOfArgs()
{ return exprListLength(BIF_LL1(thebif)); }
 
 void SgProsCallSubm::addArg(SgExpression &arg)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif),arg.thellnd); }
 
 SgExprListExp *SgProsCallSubm::args()
{ return (SgExprListExp *) LlndMapping(BIF_LL1(thebif)); }
 
// the i-th argument expression
 SgExpression * SgProsCallSubm::arg(int i)
{ return LlndMapping(getPositionInExprList(BIF_LL1(thebif), i)); }

 SgExpression * SgProsCallSubm::submachine()
{ return LlndMapping(BIF_LL2(thebif)); }
 
 SgProsCallSubm::~SgProsCallSubm()
{ RemoveFromTableBfnd((void *) this); }
 
 

// SgProcessesStmt--inlines

 SgProcessesStmt::SgProcessesStmt():SgStatement(PROCESSES_STAT)
{}

 SgProcessesStmt::~SgProcessesStmt()
{ RemoveFromTableBfnd((void *) this); }
 


// SgEndProcessesStmt--inlines
 
 SgEndProcessesStmt::SgEndProcessesStmt():SgStatement(PROCESSES_END)
{}
 
 SgEndProcessesStmt::~SgEndProcessesStmt()
{ RemoveFromTableBfnd((void *) this); }
 
 

// SgInportStmt--inlines

 SgInportStmt::SgInportStmt(SgExprListExp &name):SgStatement(INPORT_DECL)
{ BIF_LL1(thebif) = name.thellnd; }

 SgInportStmt::SgInportStmt(SgExprListExp &name, SgPortTypeExp &porttype)
                    :SgStatement(INPORT_DECL)
{
  BIF_LL1(thebif) = name.thellnd;
  BIF_LL2(thebif) = porttype.thellnd;
}

 SgInportStmt::~SgInportStmt()
{ RemoveFromTableBfnd((void *) this); }

 void SgInportStmt::addname(SgExpression &name)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), name.thellnd); }

 int SgInportStmt::numberOfNames()
{ return exprListLength(BIF_LL1(thebif)); }

 SgExprListExp * SgInportStmt::names()
{ return (SgExprListExp *) LlndMapping(BIF_LL1(thebif)); }

 SgExpression *SgInportStmt::name(int i)
{ return LlndMapping(getPositionInExprList(BIF_LL1(thebif),i)); }

 void SgInportStmt::addporttype(SgExpression &porttype)
{ BIF_LL2(thebif) = addToList(BIF_LL2(thebif), porttype.thellnd); }

 int SgInportStmt::numberOfPortTypes()
{ return exprListLength(BIF_LL2(thebif)); }

 SgPortTypeExp * SgInportStmt::porttypes()
{ return (SgPortTypeExp *) LlndMapping(BIF_LL2(thebif)); }

 SgPortTypeExp * SgInportStmt::porttype(int i)
{ return (SgPortTypeExp *) LlndMapping(getPositionInList(BIF_LL2(thebif),i)); }



// SgOutportStmt--inlines
 
 SgOutportStmt::SgOutportStmt(SgExprListExp &name)
                     :SgStatement(OUTPORT_DECL)
{ BIF_LL1(thebif) = name.thellnd; }
 
 SgOutportStmt::SgOutportStmt(SgExprListExp &name,
                                    SgPortTypeExp &porttype)
                    :SgStatement(OUTPORT_DECL)
{
  BIF_LL1(thebif) = name.thellnd;
  BIF_LL2(thebif) = porttype.thellnd;
}
 
 SgOutportStmt::~SgOutportStmt()
{ RemoveFromTableBfnd((void *) this); }
 
 void SgOutportStmt::addname(SgExpression &name)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), name.thellnd); }
 
 int SgOutportStmt::numberOfNames()
{ return exprListLength(BIF_LL1(thebif)); }
 
 SgExprListExp * SgOutportStmt::names()
{ return (SgExprListExp *) LlndMapping(BIF_LL1(thebif)); }
 
 SgExpression *SgOutportStmt::name(int i)
{ return LlndMapping(getPositionInExprList(BIF_LL1(thebif),i)); }

 void SgOutportStmt::addporttype(SgExpression &porttype)
{ BIF_LL2(thebif) = addToList(BIF_LL2(thebif), porttype.thellnd); }
 
 int SgOutportStmt::numberOfPortTypes()
{ return exprListLength(BIF_LL2(thebif)); }
 
 SgPortTypeExp * SgOutportStmt::porttypes()
{ return (SgPortTypeExp *) LlndMapping(BIF_LL2(thebif)); }
 
 SgPortTypeExp * SgOutportStmt::porttype(int i)
{ return (SgPortTypeExp *) LlndMapping(getPositionInList(BIF_LL2(thebif),i)); }

 

// SgChannelStmt--inlines

 SgChannelStmt::SgChannelStmt(SgExpression &outport, SgExpression &inport)
                     :SgStatement(CHANNEL_STAT)
{
  BIF_LL1(thebif) = outport.thellnd;
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), inport.thellnd);
}
 

 SgChannelStmt::SgChannelStmt(SgExpression &outport, SgExpression &inport,
                                    SgExpression &io_or_err)
                     :SgStatement(CHANNEL_STAT)
{
  BIF_LL1(thebif) = outport.thellnd;
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), inport.thellnd);
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), io_or_err.thellnd);
}
 

 SgChannelStmt::SgChannelStmt(SgExpression &outport, SgExpression &inport,
                                  SgExpression &iostore, SgExpression &errlabel)
                     :SgStatement(CHANNEL_STAT)
{
  BIF_LL1(thebif) = outport.thellnd;
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), inport.thellnd);
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), iostore.thellnd);
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), errlabel.thellnd);
}
 

 SgChannelStmt::~SgChannelStmt()
{ RemoveFromTableBfnd((void *) this); }


 SgExpression * SgChannelStmt::outport()
{ return LlndMapping(getPositionInList(BIF_LL1(thebif),0)); }


 SgExpression * SgChannelStmt::inport()
{ return LlndMapping(getPositionInList(BIF_LL1(thebif),1)); }


 SgExpression * SgChannelStmt::ioStore()
{
  PTR_LLND ll;
 
  ll = getPositionInList(BIF_LL1(thebif),2);
  if (!ll)
    return (SgExpression *) NULL;
  else
    if (NODE_CODE(ll) != IOSTAT_STORE) // must be ERR_LABEL
      return (SgExpression *) NULL;
    else
      return  LlndMapping(ll);
}


 SgExpression * SgChannelStmt::errLabel()
{
  PTR_LLND ll;
 
  ll = getPositionInList(BIF_LL1(thebif),2);
  if (!ll)
    return (SgExpression *) NULL;
  else
    if (NODE_CODE(ll) != ERR_LABEL) { // must be IOSTAT_STORE
      ll = NODE_OPERAND1(ll);
      if ((!ll) || (NODE_CODE(ll) != ERR_LABEL))
        return (SgExpression *) NULL;
      else
        return LlndMapping(ll);
    } else
      return  LlndMapping(ll);
}



// SgMergerStmt--inlines

 SgMergerStmt::SgMergerStmt(SgExpression &outport, SgExpression &inport):
                    SgStatement(MERGER_STAT)
{
  BIF_LL1(thebif) = outport.thellnd;
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), inport.thellnd);
}


 SgMergerStmt::SgMergerStmt(SgExpression &outport, SgExpression &inport,
                                SgExpression &io_or_err)
                   :SgStatement(MERGER_STAT)
{
  BIF_LL1(thebif) = outport.thellnd;
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), inport.thellnd);
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), io_or_err.thellnd);
}


 SgMergerStmt::SgMergerStmt(SgExpression &outport, SgExpression &inport,
                                SgExpression &iostore, SgExpression &errlabel):
                   SgStatement(MERGER_STAT)
{
  BIF_LL1(thebif) = outport.thellnd;
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), inport.thellnd);
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), iostore.thellnd);
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), errlabel.thellnd);
}


 SgMergerStmt::~SgMergerStmt()
{ RemoveFromTableBfnd((void *) this); }


 void SgMergerStmt::addOutport(SgExpression &outport)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), outport.thellnd); }
  

 void SgMergerStmt::addIoStore(SgExpression &iostore)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), iostore.thellnd); }


 void SgMergerStmt::addErrLabel(SgExpression &errlabel)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), errlabel.thellnd); }


 int SgMergerStmt::numberOfOutports()
{
  PTR_LLND ll = BIF_LL1(thebif);
  int n = 0;

  while (ll && (n != 3)) {
    if (( NODE_CODE(ll) == IOSTAT_STORE ) || ( NODE_CODE(ll) == ERR_LABEL ) ||
       ( NODE_CODE(ll) == INPORT_NAME ))
      n = n + 1;
    ll = NODE_OPERAND1(ll);
  };
  return (exprListLength(BIF_LL1(thebif)) - n);
  // double scanning the list may be improved
}


 SgExpression * SgMergerStmt::outport(int i)
{ return LlndMapping(getPositionInList(BIF_LL1(thebif),i)); }


 SgExpression * SgMergerStmt::inport()
{
  int n = numberOfOutports();
  PTR_LLND ll;
 
  ll = getPositionInList(BIF_LL1(thebif),n);
  if (!ll) {
    return (SgExpression *) NULL;
  } else
    return  LlndMapping(ll);
}
 

 SgExpression * SgMergerStmt::ioStore()
{
  int n = numberOfOutports();
  PTR_LLND ll;
 
  ll = getPositionInList(BIF_LL1(thebif),n+1);
  if (!ll)
    return (SgExpression *) NULL;
  else
    if (NODE_CODE(ll) != IOSTAT_STORE) //must be ERR_LABEL
      return (SgExpression *) NULL;
    else
      return  LlndMapping(ll);
}
 

 SgExpression * SgMergerStmt::errLabel()
{
  int n = numberOfOutports();
  PTR_LLND ll;
 
  ll = getPositionInList(BIF_LL1(thebif),n+1);
  if (!ll)
    return (SgExpression *) NULL;
  else
    if (NODE_CODE(ll) != ERR_LABEL) { // imust be IOSTAT_STORE
      ll = NODE_OPERAND1(ll);
      if (!ll)
        return (SgExpression *) NULL;
      else
        if (NODE_CODE(ll) != ERR_LABEL)
          return (SgExpression *) NULL;
        else
          return  LlndMapping(ll);
    }
    else
      return  LlndMapping(ll);
}



// SgMoveportStmt--inlines

 SgMoveportStmt::SgMoveportStmt(SgExpression &fromport,
                                      SgExpression &toport)
                     :SgStatement(MOVE_PORT)
{
  BIF_LL1(thebif) = fromport.thellnd;
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), toport.thellnd);
}
 

 SgMoveportStmt::SgMoveportStmt(SgExpression &fromport,
                                      SgExpression &toport,
                                      SgExpression &io_or_err)
                     :SgStatement(MOVE_PORT)
{
  BIF_LL1(thebif) = fromport.thellnd;
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), toport.thellnd);
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), io_or_err.thellnd);
}
 

 SgMoveportStmt::SgMoveportStmt(SgExpression &fromport,
                                      SgExpression &toport,
                                      SgExpression &iostore,
                                      SgExpression &errlabel)
                     :SgStatement(MOVE_PORT)
{
  BIF_LL1(thebif) = fromport.thellnd;
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), toport.thellnd);
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), iostore.thellnd);
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), errlabel.thellnd);
}
 

 SgMoveportStmt::~SgMoveportStmt()
{ RemoveFromTableBfnd((void *) this); }


 SgExpression * SgMoveportStmt::fromport()
{ return LlndMapping(getPositionInList(BIF_LL1(thebif),0)); }


 SgExpression * SgMoveportStmt::toport()
{ return LlndMapping(getPositionInList(BIF_LL1(thebif),1)); }


 SgExpression * SgMoveportStmt::ioStore()
{
  PTR_LLND ll;
 
  ll = getPositionInList(BIF_LL1(thebif),2);
  if (!ll)
    return (SgExpression *) NULL;
  else
    if (NODE_CODE(ll) != IOSTAT_STORE) // must be ERR_LABEL
      return (SgExpression *) NULL;
    else
      return  LlndMapping(ll);
}


 SgExpression * SgMoveportStmt::errLabel()
{
  PTR_LLND ll;
 
  ll = getPositionInList(BIF_LL1(thebif),2);
  if (!ll)
    return (SgExpression *) NULL;
  else
    if (NODE_CODE(ll) != ERR_LABEL) { // must be IOSTAT_STORE
      ll = NODE_OPERAND1(ll);
      if ((!ll) || (NODE_CODE(ll) != ERR_LABEL))
        return (SgExpression *) NULL;
      else
        return  LlndMapping(ll);
    } else
      return  LlndMapping(ll);
}



// SgSendStmt--inlines

 SgSendStmt::SgSendStmt(SgExpression &control, SgExprListExp &argument):
                   SgStatement(SEND_STAT)
{
  BIF_LL1(thebif) = control.thellnd;
  BIF_LL2(thebif) = argument.thellnd;
}


 SgSendStmt::SgSendStmt(SgExpression &outport, SgExprListExp &argument,
                              SgExpression &io_or_err): SgStatement(SEND_STAT)
{
  BIF_LL1(thebif) = outport.thellnd;
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), io_or_err.thellnd);
  BIF_LL2(thebif) = argument.thellnd;
}


 SgSendStmt::SgSendStmt(SgExpression &outport, SgExprListExp &argument,
                              SgExpression &iostore, SgExpression &errlabel):
                   SgStatement(SEND_STAT)
{
  BIF_LL1(thebif) = outport.thellnd;
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), iostore.thellnd);
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), errlabel.thellnd);
  BIF_LL2(thebif) = argument.thellnd;
}


 SgSendStmt::~SgSendStmt()
{ RemoveFromTableBfnd((void *) this); }


 void SgSendStmt::addOutport(SgExpression &outport)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), outport.thellnd); }
  

 void SgSendStmt::addIoStore(SgExpression &iostore)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), iostore.thellnd); }


 void SgSendStmt::addErrLabel(SgExpression &errlabel)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), errlabel.thellnd); }


 void SgSendStmt::addArgument(SgExpression &argument)
{ BIF_LL2(thebif) = addToExprList(BIF_LL2(thebif), argument.thellnd); }


 int SgSendStmt::numberOfOutports()
{
  PTR_LLND ll = BIF_LL1(thebif);
  int n = 0;

  while (ll && (n != 2)) {
    if (( NODE_CODE(ll) == IOSTAT_STORE ) || ( NODE_CODE(ll) == ERR_LABEL ))
      n = n + 1;
    ll = NODE_OPERAND1(ll);
  };
  return (exprListLength(BIF_LL1(thebif)) - n);
  // double scanning the list may be improved
}


 int SgSendStmt::numberOfArguments()
{ return exprListLength(BIF_LL2(thebif)); }


 SgExpression * SgSendStmt::controls()
{ return LlndMapping(BIF_LL1(thebif)); }


 SgExpression * SgSendStmt::outport(int i)
{ return LlndMapping(getPositionInList(BIF_LL1(thebif),i)); }


 SgExprListExp * SgSendStmt::arguments()
{ return (SgExprListExp *) LlndMapping(BIF_LL2(thebif)); }


 SgExpression * SgSendStmt::argument(int i)
{ return LlndMapping(getPositionInExprList(BIF_LL2(thebif),i)); }


 SgExpression * SgSendStmt::ioStore()
{
  int n = numberOfOutports();
  PTR_LLND ll;

  ll = getPositionInList(BIF_LL1(thebif),n);
  if (!ll)
    return (SgExpression *) NULL;
  else
    if (NODE_CODE(ll) != IOSTAT_STORE)
      return (SgExpression *) NULL;
    else
      return  LlndMapping(ll);
}


 SgExpression * SgSendStmt::errLabel()
{
  int n = numberOfOutports();
  PTR_LLND ll;
 
  ll = getPositionInList(BIF_LL1(thebif),n);
  if (!ll)
    return (SgExpression *) NULL;
  else
    if (NODE_CODE(ll) != ERR_LABEL) { // must be IOSTAT_STORE
      ll = NODE_OPERAND1(ll);
      if ((!ll) || (NODE_CODE(ll) != ERR_LABEL))
        return (SgExpression *) NULL;
      else
        return  LlndMapping(ll);
    } else
      return  LlndMapping(ll);
}



// SgReceiveStmt--inlines

 SgReceiveStmt::SgReceiveStmt(SgExpression &control,
                                    SgExprListExp &argument)
                     :SgStatement(RECEIVE_STAT)
{
  BIF_LL1(thebif) = control.thellnd;
  BIF_LL2(thebif) = argument.thellnd;
}


 SgReceiveStmt::SgReceiveStmt(SgExpression &inport,
                                    SgExprListExp &argument,
                                    SgExpression &e1):SgStatement(RECEIVE_STAT)
{
  BIF_LL1(thebif) = inport.thellnd;
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), e1.thellnd);
  BIF_LL2(thebif) = argument.thellnd;
}


 SgReceiveStmt::SgReceiveStmt(SgExpression &inport,
                                    SgExprListExp &argument,
                                    SgExpression &e1,
                                    SgExpression &e2):SgStatement(RECEIVE_STAT)
{
  BIF_LL1(thebif) = inport.thellnd;
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), e1.thellnd);
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), e2.thellnd);
  BIF_LL2(thebif) = argument.thellnd;
}


 SgReceiveStmt::SgReceiveStmt(SgExpression &inport,
                                    SgExprListExp &argument,
                                    SgExpression &e1,
                                    SgExpression &e2,
                                    SgExpression &e3):SgStatement(RECEIVE_STAT)
{
  BIF_LL1(thebif) = inport.thellnd;
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), e1.thellnd);
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), e2.thellnd);
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), e3.thellnd);
  BIF_LL2(thebif) = argument.thellnd;
}
 

 SgReceiveStmt::~SgReceiveStmt()
{ RemoveFromTableBfnd((void *) this); }


 void SgReceiveStmt::addInport(SgExpression &inport)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), inport.thellnd); }
  

 void SgReceiveStmt::addIoStore(SgExpression &iostore)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), iostore.thellnd); }


 void SgReceiveStmt::addErrLabel(SgExpression &errlabel)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), errlabel.thellnd); }


 void SgReceiveStmt::addEndLabel(SgExpression &endlabel)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), endlabel.thellnd); }
 

 void SgReceiveStmt::addArgument(SgExpression &argument)
{ BIF_LL2(thebif) = addToExprList(BIF_LL2(thebif), argument.thellnd); }


 int SgReceiveStmt::numberOfInports()
{
  PTR_LLND ll = BIF_LL1(thebif);
  int n = 0;

  while (ll && (n != 3)) {
    if (( NODE_CODE(ll) == IOSTAT_STORE ) || ( NODE_CODE(ll) == ERR_LABEL ) ||
       ( NODE_CODE(ll) == END_LABEL ))
      n = n + 1;
    ll = NODE_OPERAND1(ll);
  };
  return (exprListLength(BIF_LL1(thebif)) - n);
  // double scanning the list may be improved
}


 int SgReceiveStmt::numberOfArguments()
{ return exprListLength(BIF_LL2(thebif)); }


 SgExpression * SgReceiveStmt::controls()
{ return LlndMapping(BIF_LL1(thebif)); }


 SgExpression * SgReceiveStmt::inport(int i)
{ return LlndMapping(getPositionInList(BIF_LL1(thebif),i)); }


 SgExprListExp * SgReceiveStmt::arguments()
{ return (SgExprListExp *) LlndMapping(BIF_LL2(thebif)); }


 SgExpression * SgReceiveStmt::argument(int i)
{ return LlndMapping(getPositionInExprList(BIF_LL2(thebif),i)); }


 SgExpression * SgReceiveStmt::ioStore()
{
  int n = numberOfInports();
  PTR_LLND ll;

  ll = getPositionInList(BIF_LL1(thebif),n);
  if (!ll)
    return (SgExpression *) NULL;
  else
    if (NODE_CODE(ll) != IOSTAT_STORE)
      return (SgExpression *) NULL;
    else
      return  LlndMapping(ll);
}


 SgExpression * SgReceiveStmt::errLabel()
{
  int n = numberOfInports();
  PTR_LLND ll;
 
  ll = getPositionInList(BIF_LL1(thebif),n);
  if (!ll)
    return (SgExpression *) NULL;
  else
    if (NODE_CODE(ll) != ERR_LABEL) { // must be IOSTAT_STORE
      ll = NODE_OPERAND1(ll);
      if ((!ll) || (NODE_CODE(ll) != ERR_LABEL))
        return (SgExpression *) NULL;
      else
        return  LlndMapping(ll);
    } else
      return  LlndMapping(ll);
}


 SgExpression * SgReceiveStmt::endLabel()
{
  int n = numberOfInports();
  PTR_LLND ll;
 
  ll = getPositionInList(BIF_LL1(thebif),n);
  if (!ll)
    return (SgExpression *) NULL;
  else
    if (NODE_CODE(ll) != END_LABEL) { // must be IOSTAT_STORE or ERR_LABEL
      ll = NODE_OPERAND1(ll);
      if (!ll)
        return (SgExpression *) NULL;
      else
        if (NODE_CODE(ll) != END_LABEL) { // must be ERR_LABEL
          ll = NODE_OPERAND1(ll);
          if ((!ll) || (NODE_CODE(ll) != END_LABEL))
            return (SgExpression *) NULL;
          else
            return  LlndMapping(ll);
        } else
          return  LlndMapping(ll);
    } else
      return  LlndMapping(ll);
}


// SgEndchannelStmt--inlines

 SgEndchannelStmt::SgEndchannelStmt(SgExpression &outport)
                        :SgStatement(ENDCHANNEL_STAT)
{
  BIF_LL1(thebif) = outport.thellnd;
}


 SgEndchannelStmt::SgEndchannelStmt(SgExpression &outport,
                                          SgExpression &io_or_err)
                        :SgStatement(ENDCHANNEL_STAT)
{
  BIF_LL1(thebif) = outport.thellnd;
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), io_or_err.thellnd);
}


 SgEndchannelStmt::SgEndchannelStmt(SgExpression &outport,
                                          SgExpression &iostore,
                                          SgExpression &errlabel)
                        :SgStatement(ENDCHANNEL_STAT)
{
  BIF_LL1(thebif) = outport.thellnd;
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), iostore.thellnd);
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), errlabel.thellnd);
}


 SgEndchannelStmt::~SgEndchannelStmt()
{ RemoveFromTableBfnd((void *) this); }


 void SgEndchannelStmt::addOutport(SgExpression &outport)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), outport.thellnd); }
  

 void SgEndchannelStmt::addIoStore(SgExpression &iostore)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), iostore.thellnd); }


 void SgEndchannelStmt::addErrLabel(SgExpression &errlabel)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), errlabel.thellnd); }


 int SgEndchannelStmt::numberOfOutports()
{
  PTR_LLND ll = BIF_LL1(thebif);
  int n = 0;

  while (ll && (n != 2)) {
    if (( NODE_CODE(ll) == IOSTAT_STORE ) || ( NODE_CODE(ll) == ERR_LABEL ))
      n = n + 1;
    ll = NODE_OPERAND1(ll);
  };
  return (exprListLength(BIF_LL1(thebif)) - n);
  // double scanning the list may be improved
}


 SgExpression * SgEndchannelStmt::controls()
{ return LlndMapping(BIF_LL1(thebif)); }


 SgExpression * SgEndchannelStmt::outport(int i)
{ return LlndMapping(getPositionInList(BIF_LL1(thebif),i)); }


 SgExpression * SgEndchannelStmt::ioStore()
{
  int n = numberOfOutports();
  PTR_LLND ll;

  ll = getPositionInList(BIF_LL1(thebif),n);
  if (!ll)
    return (SgExpression *) NULL;
  else
    if (NODE_CODE(ll) != IOSTAT_STORE)
      return (SgExpression *) NULL;
    else
      return  LlndMapping(ll);
}


 SgExpression * SgEndchannelStmt::errLabel()
{
  int n = numberOfOutports();
  PTR_LLND ll;
 
  ll = getPositionInList(BIF_LL1(thebif),n);
  if (!ll)
    return (SgExpression *) NULL;
  else
    if (NODE_CODE(ll) != ERR_LABEL) { // must be IOSTAT_STORE
      ll = NODE_OPERAND1(ll);
      if ((!ll) || (NODE_CODE(ll) != ERR_LABEL))
        return (SgExpression *) NULL;
      else
        return  LlndMapping(ll);
    } else
      return  LlndMapping(ll);
}



// SgProbeStmt--inlines

 SgProbeStmt::SgProbeStmt(SgExpression &inport):SgStatement(PROBE_STAT)
{ BIF_LL1(thebif) = inport.thellnd; }


 SgProbeStmt::SgProbeStmt(SgExpression &inport, SgExpression &e1)
                   :SgStatement(PROBE_STAT)
{
  BIF_LL1(thebif) = inport.thellnd;
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), e1.thellnd);
}


 SgProbeStmt::SgProbeStmt(SgExpression &inport, SgExpression &e1,
                                SgExpression &e2):SgStatement(PROBE_STAT)
{
  BIF_LL1(thebif) = inport.thellnd;
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), e1.thellnd);
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), e2.thellnd);
}


 SgProbeStmt::SgProbeStmt(SgExpression &inport, SgExpression &e1,
                                SgExpression &e2, SgExpression &e3)
                   :SgStatement(PROBE_STAT)
{
  BIF_LL1(thebif) = inport.thellnd;
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), e1.thellnd);
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), e2.thellnd);
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), e3.thellnd);
}
 

 SgProbeStmt::~SgProbeStmt()
{ RemoveFromTableBfnd((void *) this); }


 void SgProbeStmt::addInport(SgExpression &inport)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), inport.thellnd); }
  

 void SgProbeStmt::addIoStore(SgExpression &iostore)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), iostore.thellnd); }


 void SgProbeStmt::addErrLabel(SgExpression &errlabel)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), errlabel.thellnd); }


 void SgProbeStmt::addEmptyStore(SgExpression &emptystore)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), emptystore.thellnd); }
 

 int SgProbeStmt::numberOfInports()
{
  PTR_LLND ll = BIF_LL1(thebif);
  int n = 0;

  while (ll && (n != 3)) {
    if (( NODE_CODE(ll) == IOSTAT_STORE ) || ( NODE_CODE(ll) == ERR_LABEL ) ||
       ( NODE_CODE(ll) == EMPTY_STORE ))
      n = n + 1;
    ll = NODE_OPERAND1(ll);
  };
  return (exprListLength(BIF_LL1(thebif)) - n);
  // double scanning the list may be improved
}


 SgExpression * SgProbeStmt::controls()
{ return LlndMapping(BIF_LL1(thebif)); }


 SgExpression * SgProbeStmt::inport(int i)
{ return LlndMapping(getPositionInList(BIF_LL1(thebif),i)); }


 SgExpression * SgProbeStmt::ioStore()
{
  int n = numberOfInports();
  PTR_LLND ll;

  ll = getPositionInList(BIF_LL1(thebif),n);
  if (!ll)
    return (SgExpression *) NULL;
  else
    if (NODE_CODE(ll) != IOSTAT_STORE)
      return (SgExpression *) NULL;
    else
      return  LlndMapping(ll);
}


 SgExpression * SgProbeStmt::errLabel()
{
  int n = numberOfInports();
  PTR_LLND ll;
 
  ll = getPositionInList(BIF_LL1(thebif),n);
  if (!ll)
    return (SgExpression *) NULL;
  else
    if (NODE_CODE(ll) != ERR_LABEL) { // must be IOSTAT_STORE
      ll = NODE_OPERAND1(ll);
      if ((!ll) || (NODE_CODE(ll) != ERR_LABEL)) // must be EMPTY_STORE
          return (SgExpression *) NULL;
      else
        return  LlndMapping(ll);
    } else
      return  LlndMapping(ll);
}


 SgExpression * SgProbeStmt::emptyStore()
{
  int n = numberOfInports();
  PTR_LLND ll;
 
  ll = getPositionInList(BIF_LL1(thebif),n);
  if (!ll)
    return (SgExpression *) NULL;
  else
    if (NODE_CODE(ll) != EMPTY_STORE) { // must be IOSTAT_STORE or ERR_LABEL
      ll = NODE_OPERAND1(ll);
      if (!ll)
        return (SgExpression *) NULL;
      else
        if (NODE_CODE(ll) != EMPTY_STORE) { // must be ERR_LABEL
          ll = NODE_OPERAND1(ll);
          if ((!ll) || (NODE_CODE(ll) != EMPTY_STORE))
            return (SgExpression *) NULL;
          else
            return  LlndMapping(ll);
        } else
          return  LlndMapping(ll);
    } else
      return  LlndMapping(ll);
}



// SgPortTypeExp--inlines

 SgPortTypeExp::SgPortTypeExp(SgType &type):SgExpression(PORT_TYPE_OP)
{ NODE_TYPE(thellnd) = type.thetype; }


 SgPortTypeExp::SgPortTypeExp(SgType &type, SgExpression &ref)
                     :SgExpression(PORT_TYPE_OP)
{
  NODE_TYPE(thellnd) = type.thetype;
  NODE_OPERAND0(thellnd) = ref.thellnd;
}


 SgPortTypeExp::SgPortTypeExp(int variant, SgExpression &porttype)
                     :SgExpression(variant)
{ NODE_OPERAND0(thellnd) = porttype.thellnd; }


 SgPortTypeExp::~SgPortTypeExp()
{ RemoveFromTableLlnd((void *) this); }


 SgType * SgPortTypeExp::type()
{ return TypeMapping(NODE_TYPE(thellnd)); }

 int SgPortTypeExp::numberOfRef()
{
  PTR_LLND ll = NODE_OPERAND0(thellnd);
  int n = 0;
  while (ll) {
      n = n + 1;
  ll = NODE_OPERAND1(ll);
  };
  return n;
}

 SgExpression * SgPortTypeExp::ref()
{ return LlndMapping(NODE_OPERAND0(thellnd)); }

 SgPortTypeExp * SgPortTypeExp::next()
{ return (SgPortTypeExp *) LlndMapping(NODE_OPERAND1(thellnd)); }


// SgControlExp--inlines
 
 SgControlExp::SgControlExp(int variant):SgExpression(variant)
{}
 
 SgControlExp::~SgControlExp()
{ RemoveFromTableLlnd((void *) this); }
 
 SgExpression * SgControlExp::exp()
{  return LlndMapping(NODE_OPERAND0(thellnd)); }
 
 

// SgInportExp--inlines

 SgInportExp::SgInportExp(SgExprListExp &exp):SgControlExp(INPORT_NAME)
{ NODE_OPERAND0(thellnd) = exp.thellnd; }

 SgInportExp::~SgInportExp()
{ RemoveFromTableLlnd((void *) this); }



// SgOutportExp--inlines
 
 SgOutportExp::SgOutportExp(SgExprListExp &exp):SgControlExp(OUTPORT_NAME)
{ NODE_OPERAND0(thellnd) = exp.thellnd; }
 
 SgOutportExp::~SgOutportExp()
{ RemoveFromTableLlnd((void *) this); }
 
 

// SgFromportExp--inlines
 
 SgFromportExp::SgFromportExp(SgExprListExp &exp)
                     :SgControlExp(FROMPORT_NAME)
{ NODE_OPERAND0(thellnd) = exp.thellnd; }
 
 SgFromportExp::~SgFromportExp()
{ RemoveFromTableLlnd((void *) this); }
 
 

// SgToportExp--inlines
 
 SgToportExp::SgToportExp(SgExprListExp &exp):SgControlExp(TOPORT_NAME)
{ NODE_OPERAND0(thellnd) = exp.thellnd; }
 
 SgToportExp::~SgToportExp()
{ RemoveFromTableLlnd((void *) this); }
 
 

// SgIO_statStoreExp--inlines
 
 SgIO_statStoreExp::SgIO_statStoreExp(SgExprListExp &exp)
                       :SgControlExp(IOSTAT_STORE)
{ NODE_OPERAND0(thellnd) = exp.thellnd; }
 
 SgIO_statStoreExp::~SgIO_statStoreExp()
{ RemoveFromTableLlnd((void *) this); }
 
 

// SgEmptyStoreExp--inlines
 
 SgEmptyStoreExp::SgEmptyStoreExp(SgExprListExp &exp)
                       :SgControlExp(EMPTY_STORE)
{ NODE_OPERAND0(thellnd) = exp.thellnd; }
 
 SgEmptyStoreExp::~SgEmptyStoreExp()
{ RemoveFromTableLlnd((void *) this); }
 
 

// SgErrLabelExp--inlines
 
 SgErrLabelExp::SgErrLabelExp(SgExprListExp &exp):SgControlExp(ERR_LABEL)
{ NODE_OPERAND0(thellnd) = exp.thellnd; }
 
 SgErrLabelExp::~SgErrLabelExp()
{ RemoveFromTableLlnd((void *) this); }
 
 

// SgEndLabelExp--inlines
 
 SgEndLabelExp::SgEndLabelExp(SgExprListExp &exp):SgControlExp(END_LABEL)
{ NODE_OPERAND0(thellnd) = exp.thellnd; }
 
 SgEndLabelExp::~SgEndLabelExp()
{ RemoveFromTableLlnd((void *) this); }
 
 

// SgDataImpliedDoExp--inlines

 SgDataImpliedDoExp::SgDataImpliedDoExp(SgExprListExp &dlist,
                                              SgSymbol &iname,
                                              SgExprListExp &ilist)
                          :SgExpression(DATA_IMPL_DO)
{
  NODE_OPERAND0(thellnd) = dlist.thellnd;
  NODE_SYMB(thellnd) = iname.thesymb;
  NODE_OPERAND1(thellnd) = ilist.thellnd;
}

 SgDataImpliedDoExp::~SgDataImpliedDoExp()
{ RemoveFromTableLlnd((void *) this); }

 void SgDataImpliedDoExp::addDataelt(SgExpression &data)
{ NODE_OPERAND0(thellnd) = addToList(NODE_OPERAND0(thellnd),data.thellnd); }

 void SgDataImpliedDoExp::addIconexpr(SgExpression &icon)
{ NODE_OPERAND1(thellnd) = addToList(NODE_OPERAND1(thellnd),icon.thellnd); }

 SgSymbol *SgDataImpliedDoExp::iname()
{ return SymbMapping(NODE_SYMB(thellnd)); }

 int SgDataImpliedDoExp::numberOfDataelt()
{ return exprListLength(NODE_OPERAND0(thellnd)); }

 SgExprListExp *SgDataImpliedDoExp::dataelts()
{ return (SgExprListExp *) LlndMapping(NODE_OPERAND0(thellnd)); }

 SgExpression *SgDataImpliedDoExp::dataelt(int i)
{ return LlndMapping(getPositionInList(NODE_OPERAND0(thellnd),i));  }

 SgExprListExp *SgDataImpliedDoExp::iconexprs()
{ return (SgExprListExp *) LlndMapping(NODE_OPERAND1(thellnd)); }

 SgExpression *SgDataImpliedDoExp::init()
{ return LlndMapping(getPositionInList(NODE_OPERAND1(thellnd),0));  }

 SgExpression *SgDataImpliedDoExp::limit()
{ return LlndMapping(getPositionInList(NODE_OPERAND1(thellnd),1));  }

 SgExpression *SgDataImpliedDoExp::increment()
{ return LlndMapping(getPositionInList(NODE_OPERAND1(thellnd),2));  }



// SgDataEltExp--inlines

 SgDataEltExp::SgDataEltExp(SgExpression &dataimplieddo)
                    :SgExpression(DATA_ELT)
{ NODE_OPERAND0(thellnd) = dataimplieddo.thellnd; }

 SgDataEltExp::SgDataEltExp(SgSymbol &name, SgExpression &datasubs,
                                  SgExpression &datarange)
                    :SgExpression(DATA_ELT)
{
  NODE_SYMB(thellnd) = name.thesymb;
  NODE_OPERAND1(datasubs.thellnd) = datarange.thellnd;
  NODE_OPERAND0(thellnd) = datasubs.thellnd;
}

 SgDataEltExp::~SgDataEltExp()
{ RemoveFromTableLlnd((void *) this); }

 SgSymbol *SgDataEltExp::name()
{ return SymbMapping(NODE_SYMB(thellnd)); }

 SgExpression *SgDataEltExp::dataimplieddo()
{
  if (NODE_SYMB(thellnd) == NULL)
    return LlndMapping(NODE_OPERAND0(thellnd));
  else
    return NULL;
}

 SgExpression *SgDataEltExp::datasubs()
{
  if (NODE_SYMB(thellnd) != NULL)
    if (NODE_CODE(NODE_OPERAND0(thellnd)) == DATA_SUBS)
      return LlndMapping(NODE_OPERAND0(thellnd));
    else
      return (SgExpression *) NULL;
  else
    return (SgExpression *) NULL;
}

 SgExpression *SgDataEltExp::datarange()
{
  if (NODE_SYMB(thellnd) != NULL)
    if (NODE_CODE(NODE_OPERAND0(thellnd)) == DATA_RANGE)
      return LlndMapping(NODE_OPERAND0(thellnd));
    else
      if (NODE_OPERAND1(NODE_OPERAND0(thellnd)) != NULL)
        return LlndMapping(NODE_OPERAND1(NODE_OPERAND0(thellnd)));
      else
        return (SgExpression *) NULL;
  else
    return (SgExpression *) NULL;
}



// SgDataSubsExp--inlines

 SgDataSubsExp::SgDataSubsExp(SgExprListExp &iconexprlist)
                     :SgExpression(DATA_SUBS)
{ NODE_OPERAND0(thellnd) = iconexprlist.thellnd; }

 SgDataSubsExp::~SgDataSubsExp()
{ RemoveFromTableLlnd((void *) this); }

 SgExprListExp *SgDataSubsExp::iconexprlist()
{ return (SgExprListExp *) LlndMapping(NODE_OPERAND0(thellnd)); }



// SgDataRangeExp--inlines

 SgDataRangeExp::SgDataRangeExp(SgExpression &iconexpr1,
                                      SgExpression &iconexpr2)
                      :SgExpression(DATA_RANGE)
{
  NODE_OPERAND0(thellnd) = iconexpr1.thellnd;
  NODE_OPERAND1(thellnd) = iconexpr2.thellnd;
}

 SgDataRangeExp::~SgDataRangeExp()
{ RemoveFromTableLlnd((void *) this); }

 SgExpression *SgDataRangeExp::iconexpr1()
{ return LlndMapping(NODE_OPERAND0(thellnd)); }

 SgExpression *SgDataRangeExp::iconexpr2()
{ return LlndMapping(NODE_OPERAND1(thellnd)); }

// SgIconExprExp--inlines

 SgIconExprExp::SgIconExprExp(SgExpression &exp):SgExpression(ICON_EXPR)
{ NODE_OPERAND0(thellnd) = exp.thellnd; }

 SgIconExprExp::~SgIconExprExp()
{ RemoveFromTableLlnd((void *) this); }

 SgExpression *SgIconExprExp::expr()
{ return LlndMapping(NODE_OPERAND0(thellnd)); }

//SgImpliedDoExp;

SgImpliedDoExp::SgImpliedDoExp():SgExpression(IMPLIED_DO)
{
  // not finished;
}

SgSymbol * SgImpliedDoExp::index()
{
  return (this->symbol());
}

int SgImpliedDoExp::numberOfTriplet()
{
  SgTripletExp * triplet = (SgTripletExp *)this->operand(2);
  return (triplet->numberOfItems());
}

SgExpression * SgImpliedDoExp::triplet(int i)
{
  SgTripletExp * triplet = (SgTripletExp *)this->operand(2);
  return (triplet->item(i));
}

SgExpression * SgImpliedDoExp::list()
{
  return LlndMapping(NODE_OPERAND0(thellnd));
}

SgImpliedDoExp::~SgImpliedDoExp()
{ RemoveFromTableLlnd((void *) this); }

// SgIOStmt--inlines
 SgIOStmt::SgIOStmt(int variant):SgExecutableStatement(variant)
{}


// SgInputOutputStmt--inlines
         
 SgInputOutputStmt::SgInputOutputStmt(int variant, SgExpression &specList, SgExpression &itemList): SgIOStmt(variant) 
{
  if (variant != READ_STAT && variant != WRITE_STAT && variant != PRINT_STAT)
    Message("illegal variant for SgInputOutputStmt",0);
  BIF_LL1(thebif) = specList.thellnd;
  BIF_LL2(thebif) = itemList.thellnd;
}
        
 SgExpression * SgInputOutputStmt::specList()
{ return LlndMapping(BIF_LL1(thebif)); }

 void SgInputOutputStmt::setSpecList(SgExpression &specList)
{ BIF_LL1(thebif) = specList.thellnd; }

 SgExpression * SgInputOutputStmt::itemList()
{ return LlndMapping(BIF_LL2(thebif)); }

 void SgInputOutputStmt::setItemList(SgExpression &itemList)
{ BIF_LL2(thebif) = itemList.thellnd; }

 SgInputOutputStmt::~SgInputOutputStmt()
{ RemoveFromTableBfnd((void *) this); }

// SgIOControlStmt--inlines

 SgExpression * SgIOControlStmt::controlSpecList()
{ return LlndMapping(BIF_LL2(thebif)); }

 void SgIOControlStmt::setControlSpecList(SgExpression &controlSpecList)
{ BIF_LL2(thebif) = controlSpecList.thellnd; }

 SgIOControlStmt::~SgIOControlStmt()
{ RemoveFromTableBfnd((void *) this); }


// SgDeclarationStatement--inlines
 SgDeclarationStatement::SgDeclarationStatement(int variant):SgStatement(variant)
{}

 SgDeclarationStatement::~SgDeclarationStatement()
{ RemoveFromTableBfnd((void *) this); }

 SgExpression * SgDeclarationStatement::varList()
{ return LlndMapping(BIF_LL1(thebif)); }

 int SgDeclarationStatement::numberOfVars()
{ return exprListLength(BIF_LL1(thebif)); }

 SgExpression * SgDeclarationStatement::var(int i)
{ 
  PTR_LLND  expr;
  expr=getPositionInExprList(BIF_LL1(thebif),i); 
  switch (expr->variant)
  {
  case ASSGN_OP:  expr=NODE_OPERAND0(expr); break;
  }
  return LlndMapping(expr);
//{ return LlndMapping(getPositionInExprList(BIF_LL1(thebif),i)); }
}

 void SgDeclarationStatement::deleteVar(int i)
{ BIF_LL1(thebif) = deleteNodeInExprList(BIF_LL1(thebif), i); }

 void SgDeclarationStatement::deleteTheVar(SgExpression &var)
{
 BIF_LL1(thebif) = deleteNodeWithItemInExprList(BIF_LL1(thebif),var.thellnd); 
}

 void SgDeclarationStatement::addVar(SgExpression &exp)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), exp.thellnd); }



// SgVarDeclStmt--inlines

 SgVarDeclStmt::SgVarDeclStmt(SgExpression &varRefValList, SgExpression &attributeList, SgType &type):SgDeclarationStatement(VAR_DECL)
{
  if ( CurrentProject->Fortranlanguage() )
    {
      BIF_LL1(thebif) = varRefValList.thellnd;
      BIF_LL2(thebif) = (PTR_LLND) newNode(TYPE_OP);
      NODE_TYPE(BIF_LL2(thebif)) =  type.thetype;
      BIF_LL3(thebif) = attributeList.thellnd;
    }
  else /* C or C++ */
    {
      BIF_LL1(thebif) = varRefValList.thellnd;
      NODE_TYPE(BIF_LL1(thebif)) = type.thetype;
    }
}

SgVarDeclStmt::SgVarDeclStmt(SgExpression &varRefValList, SgType &type):SgDeclarationStatement(VAR_DECL)
{
  // this seems unnatrual    
  if ( CurrentProject->Fortranlanguage ())
    {    
      this->addVar(varRefValList);

      SgExpression * texp = new SgExpression(TYPE_OP, NULL, NULL, NULL, &type);
      
      ((SgStatement *)this)-> setExpression (1, *texp);
    }

    /*  {
      BIF_LL1(thebif) = varRefValList.thellnd;
      BIF_LL2(thebif) = newExpr(TYPE_OP,type.thetype);
      BIF_LL3(thebif) = LLNULL;
    }
    */
  else // C or C++ 
    {
      BIF_LL1(thebif) = varRefValList.thellnd;
      NODE_TYPE(BIF_LL1(thebif)) = type.thetype;
    }
}

 SgVarDeclStmt::SgVarDeclStmt(SgExpression &varRefValList)
     :SgDeclarationStatement(VAR_DECL)
{
  if ( CurrentProject->Fortranlanguage ())
    {
      BIF_LL1(thebif) = varRefValList.thellnd;
      BIF_LL2(thebif) = LLNULL;
      BIF_LL3(thebif) = LLNULL;
    }
  else /* C or C++ */
    {
      BIF_LL1(thebif) = varRefValList.thellnd;
      NODE_TYPE(BIF_LL1(thebif)) = TYNULL;
    }
}

 SgVarDeclStmt::~SgVarDeclStmt()
{ RemoveFromTableBfnd((void *) this); }

 SgType * SgVarDeclStmt::type()  // the type
{ 
  SgType *x;

  if ( CurrentProject->Fortranlanguage() )
    {
      if (BIF_LL2(thebif))
	x = TypeMapping(NODE_TYPE(BIF_LL2(thebif))); 
      else
	x = NULL;
    }
  else /* C or C++ */
    {
       if (BIF_LL1(thebif))
	    x = TypeMapping(NODE_TYPE(BIF_LL1(thebif))); 
       else
	    x = NULL;
    }
  return x;
}


// the number of F90 attributes
 int SgVarDeclStmt::numberOfAttributes()
{ return exprListLength(BIF_LL3(thebif)); }

// the number of variables declared
 int SgVarDeclStmt::numberOfSymbols()
{ return exprListLength(BIF_LL1(thebif)); }

 SgSymbol * SgVarDeclStmt::symbol(int i)
{
  PTR_LLND  pt;
  PTR_SYMB symb = NULL;
  SgSymbol *x;

  pt = getPositionInExprList(BIF_LL1(thebif),i);
  if (pt)
    pt = giveLlSymbInDeclList(pt);
  if (pt && (symb= NODE_SYMB(pt)))
    {
      x = SymbMapping(symb);
    }
  else
    x = NULL;

  return x;
}
     
 void SgVarDeclStmt::deleteSymbol(int i)
{ BIF_LL1(thebif) = deleteNodeInExprList(BIF_LL1(thebif),i); 
  // delete the statement here !!!            add by zheng.gb
  if ( BIF_LL1(thebif) == LLNULL )
	deleteStmt();
}

 void SgVarDeclStmt::deleteTheSymbol(SgSymbol &symbol)
{
	PTR_LLND expr_list = BIF_LL1(thebif);
	int i;
	for (i=0 ; expr_list ; expr_list=NODE_OPERAND1(expr_list),i++)
		if (NODE_SYMB(NODE_OPERAND0(expr_list)) == symbol.thesymb)
		{
			deleteSymbol(i);
			return ;
		}
}

// the initial value ofthe i-th variable
 SgExpression * SgVarDeclStmt::initialValue(int i)
{ 
  PTR_LLND varRefExp;
  SgExpression *x;

  varRefExp = getPositionInExprList(BIF_LL1(thebif),i);
  if (varRefExp == LLNULL)
    x = NULL;
  else if (NODE_CODE(varRefExp) == ASSGN_OP)
    x = LlndMapping(NODE_OPERAND1(varRefExp));
  else 
    x = NULL;

  return x;
}


// SgIntentStmt--inlines

 SgIntentStmt::SgIntentStmt(SgExpression &varRefValList,
                                  SgExpression &attribute)
                    :SgDeclarationStatement(INTENT_STMT)
{
  BIF_LL1(thebif) = varRefValList.thellnd;
  BIF_LL2(thebif) = attribute.thellnd;
}

 SgIntentStmt::~SgIntentStmt()
{ RemoveFromTableBfnd((void *) this); }

 int SgIntentStmt::numberOfArgs()  // the number of arguement expressions
{ return exprListLength(BIF_LL1(thebif)); }

 void SgIntentStmt::addArg(SgExpression &arg)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif),arg.thellnd); }

 SgExpression * SgIntentStmt::args()
{ return LlndMapping(BIF_LL1(thebif)); }

 SgExpression * SgIntentStmt::arg(int i) // the i-th argument expression
{ return LlndMapping(getPositionInExprList(BIF_LL1(thebif), i)); }

 SgExpression * SgIntentStmt::attribute()
{ return LlndMapping(BIF_LL2(thebif)); }


// SgVarListDeclStmt--inlines

 SgVarListDeclStmt::~SgVarListDeclStmt()
{ RemoveFromTableBfnd((void *) this); }

// the number of variables declared
 int SgVarListDeclStmt::numberOfSymbols()
{ return exprListLength(BIF_LL1(thebif)); }

 SgSymbol * SgVarListDeclStmt::symbol(int i)   // the i-th variable
{ 
  PTR_LLND pt;
  SgSymbol *x;

  if (pt = getPositionInExprList(BIF_LL1(thebif),i))
    x = SymbMapping(NODE_SYMB(pt)); 
  else
    x = NULL;

  return x;
}

 void SgVarListDeclStmt::appendSymbol(SgSymbol &symbol)
{ 
  BIF_LL1(thebif) =  addSymbRefToExprList(BIF_LL1(thebif), symbol.thesymb); 
}

 void SgVarListDeclStmt::deleteSymbol(int i)
{ BIF_LL1(thebif) = deleteNodeInExprList(BIF_LL1(thebif), i); }

#ifdef NOT_YET_IMPLEMENTED
 void SgVarListDeclStmt::deleteTheSymbol(SgSymbol &symbol)
{ SORRY; }
#endif


// SgStructureDeclStmt--inlines

 SgStructureDeclStmt::SgStructureDeclStmt(SgSymbol &name, SgExpression &attributes, SgStatement &body):SgDeclarationStatement(STRUCT_DECL)
{
  BIF_SYMB(thebif) = name.thesymb;
  BIF_LL1(thebif) = attributes.thellnd;
  insertBfndListIn(body.thebif,thebif,thebif);             
}

 SgStructureDeclStmt::~SgStructureDeclStmt()
{ RemoveFromTableBfnd((void *) this); }


// SgNestedVarListDeclStmt--inlines

       
// varList must be of low-level variant appropriate to variant. For example,
// if the variant is COMM_STAT, listOfVarList must be of variant COMM_LIST.

 SgNestedVarListDeclStmt::~SgNestedVarListDeclStmt()
{ RemoveFromTableBfnd((void *) this); }
 
 SgExpression * SgNestedVarListDeclStmt::lists()
{ return LlndMapping(BIF_LL1(thebif)); }

 int SgNestedVarListDeclStmt::numberOfLists()
{ return exprListLength(BIF_LL1(thebif)); }

 SgExpression * SgNestedVarListDeclStmt::list(int i)
{ return LlndMapping(getPositionInExprList( BIF_LL1(thebif),i)); }

 void SgNestedVarListDeclStmt::addList(SgExpression &list)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), list.thellnd); }

 void SgNestedVarListDeclStmt::addVarToList(SgExpression &varRef)
{ 
  BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif),varRef.thellnd);
}

 void SgNestedVarListDeclStmt::deleteList(int i)
{ 
  BIF_LL1(thebif) = deleteNodeInExprList(BIF_LL1(thebif), i);
}

#ifdef NOT_YET_IMPLEMENTED
 void SgNestedVarListDeclStmt::deleteTheList(SgExpression &list)
{ 
  //            deleteNodeWithItemInExprList(BIF_LL1(thebif), list.thellnd);
  SORRY;
}
#endif

#ifdef NOT_YET_IMPLEMENTED
 void SgNestedVarListDeclStmt::deleteVarInList(int i, SgExpression &varRef)
{ 
  SORRY;
}
#endif

#ifdef NOT_YET_IMPLEMENTED
 void SgNestedVarListDeclStmt::deleteVarInTheList(SgExpression &list, SgExpression &varRef)
{ 
  SORRY;
}
#endif


// SgParameterStmt--inlines

 SgParameterStmt::SgParameterStmt(SgExpression &constants, SgExpression &values):SgDeclarationStatement(PARAM_DECL)
{
  PTR_SYMB sym;

// modified by zheng.gb
	sym=constants.thellnd->entry.Template.symbol;
	sym->entry.Template.const_value=values.thellnd;
	sym->variant=CONST_NAME;
	constants.thellnd->variant=CONST_REF;
        BIF_LL1(thebif) = newExpr(EXPR_LIST,NULL,constants.thellnd,NULL);

  //  PTR_LLND constantWithValues;
  
  //            constantWithValues = stringConstantsWithTheirValues(constants.thellnd, values.thellnd);
  //            BIF_LL1(thebif) = LlndMapping(constantWithValues);
  //SORRY;
}

 SgParameterStmt::SgParameterStmt(SgExpression &constantsWithValues):SgDeclarationStatement(PARAM_DECL)
{ BIF_LL1(thebif) = constantsWithValues.thellnd; }

 SgParameterStmt::~SgParameterStmt()
{ RemoveFromTableBfnd((void *) this); }

// the number of constants declared
 int SgParameterStmt::numberOfConstants()
{ return exprListLength(BIF_LL1(thebif)); }

// the i-th variable
 SgSymbol * SgParameterStmt::constant(int i)
{ return SymbMapping(NODE_SYMB(getPositionInExprList(BIF_LL1(thebif),i))); }

// the value of i-th variable
 SgExpression * SgParameterStmt::value(int i)
{ return LlndMapping(SYMB_VAL(NODE_SYMB(getPositionInExprList(BIF_LL1(thebif),i)))); }

#ifdef NOT_YET_IMPLEMENTED
 void SgParameterStmt::addConstant(SgSymbol &constant, SgExpression &value)

{ 
  //            SgRefExp constNode(constant, value);
  //            BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), constNode.thellnd);
  SORRY;
};
#endif
  
 void SgParameterStmt::deleteConstant(int i)
{ BIF_LL1(thebif) = deleteNodeInExprList(BIF_LL1(thebif), i); }

#ifdef NOT_YET_IMPLEMENTED
 void SgParameterStmt::deleteTheConstant(SgSymbol &constant)
{ 
  // deleteNodeWithSymbolInExprList(i, BIF_LL1(thebif)); 
  SORRY;
}
#endif


// SgImplicitStmt--inlines

 SgImplicitStmt::SgImplicitStmt(SgExpression &implicitLists):SgDeclarationStatement(IMPL_DECL)
{ BIF_LL1(thebif) = implicitLists.thellnd; }

 SgImplicitStmt::~SgImplicitStmt()
{ RemoveFromTableBfnd((void *) this); }

// the number of implicit types declared
 int SgImplicitStmt::numberOfImplicitTypes()
{ return exprListLength(BIF_LL1(thebif)); }

// the i-th implicit type
 SgType * SgImplicitStmt::implicitType(int i)
{ 
  PTR_LLND pt;
  SgType *x;

  if ( (pt = getPositionInList(BIF_LL1(thebif),i)) &&
      NODE_OPERAND0(pt))
    x = TypeMapping(NODE_TYPE(NODE_OPERAND0(pt))); 
  else
    x = NULL;

  return x;
}

// the i-th implicit type's range list eg. (A-E, G)
 SgExpression * SgImplicitStmt::implicitRangeList(int i)
{ 
  PTR_LLND pt;
  SgExpression *x;

  if ( (pt = getPositionInExprList(BIF_LL1(thebif),i)) &&
      NODE_OPERAND0(pt))
    x = LlndMapping(NODE_OPERAND0(pt)); 
  else
    x = NULL;

  return x;
}

 void SgImplicitStmt::appendImplicitNode(SgExpression &impNode)
{ BIF_LL1(thebif) = addToExprList(BIF_LL1(thebif), impNode.thellnd); }



// SgVariableSymb--inlines


 SgVariableSymb::SgVariableSymb(char *identifier, SgType &t, SgStatement &scope):SgSymbol(VARIABLE_NAME,identifier)
{
  SYMB_SCOPE(thesymb) = scope.thebif;
  SYMB_TYPE(thesymb) = t.thetype;
}

 SgVariableSymb::SgVariableSymb(char *identifier, SgType *t, SgStatement *scope):SgSymbol(VARIABLE_NAME,identifier)
{
  if (scope)
    SYMB_SCOPE(thesymb) = scope->thebif;
  if (t)
    SYMB_TYPE(thesymb) = t->thetype;
}

 SgVariableSymb::SgVariableSymb(char *identifier, 
                                SgType &t):SgSymbol(VARIABLE_NAME,identifier)
{ SYMB_TYPE(thesymb) = t.thetype; }


 SgVariableSymb::SgVariableSymb(char *identifier,
                         SgStatement &scope):SgSymbol(VARIABLE_NAME,identifier)
{    SYMB_SCOPE(thesymb) = scope.thebif;}


 SgVariableSymb::SgVariableSymb(char *identifier,
                         SgStatement *scope):SgSymbol(VARIABLE_NAME,identifier)
{    SYMB_SCOPE(thesymb) = scope->thebif;}


 SgVariableSymb::SgVariableSymb(char *identifier):
                                   SgSymbol(VARIABLE_NAME,identifier)
{}

 SgVariableSymb::~SgVariableSymb()
{ RemoveFromTableSymb((void *) this); }

/* ajm */
 SgVarRefExp *SgVariableSymb::varRef(void)
{
     return new SgVarRefExp (*this);
}


// SgConstantSymb--inlines

 SgConstantSymb::SgConstantSymb(char *identifier, SgStatement &scope, 
                                   SgExpression &value):SgSymbol(CONST_NAME,identifier, scope)
{ SYMB_VAL(thesymb) = value.thellnd; }

 SgConstantSymb::~SgConstantSymb()
{ RemoveFromTableSymb((void *) this); }

 SgExpression * SgConstantSymb::constantValue()
{ return LlndMapping(SYMB_VAL(thesymb)); }

/* zheng.gb */
 void SgConstantSymb::setConstantValue(SgExpression *value)
{ SYMB_VAL(thesymb)=value?value->thellnd:LLNULL; }


// SgFunctionSymb--inlines

 SgFunctionSymb::~SgFunctionSymb()
{ RemoveFromTableSymb((void *) this); }

 void SgFunctionSymb::addParameter(int, SgSymbol &parameters)
{ 
     SgSymbol *copy_symb = &(parameters.copy());
     SYMB_NEXT_DECL(copy_symb->thesymb) = 0;
     appendSymbToArgList (thesymb,copy_symb->thesymb);
}

 void SgFunctionSymb::insertParameter(int position, SgSymbol &symb)
{ insertSymbInArgList (this->thesymb, position, symb.thesymb);  }

 int SgFunctionSymb::numberOfParameters()
{ return lenghtOfParamList(thesymb); }

 SgSymbol * SgFunctionSymb::parameter(int i)
{ return SymbMapping(GetThParam(thesymb,i)); }

 SgSymbol * SgFunctionSymb::result()
{ return SymbMapping(SYMB_DECLARED_NAME(thesymb)); }

 void SgFunctionSymb::setResult(SgSymbol &symbol)
{ SYMB_DECLARED_NAME(thesymb) = symbol.thesymb; }

 SgStatement *SgFunctionSymb::function_hedr()
{ return  BfndMapping(thesymb->entry.func_decl.func_hedr); }

// SgMemberFuncSymb--inlines
//  status = MEMB_<see above>;
 SgMemberFuncSymb::SgMemberFuncSymb(char *identifier, SgType &t,
                                        SgStatement &cla, int status):
                       SgFunctionSymb(MEMBER_FUNC, identifier, t, cla)
{
  SYMB_ATTR(thesymb) = status;
  SYMB_MEMBER_BASENAME(thesymb) = BIF_SYMB(cla.thebif);
}

 SgMemberFuncSymb::~SgMemberFuncSymb()
{ RemoveFromTableSymb((void *) this); }

 int SgMemberFuncSymb::isMethodOfElement()
{
  int x;
  if ((int) SYMB_ATTR(thesymb) & (int) ELEMENT_FIELD)
    x = TRUE;
  else
    x = FALSE;

  return x;
}

// name of containing class;
 SgSymbol * SgMemberFuncSymb::className()
{
  return SymbMapping(SYMB_MEMBER_BASENAME(thesymb));
}

// name of containing class
 void SgMemberFuncSymb::setClassName(SgSymbol &symb)
{
  SYMB_MEMBER_BASENAME(thesymb) = symb.thesymb;
}
          

// SgFieldSymb--inlines

 SgFieldSymb::SgFieldSymb(char *identifier, SgType &t, 
        SgSymbol &structureName):SgSymbol(FIELD_NAME,identifier)
{
      SYMB_TYPE(thesymb) = t.thetype; 
      SYMB_FIELD_BASENAME(thesymb) = structureName.thesymb; 
}

 SgFieldSymb::~SgFieldSymb()
{ RemoveFromTableSymb((void *) this); }

// position in the structure
#ifdef NOT_YET_IMPLEMENTED
 int SgFieldSymb::offset()
{
  // return positionOfFieldInStruct(thesymb, SYMB_BASE_NAME(thesymb));
  SORRY;
  return 0;
}
#endif

// parent structure
 SgSymbol * SgFieldSymb::structureName()
{ return SymbMapping(SYMB_FIELD_BASENAME(thesymb)); }

 SgSymbol * SgFieldSymb::nextField()
{ return SymbMapping(getClassNextFieldOrMember(thesymb)); }

 int SgFieldSymb::isMethodOfElement()
{
  int x;

  if ((int) SYMB_ATTR(thesymb) & (int) ELEMENT_FIELD)
    x = TRUE;
  else
    x = FALSE;

  return x;
}


// SgClassSymb--inlines

 SgClassSymb::SgClassSymb(int variant, char *name, 
         SgStatement &scope):SgSymbol(variant, name, scope)
{}

 SgClassSymb::~SgClassSymb()
{ RemoveFromTableSymb((void *) this); }

// number of fields and member functions.
 int SgClassSymb::numberOfFields()
{ return lenghtOfFieldList(thesymb);}

// the i-th field or member function.
 SgSymbol * SgClassSymb::field(int i)
{ return SymbMapping(GetThOfFieldList(thesymb,i)); }


// SgLabelSymb--inlines

#ifdef NOT_YET_IMPLEMENTED
 SgLabelSymb::SgLabelSymb(char *name):SgSymbol(LABEL_NAME)
{
  SORRY;
}
#endif

 SgLabelSymb::~SgLabelSymb()
{ RemoveFromTableSymb((void *) this); }


 SgLabelVarSymb::SgLabelVarSymb(char *name, SgStatement &scope):SgSymbol(LABEL_NAME, name, scope)
{}
  
 SgLabelVarSymb::~SgLabelVarSymb()
{ RemoveFromTableSymb((void *) this); }


// SgExternalSymb--inlines
 SgExternalSymb::SgExternalSymb(char *name, SgStatement &scope):SgSymbol(ROUTINE_NAME, name, scope)
{}

 SgExternalSymb::~SgExternalSymb()
{ RemoveFromTableSymb((void *) this); }


// SgConstructSymb--inlines

 SgConstructSymb::SgConstructSymb(char *name, SgStatement &scope):SgSymbol(CONSTRUCT_NAME, name, scope)
{}
        
 SgConstructSymb::~SgConstructSymb()
{ RemoveFromTableSymb((void *) this); }


// SgInterfaceSymb--inlines

 SgInterfaceSymb::SgInterfaceSymb(char *name, SgStatement &scope):SgSymbol(INTERFACE_NAME, name, scope)
{}

 SgInterfaceSymb::~SgInterfaceSymb()
{ RemoveFromTableSymb((void *) this); }


// SgModuleSymb--inlines
 SgModuleSymb::SgModuleSymb(char *name):SgSymbol(MODULE_NAME, name, *BfndMapping(getFirstStmt()))
{}

 SgModuleSymb::~SgModuleSymb()
{ RemoveFromTableSymb((void *) this); }


// SgArrayType--inlines

 SgArrayType::SgArrayType(SgType &base_type):SgType(T_ARRAY)
{ TYPE_BASE(thetype) = base_type.thetype; }

 SgArrayType::SgArrayType(SgType &base_type,int dim, SgExpression &ranges):SgType(T_ARRAY)
{ TYPE_BASE(thetype) = base_type.thetype; 
  TYPE_RANGES(thetype) = ranges.thellnd;
  TYPE_DIM(thetype) = dim;
}

 int SgArrayType::dimension()
{ return exprListLength(TYPE_RANGES(thetype)); }

 SgExpression * SgArrayType::sizeInDim(int i)
{ return LlndMapping(getPositionInExprList(TYPE_RANGES(thetype),i)); }

 SgType * SgArrayType::baseType()
{
 return TypeMapping(lookForInternalBasetype(thetype));
 // perhaps should be return TYPE_BASE(thetype);
}

 void SgArrayType::setBaseType(SgType &bt)
{ TYPE_BASE(thetype) = bt.thetype; }

 void SgArrayType::addDimension(SgExpression *e)
{
 if(!e){
  SgExprListExp *l = new SgExprListExp();
  TYPE_RANGES(thetype) = l->thellnd;
  }
 else
  TYPE_RANGES(thetype) = addToExprList(TYPE_RANGES(thetype),e->thellnd);
}
 SgExpression * SgArrayType::getDimList()
{
  return LlndMapping(TYPE_RANGES(thetype));
}
 void SgArrayType::addRange(SgExpression &e)
{
  TYPE_RANGES(thetype) = addToExprList(TYPE_RANGES(thetype),e.thellnd);
  // For C when adding range adding one level of pointer in basetype.
  // This routine should only be used to build a dereferencing expression
  // like x[i][j] and not a declaration.  use addDimension for that.
    if (!CurrentProject->Fortranlanguage())
      {
        PTR_TYPE type;
        type = (PTR_TYPE) newNode(T_POINTER);
        TYPE_BASE(type) = TYPE_BASE(thetype);
        TYPE_BASE(thetype) = type;
      }
}

 SgArrayType::~SgArrayType()
{ RemoveFromTableType((void *) this); }


// SgPointerType--inlines

 SgType * SgPointerType::baseType()
{ return TypeMapping(TYPE_BASE(thetype)); }

 void SgPointerType::setBaseType(SgType &baseType)
{ TYPE_BASE(thetype) = baseType.thetype; }

 int SgPointerType::indirection()
{ return TYPE_TEMPLATE_DUMMY1(thetype); }

 void SgPointerType::setIndirection(int i)
{ TYPE_TEMPLATE_DUMMY1(thetype) = i; }

 SgPointerType::~SgPointerType()
{ RemoveFromTableType((void *) this); }

 int SgPointerType::modifierFlag()
{ return TYPE_TEMPLATE_DUMMY5(thetype); }

 void SgPointerType::setModifierFlag(int flag) 
{ TYPE_TEMPLATE_DUMMY5(thetype) = TYPE_TEMPLATE_DUMMY5(thetype) | flag; }


// SgFunctionType-- inlines

 SgFunctionType::SgFunctionType(SgType &ret_val):SgType(T_FUNCTION)
{ TYPE_BASE(thetype) = ret_val.thetype; }

 SgType * SgFunctionType::returnedValue()
{ return TypeMapping(TYPE_BASE(thetype)); }

 void SgFunctionType::changeReturnedValue(SgType &ret_val)
{ TYPE_BASE(thetype) = ret_val.thetype; }

 SgFunctionType::~SgFunctionType()
{ RemoveFromTableType((void *) this); }

// SgReferenceType--inlines

 SgReferenceType::SgReferenceType(SgType &base_type):SgType(T_REFERENCE)
{ TYPE_BASE(thetype) = base_type.thetype; }

 SgType * SgReferenceType::baseType()
{ return TypeMapping(TYPE_BASE(thetype)); }

 void SgReferenceType::setBaseType(SgType &baseType)
{ TYPE_BASE(thetype) = baseType.thetype; }

 SgReferenceType::~SgReferenceType()
{ RemoveFromTableType((void *) this); }

 int SgReferenceType::modifierFlag()
{ return TYPE_TEMPLATE_DUMMY5(thetype); }

 void SgReferenceType::setModifierFlag(int flag) 
{ TYPE_TEMPLATE_DUMMY5(thetype) = TYPE_TEMPLATE_DUMMY5(thetype) | flag; }


// SgDerivedType--inlines

 SgDerivedType::SgDerivedType(SgSymbol &type_name):SgType(T_DERIVED_TYPE)
{ TYPE_SYMB_DERIVE(thetype) = type_name.thesymb; }

 SgSymbol * SgDerivedType::typeName()
{ return SymbMapping(TYPE_SYMB_DERIVE(thetype)); }

 SgDerivedType::~SgDerivedType()
{ RemoveFromTableType((void *) this); }


// SgDerivedClassType--inlines

 SgDerivedClassType::SgDerivedClassType(SgSymbol &type_name):SgType(T_DERIVED_CLASS)
{ TYPE_SYMB_DERIVE(thetype) = type_name.thesymb; }

 SgSymbol * SgDerivedClassType::typeName()
{ return SymbMapping(TYPE_SYMB_DERIVE(thetype)); }

 SgDerivedClassType::~SgDerivedClassType()
{ RemoveFromTableType((void *) this); }


// SgDescriptType--inlines


 SgDescriptType::SgDescriptType(SgType &base_type, int bit_flag):SgType(T_DESCRIPT)
{
  TYPE_LONG_SHORT(thetype) =  bit_flag;
  TYPE_BASE(thetype) = base_type.thetype;
}

 int SgDescriptType::modifierFlag()
{ return TYPE_LONG_SHORT(thetype); }

 void SgDescriptType::setModifierFlag(int flag) 
{ TYPE_LONG_SHORT(thetype) = TYPE_LONG_SHORT(thetype) | flag; }

 SgDescriptType::~SgDescriptType()
{ RemoveFromTableType((void *) this); }



// SgDerivedCollectionType--inlines

 SgDerivedCollectionType::SgDerivedCollectionType(SgSymbol &s, SgType &t):SgType(T_DERIVED_COLLECTION)
{
  TYPE_COLL_BASE(thetype) = t.thetype;
  TYPE_SYMB_DERIVE(thetype) = s.thesymb;
}

 SgType * SgDerivedCollectionType::elementClass() 
{ return TypeMapping(TYPE_COLL_BASE(thetype)); }

 void SgDerivedCollectionType::setElementClass(SgType &ty)
{ TYPE_COLL_BASE(thetype) = ty.thetype; }

 SgSymbol * SgDerivedCollectionType::collectionName()
{  return SymbMapping(TYPE_SYMB_DERIVE(thetype)); }

 SgStatement * SgDerivedCollectionType::createCollectionWithElemType()
{
  return BfndMapping(LibcreateCollectionWithType(thetype,TYPE_COLL_BASE(thetype)));
}

 SgDerivedCollectionType::~SgDerivedCollectionType()
{ RemoveFromTableType((void *) this); }

// add by Liyu
inline SgDistributeAttrExp::SgDistributeAttrExp(PTR_LLND ll) : SgExpression(ll)
{}

inline SgAlignAttrExp::SgAlignAttrExp(PTR_LLND ll) : SgExpression(ll)
{}

inline SgDataDistExp::SgDataDistExp(PTR_LLND ll) : SgExpression(ll)
{}
// end of liyu addition
